/*
 * vastai_ai - driver a Vastai ai device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2021/01/25
 */

#ifndef __VASTAI_AI_H__
#define __VASTAI_AI_H__

#include "vastai_pci.h"
#include <linux/cdev.h>
#include <linux/kernel.h>

/* The ai driver version. Runtime use it to check the version
Given a version number MAJOR.MINOR.PATCH, increment the:
1. MAJOR version when you make incompatible API changes
2. MINOR version when you add functionality in a backwards compatible manner
3. PATCH version when you make backwards compatible bug fixes
Additional labels for pre-release and build metadata are available as extensions to the MAJOR.MINOR.PATCH format.
*/
#ifndef AI_DRIVER_VERSION
#define AI_DRIVER_VERSION 0x090000
#endif

#define VASTAI_ZONE_MAX_NUM 4

// device type of VASTAI
typedef enum {
	VASTAI_SV100, // SV100
	VASTAI_SV100_HV, //SV100 harvest
	VASTAI_SG100, // SG100
	VASTAI_DEV_COUNT
} vastai_dev_type_t;

// csram ddr zone type
typedef enum {
	zone_stream, // csram, stream_info
	zone_model, // ddr local share, model_info opcode_csr
	zone_operator, // ddr VDSP0 Entry1, code
	zone_share, // ddr share weight, model input/output, video
	zone_count,
} zone_type_t;

struct addr_size {
	u64 base; // base address
	u64 size; // size(Bytes)
};

// structure of bar
struct __attribute__((packed)) vastai_bar {
	u32 id; // bar id
	u32 : 32;
	u64 ddr_base; // base address of ddr
	u64 base; // base address
	u64 size; // size(Bytes)
};

struct __attribute__((packed)) vastai_zone {
	u32 id;
	u32 zone_type;
	u32 alig_ddr_start;
	u32 entry;
	u32 bank_num : 8;
	u64 : 56;
	struct addr_size bank[VASTAI_ZONE_MAX_NUM];
};

struct __attribute__((packed)) vastai_vdsp {
	/* data */
	u32 id; // vdsp id
	u32 : 32;
	u64 base; // base address
	u64 size; // size(Bytes)
};

struct __attribute__((packed)) vastai_wdbuf {
	u32 bank_num; // bank number
	u32 : 32;
	u64 bank_size; // bank size(Bytes)
	u64 base; // base address
	u64 size; // size(Bytes)
	// u64 total_size;        // wdbuf size(Bytes)
};

struct __attribute__((packed)) vastai_dlc {
	u32 id;
	u32 core_flag; // bit0 indicates whether this core is a large core
	u32 wdma_fw_entry;
	u32 : 32;
	u64 csr_base; // csr base address
	u64 csr_size; // csr size(Bytes)
	struct vastai_wdbuf wdbuf;
};

struct __attribute__((packed)) vastai_ssram {
	u32 id;
	u32 : 32;
	u64 base; // base address
	u64 size; // size(Bytes)
};

// structure of odma
struct __attribute__((packed)) vastai_odma {
	u32 id; // odma id
	u32 : 32;
	u64 base; // base address
	u64 size; // size(Bytes)
};

// soc equal to a soc in nature (total 100 bytes)
struct __attribute__((packed)) vastai_soc {
	u16 type; // vastai_dev_type_t
	u16 vacc_id; // vacc index, vacc0, vacc1
	u16 render_id; // render id
	u32 harvest_core_flag; // harvest core flag
	u8 enabled_core_num; // enabled core number
	u8 core_num; // total core number
	u8 bar_num; // bar number
	u8 zone_num; // zone number(csram, ddr)
	u8 vdsp_num; // vdsp core number
	u8 ssram_num; // ssram number
	u8 odma_num; // odma number
	u8 odma_snum; //odma src number
	u8 odma_bnum; //odma branch number
	u8 ctrans_bnum; // ctrans branch number
	u16 alig_ssram; // alignment of SRAM
	u16 alig_wdbuf_conv; // alignment of WDBUF
	u16 alig_wdbuf_fc_gemm; // alignment of WDBUF
	u32 wdma_bud; // alignment of WDMA 32M boundry (32M)
	u32 odma_bud; // alignment of ODMA 512M boundry (512M)
	u32 throttle_max_bound; //throttle max bound
	u32 capacity_ratio; // capacity ratio * CAPACITY_RATIO_BASE
	u32 capacity_base; // capacity ratio base
	u64 wdbuf_broadcast_base; // base address of wdbuf for odma
	u64 ddr_total_size; // total size of DDR
	u64 ddr_2_bit_ecc_start; //ddr_2_bit_ecc start
	u64 video_phys_addr; // physical address for video
	u64 video_soc_addr; // soc address for video
	u64 video_size; // size for video
	u8 reserved[58]; // reserved field
};

// struct process_node {
// 	struct list_head node; // node in process_head list
// 	pid_t pid;
// };

typedef enum {
	mmap_idle,
	mmap_initializing,
	mmap_initialized,
} mmap_init_t;

struct mmap_page {
	struct page *page;
	struct list_head node;
};

struct mmap_init {
	wait_queue_head_t wait;
	struct mutex mutex;
	pid_t pid;
	// 0 idle, 1 curren process init mmap
	// 2 has been initialized
	u32 state;
	struct list_head head; //mmap_page
	u64 addr;
	u32 size;
};

struct __attribute__((packed)) dmabuf_alloc_cmd {
	u32 size;
	int dma_buf_fd;
	u64 dma_addr_t;
};

struct __attribute__((packed)) dmabuf_start_cmd {
	u32 is_dev_to_host;
	u32 dma_buf_fd;
	u64 axi_addr;
	u32 size;
	u32 die_index;
};

/* reset device*/
#define VACC_RESET_DEV _IO('V', 0x01)
/* run model*/
#define VACC_RUN_MODEL _IOW('V', 0x02, union vastai_cmd)
/* run stream*/
#define VACC_RUN_STREAM _IOW('V', 0x03, union vastai_cmd)
/* load model*/
#define VACC_LOAD_MODEL _IOW('V', 0x04, union vastai_cmd)
/* destroy model */
#define VACC_DESTROY_MODEL _IOW('V', 0x05, union vastai_cmd)
/* load operator */
#define VACC_LOAD_OPERATOR _IOW('V', 0x06, union vastai_cmd)
/* get current process id for container in host */
#define VACC_GET_CURRENT_PID _IOR('V', 0x08, union vastai_cmd)
/* get process state */
#define VACC_GET_PROCESS _IOWR('V', 0x09, union vastai_cmd)
/* release process resource */
#define VACC_RELEASE_PROCESS _IOWR('V', 0x10, union vastai_cmd)
/* get out desc */
#define VACC_GET_OUT_DESC _IOWR('V', 0x11, union vastai_cmd)
/* read device odma info*/
#define VACC_ECC2BIT _IOWR('V', 0x12, bool)
/* move data device 2 device*/
#define VACC_CP_DIE2DIE _IOW('V', 0x13, union vastai_cmd)
/* read ai version */
#define VACC_RD_VERSION _IOR('V', 0x14, u32)
/* read fw status */
#define VACC_RD_FW_STATUS _IOR('V', 0x15, u32)

/* read device soc info*/
#define VACC_RD_SOC _IOR('V', 0x21, struct vastai_soc)
/* read device zone info*/
#define VACC_RD_ZONE _IOWR('V', 0x22, struct vastai_zone)
/* read device vdsp info*/
#define VACC_RD_VDSP _IOWR('V', 0x23, struct vastai_vdsp)
/* read device dlc info*/
#define VACC_RD_DLC _IOWR('V', 0x24, struct vastai_dlc)
/* read device ssram info*/
#define VACC_RD_SSRAM _IOWR('V', 0x25, struct vastai_ssram)
/* read device odma info*/
#define VACC_RD_ODAM _IOWR('V', 0x26, struct vastai_odma)
/* read device bar info*/
#define VACC_RD_BAR _IOWR('V', 0x28, struct vastai_bar)
/* ddr overflow*/
#define VACC_DDR_OVERFLOW _IO('V', 0x30)

/* init mmap */
#define VACC_MMAP_INIT _IOWR('V', 0x80, union vastai_cmd)

/* zero-copy by dma-buf */
#define VACC_DMA_BUF_ALLOC    _IOWR('V', 0xA0, struct dmabuf_alloc_cmd)
#define VACC_DMA_BUF_START    _IOWR('V', 0xA1, struct dmabuf_start_cmd)
#define VACC_DMA_BUF_TRANS    _IOWR('V', 0xA2, union vastai_cmd)
#define VACC_DMA_BUF_TRANS_SG _IOWR('V', 0xA3, union vastai_cmd)

#define VACC_DEVICE_NAME_LEN 32

struct vastai_cdev {
	unsigned int dev_id; // device id
	unsigned int die_id; // die id (for multi-die in one device)
	unsigned int die_index; // die index(for multi-die in multi-device)
	struct vastai_pci_info *pci_info; // the parent pci device
	struct vastai_file_info file_info;
	struct fasync_struct *fasync;
	wait_queue_head_t r_wait;
	wait_queue_head_t w_wait;
	wait_queue_head_t wait_runsteam;
	struct mmap_init mmap;
	// int dev_type;
	u32 ai_version; // ai version
	u32 fw_status; // fw status, such as CORE_POINT_CMCU
	struct vastai_soc soc;
	struct vastai_zone *zone;
	struct vastai_dlc *dlc;
	struct vastai_ssram *ssram;
	struct vastai_bar *bars;
	struct vastai_odma *odmas;
	struct vastai_vdsp *vdsp;
	struct dev_trans *dev_trans_ai;
	struct list_head vacc_dev_node; // node in device list
	// struct mutex process_mutex;
	// struct list_head process_head; //process_node
};

int vacc_init(void *pcie_dev, unsigned int die_index);
void vacc_exit(void *pcie_dev);

#endif /* end of __VASTAI_AI_H__ */
