/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/delay.h>
#include <linux/workqueue.h>
#include <linux/firmware.h>

#include "vastai_pci_boot.h"
#include "vastai_pci_test.h"
#include "vastai_dmi_table.h"
#define CONFIG_VASTAI_DOWNLOAD_FW_FROM_HEX
#ifdef CONFIG_VASTAI_DOWNLOAD_FW_FROM_HEX
#include "vastai_fw_hex.h"
#endif
#include "vastai_version.h"
#include "vastai_udma_engine.h"
#include "vastai_die.h"
#include "vastai_avfs.h"
#include "vastai_state.h"

#define vastai_bl0_hexcode	"../fw/vastai_bl0.bin.hex"
#define vastai_bl0_app_hexcode	"../fw/vastai_bl0_app.bin.hex"
#define vastai_bl1_hexcode	"../fw/vastai_bl1.bin.hex"
#define vastai_fw_hexcode	"../fw/vastai_fw.bin.hex"
#define vastai_fw_ddrbw_hexcode "../fw/vastai_fw_ddrbw.bin.hex"
#define vastai_bmcu_hexcode "../fw/vastai_bmcu.bin.hex"

const unsigned char vastai_bl0_hex_buf[] = {
#include vastai_bl0_hexcode
};
const unsigned char vastai_bl0_app_hex_buf[] = {
#include vastai_bl0_app_hexcode
};
const unsigned char vastai_bl1_hex_buf[] = {
#include vastai_bl1_hexcode
};
const unsigned char vastai_fw_hex_buf[] = {
#include vastai_fw_hexcode
};
const unsigned char vastai_fw_ddrbw_hex_buf[] = {
#include vastai_fw_ddrbw_hexcode
};
const unsigned char vastai_bmcu_hex_buf[] = {
#include vastai_bmcu_hexcode
};

/*
 * smcu boot stage:
 * bit31(gpio3/pkg sel), bit30(gpio2/last die), bit29(dieid/slave die)
 * die_num die_id val      bit31 bit30 bit29
 * 1-die   die0   3'h2     0     1     0
 * 2-die   die0   3'h0     0     0     0
 * 2-die   die1   3'h3     0     1     1
 * 4-die   die0   3'h4     1     0     0
 * 4-die   die1   3'h1     0     0     1
 * 4-die   die2   3'h5     1     0     1
 * 4-die   die3   3'h7     1     1     1
 */
struct vastai_die_cfg die_cfg[VASTAI_DIE_CONFIG_TABLE_NUM] = {
	{ 2, VASTAI_SINGLE_DIE, VASTAI_DIE0 },
	{ 0, VASTAI_TWO_DIE, VASTAI_DIE0 },
	{ 3, VASTAI_TWO_DIE, VASTAI_DIE1 },
	{ 4, VASTAI_FOUR_DIE, VASTAI_DIE0 },
	{ 1, VASTAI_FOUR_DIE, VASTAI_DIE1 },
	{ 5, VASTAI_FOUR_DIE, VASTAI_DIE2 },
	{ 7, VASTAI_FOUR_DIE, VASTAI_DIE3 }
};

struct vastai_core_sync core_sync[VASTAI_CORE_SYNC_TOTAL_NUM] = {
	{ 0, "smcu "},
	{ 1, "cmcu "},
	{ 2, "lmcu0 "},
	{ 3, "lmcu1 "},
	{ 4, "lmcu2 "},
	{ 5, "lmcu3 "},
	{ 6, "lmcu4 "},
	{ 7, "lmcu5 "},
	{ 8, "lmcu6 "},
	{ 9, "lmcu7 "},
	{ 10, "odsp0 "},
	{ 11, "odsp1 "},
	{ 12, "odsp2 "},
	{ 13, "odsp3 "},
	{ 14, "odsp4 "},
	{ 15, "odsp5 "},
	{ 16, "odsp6 "},
	{ 17, "odsp7 "},
	{ 18, "vdmcu0 "},
	{ 19, "vdmcu1 "},
	{ 20, "vdmcu2 "},
	{ 21, "vemcu0 "},
	{ 22, "vemcu1 "},
	{ 23, "vemcu2 "},
	{ 24, "vemcu3 "},
	{ 25, "vdsp0 "},
	{ 26, "vdsp1 "},
	{ 27, "vdsp2 "},
	{ 28, "vdsp3 "},
	{ 29, "fwtotal"}
};

u8 vastai_pci_read_die_num(struct vastai_pci_info *priv)
{
	int die_num = VASTAI_SINGLE_DIE;
#ifndef CONFIG_VASTAI_PCI_ONLY_DOWNLOAD_BL1
	union smcu_boot_stage reg;
	u64 reg_addr = VASTAI_SMCU_BOOT_STAGE_REG;
	u32 i;

	vastai_pci_mem_read_direct(priv, -1, reg_addr, &(reg.val), 4);
	VASTAI_PCI_DBG_RATELIMIT("%s [read] addr=0x%llx, len=4, val=0x%08x\n",
				 __func__, reg_addr, reg.val);
	/*
	 * VASTAI_SMCU_BOOT_STAGE_REG, 0x6E4
	 * bit[31:29]: die config, refer to struct vastai_die_cfg table
	 */
	for (i = 0; i < VASTAI_DIE_CONFIG_TABLE_NUM; i++) {
		if (die_cfg[i].val == reg.bit.die_cfg) {
			die_num = die_cfg[i].die_num;
			break;
		}
	}
	if (i == VASTAI_DIE_CONFIG_TABLE_NUM) {
		die_num = VASTAI_SINGLE_DIE;
		VASTAI_PCI_ERR(
			priv, DUMMY_DIE_ID,
			"%s cfg=%d not in the cfg tbl, set to single-die.\n",
			__func__, reg.bit.die_cfg);
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s=%d, cfg=%d\n", __func__,
			die_num, reg.bit.die_cfg);
#endif
	return die_num;
}

/**
 * @brief gen a string include fw version information
 *
 * @param pci_info: vastai sv100 pci base struct.
 * @param mask: bitmap for which mcu's version need print
 * @return char point: string buffer. It's need free after used.
 */
void vastai_pci_show_fw_ver(struct vastai_pci_info *pci_info, u32 mask)
{
	if (mask & 0x1)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n",
				pci_info->dies[0].fw_ver.ver);
	if (mask & 0x2)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n",
				pci_info->dies[0].fw_ver.smcu_ver);
	if (mask & 0x4) {
		struct vastai_fw_bmcu_xspi_version bmcu_xspi_ver;

		memcpy(bmcu_xspi_ver.bmcu_ver,
		       pci_info->dies[0].fw_ver.bmcu_ver,
		       VASTAI_VER_COMMON_LEN);
		bmcu_xspi_ver.bmcu_sub_ver[VASTAI_VER_BMCU_XSPI_LEN - 1] = 0;

		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s] [active]\n",
				bmcu_xspi_ver.bmcu_sub_ver);

		memcpy(bmcu_xspi_ver.bmcu_ver,
		       &(pci_info->dies[0].fw_ver.bmcu_backup_ver),
		       VASTAI_VER_COMMON_LEN);
		bmcu_xspi_ver.bmcu_sub_ver[VASTAI_VER_BMCU_XSPI_LEN - 1] = 0;

		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s] [backup]\n",
				bmcu_xspi_ver.bmcu_sub_ver);
	}
	if (mask & 0x8)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n",
				pci_info->dies[0].fw_ver.vdmcu_ver);
	if (mask & 0x10)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n",
				pci_info->dies[0].fw_ver.vemcu_ver);
	if (mask & 0x20)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n",
				pci_info->dies[0].fw_ver.vdsp_ver);
	if (mask & 0x40)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n",
				pci_info->dies[0].fw_ver.cmcu_ver);
	if (mask & 0x80)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n",
				pci_info->dies[0].fw_ver.lmcu_ver);
	if (mask & 0x100)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n",
				pci_info->dies[0].fw_ver.odsp_ver);
	if (mask & 0x200)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n",
				pci_info->dies[0].fw_ver.driver_ver);

	if (mask & 0x400) {
		int i = 0;
		for (i = 0; i < pci_info->die_num; i++) {
			struct vastai_fw_bmcu_xspi_version *bmcu_xspi_ver;
			bmcu_xspi_ver = (struct vastai_fw_bmcu_xspi_version *)
					pci_info->dies[i].fw_ver.bmcu_ver;

			VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID,
					"[die %d] [%s] [active]\n",
					i, bmcu_xspi_ver->xspi_sub_ver);

			bmcu_xspi_ver = (struct vastai_fw_bmcu_xspi_version *)
				pci_info->dies[i].fw_ver.bmcu_backup_ver;
			VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID,
					"[die %d] [%s] [backup]\n",
					i, bmcu_xspi_ver->xspi_sub_ver);
		}
	}
	if (mask & 0x800) {
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n",
				pci_info->dies[0].fw_ver.phy_ver);
	}
}

void vastai_pci_get_fw_ver(struct vastai_pci_info *pci_info, u32 mask)
{
	int ret = 0;
	int i;

	for (i = 0; i < pci_info->die_num; i++) {
		unsigned int die_index = pci_info->dies[i].die_index;
		struct vastai_sv100_die *die = &(pci_info->dies[i]);

		ret = vastai_pci_mem_write(pci_info, die_index, VSERSION_INFO_BASE,
					   VASTAI_VERSION, VASTAI_VER_COMMON_LEN);
		if (ret < 0) {
			VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				       "%s write version error:%d\n",
				       __func__, ret);
			break;
		}

		ret = vastai_pci_mem_read(pci_info, die_index, VSERSION_INFO_BASE,
					  &die->fw_ver,
					  sizeof(struct vastai_fw_version));
		if (ret < 0) {
			VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				       "%s read firmware version error:%d\n",
				       __func__, ret);
			break;
		}
		ret = vastai_pci_mem_read(pci_info, die_index, STORE_INFO_ADDR_BMCU_BK,
					  die->fw_ver.bmcu_backup_ver,
					  sizeof(struct vastai_fw_bmcu_xspi_version ));
		if (ret < 0) {
			VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				       "%s read bmcu backup firmware version error:%d\n",
				       __func__, ret);
			break;
		}

		die->fw_ver.ver[VASTAI_VER_COMMON_LEN - 1] = 0;
		die->fw_ver.smcu_ver[VASTAI_VER_COMMON_LEN - 1] = 0;
		die->fw_ver.bmcu_ver[VASTAI_VER_COMMON_LEN - 1] = 0;
		die->fw_ver.vdmcu_ver[VASTAI_VER_COMMON_LEN - 1] = 0;
		die->fw_ver.vemcu_ver[VASTAI_VER_COMMON_LEN - 1] = 0;
		die->fw_ver.vdsp_ver[VASTAI_VER_COMMON_LEN - 1] = 0;
		die->fw_ver.cmcu_ver[VASTAI_VER_COMMON_LEN - 1] = 0;
		die->fw_ver.lmcu_ver[VASTAI_VER_COMMON_LEN - 1] = 0;
		die->fw_ver.odsp_ver[VASTAI_VER_COMMON_LEN - 1] = 0;
		snprintf(die->fw_ver.driver_ver, VASTAI_VER_COMMON_LEN,
			 "Driver: %s %s %s %s", VASTAI_PCI_DRIVER_VERSION,
			 VASTAI_PCI_GIT_BRANCH_NAME, VASTAI_PCI_GIT_COMMIT_HASH,
			 VASTAI_PCI_MKPACKAGE_TIME);
		/* pcie_phy version info not in csram */
		/* pcie_hpy version in xspiflash 0xd000 */
		/* we read from app partition, relative position */
#ifndef CONFIG_VASTAI_AI_EMU
		vastai_pci_read_xspi(pci_info, i, 0x6000, die->fw_ver.phy_ver,
					sizeof(die->fw_ver.phy_ver), 0);
#endif
		if (die->fw_ver.phy_ver[0] == 0xFF)
			strcpy(die->fw_ver.phy_ver, "no pcie phy version");
	}
	if (!ret) {
		vastai_pci_show_fw_ver(pci_info, mask);
	}
}

static int vastai_pci_poll_smcu_stage(struct vastai_pci_info *priv, u8 die_id,
				      u32 smcu_bootstage)
{
	int ret;
	union smcu_boot_stage reg;
	u64 reg_addr = VASTAI_SMCU_BOOT_STAGE_REG;
	u64 tm_begin = ktime_to_ms(ktime_get()), tm_end, tm_total_ms;

	do {
		ret = vastai_pci_mem_read_direct(priv, vastai_pci_get_die_index(priv, die_id),
					  reg_addr, &(reg.val), 4);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, die_id,
				       "%s [read] 0x%llx, len=4, error:%d\n",
				       __func__, reg_addr, ret);
			break;
		}
		VASTAI_PCI_DBG_RATELIMIT("%s [read] addr=0x%llx, len=4, "
					 "val=0x%08x, poll die%d stage%d\n",
					 __func__, reg_addr, reg.val, die_id,
					 smcu_bootstage);
		/*
		 * VASTAI_SMCU_BOOT_STAGE_REG, 0x6E4
		 * bit[26:24]: smcu boot stage.
		 *             3'h0: initial state
		 *             3'h1: wait for downloading bl1
		 *             3'h2: bl1 verify pass
		 *             3'h3: wait for downloading fw
		 *             3'h4: fw verify pass
		 */
		if (smcu_bootstage < VASTAI_SMCU_BOOT_STAGE_DIE_DONE) {
			if (reg.bit.stage == smcu_bootstage) {
				VASTAI_PCI_DBG(priv, die_id,
					       "%s smcu.bootstage=0x%x, ready "
					       "to download. "
					       "(1:bl1_ready,2:bl1_done,"
					       "3:fw_ready,4:fw_done)",
					       __func__, reg.bit.stage);
				break;
			}
		} else {
			if ((reg.bit.stage >= smcu_bootstage) &&
			    (reg.bit.stage != VASTAI_SMCU_BOOT_STAGE_DUMMY)) {
				VASTAI_PCI_DBG(priv, die_id,
					       "%s smcu.bootstage=0x%x, ready "
					       "to download. "
					       "(1:bl1_ready,2:bl1_done,"
					       "3:fw_ready,4:fw_done)",
					       __func__, reg.bit.stage);
				break;
			}
		}

		tm_end = ktime_to_ms(ktime_get());
		tm_total_ms = tm_end - tm_begin;
		if (tm_total_ms > CONFIG_VASTAI_POLL_BOOT_STAGE_MS) {
			VASTAI_PCI_ERR(
				priv, die_id,
				"%s poll stage%d, but stage is %d, %ds timeout!\n",
				__func__, smcu_bootstage, reg.bit.stage,
				(CONFIG_VASTAI_POLL_BOOT_STAGE_MS / 1000));
			if (reg.bit.stage == VASTAI_SMCU_BOOT_STAGE_BL1_READY)
				return -EAGAIN;
			else
				return -ETIMEDOUT;
		}
		msleep(VASTAI_BOOT_POLL_SLEEP_MS);
	} while (1);

	return ret;
}

static int vastai_pci_poll_die_linkup(struct vastai_pci_info *priv, u8 die_id)
{
	int ret;
	union smcu_boot_stage reg;
	u64 reg_addr = VASTAI_SMCU_BOOT_STAGE_REG;
	u64 tm_begin = ktime_to_ms(ktime_get()), tm_end, tm_total_ms;

	do {
		ret = vastai_pci_mem_read_direct(priv, vastai_pci_get_die_index(priv, die_id),
					  reg_addr, &(reg.val), 4);
		if (ret < 0)
			VASTAI_PCI_ERR(priv, die_id,
				       "%s [read] 0x%llx, len=4, error:%d\n",
				       __func__, reg_addr, ret);
		VASTAI_PCI_DBG_RATELIMIT("%s [read] addr=0x%llx, len=4, "
					 "val=0x%08x, poll die%d linkup\n",
					 __func__, reg_addr, reg.val, die_id);
		/*
		 * VASTAI_SMCU_BOOT_STAGE_REG, 0x6E4
		 * bit[28]: rc_link_sts:
		 *          current die ep and front die or host  rc link up
		 * flag bit[27]: ep_link_sts: current die rc and next die ep
		 * link up flag
		 */
		if (reg.bit.rc_link_sts == 1) {
			if (reg.bit.d2d_link_retry_cnt != 0)
				VASTAI_PCI_INFO(priv, die_id,
						"%s die%d retry %d times "
						"to linkup",
						__func__, (die_id + 1),
						reg.bit.d2d_link_retry_cnt);
			else
				VASTAI_PCI_INFO(priv, die_id,
						"%s die%d linkup",
						__func__, (die_id + 1));
			break;
		}

		tm_end = ktime_to_ms(ktime_get());
		tm_total_ms = tm_end - tm_begin;
		if (tm_total_ms > CONFIG_VASTAI_POLL_BOOT_STAGE_MS) {
			VASTAI_PCI_ERR(priv, die_id,
				       "%s poll linkup, but link status is %d,"
				       "retry %d times, %ds timeout!\n",
				       __func__, reg.bit.rc_link_sts,
				       reg.bit.d2d_link_retry_cnt,
				       (CONFIG_VASTAI_POLL_BOOT_STAGE_MS / 1000));
			return -ETIMEDOUT;
		}
		msleep(VASTAI_BOOT_POLL_SLEEP_MS);
	} while (1);

	return ret;
}

static int vastai_pci_check_core_sync(struct vastai_pci_info *priv, u8 die_id)
{
	int ret, i, j;
	u64 reg_addr = CORE_STATUS_SYNC;
	u32 reg_val;

	memset(priv->dies[die_id].core_sync_result, 0,
	       VASTAI_CORE_SYNC_TOTAL_NUM * 8);
	for (i = 0; i < VASTAI_CORE_SYNC_TOTAL_NUM; i += 4) {
		ret = vastai_pci_mem_read_direct(priv, vastai_pci_get_die_index(priv, die_id),
					  (reg_addr + i), &reg_val, 4);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, die_id,
				       "%s [read] 0x%llx, len=4, error:%d\n",
				       __func__, (reg_addr + i), ret);
			return ret;
		}
		for (j = 0; j < 4; j++) {
			if ((i + j) >= VASTAI_CORE_SYNC_TOTAL_NUM)
				break;
			if (((reg_val >> (j * 8)) & 0xFF) != VASTAI_FW_BOOT_OK)
				strcat(priv->dies[die_id].core_sync_result,
				       core_sync[i + j].str);
		}
	}
	return ret;
}

#if 0 /* This action (set bar4 size to 256GB) has been moved to smcu side. */
static int vastai_pci_set_bar4_size_256g(struct vastai_pci_info *priv, u8 die_id)
{
	int ret = 0;
	u32 reg_val;
	u64 reg_addr;

	reg_addr = VASTAI_PCI_CDN_PHY_FUNC_BAR_CFG_REG1;
	ret = vastai_pci_mem_read(priv, vastai_pci_get_die_index(priv, die_id), reg_addr, &reg_val, 4);
	/* bar4: 64bit memory BAR, prefetchable, 256GB */
	reg_val = ((reg_val & 0xFFFF0000) | 0x05FF);
	ret = vastai_pci_mem_write(priv, vastai_pci_get_die_index(priv, die_id), reg_addr, &reg_val, 4);
	/* Add delay to solve atu configuration not work issue. */
	mdelay(10);

	return ret;
}
#endif

static int vastai_pci_refresh_msix_table(struct vastai_pci_info *priv)
{
	int ret = 0; //, i;

#ifndef CONFIG_TRAP_MSIX_ACCESS
	ret = vastai_pci_mem_read_direct(
		priv, vastai_pci_get_die_index(priv, 0), PCIE_MSIX_PF_TABLE,
		&priv->msix_table,
		(VASTAI_PCIE_MSIX_NUM * sizeof(struct msi_x_table)));
#else
	unsigned long saved_atu_addr = priv->bar[VASTAI_PCI_BAR2].at_addr;
	u32 tmp_data = 0;

	ret = vastai_pci_mem_bar2_read(priv, -1, PCIE_MSIX_PF_TABLE, priv->msix_table,
					(VASTAI_PCIE_MSIX_NUM * sizeof(struct msi_x_table)));
	/*switch back bar2*/
	ret = vastai_pci_mem_bar2_read(priv, -1, saved_atu_addr, &tmp_data,
					4);
#endif

#if 0
	/* Caution:
	 * [2021/5/31]
	 * On the Zebu platform, the host updates the MSI-X table by means of
	 * PCIE memwr, which causes the emulated environment to assume that
	 * the attributes of the MSI-X are being updated at the same time,
	 * resulting in failure to receive subsequent interrupts.
	 */
	/* The first 32 items in the MSI-X table are refreshed with the new
	 * value (add address base 0x7000_0000) of HOST, and the parameters
	 * need to be fetched starting from the 33rd item for mcu.
	 */
	ret = vastai_pci_mem_write(priv, vastai_pci_get_die_index(priv, die_id), (VASTAI_PCIE_MSIX_TABLE + (VASTAI_PCIE_MSIX_NUM_PER_DIE * sizeof(struct msi_x_table))), &priv->msix_table[0], sizeof(struct msi_x_table));
#endif
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"%s msg_addr:0x%x 0x%x, msg_data:0x%x,"
			" vector_control:0x%x\n",
			__func__, priv->msix_table[0].msg_addr,
			priv->msix_table[0].msg_upper_addr,
			priv->msix_table[0].msg_data,
			priv->msix_table[0].vector_control);
#if 0
	/* mcu write address 0x7000_0000, will be routed to address 0 to
	 * PCIe ep master via NoC. */
	for (i = 0; i < VASTAI_PCIE_MSIX_NUM; i++)
		priv->msix_table[i].msg_addr = (priv->msix_table[i].msg_addr & 0x0FFFFFFF) | VASTAI_PCIE_EP_SLAVE_BASE;
	ret = vastai_pci_mem_write(priv, vastai_pci_get_die_index(priv, die_id), VASTAI_PCIE_MSIX_TABLE, &priv->msix_table[0], (VASTAI_PCIE_MSIX_NUM_PER_DIE * sizeof(struct msi_x_table)));
#endif
	return ret;
}

static int vastai_pci_die_enumeration(struct vastai_pci_info *priv, u8 die_id)
{
	int ret = 0;
	u32 reg_val;
	u64 reg_addr;

	/* config bar0 ~ bar5 */
	if (priv->bar[VASTAI_PCI_BAR0].minish_bar_size)
		reg_val = VASTAI_PCIE_BAR0_PCIE_ADDR_8MB;
	else
		reg_val = VASTAI_PCIE_BAR0_PCIE_ADDR_32MB;
	reg_addr = VASTAI_PCIE_EP_CONFIG_BAR0;
	ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id), reg_addr,
				   &reg_val, 4);
	reg_val = VASTAI_PCIE_BAR1_PCIE_ADDR_32MB;
	reg_addr = VASTAI_PCIE_EP_CONFIG_BAR1;
	ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id), reg_addr,
				   &reg_val, 4);
	if (priv->bar[VASTAI_PCI_BAR2].minish_bar_size)
		reg_val = VASTAI_PCIE_BAR2_PCIE_ADDR_32MB & VASTAI_VAL_DW_MASK;
	else
		reg_val = VASTAI_PCIE_BAR2_PCIE_ADDR_32GB & VASTAI_VAL_DW_MASK;
	reg_addr = VASTAI_PCIE_EP_CONFIG_BAR2;
	ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id), reg_addr,
				   &reg_val, 4);
	if (priv->bar[VASTAI_PCI_BAR2].minish_bar_size)
		reg_val = (VASTAI_PCIE_BAR2_PCIE_ADDR_32MB >> 32) &
			  VASTAI_VAL_DW_MASK;
	else
		reg_val = (VASTAI_PCIE_BAR2_PCIE_ADDR_32GB >> 32) &
			  VASTAI_VAL_DW_MASK;
	reg_addr = VASTAI_PCIE_EP_CONFIG_BAR3;
	ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id), reg_addr,
				   &reg_val, 4);
	reg_val = VASTAI_PCIE_BAR4_PCIE_ADDR_256GB & VASTAI_VAL_DW_MASK;
	reg_addr = VASTAI_PCIE_EP_CONFIG_BAR4;
	ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id), reg_addr,
				   &reg_val, 4);
	reg_val = (VASTAI_PCIE_BAR4_PCIE_ADDR_256GB >> 32) & VASTAI_VAL_DW_MASK;
	reg_addr = VASTAI_PCIE_EP_CONFIG_BAR5;
	ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id), reg_addr,
				   &reg_val, 4);

	/* enable mem-space */
	reg_val = 0x00100507;
	reg_addr = VASTAI_PCIE_EP_CONFIG_CMD_STATUS;
	ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id), reg_addr,
				   &reg_val, 4);

#if 0 /* This action (set bar4 size to 256GB) has been moved to smcu side. */
	if (priv->minish_bar_size == 0)
		vastai_pci_set_bar4_size_256g(priv, die_id);
#endif

	/* enable msi */
	if (priv->irq_type == VASTAI_MSI_IRQ) {
		reg_addr = VASTAI_PCIE_EP_CONFIG_MSI_CTRL;
		ret = vastai_pci_mem_read_direct(priv, vastai_pci_get_die_index(priv, die_id),
					  reg_addr, &reg_val, 4);
		reg_val |= 0x10000;
		ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id),
					   reg_addr, &reg_val, 4);
		reg_val = VASTAI_MULTI_DIE_MSI_ADDR & VASTAI_VAL_DW_MASK;
		reg_addr = VASTAI_PCIE_EP_CONFIG_MSI_MSG_L;
		ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id),
					   reg_addr, &reg_val, 4);
		reg_val =
			(VASTAI_MULTI_DIE_MSI_ADDR >> 32) & VASTAI_VAL_DW_MASK;
		reg_addr = VASTAI_PCIE_EP_CONFIG_MSI_MSG_H;
		ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id),
					   reg_addr, &reg_val, 4);
	}

#ifndef CONFIG_TRAP_MSIX_ACCESS
	/* enable msi-x and write msix table */
	if (priv->irq_type == VASTAI_MSIX_IRQ) {
		reg_addr = VASTAI_PCIE_EP_CONFIG_MSIX_CTRL;
		ret = vastai_pci_mem_read_direct(priv, vastai_pci_get_die_index(priv, die_id),
					  reg_addr, &reg_val, 4);
		reg_val |= 0x80000000;
		ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id),
					   reg_addr, &reg_val, 4);
		ret = vastai_pci_mem_write_direct(
			priv, vastai_pci_get_die_index(priv, die_id), PCIE_MSIX_PF_TABLE,
			&priv->msix_table[VASTAI_PCIE_MSIX_NUM_PER_DIE * die_id],
			(VASTAI_PCIE_MSIX_NUM_PER_DIE *
			 sizeof(struct msi_x_table)));
	}
#endif

	return ret;
}

static int vastai_pci_die_enumeration_bh(struct vastai_pci_info *priv,
					 u8 die_id)
{
	int ret = 0;
	u32 reg_val;
	u64 reg_addr;

	/* config bar0 ~ bar5 */
	if (priv->bar[VASTAI_PCI_BAR0].minish_bar_size) {
		reg_val = VASTAI_PCIE_BAR0_PCIE_ADDR_256MB;
		reg_addr = VASTAI_PCIE_EP_CONFIG_BAR0;
		ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id),
					   reg_addr, &reg_val, 4);
	}

	if (priv->bar[VASTAI_PCI_BAR2].minish_bar_size) {
		reg_val = VASTAI_PCIE_BAR2_PCIE_ADDR_256GB & VASTAI_VAL_DW_MASK;
		reg_addr = VASTAI_PCIE_EP_CONFIG_BAR2;
		ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id),
					   reg_addr, &reg_val, 4);
		reg_val = (VASTAI_PCIE_BAR2_PCIE_ADDR_256GB >> 32) &
			  VASTAI_VAL_DW_MASK;
		reg_addr = VASTAI_PCIE_EP_CONFIG_BAR3;
		ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id),
					   reg_addr, &reg_val, 4);
	}

	return ret;
}

static int vastai_pci_poll_host_stage(struct vastai_pci_info *priv, u8 die_id,
				      u32 host_bootstage)
{
	int ret;
	union host_boot_stage reg;
	u64 reg_addr = VASTAI_HOST_BOOT_STAGE_REG;
	u64 tm_begin = ktime_to_ms(ktime_get()), tm_end, tm_total_ms;

	do {
		ret = vastai_pci_mem_read_direct(priv, vastai_pci_get_die_index(priv, die_id),
					  reg_addr, &(reg.val), 4);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, die_id,
				       "%s [read] 0x%llx, len=4, error:%d\n",
				       __func__, reg_addr, ret);
			break;
		}
		VASTAI_PCI_DBG_RATELIMIT("%s [read] addr=0x%llx, len=4, "
					 "val=0x%08x, poll die%d stage%d\n",
					 __func__, reg_addr, reg.val, die_id,
					 host_bootstage);
		if (reg.bit.stage == host_bootstage) {
			VASTAI_PCI_DBG(priv, die_id, "%s host.bootstage=0x%x",
				       __func__, reg.bit.stage);
			break;
		}

		tm_end = ktime_to_ms(ktime_get());
		tm_total_ms = tm_end - tm_begin;
		if (tm_total_ms > CONFIG_VASTAI_POLL_BOOT_STAGE_MS) {
			VASTAI_PCI_ERR(
				priv, die_id,
				"%s poll stage%d, but stage is %d, %ds timeout!\n",
				__func__, host_bootstage, reg.bit.stage,
				(CONFIG_VASTAI_POLL_BOOT_STAGE_MS / 1000));
			return -ETIMEDOUT;
		}
		msleep(VASTAI_BOOT_POLL_SLEEP_MS);
	} while (1);

	return ret;
}

static int vastai_pci_set_host_stage(struct vastai_pci_info *priv, u8 die_id,
				     u32 host_bootstage)
{
	int ret;
	union host_boot_stage reg;
	u64 reg_addr = VASTAI_HOST_BOOT_STAGE_REG;

	ret = vastai_pci_mem_read_direct(priv, vastai_pci_get_die_index(priv, die_id), reg_addr,
				  &(reg.val), 4);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, die_id,
			       "%s [read] 0x%llx, len=4, error:%d\n", __func__,
			       reg_addr, ret);
	/*
	 * VASTAI_HOST_BOOT_STAGE_REG, 0x6E0
	 * bit[26:24]: host boot stage.
	 *             3'h0: initial state
	 *             3'h1: bl1 download finish
	 *             3'h2: fw download finish
	 *             3'h3: all dies download finish, host ready
	 */
	reg.bit.stage = host_bootstage;
	ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id), reg_addr,
				   &(reg.val), 4);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, die_id,
			       "%s [write] 0x%llx, len=4, val=0x%x, error:%d\n",
			       __func__, reg_addr, reg.val, ret);

	return ret;
}

static int vastai_pci_set_current_die(struct vastai_pci_info *priv, u8 die_id)
{
	int ret;
	union host_boot_stage reg;
	u64 reg_addr = VASTAI_AXI_CSR_BASE_ADDR + 0x6E0;

	ret = vastai_pci_mem_read_direct(priv, vastai_pci_get_die_index(priv, die_id), reg_addr,
				  &(reg.val), 4);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, die_id,
			       "%s [read] 0x%llx, len=4, error:%d\n", __func__,
			       reg_addr, ret);
	priv->vadev.dev_type = VASTAI_PROCESSING_ACCELERATORS_DEVICE;
	reg.bit.hw_type = priv->vadev.dev_cap;
#ifdef CONFIG_VASTAI_HW_TYPE_BY_HOST
	reg.bit.host_set_hw_type = 1;
	if (die_id == VASTAI_DIE0)
		VASTAI_PCI_INFO(priv, die_id, "Host set hw type=%d.\n",
				reg.bit.hw_type);
#else
	reg.bit.host_set_hw_type = 0;
#endif

#ifdef CONFIG_VASTAI_CUS_TYPE
	reg.bit.cus_type = CONFIG_VASTAI_CUS_TYPE;
	if (die_id == VASTAI_DIE0)
		VASTAI_PCI_INFO(priv, die_id, "Host set customer type=%d.\n",
				reg.bit.cus_type);
#else
	reg.bit.cus_type = 0;
#endif

#ifdef CONFIG_VASTAI_AI_EMU
	reg.bit.env_type = VASTAI_EMU;
	VASTAI_PCI_INFO(priv, die_id, "This is emulation environment.\n");
#else
	reg.bit.env_type = VASTAI_SILICON;
#endif
#ifdef CONFIG_VASTAI_PCI_FORCE_SINGLE_DIE
	reg.bit.force_single_die = 1;
	VASTAI_PCI_INFO(priv, die_id, "This is forced single die.\n");
#else
	if (priv->bar[VASTAI_PCI_BAR0].mmio_len == (2 * 1024 * 1024))
		reg.bit.force_single_die = 1;
	else
		reg.bit.force_single_die = 0;
#endif
	if (priv->ddr_bw_test) {
		reg.bit.hw_type = VASTAI_HW_TYPE_DEFAULT;
	}
	ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id), reg_addr,
				   &(reg.val), 4);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, die_id,
			       "%s [write] 0x%llx, len=4, val=0x%x, error:%d\n",
			       __func__, reg_addr, reg.val, ret);

	return ret;
}

static int vastai_pci_write_fw(struct vastai_pci_info *priv,
			       unsigned int die_id, struct vastai_fw *vfw)
{
	int ret;

	/* NTD: judge length of fw */
	ret = vastai_pci_mem_write_direct(priv, vastai_pci_get_die_index(priv, die_id),
				   vfw->addr, (void *)(vfw->data), vfw->size);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, die_id,
			       "%s [write] 0x%llx, len=%d, error:%d\n",
			       __func__, vfw->addr, (int)(vfw->size), ret);
	return ret;
}

static int vastai_pci_download_fw(struct vastai_pci_info *priv,
				  unsigned int die_id,
				  const unsigned char *fw_hex_buf, size_t fw_hex_size,
				  u64 fw_addr)
{
	int ret;
	struct vastai_fw vfw;

	vfw.data = fw_hex_buf;
	vfw.size = fw_hex_size;
	vfw.priv = fw_hex_buf;
	vfw.addr = fw_addr;

	ret = vastai_pci_write_fw(priv, die_id, &vfw);
	if (ret) {
		VASTAI_PCI_ERR(priv, die_id, "%s write bl1/fw error\n",
			       __func__);
	}

	return ret;
}

/**
 * @brief download fw. If path_name is given, will try to
 * 	load fw from /lib/firmware first.
 *
 * @param pci_info: vastai sv100 pci base struct.
 * @param die_id: the destinat die id. not the die_index
 * @param path_name: fw path at /lib/firmware/. If not NULL, this is most priority.
 * @param data: fw.bin data.
 * @param size: fw.bin data's size.
 * @param fw_addr: fw destinat addr.
 * @return int: 0 is success. other is failed.
 */
int vastai_process_download_request(struct vastai_pci_info *pci_info,
				    u8 die_id, u8 *path_name, const u8 *data,
				    size_t size, u64 fw_addr)
{
	int request_ret = 0, ret = 0;
	const struct firmware *fw_entry = NULL;
	const u8 *__data = data;
	size_t __size = size;

	/* most priority */
	if (path_name) {
		request_ret = request_firmware(&fw_entry, path_name, &(pci_info->dev->dev));
		if (request_ret != 0)
			VASTAI_PCI_INFO(pci_info, die_id,
					"%s ret[%d]: can't find fw/bl from path[/lib/firmware/%s], it will download fw/bl from ko file\n",
					__func__, request_ret, path_name);
	}

	/* get fw from /lib/firmarw success */
	if (!request_ret && fw_entry) {
		VASTAI_PCI_INFO(pci_info, die_id,
				"%s get fw/bl from path[/lib/firmware/%s] success, download fw/bl from this path\n",
				__func__, path_name);
		__data = fw_entry->data;
		__size = fw_entry->size;
	}

	/* this is only for xspi update and bmcu update in normal time */
	/* background : 1. xspi flash and bmcu update is use AI's csram space.
			2. This space size is CHN_INFO_BASE - MODULE_INFO_BASE
	therefore, Need to check the bin's size when do xspi udpate and bmcu
	update.
	NOTE THAT: This code can be remove after using ddr space to do xpsi
			flash and bmcu udpate */
	if ((atomic_read(&(pci_info->pci_state)) != VASTAI_BOOT_STATE)
		&& (fw_addr == MODULE_INFO_BASE)
		&& (__size >= (CHN_INFO_BASE - MODULE_INFO_BASE))) {
		VASTAI_PCI_ERR(pci_info, die_id,
			       "fw bin is too long 0x%lx\n", __size);

		VASTAI_PCI_ERR(pci_info, die_id,
			       "scram buf begin 0x%x end 0x%x\n",
			       MODULE_INFO_BASE,
			       CHN_INFO_BASE);
		return -ENOSPC;
	}

	/* Not get any fw */
	if (!__data) {
		VASTAI_PCI_ERR(pci_info, die_id,
				"%s get fw/bl failed! path_name %s, data_addr %p\n",
				__func__, path_name, data);
		return -ENOENT;
	}

	/* load fw */
	ret = vastai_pci_download_fw(pci_info, die_id, __data,
				     __size, fw_addr);
	/* we need free fw_entry */
	if (fw_entry)
		release_firmware(fw_entry);
	return ret;
}

static int vastai_wait_rw_xspi(struct vastai_pci_info *priv, u32 die_id)
{
	struct vastai_sv100_die *die;
	int ret;

	if (!priv || (priv->die_num <= die_id)) {
		return -EINVAL;
	}

	die = &(priv->dies[die_id]);

	/* wait completion int from smcu */
	ret = wait_for_completion_interruptible(&(die->rw_xspi));
	reinit_completion(&(die->rw_xspi));
	if (die->rw_xspi_msg != VASTAI_XSPI_SUCCESS) {
		VASTAI_PCI_ERR(
			priv, die_id,
			"rw xspi result is 0x%x (1: erase err; 2: read err; 3: write err; 4: unlock err; 5: lock err; 6: chksum err; 7: flag err; 8: wp err; 9: type err)\n",
			die->rw_xspi_msg);
		return -(die->rw_xspi_msg);
	}

	return ret;
}

static int send_xspi_flash_cmd(struct vastai_pci_info *priv, u32 die_id,
				enum VASTAI_PCIE_SUBCMD cmd, u32 buf_size)
{
	int ret;

	/* write length of xspi bin to csram */
	if ((cmd == VASTAI_PCIE_SUB_FLASH_BASE_XSPI)
			|| (cmd == VASTAI_PCIE_SUB_FLASH_XSPI)) {
		ret = vastai_pci_mem_write(priv, priv->dies[die_id].die_index,
					   (VASTAI_XSPI_DL_ADDR - 4), &buf_size, 4);
	} else {
		ret = vastai_pci_mem_write(priv, priv->dies[die_id].die_index,
					   (XSPI_READ_BASE - 4), &buf_size, 4);
	}
	if (ret) {
		VASTAI_PCI_ERR(priv, die_id, "%s write xspi flash size failed %d\n",
				__func__, ret);
		return ret;
	}

	/* send read xspi cmd to smcu */
	ret = vastai_send_pcie_cmd(priv, priv->dies[die_id].die_index, cmd, 0);
	if (ret) {
		VASTAI_PCI_ERR(priv, die_id, "%s send xspi flash cmd failed %d\n",
				__func__, ret);
		return ret;
	}
	ret = vastai_wait_rw_xspi(priv, die_id);
	if (ret) {
		VASTAI_PCI_ERR(priv, die_id, "%s wait flash compltion failed %d\n",
				__func__, ret);
	}
	return ret;
}

int vastai_pci_read_xspi(struct vastai_pci_info *priv, u32 die_id,
			 unsigned long addr, void *buf, u32 len,
			 bool is_base_partition)
{
	int ret;

	if (addr >= BL0_BASE_TOTAL_SIZE ||
		(addr + len) > BL0_BASE_TOTAL_SIZE) {
		VASTAI_PCI_ERR(priv, die_id,
				"%s flash read failed with error ddr. addr 0x%lx len 0x%x",
				__func__, addr, len);
		return -EFAULT;
	}

	if (is_base_partition)
		ret = send_xspi_flash_cmd(priv, die_id,
					VASTAI_PCIE_SUB_READ_BASE_XSPI,
					BL0_BASE_TOTAL_SIZE);
	else
		ret = send_xspi_flash_cmd(priv, die_id,
					VASTAI_PCIE_SUB_READ_XSPI,
					BL0_APP_TOTAL_SIZE);
	if (ret) {
		VASTAI_PCI_ERR(priv, die_id, "%s send cmd failed%d\n",
				__func__, ret);
		return ret;
	}

	ret = vastai_pci_mem_read(priv, priv->dies[die_id].die_index,
				  addr + XSPI_READ_BASE, buf, len);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s [read] 0x%lx, len=%d, error:%d\n", __func__,
			       addr, len, ret);
	return ret;
}

int vastai_pci_flash_bmcu_bin(struct vastai_pci_info *priv, u32 die_id,
			      u8 *path_name, const void* bin_buf, size_t buf_size)
{
	int ret = 0;
	u32 reg_val;

	/* vastai_pci_send_pcie_cmd need smcu to response. Therefore, to check
		pcie_state is VASTAI_NORMAL_STATE before turn it to
		VASTAI_UPDATE_STATE */
	/* This pci_state will turn to be VASTAI_UPDATE_NORMAL after recive MSIX
		after bmcu update is done. reference to
		msix_handle: VASTAI_PCIE_MSG1_BMCU_XSPI_FINISH */

	ret = vastai_pci_state_update_start(priv, die_id);
	if (ret) return ret;

	VASTAI_PCI_INFO(priv, die_id, "%s die%d (1/2) start.\n", __func__, die_id);
	/* write length of bmcu bin to csram */
	reg_val = VASTAI_BMCU_HEX_SIZE;
	ret = vastai_pci_mem_write(priv, vastai_pci_get_die_index(priv, die_id),
				(VASTAI_BMCU_DL_ADDR - 4), &reg_val, 4);

	/* download bmcu code to csram. */
	ret = vastai_process_download_request(priv, die_id,
					      path_name, bin_buf,
					      buf_size, VASTAI_BMCU_DL_ADDR);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id,
			       "%s die%d (2/2) download bmcu error, size=%ld, addr=%d\n",
			       __func__, die_id,
			       VASTAI_BMCU_HEX_SIZE,
			       VASTAI_BMCU_DL_ADDR);
		return ret;
	}
	VASTAI_PCI_INFO(priv, die_id, "%s die%d (2/2) download bmcu success.\n", __func__, die_id);

	/* send upgrade cmd to smcu */
	ret = vastai_send_pcie_cmd(priv, vastai_pci_get_die_index(priv, die_id),
				   VASTAI_PCIE_SUB_FLASH_BMCU, 0);
	return ret;
}

int vastai_pci_flash_bmcu(struct vastai_pci_info *priv, u32 die_id)
{
	int ret = 0;
	ret = vastai_pci_flash_bmcu_bin(priv, die_id, BMCU_PATH,
					vastai_bmcu_hex_buf, VASTAI_BMCU_HEX_SIZE);
	if (ret) {
		return ret;
	}
	ssleep(16);
	vastai_pci_get_fw_ver(priv, 0x4);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "flash BMCU by pcie finished!\n");
	return ret;
}

int vastai_pci_update_flag_xspi(struct vastai_pci_info *priv, u32 die_id,
				bool is_base_partition)
{
	int ret;

	/* send read xspi cmd to smcu */
	if (is_base_partition)
		ret = send_xspi_flash_cmd(priv, die_id,
			     VASTAI_PCIE_SUB_FLAG_BASE_XSPI,
			     BL0_BASE_TOTAL_SIZE);
	else
		ret = send_xspi_flash_cmd(priv, die_id,
			     VASTAI_PCIE_SUB_FLAG_XSPI,
			     BL0_APP_TOTAL_SIZE);
	if (ret) {

		VASTAI_PCI_ERR(priv, die_id, "%s send cmd failed%d\n",
				__func__, ret);
		return ret;
	}
	return ret;
}

int vastai_pci_flash_xspi_bin(struct vastai_pci_info *priv, u32 die_id,
			      u8 *path_name, const void *bin_buf,
			      size_t buf_size, bool is_base_partition)
{
	int ret = 0;

	/* vastai_pci_send_pcie_cmd need smcu to response. Therefore, to check
		pcie_state is VASTAI_NORMAL_STATE before turn it to
		VASTAI_UPDATE_STATE */
	/* This pci_state will turn to be VASTAI_UPDATE_NORMAL after flash xspi
		is done. For now, thie function is block, pci_state will turn
		to VASTAI_NORMAL_STATE after vastai_wait_rw_xspi */
	ret = vastai_pci_state_update_start(priv, die_id);
	if(ret) return ret;

	VASTAI_PCI_INFO(priv, die_id, "%s die%d (1/5) start.\n",
			__func__, die_id);

	/* download xspi code to csram. */
	ret = vastai_process_download_request(priv, die_id, path_name, bin_buf,
					      buf_size, VASTAI_XSPI_DL_ADDR);
	if (ret < 0) {
		VASTAI_PCI_ERR(
			priv, die_id,
			"%s die%d (2/5) download xspi error, size=%ld, addr=%d\n",
			__func__, die_id, buf_size,
			VASTAI_XSPI_DL_ADDR);
		return ret;
	}

	/* send flash xspi cmd to smcu */
	if (is_base_partition) {
		ret = send_xspi_flash_cmd(priv, die_id,
			     VASTAI_PCIE_SUB_FLASH_BASE_XSPI, BL0_BASE_TOTAL_SIZE);
	} else {
		ret = send_xspi_flash_cmd(priv, die_id,
			     VASTAI_PCIE_SUB_FLASH_XSPI, BL0_APP_TOTAL_SIZE);
	}
	if (ret) {
		VASTAI_PCI_ERR(priv, die_id,
			       "%s die%d (2/5) write flash cmd err=%d\n",
			       __func__, die_id, ret);
		return ret;
	}

#if 0
	atomic_set(&(priv->pci_state), VASTAI_NORMAL_STATE);
#else
	vastai_pci_state_update_done(priv, die_id);
#endif

	VASTAI_PCI_INFO(priv, die_id, "%s die%d (2/5) write successfully.\n",
			__func__, die_id);

	dmi_set_device(priv);

	return ret;
}

int vastai_pci_flash_xspi(struct vastai_pci_info *priv, u32 die_id,
			struct char_drv_info *dev, bool is_base_partition)
{
	int ret = 0;
	int i = 0;
	char *path_name;
	const char *buf;
	size_t buf_size;

	if (is_base_partition) {
		path_name = BL0_PATH;
		buf = vastai_bl0_hex_buf;
		buf_size = VASTAI_BL0_HEX_SIZE;
	} else {
		path_name = BL0_APP_PATH;
		buf = vastai_bl0_app_hex_buf;
		buf_size = VASTAI_BL0_APP_HEX_SIZE;
	}

	ret = vastai_pci_flash_xspi_bin(priv, die_id, path_name, buf,
					buf_size, is_base_partition);
	if (ret) {
		return ret;
	}

	/* read xspi */
	dev->valid_data_length = buf_size;
	ret = vastai_pci_read_xspi(priv, die_id, 0, dev->dm->vir,
					dev->valid_data_length,
					is_base_partition);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id,
			       "%s die%d (3/5) read xspi flash err=%d\n",
			       __func__, die_id, ret);
		return ret;
	}
	VASTAI_PCI_INFO(priv, die_id, "%s die%d (3/5) read successfully.\n",
			__func__, die_id);

	/* verify xspi */
	for (i = 0; i < buf_size; i++) {
		if (buf[i] !=
		    *((char *)(dev->dm->vir) + i)) {
			VASTAI_PCI_ERR(
				priv, die_id,
				"%s die%d (4/5) verify xspi flash errpos=0x%x\n",
				__func__, die_id, i);
			return ret;
		}
	}

	VASTAI_PCI_INFO(priv, die_id, "%s die%d (4/5) verify successfully.\n",
			__func__, die_id);

	/* update flag */
	ret = vastai_pci_update_flag_xspi(priv, die_id, is_base_partition);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id,
			       "%s die%d (5/5) update flag xspi flash err=%d\n",
			       __func__, die_id, ret);
		return ret;
	}
	VASTAI_PCI_INFO(priv, die_id, "%s die%d (5/5) update flag successfully, finished!\n",
			__func__, die_id);

	return ret;
}

int vastai_pci_update_flash_wp(struct vastai_pci_info *priv, u32 die_id,
			 struct char_drv_info *dev, u32 protect_cfg)
{
	int ret;

	if (protect_cfg == 0)
		vastai_send_pcie_cmd(priv, priv->dies[die_id].die_index,
			     VASTAI_PCIE_SUB_FLASH_WPDIS, 0);
	else
		vastai_send_pcie_cmd(priv, priv->dies[die_id].die_index,
			     VASTAI_PCIE_SUB_FLASH_WPEN, 0);

	ret = vastai_wait_rw_xspi(priv, die_id);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id, "%s die%d failed!\n", __func__,
			       die_id);
		return ret;
	}
	return ret;
}

int vastai_pci_set_flash_wp(struct vastai_pci_info *priv, u32 die_id, struct char_drv_info *dev, u32 cfg)
{
	int ret = 0;

	ret = vastai_pci_update_flash_wp(priv, die_id, dev, cfg);
	if (ret < 0) {
		if (cfg == 0)
			VASTAI_PCI_ERR(priv, die_id,
			    "%s die%d disable flash write protect err=%d\n",
			        __func__, die_id, ret);
		else
			VASTAI_PCI_ERR(priv, die_id,
			    "%s die%d enable flash write protect err=%d\n",
			        __func__, die_id, ret);
		return ret;
	}
	VASTAI_PCI_INFO(priv, die_id, "%s die%d update flash write protect successfully!\n",
			__func__, die_id);

	return ret;
}

int vastai_pci_boot_die(struct vastai_pci_info *priv, u8 die_id)
{
	int ret = 0, retry_cnt = 0;

	/* polling smcu.bootstage=1, smcu is ready for downloading bl1. */
	ret = vastai_pci_poll_smcu_stage(priv, die_id,
					 VASTAI_SMCU_BOOT_STAGE_BL1_READY);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id, "%s poll smcu stage[%d] error\n",
			       __func__, VASTAI_SMCU_BOOT_STAGE_BL1_READY);
		return ret;
	}

	ret = vastai_pci_set_current_die(priv, die_id);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id, "%s set current die error\n",
			       __func__);
		return ret;
	}

DOWNLOAD_AGAIN:
	if (retry_cnt > VASTAI_BOOT_RETRY_CNT_MAX)
		return -ETIMEDOUT;

	/* download bl1 code to csram. */
	ret = vastai_process_download_request(priv, die_id,
					      BL1_PATH, vastai_bl1_hex_buf,
					      VASTAI_BL1_HEX_SIZE, VASTAI_BL1_DL_ADDR);

	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id,
			       "%s download bl1 error, size=%ld, addr=%d\n",
			       __func__, VASTAI_BL1_HEX_SIZE,
			       VASTAI_BL1_DL_ADDR);
		return ret;
	}
	/* PCIe host driver set host.bootstage=1 and addr of bl1 run. */
	ret = vastai_pci_set_host_stage(priv, die_id,
					VASTAI_HOST_BOOT_STAGE_BL1_DONE);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id, "%s set host stage[%d] error\n",
			       __func__, VASTAI_HOST_BOOT_STAGE_BL1_DONE);
		return ret;
	}
	VASTAI_PCI_INFO(priv, die_id, "%s%d bl1 download finish!\n", __func__,
			die_id);
	/* polling smcu.bootstage=3, smcu is ready for downloading fw. */
	ret = vastai_pci_poll_smcu_stage(priv, die_id,
					 VASTAI_SMCU_BOOT_STAGE_FW_READY);
	if (ret < 0) {
		if (ret == -EAGAIN) {
			retry_cnt++;
			VASTAI_PCI_INFO(priv, die_id,
					"%s now reboot again, cnt=%d\n",
					__func__, retry_cnt);
			goto DOWNLOAD_AGAIN;
		} else {
			VASTAI_PCI_ERR(priv, die_id,
				       "%s poll smcu stage[%d] error\n",
				       __func__,
				       VASTAI_SMCU_BOOT_STAGE_FW_READY);
			return ret;
		}
	}
#ifdef CONFIG_VASTAI_PCI_LINK_MONITOR
	if (atomic_read(&priv->pci_state) == VASTAI_BOOT_STATE && die_id == 1 &&
		atomic_read(&(priv->pci_link_stat.is_reseting))) {
		memcpy(priv->dev->saved_config_space, priv->pci_link_stat.saved_config_space,
			sizeof(priv->pci_link_stat.saved_config_space));
		vastai_pci_config_write(priv, 4*8, (char *)&priv->pci_link_stat.saved_config_space[8], 4);
		vastai_pci_config_write(priv, 4*9, (char *)&priv->pci_link_stat.saved_config_space[9], 4);
		msleep(1);
	}
#endif
#ifdef CONFIG_VASTAI_PCI_ONLY_DOWNLOAD_BL1
	return ret;
#endif

	/* download all fw packets. */
	if (priv->ddr_bw_test) {
		ret = vastai_pci_download_fw(priv, die_id,
					     vastai_fw_ddrbw_hex_buf,
					     VASTAI_FW_DDRBW_HEX_SIZE,
					     VASTAI_FW_DL_ADDR);
		if (ret < 0) {
			VASTAI_PCI_ERR(
				priv, die_id,
				"%s download fw error, size=%ld, addr=%ld\n",
				__func__, VASTAI_FW_DDRBW_HEX_SIZE,
				VASTAI_FW_DL_ADDR);
			return ret;
		}
	} else {
		ret = vastai_process_download_request(priv, die_id,
						      FW_PATH, vastai_fw_hex_buf,
						      VASTAI_FW_HEX_SIZE, VASTAI_FW_DL_ADDR);

		if (ret < 0) {
			VASTAI_PCI_ERR(
				priv, die_id,
				"%s download fw error, size=%ld, addr=%ld\n",
				__func__, VASTAI_FW_HEX_SIZE,
				VASTAI_FW_DL_ADDR);
			return ret;
		}
	}

	/* PCIe host driver set host.bootstage=2 and addr of fw download. */
	ret = vastai_pci_set_host_stage(priv, die_id,
					VASTAI_HOST_BOOT_STAGE_FW_DONE);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id, "%s set host stage[%d] error\n",
			       __func__, VASTAI_HOST_BOOT_STAGE_FW_DONE);
		return ret;
	}
	/* polling smcu.bootstage=5, smcu download fw done. */
	ret = vastai_pci_poll_smcu_stage(priv, die_id,
					 VASTAI_SMCU_BOOT_STAGE_DIE_DONE);
	if (ret < 0) {
		if (ret == -EAGAIN) {
			retry_cnt++;
			VASTAI_PCI_INFO(priv, die_id,
					"%s now reboot again, cnt=%d\n",
					__func__, retry_cnt);
			goto DOWNLOAD_AGAIN;
		} else {
			VASTAI_PCI_ERR(priv, die_id,
				       "%s poll smcu stage[%d] error\n",
				       __func__,
				       VASTAI_SMCU_BOOT_STAGE_DIE_DONE);
			return ret;
		}
	}
	VASTAI_PCI_INFO(priv, die_id, "%s%d fw download finish!\n", __func__,
			die_id);

	/* polling smcu.bootstage=6, all mcu and dsp of current die init finish. */
	ret = vastai_pci_poll_smcu_stage(
		priv, die_id, VASTAI_SMCU_BOOT_STAGE_BOOT_DONE);
	if (ret < 0) {
		vastai_pci_check_core_sync(priv, die_id);
		VASTAI_PCI_ERR(priv, die_id,
			       "%s dev%d-die%d (%s) is not ready!\n",
			       __func__, priv->dev_id, die_id,
			       priv->dies[die_id].core_sync_result);
		return ret;
	}
	return ret;
}

int vastai_device_send_reset(struct vastai_pci_info *priv)
{
	int ret = 0;
	int die_id = 0;

	for (die_id = priv->die_num - 1; die_id >= 0; die_id--) {
		ret = vastai_send_pcie_cmd(priv,
				vastai_pci_get_die_index(priv, die_id),
				VASTAI_PCIE_SUB_RESET, 0);
		if (ret)
			return ret;
	}
	return ret;
}

int vastai_device_reset(struct vastai_pci_info *priv, u32 flags)
{
	int ret = 0;

	if (flags & VASTAI_RESET_MSG_TRIGER
		&& atomic_read(&priv->pci_state) != VASTAI_RESET_STATE) {
		u32 host_boot_stage;
		int count = 200;

#ifdef CONFIG_VASTAI_PCI_LINK_MONITOR
		vastai_enable_pcie_link_monitor(priv, false);
#endif

		/* reset cmd need smcu response. It only can work in
			VASTAI_NORMAL_STATE. Will try to reset when state is
			VASTAI_ERROR_STATE, but not sure will be success */
		/* if reset is success, pcie_state will turn to
			VASTAI_NO_RUNNING_STATE */
		if ((atomic_read(&priv->pci_state) != VASTAI_NORMAL_STATE)
			&& (atomic_read(&priv->pci_state) != VASTAI_ERROR_STATE)) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"%s: reset from %s is Unexpected\n", __func__,
				state_name(vastai_get_pci_state(priv)));
			ret = -EIO;
			goto failed;
		}
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"%s: triger device reset by msg\n", __func__);
		ret = vastai_device_send_reset(priv);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s: send reset msg error, %d\n",
				       __func__, ret);
		}
#if 0
		atomic_set(&priv->pci_state, VASTAI_RESET_STATE);
#else
		vastai_pci_state_reset_start(priv);
#endif
		ssleep(VASTAI_WAITING_RESET);
		do {
			msleep(100);
			vastai_pci_mem_read_direct(priv, priv->dies[0].die_index,
						    VASTAI_HOST_BOOT_STAGE_REG,
						    &host_boot_stage, 4);
		} while ((host_boot_stage != 0) && count--);

		if (count > 0) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "reset success\n");
#if 0
			atomic_set(&priv->pci_state, VASTAI_NO_RUN_STATE);
#else
			vastai_pci_state_reset_done(priv);
#endif
		} else {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "reset failed\n");
			ret = -EIO;
			goto failed;
		}
	}

	/* clear all completion and workqueue */
	/* TODO: kill all task */
	/* re-initial data */

	// vastai_pci_resource_recycling(priv);
	if (flags & VASTAI_RESET_BOOT_DEV) {
		int i;
		if (vastai_get_pci_state(priv) != VASTAI_NO_RUN_STATE) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s: current device state is %s, can't boot again\n",
				__func__, state_name(vastai_get_pci_state(priv)));
			return -EPERM;
		}
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s: reboot device\n",
				__func__);
#ifdef CONFIG_VASTAI_PCI_BOOT
		/* move to reset section */
		for (i = 0; i < VASTAI_PCI_BAR_NUM; i++) {
			int flag = pci_resource_flags(priv->dev, i);
			if (!(flag & IORESOURCE_MEM))
				continue;
			vastai_pci_init_bar_at(priv, i);
		}
		/* remove this */
		ret = vastai_pci_poll_smcu_stage(priv, VASTAI_DIE0,
					 VASTAI_SMCU_BOOT_STAGE_BL1_READY);
		if (ret) {
			VASTAI_PCI_ERR(priv, VASTAI_DIE0,
				       "%s poll stage[%d] error, %d\n", __func__,
				       VASTAI_SMCU_BOOT_STAGE_BL1_READY, ret);
			return -EIO;
		}
		ret = vastai_pci_boot(priv);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s: reboot device error, %d\n",
				       __func__, ret);
			goto failed;
		}
#endif
#ifdef CONFIG_VASTAI_PCI_LINK_MONITOR
		if (atomic_read(&priv->pci_state) == VASTAI_NORMAL_STATE &&
			priv->ddr_bw_test == 0)
			vastai_enable_pcie_link_monitor(priv, true);
#endif
	}

	return 0;

failed:
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s: disable device\n", __func__);

	return ret;
}

void vastai_pci_reinit(struct vastai_pci_info *pci_info)
{
	int ret;
	union smcu_boot_stage smcu_reg;
	int i;

	ret = vastai_pci_mem_read_direct(pci_info, -1, VASTAI_SMCU_BOOT_STAGE_REG,
					&(smcu_reg.val), 4);
	if (ret) {
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "%s:can not read smcu stage.\n",
			__FUNCTION__);
		return;
	}

	if (smcu_reg.bit.stage != VASTAI_SMCU_BOOT_STAGE_BOOT_DONE) {
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "%s:can not read smcu stage.\n",
			__FUNCTION__);
		return;
	}

	vastai_pci_state_boot_done(pci_info);

	pci_info->die_num = vastai_pci_read_die_num(pci_info);
	vastai_pci_init_all_die(pci_info);

	ret = vastai_device_reset(pci_info, VASTAI_RESET_MSG_TRIGER);
	if (ret) {
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "%s:reset failed.\n",
			__FUNCTION__);
		return;
	}
	for (i = 0; i < VASTAI_PCI_BAR_NUM; i++) {
		int flag = pci_resource_flags(pci_info->dev, i);
		if (!(flag & IORESOURCE_MEM))
			continue;
		vastai_pci_init_bar_at(pci_info, i);
	}
}

#ifdef CONFIG_TRAP_MSIX_ACCESS
int vastai_set_die_msix_table(struct vastai_pci_info *priv)
{
	u32 reg_val;
	u64 reg_addr;
	int i;
	int ret = 0;
	u32 die_index;
	u32 tmp_data = 0;
	unsigned long saved_atu_addr = priv->bar[VASTAI_PCI_BAR2].at_addr;

	for (i = VASTAI_DIE1; i < priv->die_num; i++) {
		die_index = VASTAI_GET_DIE_INDEX(priv, i);

		reg_addr = VASTAI_PCIE_EP_CONFIG_MSIX_CTRL;
		ret = vastai_pci_mem_read_direct(priv, die_index,
					  reg_addr, &reg_val, 4);
		reg_val |= 0x80000000;
		ret = vastai_pci_mem_write_direct(priv, die_index,
					   reg_addr, &reg_val, 4);
		ret = vastai_pci_mem_write_direct(priv, die_index, PCIE_MSIX_PF_TABLE,
			&priv->msix_table[VASTAI_PCIE_MSIX_NUM_PER_DIE * i],
			(VASTAI_PCIE_MSIX_NUM_PER_DIE * sizeof(struct msi_x_table)));
	}
	/*switch back bar2*/
	ret = vastai_pci_mem_bar2_read(priv, -1, saved_atu_addr, &tmp_data, 4);

	return ret;
}
#endif

int vastai_pci_boot(struct vastai_pci_info *priv)
{
	u8 i = 0, pre_die;
	int ret;

	VASTAI_PCI_FUNC_ENTERY;
	vastai_pci_reinit(priv);
#if 0
	ret = atomic_cmpxchg(&priv->pci_state, VASTAI_NO_RUN_STATE,
				VASTAI_BOOT_STATE);
	if (ret != VASTAI_NO_RUN_STATE) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"%s: boot device must from NO_RUN_STATE. %d\n",
			__FUNCTION__, ret);
		return -EIO;
	}
#else
	ret = vastai_pci_state_boot_start(priv);
	if (ret != VASTAI_STATE_TRANSIT_SUCCESS) return ret;
#endif

	/* boot die0 */
	ret = vastai_pci_boot_die(priv, VASTAI_DIE0);
	if (ret) {
		VASTAI_PCI_ERR(priv, VASTAI_DIE0,
			       "%s dev%d-die%d boot error\n",
			       __func__, priv->dev_id, VASTAI_DIE0);
		return ret;
	}
	//vastai_get_boot_mode(priv);

	vastai_die_init(&(priv->dies[VASTAI_DIE0]), priv, VASTAI_DIE0);
	/* get die number from die0 */
	priv->die_num = vastai_pci_read_die_num(priv);
	vastai_die_get_pcie_attr(&(priv->dies[VASTAI_DIE0]));
	VASTAI_PCI_INFO(priv, VASTAI_DIE0,
			"%s dev%d-die%d(%d/%d) GEN%dx%d boot finish!\n",
			__func__, priv->dev_id,
			VASTAI_DIE0, (VASTAI_DIE0 + 1), priv->die_num,
			priv->dies[VASTAI_DIE0].pcie_nls,
			priv->dies[VASTAI_DIE0].pcie_nlw);

#ifndef CONFIG_TRAP_MSIX_ACCESS
	if (priv->irq_type == VASTAI_MSIX_IRQ)
		vastai_pci_refresh_msix_table(priv);
#endif

	/* boot other dies */
	for (i = VASTAI_DIE1; i < priv->die_num; i++) {
		pre_die = i - 1;

		ret = vastai_pci_poll_die_linkup(priv, pre_die);
		if (ret < 0)
			return ret;

		/* set master die bar2 atu to 0x10_0800_0000, 0x20_0800_0000, 0x30_0800_0000 */
		ret = vastai_send_pcie_cmd_spe(priv,
				vastai_pci_get_die_index(priv, 0),
				VASTAI_PCIE_SUB_BAR2_ATU, VASTAI_PCIE_EP_IO_SPACE_ADDR(i));
		ret = vastai_send_pcie_cmd_spe(priv,
				vastai_pci_get_die_index(priv, 0),
				VASTAI_PCIE_SUB_BAR4_ATU, VASTAI_DDR_FW_DL_ADDR(i));
		ret = vastai_pci_poll_host_stage(
			priv, VASTAI_DIE0, VASTAI_HOST_BOOT_STAGE_INIT);
		if (ret < 0) {
			VASTAI_PCI_ERR(
				priv, DUMMY_DIE_ID,
				"%s set master die bar2 atu fail when boot die%d!\n",
				__func__, i);
			return ret;
		}
		if(priv->bar[VASTAI_PCI_BAR2].minish_bar_size == 1)
			priv->bar[VASTAI_PCI_BAR2].at_addr =
				VASTAI_PCIE_EP_IO_SPACE_ADDR(i);
		if(priv->bar[VASTAI_PCI_BAR4].minish_bar_size == 1)
			priv->bar[VASTAI_PCI_BAR4].at_addr =
				VASTAI_DDR_FW_DL_ADDR(i);

		ret = vastai_pci_die_enumeration(priv, i);
		if (ret) {
			VASTAI_PCI_ERR(priv, i,
				       "%s dev%d-die%d enumeration error\n",
				       __func__, priv->dev_id, i);
			return ret;
		}
		ret = vastai_pci_boot_die(priv, i);
		if (ret) {
			VASTAI_PCI_ERR(priv, i,
				       "%s dev%d-die%d boot error\n",
				       __func__, priv->dev_id, i);
			return ret;
		}
		vastai_die_init(&(priv->dies[i]), priv, i);
		vastai_die_get_pcie_attr(&(priv->dies[i]));
		VASTAI_PCI_INFO(priv, i,
				"%s dev%d-die%d(%d/%d) GEN%dx%d boot finish!\n",
				__func__, priv->dev_id,
				i, (i + 1), priv->die_num,
				priv->dies[i].pcie_nls,
				priv->dies[i].pcie_nlw);

		VASTAI_PCI_INFO(priv, i,
				"%s dev%d-die%d(%d/%d) boot finish!\n",
				__func__, priv->dev_id,
				i, (i + 1), priv->die_num);

		/* set slave die atu */
		vastai_pci_set_host_stage(
			priv, i, VASTAI_HOST_BOOT_STAGE_SET_BAR_PHASE1);
		msleep(1);
		vastai_pci_die_enumeration_bh(priv, i);
		ret = vastai_pci_poll_host_stage(
			priv, i, VASTAI_HOST_BOOT_STAGE_INIT);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, i,
				       "%s set slave die atu fail!\n",
				       __func__);
			return ret;
		}
	}

	/* init software manage source */

	/* boot finish, set host boot stage */
	for (i = VASTAI_DIE0; i < priv->die_num; i++)
		vastai_pci_set_host_stage(priv, i,
					  VASTAI_HOST_BOOT_STAGE_READY);

	/* After smcu poll host satge 4, smcu will re-set bar2/bar4 atu. */
	if(priv->bar[VASTAI_PCI_BAR2].minish_bar_size)
		priv->bar[VASTAI_PCI_BAR2].at_addr =
			VASTAI_PCIE_BAR2_AXI_ADDR_32MB_2;
	if(priv->bar[VASTAI_PCI_BAR4].minish_bar_size)
		priv->bar[VASTAI_PCI_BAR4].at_addr =
			VASTAI_PCIE_BAR4_AXI_ADDR_128MB_2;

#if 0
	atomic_set(&(priv->pci_state), VASTAI_NORMAL_STATE);
#else
	vastai_pci_state_boot_done(priv);
#endif
#ifdef CONFIG_TRAP_MSIX_ACCESS
	/* To switch BAR 2/3, sw resources need to be ready */
	if (priv->irq_type == VASTAI_MSIX_IRQ) {
		msleep(2); // smcu may not reset bar2 and bar4's addr immediately, wait for while
		vastai_pci_refresh_msix_table(priv);
		vastai_set_die_msix_table(priv);
	}

#endif
#ifndef CONFIG_VASTAI_AI_EMU
	ret = avfg_set_die_avfs(priv);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s:set die avfs is fail %d\n",
			       pci_name(priv->dev), ret);
	}
#endif
	VASTAI_PCI_FUNC_EXIT;
	return 0;
}

void vastai_pci_boot_wq(struct work_struct *work)
{
	struct delayed_work *delayed_work = to_delayed_work(work);
	struct vastai_pci_info *priv =
		container_of(delayed_work, struct vastai_pci_info, boot_wq);
	vastai_pci_boot(priv);
}
