/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/aer.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/version.h>
#include <linux/rtc.h>

#define CONFIG_VASTAI_PCI_PERFORMANCE 0

#include <linux/timer.h>

#include "vastai_pci.h"
#include "vastai_pci_api.h"
#ifdef CONFIG_VASTAI_PCI_BOOT
#include "vastai_pci_boot.h"
#endif
#include "vastai_pci_test.h"
#include "vastai_udma_engine.h"
#include "vastai_version.h"
#include "vastai_msgq.h"

#ifdef CONFIG_VASTAI_AI
#include "vastai_ai.h"
#endif

#ifdef CONFIG_VASTAI_LOGSYS
#include "vastai_logsys.h"
#endif

#ifdef CONFIG_VASTAI_TOOLS_SUITE
#include "vatools_driver.h"
#endif

#ifdef CONFIG_VASTAI_VIDEO
#include "vastai.h"
#include "vastai_enc.h"
#include "vastai_render.h"

#endif

#include <linux/fs.h>
#include <linux/uaccess.h>
#include "vastai_dmabuf.h"
#include "vastai_dmi_table.h"
#include "vastai_fifo.h"
#include "vastai_die.h"
#include "vastai_avfs.h"
#include "vastai_state.h"
#include "vatools.h"

#ifdef CONFIG_VASTAI_KUNIT
#include "stub.h"
#endif

#define VASTAI_PCIE_2_SMCU 0
#define VASTAI_PCIE_2_VDSP 1
#define VASTAI_PCIE_2_CMCU 1

#define VASTAI_HEARTBEAT_ENABLE 1

static bool msi_enabled = 1;

module_param(msi_enabled, bool, 0644);


static int vf_num = 0;
module_param(vf_num, int, 0);


unsigned char vastai_pci_log_level =
	VASTAI_PCI_NORMAL_LEVEL | VASTAI_PCI_DATA_LEVEL | VASTAI_PCI_PERF_LEVEL;
module_param(vastai_pci_log_level, byte, 0644);
MODULE_PARM_DESC(vastai_pci_log_level,
		 "Vastai's log level, please select in 0xf/0xd/0x0 ");

bool vastai_pci_reset_out = 1;
module_param(vastai_pci_reset_out, bool, 0644);
MODULE_PARM_DESC(vastai_pci_reset_out, "Vastai Exit with reset, bool ");

static bool core_exception_en = 1;
module_param(core_exception_en, bool, 0644);
MODULE_PARM_DESC(core_exception_en , "Vastai core exeption enable ");

#ifdef CONFIG_VASTAI_KUNIT
struct list_head vastai_pci_info_list;
#else
static struct list_head vastai_pci_info_list;
#endif

u32 ai_group[] = {HEARTBEAT_CMCU, HEARTBEAT_LMCU0, HEARTBEAT_LMCU1, HEARTBEAT_LMCU2,
		  HEARTBEAT_LMCU3, HEARTBEAT_LMCU4, HEARTBEAT_LMCU5, HEARTBEAT_LMCU6,
		  HEARTBEAT_LMCU7, HEARTBEAT_ODSP0, HEARTBEAT_ODSP1, HEARTBEAT_ODSP2,
		  HEARTBEAT_ODSP3, HEARTBEAT_ODSP4, HEARTBEAT_ODSP5, HEARTBEAT_ODSP6,
		  HEARTBEAT_ODSP7};

vastai_reset_handle g_reset_ai_handle = NULL;

static struct msg_addr_entry {
	u64 interrupt_addr;
	u64 msg_addr;
	u32 is_use_ring_buf;
} msg_addr_table[32] = {
#ifdef CONFIG_VASTAI_CMCU_RINGBUF
	{CMCU_GRP12_WR_REG, MSG_PCIE_TO_CMCU, 1}, /*cmcu 0*/
#else
	{ CMCU_GRP12_WR_REG, MSG_PCIE_TO_CMCU, 0 }, /*cmcu 0*/
#endif
	{ 0, 0, 0 }, /*RSVD _1*/
	{ 0, 0, 0 }, /*RSVD _2*/
	{ 0, 0, 0 }, /*RSVD _3*/
	{ 0, 0, 0 }, /*RSVD _4*/
	{ 0, 0, 0 }, /*RSVD _5*/
	{ 0, 0, 0 }, /*RSVD _6*/
	{ 0, 0, 0 }, /*RSVD _7*/

	{ VDSP0_GRP5_WR_REG, MSG_PCIE_TO_VDSP0, 1 }, /*vdsp 0*/
	{ VDSP1_GRP5_WR_REG, MSG_PCIE_TO_VDSP1, 1 }, /*vdsp 1*/
	{ VDSP2_GRP5_WR_REG, MSG_PCIE_TO_VDSP2, 1 }, /*vdsp 2*/
	{ VDSP3_GRP5_WR_REG, MSG_PCIE_TO_VDSP3, 1 }, /*vdsp 3*/
	{ 0, 0, 0 }, /*RSVD _4*/
	{ 0, 0, 0 }, /*RSVD _5*/
	{ 0, 0, 0 }, /*RSVD _6*/
	{ 0, 0, 0 }, /*RSVD _7*/

	{ 0, 0, 0 }, /*RSVD _0*/
	{ 0, 0, 0 }, /*RSVD _1*/
	{ 0, 0, 0 }, /*RSVD _2*/
	{ 0, 0, 0 }, /*RSVD _3*/
	{ 0, 0, 0 }, /*RSVD _4*/
	{ 0, 0, 0 }, /*RSVD _5*/
	{ 0, 0, 0 }, /*RSVD _6*/
	{ 0, 0, 0 }, /*RSVD _7*/

	{ VDMCU0_GRP5_WR_REG, MSG_PCIE_TO_VDMCU0, 0 }, /*vdmcu 0*/
	{ VDMCU1_GRP4_WR_REG, MSG_PCIE_TO_VDMCU1, 0 }, /*vdmcu 1*/
	{ VDMCU2_GRP5_WR_REG, MSG_PCIE_TO_VDMCU2, 0 }, /*vdmcu 2*/
	{ 0, 0, 0 }, /*RSVD _3*/
	{ 0, 0, 0 }, /*RSVD _4*/
	{ 0, 0, 0 }, /*RSVD _5*/
	{ 0, 0, 0 }, /*RSVD _6*/
	{ 0, 0, 0 }, /*RSVD _7*/
};

char* core_name(int core_id)
{
	switch(core_id)
	{
		case    HEARTBEAT_SMCU_THREAD : return "HEARTBEAT_SMCU_THREAD";
		case	HEARTBEAT_SMCU        : return "HEARTBEAT_SMCU";
		case	HEARTBEAT_CMCU        : return "HEARTBEAT_CMCU";
		case	HEARTBEAT_LMCU0       : return "HEARTBEAT_LMCU0";
		case	HEARTBEAT_LMCU1       : return "HEARTBEAT_LMCU1";
		case	HEARTBEAT_LMCU2       : return "HEARTBEAT_LMCU2";
		case	HEARTBEAT_LMCU3       : return "HEARTBEAT_LMCU3";
		case	HEARTBEAT_LMCU4       : return "HEARTBEAT_LMCU4";
		case	HEARTBEAT_LMCU5       : return "HEARTBEAT_LMCU5";
		case	HEARTBEAT_LMCU6       : return "HEARTBEAT_LMCU6";
		case	HEARTBEAT_LMCU7       : return "HEARTBEAT_LMCU7";
		case	HEARTBEAT_ODSP0       : return "HEARTBEAT_ODSP0";
		case	HEARTBEAT_ODSP1       : return "HEARTBEAT_ODSP1";
		case	HEARTBEAT_ODSP2       : return "HEARTBEAT_ODSP2";
		case	HEARTBEAT_ODSP3       : return "HEARTBEAT_ODSP3";
		case	HEARTBEAT_ODSP4       : return "HEARTBEAT_ODSP4";
		case	HEARTBEAT_ODSP5       : return "HEARTBEAT_ODSP5";
		case	HEARTBEAT_ODSP6       : return "HEARTBEAT_ODSP6";
		case	HEARTBEAT_ODSP7       : return "HEARTBEAT_ODSP7";
		case	HEARTBEAT_VDMCU0      : return "HEARTBEAT_VDMCU0";
		case	HEARTBEAT_VDMCU1      : return "HEARTBEAT_VDMCU1";
		case	HEARTBEAT_VDMCU2      : return "HEARTBEAT_VDMCU2";
		case	HEARTBEAT_VEMCU0      : return "HEARTBEAT_VEMCU0";
		case	HEARTBEAT_VEMCU1      : return "HEARTBEAT_VEMCU1";
		case	HEARTBEAT_VEMCU2      : return "HEARTBEAT_VEMCU2";
		case	HEARTBEAT_VEMCU3      : return "HEARTBEAT_VEMCU3";
		case	HEARTBEAT_VDSP0       : return "HEARTBEAT_VDSP0";
		case	HEARTBEAT_VDSP1       : return "HEARTBEAT_VDSP1";
		case	HEARTBEAT_VDSP2       : return "HEARTBEAT_VDSP2";
		case	HEARTBEAT_VDSP3       : return "HEARTBEAT_VDSP3";
		case	HEARTBEAT_RESERVE0    : return "HEARTBEAT_RESERVE0";
		case	HEARTBEAT_RESERVE1    : return "HEARTBEAT_RESERVE1";
		default: break;
	}

	return "Wrong heartbeat input.";
}

struct msg_addr_entry  *get_msg_addr_entry(void)
{
	return msg_addr_table;
}
int get_vastai_die_info(struct vastai_pci_info *priv, int die_index[],
			int *die_num)
{
	int i = 0;

	if (!die_index || !priv) {
		return -EINVAL;
	}
	for (i = 0; i < priv->die_num; i++) {
		die_index[i] = priv->dies[i].die_index;
	}
	*die_num = priv->die_num;
	return 0;
}


int vastai_pci_get_sriov_numvfs(char dev_id)
{
	int vf_num = 0;
	struct vastai_pci_info *pcie_dev = NULL;
	pcie_dev = get_vastai_pci_device_info(dev_id);
	if (pcie_dev != NULL) {
		if (pcie_dev->is_physfn) {
			vf_num = pci_num_vf(pcie_dev->dev);
			VASTAI_PCI_DBG(pcie_dev, DUMMY_DIE_ID,
					"%s Get sriov_numvfs: %d\n", __func__, vf_num);
			return vf_num;
		} else {
			VASTAI_PCI_DBG(pcie_dev, DUMMY_DIE_ID,
					"%s Please go back to PF driver to query again!\n", __func__);
			return 0;
		}
	} else {
		VASTAI_PCI_DBG(pcie_dev, DUMMY_DIE_ID,
				"%s Get vastai_pci_info faile!\n", __func__);
		return 0;
	}
}


struct vastai_pci_info *get_vastai_pci_device_info(char dev_id)
{
	struct vastai_pci_info *pci_info = NULL;
	if (list_empty(&vastai_pci_info_list)) {
		return NULL;
	}

	/* for now, we only get first entry */
	list_for_each_entry(pci_info, &vastai_pci_info_list, dev_list) {
		if (pci_info->dev_id == dev_id) {
			return pci_info;
		}
	}
	return NULL;
}

u64 vastai_get_host_time_ns(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,6,0)
	struct timespec time;
#else
	u64 time;
#endif
	u64 time_ns;


#if LINUX_VERSION_CODE < KERNEL_VERSION(5,6,0)
	/* getnstimeofday function is removed from kernel5.6 */
	getnstimeofday(&time);
#else
	time = ktime_get_real_ns();
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,6,0)
	time_ns = timespec_to_ns(&time);
#else
	time_ns = time;
#endif

	return time_ns;
}

int get_vastai_pci_device_number(void)
{
	struct list_head *ptr = NULL;
	int device_num = 0;

	if (list_empty(&vastai_pci_info_list)) {
		return 0;
	}

	list_for_each(ptr, &vastai_pci_info_list) {
		device_num++;
	}
	return device_num;
}

static unsigned char *vastai_pci_bar_vmem(struct vastai_pci_info *priv, int bar)
{
	if (!priv) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "vastai pcie_dev is NULL\n");
		return NULL;
	}

	return priv->bar[bar].vmem;
}

int vastai_pci_config_read(struct vastai_pci_info *priv, int offset, char *buf,
			   int len)
{
	int i = 0, ret = 0;

	if ((!priv) || (!priv->dev)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s priv 0x%p or dev 0x%p is NULL\n", __func__,
			       priv, priv->dev);
		return -ENXIO;
	}
	for (i = 0; i < len; i++) {
		ret = pci_read_config_byte(priv->dev, i + offset, &(buf[i]));
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s err=%d\n",
				       __func__, ret);
			return -1;
		}
	}
	return 0;
}

int vastai_pci_config_write(struct vastai_pci_info *priv, int offset, char *buf,
			    int len)
{
	int i = 0, ret = 0;

	if ((!priv) || (!priv->dev)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s priv 0x%p or dev 0x%p is NULL\n", __func__,
			       priv, priv->dev);
		return -ENXIO;
	}
	for (i = 0; i < len; i++) {
		ret = pci_write_config_byte(priv->dev, i + offset, buf[i]);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s err=%d\n",
				       __func__, ret);
			return -1;
		}
	}
	return 0;
}

void vastai_pci_init_bar_at(struct vastai_pci_info *priv, int bar)
{
	if (priv->is_virtfn)
		return vastai_pci_config_vf_bar_at(priv, bar);

	switch (bar) {
	case VASTAI_PCI_BAR0:
		if (priv->bar[VASTAI_PCI_BAR0].mmio_len == (256 * 1024 * 1024))
			priv->bar[VASTAI_PCI_BAR0].at_addr =
				VASTAI_PCIE_BAR0_AXI_ADDR_256MB;
		else if (priv->bar[VASTAI_PCI_BAR0].mmio_len ==
			 (32 * 1024 * 1024))
			priv->bar[VASTAI_PCI_BAR0].at_addr =
				VASTAI_PCIE_BAR0_AXI_ADDR_32MB;
		else if (priv->bar[VASTAI_PCI_BAR0].mmio_len == (2 * 1024 * 1024)) {
			priv->bar[VASTAI_PCI_BAR0].at_addr =
				VASTAI_PCIE_BAR0_AXI_ADDR_2MB;
			priv->bar[VASTAI_PCI_BAR0].minish_bar_size = 1;
		}
		else {
			priv->bar[VASTAI_PCI_BAR0].at_addr =
				VASTAI_PCIE_BAR0_AXI_ADDR_8MB;
			priv->bar[VASTAI_PCI_BAR0].minish_bar_size = 1;
		}
		break;
	case VASTAI_PCI_BAR1:
		if (priv->bar[VASTAI_PCI_BAR1].mmio_len == (32 * 1024 * 1024))
			priv->bar[VASTAI_PCI_BAR1].at_addr =
				VASTAI_PCIE_BAR1_AXI_ADDR_32MB;
		else if (priv->bar[VASTAI_PCI_BAR1].mmio_len == (4 * 1024 * 1024)) {
			priv->bar[VASTAI_PCI_BAR1].at_addr =
				VASTAI_PCIE_BAR1_AXI_ADDR_4MB;
			priv->bar[VASTAI_PCI_BAR1].minish_bar_size = 1;
		}
		else {
			priv->bar[VASTAI_PCI_BAR1].at_addr =
				VASTAI_PCIE_BAR1_AXI_ADDR_8MB;
			priv->bar[VASTAI_PCI_BAR1].minish_bar_size = 1;
		}
		priv->bar[VASTAI_PCI_BAR1].map_in_addr = VASTAI_PCIE_BAR1_BASEX_IN;
		priv->bar[VASTAI_PCI_BAR1].map_out_addr = VASTAI_PCIE_BAR1_BASEX_OUT;
		break;
	case VASTAI_PCI_BAR2:
		if (priv->bar[VASTAI_PCI_BAR2].mmio_len == (32 * 1024 * 1024)) {
			priv->bar[VASTAI_PCI_BAR2].at_addr =
				VASTAI_PCIE_BAR2_AXI_ADDR_32MB_1;
			priv->bar[VASTAI_PCI_BAR2].minish_bar_size = 1;
		}
		else
			priv->bar[VASTAI_PCI_BAR2].at_addr =
				VASTAI_PCIE_BAR2_AXI_ADDR_32GB;
		break;
	case VASTAI_PCI_BAR4:
		if (priv->bar[VASTAI_PCI_BAR4].mmio_len == (128 * 1024 * 1024)) {
			priv->bar[VASTAI_PCI_BAR4].at_addr =
				VASTAI_PCIE_BAR4_AXI_ADDR_128MB_1;
			priv->bar[VASTAI_PCI_BAR4].minish_bar_size = 1;
		}
		else
			priv->bar[VASTAI_PCI_BAR4].at_addr =
				VASTAI_PCIE_BAR4_AXI_ADDR_32GB;
		priv->bar[VASTAI_PCI_BAR4].map_in_addr = VASTAI_PCIE_BAR4_BASEX_IN;
		priv->bar[VASTAI_PCI_BAR4].map_out_addr = VASTAI_PCIE_BAR4_BASEX_OUT;
		break;
	default:
		break;
	}
}

/* generate a die_index */
u32 vastai_pci_gen_die_index(void *pci_info, u32 die_id)
{
	union die_index_data die_index_temp = { 0 };

	die_index_temp.dev_id = ((struct vastai_pci_info *)pci_info)->dev_id;
	die_index_temp.die_id = die_id;
	die_index_temp.seq_num = vastai_get_seqnum(pci_info, die_id);

	return die_index_temp.val;
}

/* check a die_index */
u32 vastai_pci_get_die_index(void *pci_info, u32 die_id)
{
	struct vastai_pci_info *__pci_info = pci_info;

	if (die_id >= VASTAI_SV100_MAX_DIE_NUM)
		return UINT_MAX;
	if (atomic_read(&__pci_info->pci_state) != VASTAI_NORMAL_STATE
	    || !__pci_info->dies) {
		return vastai_pci_gen_die_index(pci_info, die_id);
	}

	return vastai_die_get_die_index(&(__pci_info->dies[die_id]));

}

#ifdef CONFIG_VASTAI_PCI_SET_BAR
static int vastai_pci_set_bar_at(struct vastai_pci_info *priv, int bar,
				 u64 addr, int die_index)
{
	int ret = 0;
	u64 reg_addr = 0;
	u32 val_l = addr & 0xFFFFFFFF, val_h = addr >> 32, cnt = 0;

	if (priv->bar[VASTAI_PCI_BAR0].minish_bar_size == 0) {
		/* set bar by host */
		reg_addr =
			VASTAI_PCI_CDN_FUNC0_BAR0_EP_INBOUND_ADDR_TRANS_REG0 +
			bar * 0x8;
		ret = vastai_pci_mem_write_direct(priv, die_index, reg_addr, &val_l,
					   4);
		ret = vastai_pci_mem_write_direct(priv, die_index, reg_addr + 4,
					   &val_h, 4);
		while (1) {
			/* read check to ensure atu configuration is working. */
			vastai_pci_mem_read_direct(priv, die_index, reg_addr, &val_l,
					    4);
			vastai_pci_mem_read_direct(priv, die_index, reg_addr + 4,
					    &val_h, 4);
			if ((val_l == (addr & 0xFFFFFFFF)) &&
			    (val_h == (addr >> 32))) {
				priv->bar[bar].at_addr = addr;
				break;
			}
			mdelay(1);
			cnt++;
			if (cnt > 10) {
				ret = -EIO;
				VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
					       "%s err=%d\n", __func__, ret);
				return ret;
			}
		}
	} else {
		/* send set bar at cmd to smcu */
		ret = vastai_send_pcie_cmd(priv, priv->dies[0].die_index,
					   VASTAI_PCIE_SUB_SET_BAR, addr);
		if (ret < 0)
			return ret;
		/* wait completion int from smcu */
		if (wait_for_completion_timeout(
			    &(priv->set_bar_done),
			    msecs_to_jiffies(
				    VASTAI_PCIE_WAIT_SMCU_MSG1_TIME)) <= 0) {
			VASTAI_PCI_ERR(
				priv, die_index,
				"%s BAR(%d) at_addr=0x%lx to addr=0x%llx timeout!\n",
				__func__, bar, priv->bar[bar].at_addr, addr);
			return -ETIMEDOUT;
		}
		reinit_completion(&(priv->set_bar_done));
		priv->bar[bar].at_addr = addr;
	}

	return ret;
}

/* only bar2 need be reset, becase only bar2 will be skip */
void vastai_pci_reset_bar_at(struct vastai_pci_info *priv, int die_index)
{
	vastai_pci_set_bar_at(priv, VASTAI_PCI_BAR2,
				VASTAI_PCIE_BAR2_AXI_ADDR_32MB_2, die_index);
}

#endif

static int vastai_pci_select_bar(struct vastai_pci_info *priv, u64 addr)
{
	if ((addr >= priv->bar[VASTAI_PCI_BAR0].at_addr) &&
	    (addr < priv->bar[VASTAI_PCI_BAR0].at_addr +
			    priv->bar[VASTAI_PCI_BAR0].mmio_len)) {
		return VASTAI_PCI_BAR0;
	} else if ((addr >= priv->bar[VASTAI_PCI_BAR1].at_addr) &&
		   (addr < priv->bar[VASTAI_PCI_BAR1].at_addr +
				   priv->bar[VASTAI_PCI_BAR1].mmio_len)) {
		return VASTAI_PCI_BAR1;
	} else if ((addr >= priv->bar[VASTAI_PCI_BAR4].at_addr) &&
		   (addr < priv->bar[VASTAI_PCI_BAR4].at_addr +
				   priv->bar[VASTAI_PCI_BAR4].mmio_len)) {
		return VASTAI_PCI_BAR4;
	} else {
		/* using bar2 to switch bar dynamically  */
		return VASTAI_PCI_BAR2;
	}
}

static bool is_in_map_range(u64 addr, u64 start_addr, u64 size)
{
	if((addr >= start_addr) && (addr < (start_addr+size)))
		return true;
	else
		return false;
}

/*
	bar1:             map in addr                  :          map out addr

	die 1: 0x10_08B0_0000 ~ (0x10_08B0_0000 + 2M)  :  0x10_0A00_0000       ~ (0x10_0A00_0000 + 2M)
	die 2: 0x20_08B0_0000 ~ (0x20_08B0_0000 + 2M)  : (0x10_0A00_0000 + 2M) ~ (0x10_0A00_0000 + 4M)
	die 3: 0x30_08B0_0000 ~ (0x20_08B0_0000 + 2M)  : (0x10_0A00_0000 + 4M) ~ (0x10_0A00_0000 + 6M)

	bar4:             map in addr                  :          map out addr

	die 1: 0x18_1400_0000 ~ (0x18_1400_0000 + 32M) : 0x14_0000_0000         ~ (0x14_0000_0000 + 32M)
	die 2: 0x28_1400_0000 ~ (0x28_1400_0000 + 32M) : (0x14_0000_0000 + 32M) ~ (0x14_0000_0000 + 64M)
	die 3: 0x38_1400_0000 ~ (0x38_1400_0000 + 32M) : (0x14_0000_0000 + 64M) ~ (0x14_0000_0000 + 96M)
*/
static void vastai_pci_bar_trans_addr(struct vastai_pci_info *priv, int die_id, u32 bar_id,
						u64 *addr)
{
	u64 map_in_start_addr = priv->bar[bar_id].map_in_addr;
	u64 map_out_start_addr = priv->bar[bar_id].map_out_addr;
	u64 range_len = priv->bar[bar_id].mmio_len/VASTAI_DIE_MAX_NUM;

	if (!priv->bar[bar_id].minish_bar_size)
		return;

	if(is_in_map_range(*addr, map_in_start_addr, range_len)) {
		u64 mapBaseAddr = map_out_start_addr + range_len * (die_id - 1);
		u64 offset = *addr - map_in_start_addr;
		*addr = mapBaseAddr + offset;
	}
}

static u64 vastai_pci_cal_absolute_addr(struct vastai_pci_info *priv, int die_id,
					u64 rel_addr, u64 abs_addr)
{
	u64 addr = rel_addr;

	if(die_id == 0)
		return rel_addr;

	vastai_pci_bar_trans_addr(priv, die_id, VASTAI_PCI_BAR1, &addr);
	vastai_pci_bar_trans_addr(priv, die_id, VASTAI_PCI_BAR4, &addr);

	if (rel_addr != addr)
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
			       "%s: addr_in:0x%llx addr_out:0x%llx\n", __func__,
			       abs_addr, addr);
	else {
		return abs_addr;
	}

	return addr;
}

static int vastai_pci_bar_read(struct vastai_pci_info *priv, int bar,
			       u64 offset, void *buf, unsigned int len)
{
	volatile unsigned char *mem = vastai_pci_bar_vmem(priv, bar);
	unsigned int i = 0;

	if (unlikely(!mem)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s: bar%d vmem is NULL, return\n", __func__,
			       bar);
		return -ENXIO;
	}
	if (unlikely(!buf)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s: buff is NULL, return\n",
			       __func__);
		return -ENOMEM;
	}
	mem += offset;

	if (unlikely(((((u64)mem) % 4) != 0) && (len == 4))) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s: read not align offset 0x%llx len 0x%x\n",
			       __func__, offset, len);
		return -EINVAL;
	}

	/* Align any unaligned source IO */
	if (unlikely(3 & (unsigned long)mem)) {
		u32 temp_len = ((u64)mem) & 3;
		memcpy_fromio(buf, mem, temp_len);
		i += temp_len;
	}

	/* rep copy */
	while ((len - i) >= 4) {
		#ifndef CONFIG_VASTAI_KUNIT
		*((unsigned int *)((u64)buf + i)) = readl(mem + i);
		#endif
		i += sizeof(unsigned int);
	}

	/* Align any unaligned source IO len */
	if (unlikely(3 & (len - i))) {
		u32 temp_len = (len - i) & 3;
		memcpy_fromio(buf + i, mem + i, temp_len);
		i += temp_len;
	}

	BUG_ON((len - i) != 0);
	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
		       "%s(bar=%d, offset=0x%llx, val=0x%x)\n", __func__, bar,
		       offset, *((int *)buf));

	return 0;
}

static int vastai_pci_bar_write(struct vastai_pci_info *priv, int bar,
				u64 offset, const void *buf, unsigned int len)
{
	unsigned char *mem = vastai_pci_bar_vmem(priv, bar);
	unsigned int i = 0;

	if (unlikely(!mem)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s: bar%d vmem is NULL, return\n", __func__,
			       bar);
		return -ENXIO;
	}
	if (unlikely(!buf)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s: buff is NULL, return\n",
			       __func__);
		return -ENOMEM;
	}
	mem += offset;

	if (unlikely(((((u64)mem) % 4) != 0) && (len == 4))) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s: read not align offset 0x%llx len 0x%x\n",
			       __func__, offset, len);
		return -EINVAL;
	}

	/* Align any unaligned source IO */
	if (unlikely(3 & (unsigned long)mem)) {
		u32 temp_len = ((u64)mem) & 3;
		memcpy_toio(mem, buf, temp_len);
		i += temp_len;
	}

	/* rep copy */
	while ((len - i) >= 4) {
		#ifndef CONFIG_VASTAI_KUNIT
		writel(*((unsigned int *)((u64)buf + i)), (mem + i));
		#endif
		i += sizeof(unsigned int);
	}

	/* Align any unaligned source IO len */
	if (unlikely(3 & (len - i))) {
		u32 temp_len = (len - i) & 3;
		memcpy_toio(mem + i, buf + i, temp_len);
		i += temp_len;
	}

	BUG_ON((len - i) != 0);
	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
		       "%s(bar=%d, offset=0x%llx, val=0x%x)\n", __func__, bar,
		       offset, *((int *)buf));

	return 0;
}

int vastai_pci_move_bar_window(struct vastai_pci_info *priv, int die_index, int bar, u64 absolute_addr)
{
	int ret = 0;
	u64 addr = 0;

	addr = absolute_addr / priv->bar[bar].mmio_len *
	       priv->bar[bar].mmio_len;
	if (addr != priv->bar[bar].at_addr) {
#ifdef CONFIG_VASTAI_PCI_SET_BAR
#ifdef CONFIG_SRIOV_ENABLE
		if (priv->is_virtfn) {
			VASTAI_PCI_ERR(
			       priv, die_index,
			       "%s [dangerous] 0x%llx, set BAR(%d) at_addr=0x%lx to addr=0x%llx!\n",
			       __func__, absolute_addr, bar,
			       priv->bar[bar].at_addr, addr);
			dump_stack();
			return -1;
		}
#endif
		ret = vastai_pci_set_bar_at(priv, bar, addr,
					    priv->dies[0].die_index);
		if (ret < 0) {
			return ret;
		}
		VASTAI_PCI_DBG(
			priv, die_index,
			"%s [dangerous] 0x%llx, set BAR(%d) at_addr=0x%lx to addr=0x%llx!\n",
			__func__, absolute_addr, bar,
			priv->bar[bar].at_addr, addr);
#else
		VASTAI_PCI_ERR(
			priv, die_index,
			"%s [dangerous] 0x%llx, set BAR(%d) at_addr=0x%lx to addr=0x%llx!\n",
			__func__, absolute_addr, bar,
			priv->bar[bar].at_addr, addr);
		dump_stack();
		return -1;
#endif
	}

	return ret;
}

int vastai_pci_mem_bar2_read(struct vastai_pci_info *priv, int die_index, u64 absolute_addr,
				 void *buf, unsigned int len)
{
	int ret = 0;
	u64 base_addr = 0;
	u64 offset = 0;

	while(len) {
		u32 tmp_len;
		ret = vastai_pci_move_bar_window(priv, die_index, VASTAI_PCI_BAR2, absolute_addr);
		if(ret < 0) {
			return ret;
		}

		base_addr = priv->bar[VASTAI_PCI_BAR2].at_addr;
		offset = absolute_addr - base_addr;
		tmp_len = ((offset + len) > priv->bar[VASTAI_PCI_BAR2].mmio_len) ?
				(priv->bar[VASTAI_PCI_BAR2].mmio_len - offset) : len;
		ret = vastai_pci_bar_read(priv, VASTAI_PCI_BAR2, offset, buf, tmp_len);
		if (ret)
			break;
		len -= tmp_len;
		if (len) {
			buf += tmp_len;
			absolute_addr += priv->bar[VASTAI_PCI_BAR2].mmio_len;
		}
	}

	return ret;
}

/* if die_index == -1, function will default use die_id == 0 */
int vastai_pci_mem_read_direct(struct vastai_pci_info *priv, int die_index,
			u64 relative_addr, void *buf, unsigned int len)
{
	int die_id =
		(die_index != -1) ? vastai_pci_get_die_id(priv, die_index) : 0;
	unsigned long long die_offset = die_id * VASTAI_DIE1_BASE;
	int ret = 0, bar = 0;
	u64 base_addr = 0, offset = 0;
	u64 absolute_addr = die_offset | relative_addr;

	if (die_id < 0) {
		return die_id;
	}

	if ((!priv) || (!priv->dev)) {
		VASTAI_PCI_ERR(priv, die_id,
			       "%s priv 0x%p or dev 0x%p is NULL\n", __func__,
			       priv, priv->dev);
		return -ENXIO;
	}

	absolute_addr =
		vastai_pci_cal_absolute_addr(priv, die_id, relative_addr, absolute_addr);
	if ((bar = vastai_pci_select_bar(priv, absolute_addr)) ==
	    VASTAI_PCI_BAR2) {
		ret = vastai_pci_mem_bar2_read(priv, die_index, absolute_addr, buf, len);
	} else {
		u32 tmp_len = 0;
		base_addr = priv->bar[bar].at_addr;
		offset = absolute_addr - base_addr;

		tmp_len = ((offset + len) > priv->bar[bar].mmio_len) ?
				(priv->bar[bar].mmio_len - offset) : len;
		ret = vastai_pci_bar_read(priv, bar, offset, buf, tmp_len);
		if(ret)
			return ret;

		len -= tmp_len;
		if (len) {
			ret = vastai_pci_mem_read_direct(priv, die_index, relative_addr+tmp_len,
							 buf+tmp_len, len);
		}
	}

	VASTAI_PCI_DBG(priv, die_id,
		       "%s: bar=%d, base=0x%llx, offset=0x%llx, len=0x%x\n",
		       __func__, bar, base_addr, offset, len);

	return ret;
}

int vastai_pci_mem_read(struct vastai_pci_info *priv, int die_index,
			u64 relative_addr, void *buf, unsigned int len)
{
	u32 pci_state = atomic_read(&priv->pci_state);

	if (pci_state == VASTAI_RESET_STATE || pci_state == VASTAI_NO_RUN_STATE || pci_state == VASTAI_BOOT_STATE) {
		//TODO: print
		return -EBUSY;
	}

	return vastai_pci_mem_read_direct(priv, die_index, relative_addr, buf, len);
}

int vastai_pci_get_die_id(struct vastai_pci_info *pcie_dev, u32 die_index)
{
	return ((union die_index_data)die_index).die_id;
}

int vastai_pci_mem_bar2_write(struct vastai_pci_info *priv, int die_index, u64 absolute_addr,
				const void *buf, unsigned int len)
{
	int ret = 0;
	u64 base_addr = 0;
	u64 offset = 0;

	while(len) {
		u32 tmp_len;
		ret = vastai_pci_move_bar_window(priv, die_index, VASTAI_PCI_BAR2, absolute_addr);
		if(ret < 0) {
			return ret;
		}

		base_addr = priv->bar[VASTAI_PCI_BAR2].at_addr;
		offset = absolute_addr - base_addr;
		tmp_len = ((offset + len) > priv->bar[VASTAI_PCI_BAR2].mmio_len) ?
				(priv->bar[VASTAI_PCI_BAR2].mmio_len - offset) : len;
		ret = vastai_pci_bar_write(priv, VASTAI_PCI_BAR2, offset, buf, tmp_len);
		if (ret)
			break;
		len -= tmp_len;
		if (len) {
			buf += tmp_len;
			absolute_addr += priv->bar[VASTAI_PCI_BAR2].mmio_len;
		}
	}

	return ret;
}

/* if die_index == -1, function will default use die_id == 0 */
int vastai_pci_mem_write_direct(struct vastai_pci_info *priv, int die_index,
			 u64 relative_addr, const void *buf, unsigned int len)
{
	int die_id =
		(die_index != -1) ? vastai_pci_get_die_id(priv, die_index) : 0;
	unsigned long long die_offset = die_id * VASTAI_DIE1_BASE;
	int ret = 0, bar = 0;
	u64 base_addr = 0, offset = 0;
	u64 absolute_addr = die_offset | relative_addr;

	if (die_id < 0) {
		return die_id;
	}

	if ((!priv) || (!priv->dev)) {
		VASTAI_PCI_ERR(priv, die_id,
			       "%s priv 0x%p or dev 0x%p is NULL\n", __func__,
			       priv, priv->dev);
		return -ENXIO;
	}

	absolute_addr =
		vastai_pci_cal_absolute_addr(priv, die_id, relative_addr, absolute_addr);
	if ((bar = vastai_pci_select_bar(priv, absolute_addr)) ==
	    VASTAI_PCI_BAR2) {
		ret = vastai_pci_mem_bar2_write(priv, die_index, absolute_addr, buf, len);
	} else {
		u32 tmp_len = 0;

		base_addr = priv->bar[bar].at_addr;
		offset = absolute_addr - base_addr;

		tmp_len = ((offset + len) > priv->bar[bar].mmio_len) ?
				(priv->bar[bar].mmio_len - offset) : len;
		ret = vastai_pci_bar_write(priv, bar, offset, buf, tmp_len);

		len -= tmp_len;
		if (len) {
			ret = vastai_pci_mem_write_direct(priv, die_index, relative_addr+tmp_len,
							 buf+tmp_len, len);
		}
	}

	VASTAI_PCI_DBG(priv, die_id,
		       "%s: bar=%d, base=0x%llx, offset=0x%llx, len=0x%x\n",
		       __func__, bar, base_addr, offset, len);

	return ret;
}

int vastai_pci_mem_write(struct vastai_pci_info *priv, int die_index,
			 u64 relative_addr, const void *buf, unsigned int len)
{
	u32 pci_state = atomic_read(&priv->pci_state);

	if (pci_state == VASTAI_RESET_STATE || pci_state == VASTAI_NO_RUN_STATE || pci_state == VASTAI_BOOT_STATE)
		return -EBUSY;

	return vastai_pci_mem_write_direct(priv, die_index, relative_addr, buf, len);

}

int vastai_pci_generation_sel(struct vastai_pci_info *priv, u8 gen)
{
	u64 reg_addr = 0;
	u32 reg_val = 0;
	u8 lnksta = 0, lnkcap = 0, lnkcnt = 0;
	unsigned int die_index = priv->dies[0].die_index;
	int ret = 0;

	if ((gen < VASTAI_PCI_GEN1) || (gen > VASTAI_PCI_GEN4)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s cann't switch Host/Die0 to GEN%d\n",
			       __func__, gen);
		return -EINVAL;
	}

	/* try to change generation up to 10 times */
	while (lnkcnt < 10) {
		/* get negotiated link speed */
		ret = pci_read_config_dword(priv->dev,
					    PCIE_LINK_CONTROL_STATUS_REG_OFFSET,
					    &reg_val);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s err=%d\n",
				       __func__, ret);
			return -1;
		}
		lnksta = (reg_val & 0x000F0000) >> 16;
		if (gen == lnksta) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
					"%s Host/Die0 lnksta=GEN%d.\n",
					__func__, gen);
			return 0;
		}

		/* configurate pcie_generation_sel(bit18:17) to GEN4 */
		reg_addr = VASTAI_PCI_CONFIGURATION_IO_0;
		ret = vastai_pci_mem_read(priv, die_index, reg_addr, &reg_val,
					  4);
		if (ret < 0) {
			return -EIO;
		}
		lnkcap = ((reg_val & 0x00060000) >> 17) + 1;
		if (gen > lnkcap) {
			reg_val = (reg_val & 0xFFF9FFFF) | (3 << 17);
			vastai_pci_mem_write(priv, die_index, reg_addr,
					     &reg_val, 4);
			VASTAI_PCI_DBG(
				priv, DUMMY_DIE_ID,
				"configurate Host/Die0 pcie_generation_sel to GEN4\n");
		}

		/* configurate target link speed (LnkCtl2) */
		ret = pci_read_config_dword(
			priv->dev, PCIE_LINK_CONTROL_STATUS_REG2_OFFSET,
			&reg_val);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s Host/Die0 err=%d\n", __func__, ret);
			return -1;
		}
		reg_val = (reg_val & 0xFFFFFFF0) | gen;
		ret = pci_write_config_dword(
			priv->dev, PCIE_LINK_CONTROL_STATUS_REG2_OFFSET,
			reg_val);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s Host/Die0 err=%d\n", __func__, ret);
			return -1;
		}

		/* trigger link retraining */
		reg_addr = VASTAI_PCI_CDN_LINKWIDTH_CONTROL_REG;
		while (1) {
			ret = vastai_pci_mem_read(priv, die_index, reg_addr,
						  &reg_val, 4);
			if (ret < 0) {
				return -EIO;
			}
			/* sw must wait for retrain link bit to be clear before
			 * setting it again. */
			if ((reg_val & 0x80000000) == 0)
				break;
		}
		reg_val &= 0xFCFFFFFF;
		reg_val |= ((gen - 1) << 24) | (1u << 31);
		vastai_pci_mem_write(priv, die_index, reg_addr, &reg_val, 4);

		lnkcnt++;
	}
	return 0;
}

int vastai_pci_generation_sel_slave_die(struct vastai_pci_info *priv, u8 gen,
					u32 die_id)
{
	u64 reg_addr = 0;
	u32 reg_val = 0, next_die_id = die_id + 1;
	u8 lnksta = 0, lnkcap = 0, lnkcnt = 0;
	int ret = 0;

	if ((gen < VASTAI_PCI_GEN1) || (gen > VASTAI_PCI_GEN4)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s cann't switch Die%d/Die%d to GEN%d\n",
			       __func__, die_id, next_die_id, gen);
		return -EINVAL;
	}

	/* try to change generation up to 10 times */
	while (lnkcnt < 10) {
		/* get negotiated link speed */
		reg_addr = VASTAI_PCI_LINK_CONTROL_STATUS_REG;
		ret = vastai_pci_mem_read(priv,
					  priv->dies[next_die_id].die_index,
					  reg_addr, &reg_val, 4);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s Die%d/Die%d err=%d\n", __func__,
				       die_id, next_die_id, ret);
			return -1;
		}
		lnksta = (reg_val & 0x000F0000) >> 16;
		if (gen == lnksta) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
					"%s Die%d/Die%d lnksta=GEN%d.\n",
					__func__, die_id, next_die_id, gen);
			return 0;
		}

		/* configurate pcie_generation_sel(bit18:17) to GEN4 */
		reg_addr = VASTAI_PCI_CONFIGURATION_IO_0;
		ret = vastai_pci_mem_read(priv,
					  priv->dies[next_die_id].die_index,
					  reg_addr, &reg_val, 4);
		if (ret < 0) {
			return -EIO;
		}
		lnkcap = ((reg_val & 0x00060000) >> 17) + 1;
		if (gen > lnkcap) {
			reg_val = (reg_val & 0xFFF9FFFF) | (3 << 17);
			vastai_pci_mem_write(priv,
					     priv->dies[next_die_id].die_index,
					     reg_addr, &reg_val, 4);
			VASTAI_PCI_DBG(
				priv, DUMMY_DIE_ID,
				"configurate Die%d/Die%d pcie_generation_sel to GEN4\n",
				die_id, next_die_id);
		}

		/* configurate target link speed (LnkCtl2) */
		reg_addr = VASTAI_PCI_LINK_CONTROL_STATUS_REG2;
		ret = vastai_pci_mem_read(priv,
					  priv->dies[next_die_id].die_index,
					  reg_addr, &reg_val, 4);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s Die%d/Die%d err=%d\n", __func__,
				       die_id, next_die_id, ret);
			return -1;
		}
		reg_val = (reg_val & 0xFFFFFFF0) | gen;
		vastai_pci_mem_write(priv, priv->dies[next_die_id].die_index,
				     reg_addr, &reg_val, 4);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s Die%d/Die%d err=%d\n", __func__,
				       die_id, next_die_id, ret);
			return -1;
		}

		/* trigger link retraining */
		reg_addr = VASTAI_PCI_CDN_LINKWIDTH_CONTROL_REG;
		while (1) {
			ret = vastai_pci_mem_read(
				priv, priv->dies[next_die_id].die_index,
				reg_addr, &reg_val, 4);
			if (ret < 0) {
				return -EIO;
			}
			/* sw must wait for retrain link bit to be clear before
			 * setting it again. */
			if ((reg_val & 0x80000000) == 0)
				break;
		}
		reg_val &= 0xFCFFFFFF;
		reg_val |= ((gen - 1) << 24) | (1u << 31);
		vastai_pci_mem_write(priv, priv->dies[next_die_id].die_index,
				     reg_addr, &reg_val, 4);

		lnkcnt++;
	}
	return 0;
}

static int vastai_pci_get_out_desc(struct vastai_pci_info *priv, int irq,
				   struct out_desc *out)
{
	u64 reg_addr = 0;
	u32 die_index;
	u32 die_irq = irq % VASTAI_PCIE_MSIX_NUM_PER_DIE;
	u32 die_id = irq / VASTAI_PCIE_MSIX_NUM_PER_DIE;
	vastai_intr_handle handle;
	void *pd;
	u32 vdsp_id;


	if (priv->is_virtfn) {
		die_irq = irq;
		die_id = VASTAI_DIE0;
	}


	die_index = priv->dies[die_id].die_index;
	handle = priv->dies[die_id].ivt[die_irq].intr_handle;
	pd = priv->dies[die_id].ivt[die_irq].private_data;

	if ((die_irq >= VDSP0_2_HOST_INT0) && (die_irq <= VDSP3_2_HOST_INT1)) {
		/* TODO: move this into ai callback */
		/* run stream cmd irq from vdsp */
		if ((die_irq == VDSP0_2_HOST_INT0) ||
		    (die_irq == VDSP0_2_HOST_INT1)) {
			vdsp_id = 0;
			reg_addr = MSG_VDSP0_TO_PCIE;
		} else if ((die_irq == VDSP1_2_HOST_INT0) ||
			   (die_irq == VDSP1_2_HOST_INT1)) {
			vdsp_id = 1;
			reg_addr = MSG_VDSP1_TO_PCIE;
		} else if ((die_irq == VDSP2_2_HOST_INT0) ||
			   (die_irq == VDSP2_2_HOST_INT1)) {
			vdsp_id = 2;
			reg_addr = MSG_VDSP2_TO_PCIE;
		} else {
			vdsp_id = 3;
			reg_addr = MSG_VDSP3_TO_PCIE;
		}

		while (!vastai_fifo_pop_elem(priv, die_index, reg_addr,
					     (void *)out, NULL)) {
			priv->dies[die_id].pcie_rx_vdsp_count[vdsp_id] += 1;
			if (handle) {
				handle(priv, die_index, die_irq, pd, out);
			}
			VASTAI_PCI_DBG(
				priv, die_id,
				"%s dev:0x%x, data_id:0x%llx, context_id:0x%x, stream_addr:0x%llx\n",
				__func__, out->dev, out->data_id,
				out->context_id, out->stream_addr);
		}
	} else if ((die_irq >= CMCU_2_HOST_INT0) &&
		   (die_irq <= CMCU_2_HOST_INT1)) {
		/* TODO: move this into ai callback */
		/* run model cmd irq from cmcu */
		reg_addr = MSG_CMCU_TO_PCIE;
#ifdef CONFIG_VASTAI_CMCU_RINGBUF
		while (!vastai_fifo_pop_elem(priv, die_index, reg_addr, (void *)out, NULL)) {
			if (handle) {
				handle(priv, die_index, die_irq, pd, out);
			}
		}
#else
		vastai_pci_mem_read(priv, die_index, reg_addr, (void *)out,
				    sizeof(struct out_desc));
		if (handle) {
			handle(priv, die_index, die_irq, pd, out);
		}
#endif
	} else if ((die_irq >= VDMCU0_2_HOST_INT) &&
		   (die_irq <= VDMCU2_2_HOST_INT)) {
		/* video irq */
		if (die_irq == VDMCU0_2_HOST_INT) {
			reg_addr = MSG_VDMCU0_TO_PCIE;
		} else if (die_irq == VDMCU1_2_HOST_INT) {
			reg_addr = MSG_VDMCU1_TO_PCIE;
		} else {
			reg_addr = MSG_VDMCU2_TO_PCIE;
		}
		/* get rd num */
		while (!vastai_fifo_pop_elem(priv, die_index, reg_addr,
					     (void *)out, NULL)) {
			if (handle)
				handle(priv, die_index, die_irq, pd, out);
		}
	} else {
		if (die_irq == VEMCU0_2_HOST_INT) {
			reg_addr = MSG_VEMCU0_TO_PCIE;
		} else if (die_irq == VEMCU1_2_HOST_INT) {
			reg_addr = MSG_VEMCU1_TO_PCIE;
		} else if (die_irq == VEMCU2_2_HOST_INT) {
			reg_addr = MSG_VEMCU2_TO_PCIE;
		} else {
			reg_addr = MSG_VEMCU3_TO_PCIE;
		}
		while (!vastai_fifo_pop_elem(priv, die_index, reg_addr,
					     (void *)out, NULL)) {
			if (handle)
				handle(priv, die_index, die_irq, pd, out);
		}
	}
#if 0
	VASTAI_PCI_INFO(priv, die_id,
			"%s dev:0x%x, data_id:0x%llx, context_id:0x%x, stream_addr:0x%llx\n",
			__func__, out->dev, out->data_id, out->context_id, out->stream_addr);
#endif
	return 0;
}

static int vastai_pci_legacy_irq(int irq, void *arg)
{
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "legacy irq called %d\n", irq);

	// legacy_irq_handle(irq);

	return IRQ_HANDLED;
}

static void vastai_print_ecc_entry(void *priv, u64 fifo_addr, u32 die_index)
{
	struct vastai_fifo fifo = { 0 };
	u64 ecc_addr = 0x0;
	struct ecc_entry entry = { 0 };
	union die_index_data die_temp = { .val = die_index };

	vastai_pci_mem_read(priv, die_index, fifo_addr, &fifo, sizeof(fifo));
	ecc_addr = ((fifo.wr - 1) * fifo.elem_size) + fifo_addr + 0x10;

	vastai_pci_mem_read(priv, die_index, ecc_addr, &entry, sizeof(entry));

	VASTAI_PCI_ERR(priv, die_temp.die_id,
		       "[vaerr] Get DDR ECC error addr_l 0x%x addr_h 0x%x\n",
		       entry.addr_1, entry.addr_h);
}

static u32 vastai_check_core_status(struct vastai_pci_info *priv, int irq)
{
	u32 die_irq = irq % VASTAI_PCIE_MSIX_NUM_PER_DIE;
	u32 die_id = irq / VASTAI_PCIE_MSIX_NUM_PER_DIE;
	u32 core_bit = 0;

	if (!priv->dies[die_id].core_status_bitmap)
		return 0;

	if ((die_irq >= VDSP0_2_HOST_INT0) && (die_irq <= VDSP3_2_HOST_INT1)) {
		if ((die_irq == VDSP0_2_HOST_INT0) ||
		    (die_irq == VDSP0_2_HOST_INT1)) {
			core_bit = CORE_POINT_VDSP0;
		} else if ((die_irq == VDSP1_2_HOST_INT0) ||
			   (die_irq == VDSP1_2_HOST_INT1)) {
			core_bit = CORE_POINT_VDSP1;
		} else if ((die_irq == VDSP2_2_HOST_INT0) ||
			   (die_irq == VDSP2_2_HOST_INT1)) {
			core_bit = CORE_POINT_VDSP2;
		} else {
			core_bit = CORE_POINT_VDSP3;
		}
	} else if ((die_irq >= CMCU_2_HOST_INT0) &&
		   (die_irq <= CMCU_2_HOST_INT1)) {
		core_bit = CORE_POINT_CMCU;
	} else if ((die_irq >= VDMCU0_2_HOST_INT) &&
		   (die_irq <= VDMCU2_2_HOST_INT)) {
		core_bit = die_irq - VDMCU0_2_HOST_INT + CORE_POINT_VDMCU0;
	} else if ((die_irq >= VEMCU0_2_HOST_INT) &&
		   (die_irq <= VEMCU3_2_HOST_INT)) {
		core_bit = die_irq - VEMCU0_2_HOST_INT + CORE_POINT_VEMCU0;
	}
	return priv->dies[die_id].core_status_bitmap & (1 << core_bit);
}

u8 core_point_to_heartbeat(u8 core_point)
{
	return core_point+1;
}

void event_notify(u32 die_index, u32 core_bit, reset_event event)
{
#ifdef CONFIG_VASTAI_TOOLS_SUITE
	vatools_device_event_callback(die_index, core_bit, event);
#endif
	if (g_reset_ai_handle) {
		g_reset_ai_handle(die_index, core_bit, event);
	}

	return;
}

static int vastai_pci_msi_irq(int irq, void *arg)
{
	struct msix_entry *msix = arg;
	u32 die_id = 0;
	u32 die_irq = 0;
	struct vastai_pci_info *priv;
	struct vastai_sv100_die *die;

#if CONFIG_VASTAI_PCI_PERFORMANCE
	struct timespec tm_begin, tm_end;
	unsigned long time_total_ns;
	u32 reg_val;

	getnstimeofday(&tm_begin);
	getnstimeofday(&tm_end);
	time_total_ns = timespec_to_ns(&tm_end) - timespec_to_ns(&tm_begin);
	VASTAI_PCI_INFO(priv, die_id, "msi irq called %d, MRd:%ld\n", irq,
			time_total_ns);

	getnstimeofday(&tm_begin);
	getnstimeofday(&tm_end);
	time_total_ns = timespec_to_ns(&tm_end) - timespec_to_ns(&tm_begin);
	VASTAI_PCI_INFO(priv, die_id, "msi irq called %d, MWr:%ld\n", irq,
			time_total_ns);
#endif
	if (!msix) {
		pr_err("irq handle arg is NULL. irq %d\n", irq);
		return IRQ_HANDLED;
	}
	irq = msix->entry;

	priv = container_of(msix, typeof(*priv), msix[msix->entry]);
	if (!priv) {
		pr_err("irq handle pcie_info is NULL. irq %d arg %p msix %p\n",
				irq, arg, msix);
		return IRQ_HANDLED;
	}

	if (atomic_read(&(priv->pci_state)) == VASTAI_EXIT_STATE)
		return IRQ_HANDLED;

	if (priv->is_virtfn) {
		die_irq = irq;
		die_id = VASTAI_DIE0;
	} else {
		die_irq = irq % VASTAI_PCIE_MSIX_NUM_PER_DIE;
		die_id = irq / VASTAI_PCIE_MSIX_NUM_PER_DIE;
	}

	die = &priv->dies[die_id];

	if (die_id > ARRAY_SIZE(priv->dies) || (die_id >= priv->die_num)) {
		VASTAI_PCI_INFO(
			priv, DUMMY_DIE_ID,
			"msi called with miss die! irq[%d] die_id[%d] die_irq[%d]\n",
			irq, die_id, die_irq);
		return IRQ_HANDLED;
	}
	// msi_irq_handle(irq);
	VASTAI_PCI_DBG(priv, die_id, "msi irq called %d die_id %d die_irq %d\n",
		       irq, die_id, die_irq);

	// NTD
	if (die_irq == 0) {
		queue_work(priv->dies[die_id].udma_wq,
			   &priv->dies[die_id].dma_done);

	} else if (die_irq == HOST_GUEST_INT) {
		vastai_proc_peer_msg(priv, die_id);

	} else if (die_irq == 1) {
		u32 die_index = priv->dies[die_id].die_index;
		struct smcu_to_host_msg1 msg;
		unsigned long flags = 0;
		u64 msg1_addr = MSG_SMCUQ1_TO_PCIE;
		u64 msg1_spe_addr = MSG_SMCUQ1_TO_PCIE + VASTAI_PCIE_MSGQ_1_SIZE;

		if (priv->is_virtfn)
			msg1_addr += VASTAI_PCIE_SMCU_1_VF_OFFSET;

		spin_lock_irqsave(&priv->msgq_1_lock, flags);

		while (!vastai_fifo_pop_elem_spe(priv, die_index,
					msg1_spe_addr, &msg, NULL)) {
			switch (msg.type) {
				case VASTAI_PCIE_MSG1_SET_BAR_DONE: {
					VASTAI_PCI_INFO(priv, die_id, "set bar done\n");
					complete(&priv->set_bar_done);
					break;
				}
				default:
					break;
			}
		}

		while (!vastai_fifo_pop_elem(priv, die_index,
					     msg1_addr, &msg, NULL)) {
			switch (msg.type) {
			case VASTAI_PCIE_MSG1_LOG_READ: {
				struct vastai_sv100_die *dies =
					&(priv->dies[die_id]);

				complete(&dies->log_comp);
				break;
			}
			case VASTAI_PCIE_MSG1_ECC_2BIT_ERR: {
				struct vastai_sv100_die *dies =
					&(priv->dies[die_id]);
				complete(&dies->ecc_2bit_err);

				break;
			}
			case VASTAI_PCIE_MSG1_ECC_1BIT_ERR: {
				vastai_print_ecc_entry(priv, ECC_1BIT_BUF,
						       die_index);
				break;
			}
			case VASTAI_PCIE_MSG1_ECC_2BIT_FROM_FLASH: {
				vastai_print_ecc_entry(priv, ECC_2BIT_BUF,
						       die_index);
				break;
			}
			case VASTAI_PCIE_MSG1_FREQ_REDUCE_BY_TEMP: {
				if (priv->dies[die_id].dev_freq > msg.info[0]) {
					VASTAI_PCI_INFO(
						priv, die_id,
						"[vawarn] Frequency reduced to  %d%% due to over-temperature!\n",
						msg.info[0]);
				} else {
					VASTAI_PCI_INFO(
						priv, die_id,
						"[vawarn] Frequency increased to  %d%% as temperature returned to normal.\n",
						msg.info[0]);
				}
				priv->dies[die_id].dev_freq = msg.info[0];
				break;
			}
			case VASTAI_PCIE_MSG1_FREQ_REDUCE_BY_VOLT: {
				if (priv->dies[die_id].dev_freq > msg.info[0]) {
					VASTAI_PCI_INFO(
						priv, die_id,
						"[vawarn] Frequency reduced to  %d%% due to over-current!",
						msg.info[0]);
				} else {
					VASTAI_PCI_INFO(
						priv, die_id,
						"[vawarn] Frequency increased to  %d%% as current returned to normal.\n",
						msg.info[0]);
				}
				priv->dies[die_id].dev_freq = msg.info[0];
				break;
			}
			case VASTAI_PCIE_MSG1_PCIE_ERR: {
				struct vastai_pcie_link_err pcie_link_err = {0};
				int i;

				/* FIXME: read pcie_link_err from csram */
				for (i = 0;
				     i < ARRAY_SIZE(pcie_link_err.err_list);
				     i++) {
					if (((pcie_link_err.error_type >>
					      (i * 4)) &
					     0xf) ==
					    2) { /* type 2 is pcie_link error */
						switch (pcie_link_err
								.err_list[i]
								.sub_err_type) {
						case PCIE_LOCAL_ERROR:
							VASTAI_PCI_ERR(
								priv, die_id,
								"[vaerr] Get PCIe local error, error status = %d\n",
								pcie_link_err
									.err_list[i]
									.AddrLOrErrSt
									.ErrSt);
							break;
						case PCIE_LINK_ERROR:
							VASTAI_PCI_ERR(
								priv, die_id,
								"[vaerr] Get PCIe link error, error status = %d\n",
								pcie_link_err
									.err_list[i]
									.AddrLOrErrSt
									.ErrSt);
							break;
						case PCIE_DMA_ERROR:
							VASTAI_PCI_ERR(
								priv, die_id,
								"[vaerr] Get PCIe dma error, error status = %d\n",
								pcie_link_err
									.err_list[i]
									.AddrLOrErrSt
									.ErrSt);
							break;
						}
					}
				}
				break;
			}
			case VASTAI_PCIE_MSG1_RW_XSPI: {
				struct vastai_sv100_die *dies =
					&(priv->dies[die_id]);
				dies->rw_xspi_msg = msg.val >> 8;
				complete(&dies->rw_xspi);
				break;
			}
			case VASTAI_PCIE_MSG1_CTRL_COMPLETION: {
				vastai_ctrl_cmd_compeletion(
					priv, priv->dies[die_id].die_index,
					&msg);
				break;
			}
			case VASTAI_PCIE_MSG1_DMA_LIST_BROKEN: {
				VASTAI_PCI_ERR(priv, die_id, "Dma list can't append! pls check.\n");
				break;
			}
			case VASTAI_PCIE_MSG1_POWER_DOWN_BY_BMCU: {
				int i;
				/* set smcu machine state for each die */
				VASTAI_PCI_ERR(priv, die_id,
					"bmcu power down alert occurs, ret=%d.\n", msg.info[0]);
				for (i = 0; i < priv->die_num; i++) {
					vastai_send_pcie_cmd(priv,
							    priv->dies[i].die_index,
							    VASTAI_PCIE_SUB_SET_MACHINE_STATE,
							    1);
#if 0
					atomic_set(&priv->pci_state, VASTAI_EXIT_STATE);
#else
					vastai_pci_state_power_off(priv);
#endif
				}
				break;
			}
			case VASTAI_PCIE_MSG1_BMCU_XSPI_FINISH: {
#if 0
				u32 old;
				old = atomic_cmpxchg(&(priv->pci_state),
							VASTAI_UPDATE_STATE,
							VASTAI_NORMAL_STATE);
				if (old != VASTAI_UPDATE_STATE) {
					VASTAI_PCI_INFO(priv, die_id,
						"pcie state error: %d: update bmcu completion back, but state is not update state\n",
						old);
				}
#else
				vastai_pci_state_update_done(priv, die_id);
#endif
				break;
			}

			case VASTAI_PCIE_MSG1_SMI_ACK: {
				struct vastai_sv100_die *dies =
					&(priv->dies[die_id]);
				u8 bc = (msg.val >> 16) & 0xFF;

				VASTAI_PCI_DBG(
					priv, die_id,
					"recv VASTAI_PCIE_MSG1_SMI_ACK block_id=%d",
					bc);

				if (bc >= 2) {
					VASTAI_PCI_ERR(priv, die_id,
						       "[ERROR] SMI block count > 2 ");
					dies->smi_ack_msg[0] = 0;
					dies->smi_ack_msg[1] = 0;
					complete(&dies->smi_ack[0]);
					complete(&dies->smi_ack[1]);
					break;
				}
				dies->smi_ack_msg[bc] = (msg.val >> 8) & 0xFF;
				complete(&dies->smi_ack[bc]);
				break;
			}
			case VASTAI_PCIE_MSG1_CORE_TIME_SYNC_ACK: {
				struct vastai_sv100_die *dies =
					&(priv->dies[die_id]);

				VASTAI_PCI_DBG(
					priv, die_id,
					"recv VASTAI_PCIE_MSG1_CORE_TIME_SYNC_ACK ");
				complete(&dies->core_time_sync_ack);
				break;
			}
			case VASTAI_PCIE_MSG1_EXCEPTION: {
				if(msg.info[0] == CORE_2_CORE_RESET_AI_SYS_HANG_SUBCMD) {
					int i;
					u32 ai_mask = 0;
					for(i = 0; i < sizeof(ai_group)/sizeof(ai_group[0]); i++)
						ai_mask |= (1<<ai_group[i]);

					priv->dies[die_id].heartbeat_mask |= ai_mask;
					priv->dies[die_id].core_status_bitmap |= 1 << CORE_POINT_CMCU;
					event_notify(
						priv->dies[die_id].die_index,
						CORE_POINT_CMCU, RESET_START);
					schedule_work(&priv->dies[die_id].exception_work);
				}
				else if (msg.info[0] == CORE_2_CORE_RESET_AI_FINISH_SUBCMD) {
					//设置heart beat
					int i;
					VASTAI_PCI_INFO(priv, die_id,
						       "recv RESET_AI_FINISH_SUBCMD \n");

					if(priv->vadev.dev_cap != VASTAI_HW_TYPE_VIDEO_ONLY)
						for(i = 0; i<sizeof(ai_group)/sizeof(ai_group[0]); i++)
							priv->dies[die_id].heartbeat_mask &= ~(1<<ai_group[i]);
					event_notify(
						priv->dies[die_id].die_index,
						CORE_POINT_CMCU, RESET_END);
				}
				else if (msg.info[0] == CORE_2_CORE_RESET_VDSP_HANG_SUBCMD)
				{
					u32 vdsp_mask =  1<<(HEARTBEAT_VDSP0+(msg.info[1]-CORE_POINT_VDSP0));
					//设置heart beat
					VASTAI_PCI_INFO(priv, die_id,
						       "recv vdsp hang subcmd core[%u]", msg.info[1]);
					event_notify(
						priv->dies[die_id].die_index,
						msg.info[1], RESET_START);

					priv->dies[die_id].heartbeat_mask |= vdsp_mask;
					priv->dies[die_id].core_status_bitmap |= 1 << msg.info[1];
					schedule_work(&priv->dies[die_id].exception_work);
				}
				else if (msg.info[0] == CORE_2_CORE_RESET_VDSP_FINISH_SUBCMD)
				{
					//设置heart beat
					VASTAI_PCI_INFO(priv, die_id,
						       "recv RESET_vdsp%u_FINISH_SUBCMD\n", msg.info[1] - CORE_VDSP0);

					priv->dies[die_id].heartbeat_mask &= ~(1<<(msg.info[1] - CORE_VDSP0 + HEARTBEAT_VDSP0));
					event_notify(
						priv->dies[die_id].die_index,
						msg.info[1], RESET_END);
				}
				break;
			}
			case VASTAI_PCIE_MSG1_CORE_EXCEPTION: {
				struct vastai_sv100_die *dies =
					&(priv->dies[die_id]);

				event_notify(priv->dies[die_id].die_index,
					     msg.info[0], RESET_START);
				dies->core_status_bitmap |= 1 << msg.info[0];
				schedule_work(&dies->exception_work);
				break;
			}
			case VASTAI_PCIE_MSG1_CORE_RESET_ACK: {
				struct vastai_sv100_die *dies =
					&(priv->dies[die_id]);
				dies->core_status_bitmap &= ~(1 << msg.info[0]);
				dies->heartbeat_mask &= ~(1 << core_point_to_heartbeat(msg.info[0]));

				event_notify(priv->dies[die_id].die_index,
					     msg.info[0], RESET_END);
				VASTAI_PCI_INFO(priv, die_id, "receive core reset ack core:%s, ret:%x\n",
								core_name(core_point_to_heartbeat(msg.info[0])), msg.info[1]);
				break;
			}
			case VASTAI_PCIE_MSG1_VE1_FAN_WARN: {
				if (msg.info[0] == 0)
					VASTAI_PCI_INFO(priv, die_id,
						"ve1s fan speed warn clear.\n");
				else
					VASTAI_PCI_INFO(priv, die_id,
						"ve1s fan speed warn, ret=%d%%.\n", msg.info[0]);
				break;
			}
			case VASTAI_PCIE_MSG1_VIDEO_EXCEPTION: {
				//info[0]:core_id, info[1]:exception num
				event_notify(priv->dies[die_id].die_index,
					     msg.info[0], msg.info[1]);
				VASTAI_PCI_DBG(priv, die_id,
						       "recv core_id:%x, exp_num:%x\n", msg.info[0], msg.info[1]);
				break;
			}
			case VASTAI_PCIE_MSG1_GET_DLC_CORE_DONE: {
				u16 tmp = (msg.info[1]<<8)|msg.info[0];
				u8 type = (tmp>>12)&0xf;
				u8 cfg = msg.info[2];

				priv->harvest_info[die_id].type = type;
				priv->harvest_info[die_id].dlc_cores = tmp;

				if (type == 0)
					VASTAI_PCI_INFO(priv, die_id,
						"Harvest: All DLC & ODSP core are enabled.\n");
				else if (type == 1 || type == 4)
					VASTAI_PCI_INFO(priv, die_id,
						"Harvest: CodecDis=[%d] OakCfgId=[%d] Type=[%d] Harvest_Core=[0x%x] (DLC%d DLC%d DLC%d DLC%d).\n", \
						((cfg>>3)&0x1), (cfg&0x7), type, (tmp&0xfff), ((tmp>>9)&0x7), ((tmp>>6)&0x7), ((tmp>>3)&0x7), (tmp&0x7));
				else if (type == 2) {
					if (tmp & 0x1)
						VASTAI_PCI_INFO(priv, die_id,
							"Harvest: All DLC & ODSP core are not available and disabled.\n");
					else
						VASTAI_PCI_INFO(priv, die_id,
							"Harvest: All DLC core are not available and disabled, All ODSP core are enabled.\n");

				}
				else if (type == 3)
					VASTAI_PCI_INFO(priv, die_id,
						"Harvest: All DLC & ODSP core are not available and disabled.\n");
				else
					VASTAI_PCI_INFO(priv, die_id,
						"Harvest: Invalid type.\n");
				complete(&priv->dies[die_id].get_dlc_core_done);
				break;
			}

			case VASTAI_PCIE_MSG1_ACTIVATE_VF: {
				complete(&priv->activate_vf_done);
				break;
			}
			case VASTAI_PCIE_MSG1_DEACTIVATE_VF: {
				complete(&priv->deactivate_vf_done);
				break;
			}

			default:
				break;
			}
		}
		spin_unlock_irqrestore(&priv->msgq_1_lock, flags);
	} else {
		unsigned long flags = 0;
		/*if core is exception, do not process the core's fifo*/
		if (core_exception_en) {
			if (vastai_check_core_status(priv, irq)) {
				goto out;
			}
		}

		/* Maybe we don't need keep desc_out this data struct */
		spin_lock_irqsave(&priv->trans_lock, flags);
		vastai_pci_get_out_desc(priv, irq,
					&(priv->dies[die_id].ivt[die_irq].out));
		spin_unlock_irqrestore(&priv->trans_lock, flags);
	}
out:
	VASTAI_PCI_DBG(priv, die_id, "%s irq:%d done\n", __func__,irq);

	return IRQ_HANDLED;
}

static int vastai_pci_disable_irq(struct vastai_pci_info *priv)
{
	int i = 0;

	if (priv->irq_type == VASTAI_MSIX_IRQ) {
		for (i = 0; i < priv->irq_num; i++) {
			disable_irq(priv->msix[i].vector);
			free_irq(priv->msix[i].vector, (void *)&(priv->msix[i]));
			VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
				       "%s disable_irq(%d) ok\n", __func__,
				       priv->msix[i].vector);
		}
		pci_disable_msix(priv->dev);
	}

	if (priv->irq_type == VASTAI_MSI_IRQ) {
		for (i = 0; i < priv->irq_num; i++) {
			if (priv->irq) {
				disable_irq(priv->irq + i);
				if (!free_irq(priv->irq + i, (void *)&(priv->msix[i])))
					return -1;
			}
		}

		pci_disable_msi(priv->dev);
	}

	return 0;
}

int vastai_irq_register(struct vastai_pci_info *pcie_dev, int die_index,
			int irq, /* reference vastai_sv100_reg.h +272 */
			vastai_intr_handle handle,
			void *private_data) /* for user keep self data */
{
	int die_id = vastai_pci_get_die_id(pcie_dev, die_index);

	if (die_id < 0) {
		return -ENODEV;
	}
	pcie_dev->dies[die_id].ivt[irq].intr_handle = handle;
	pcie_dev->dies[die_id].ivt[irq].private_data = private_data;
	return 0;
}

int vastai_irq_unregister(struct vastai_pci_info *pcie_dev, int die_index,
			  int irq) /* reference vastai_sv100_reg.h +272 */
{
	int die_id = vastai_pci_get_die_id(pcie_dev, die_index);

	if (die_id < 0) {
		return -ENODEV;
	}
	pcie_dev->dies[die_id].ivt[irq].intr_handle = NULL;
	pcie_dev->dies[die_id].ivt[irq].private_data = NULL;
	return 0;
}

bool isCoreReseting(struct vastai_pci_info *priv, u32 die_id, int core_index)
{
	/*cmcu*/
	if(core_index == 0) {
		if(priv->vadev.dev_cap != VASTAI_HW_TYPE_VIDEO_ONLY)
			if(priv->dies[die_id].heartbeat_mask & (1<<HEARTBEAT_CMCU))
				return true;
	} else if (core_index>=8 && core_index<=11) { /*vdsp*/
		if(priv->dies[die_id].heartbeat_mask & (1<<(HEARTBEAT_VDSP0 + core_index - 8)))
			return true;
	}

	return false;
}


/*==================== DMA transfer && MSG api ====================*/
/* Don't call this function at irq, This only use spin_lock */
int vastai_pci_send_msg_pid(void *pcie_dev, int die_index,
			union core_bitmap core_id, void *msg_buf, u32 msg_len, pid_t pid)
{
	u32 temp_core_bitmap = core_id.val;
	int ret = 0;
	u32 die_id;
	profiler_msg_checkpoint_t msg_profiler;
	struct vastai_pci_info *priv;
	struct vastai_sv100_die *die;
	VASTAI_PCI_DBG(pcie_dev, vastai_pci_get_die_id(pcie_dev, die_index),
		       "%s: die=%d, core_id=0x%x, msg_buf=0x%p, len=0x%x\n",
		       __func__, die_index, core_id.val, msg_buf, msg_len);
	memset(&msg_profiler, 0, sizeof(profiler_msg_checkpoint_t));
	msg_profiler.dev_id = vastai_pci_info_get_dev_id(pcie_dev);
	msg_profiler.die_id = vastai_pci_get_die_id(pcie_dev, die_index);
	msg_profiler.die_index = die_index;
	msg_profiler.time_of_host_send_msg = vastai_get_host_time_ns();
	msg_profiler.pid = pid;


	while (temp_core_bitmap) {
		int core_index = ffs(temp_core_bitmap) - 1;
		u64 int_addr = msg_addr_table[core_index].interrupt_addr;
		const u64 msg_addr = msg_addr_table[core_index].msg_addr;
		const u32 msg_addr_dev = 0x80000000 | msg_addr;
		u32 is_use_ring_buf =
			msg_addr_table[core_index].is_use_ring_buf;
		temp_core_bitmap &= ~(BIT(core_index));

		if (!int_addr || !msg_addr) {
			VASTAI_PCI_ERR(
				pcie_dev,
				vastai_pci_get_die_id(pcie_dev, die_index),
				"send_msg with err addr. core_id[0x%x] intr[0x%llx] msg[0x%llx]\n",
				core_id.val, int_addr, msg_addr);
			continue;
		}

		if (is_use_ring_buf) {
			if(isCoreReseting(pcie_dev, vastai_pci_get_die_id(pcie_dev, die_index), core_index)) {
				ret = -ENOMEM;
				break;
			}
			/* msg is a ring buf. There is fifo info at ring head. */
			priv = pcie_dev;
			die_id = vastai_pci_get_die_id(pcie_dev, die_index);
			die = &(priv->dies[die_id]);

			spin_lock(&(die->msg_ring_buf_lock));

			ret = vastai_fifo_push_elem(pcie_dev, die_index,
						    msg_addr, msg_buf, NULL);

			/* get performance info */
			if (core_index >= 8 && core_index < (8 + 4) && !ret) {
				die->pcie_tx_vdsp_count[core_index % 8] += 1;
			}

			spin_unlock(&(die->msg_ring_buf_lock));

		} else {
			ret = vastai_pci_mem_write(pcie_dev, die_index,
						   msg_addr, msg_buf, msg_len);
		}

		if (ret == -ENOMEM) {
			break;
		} else if (ret) {
			VASTAI_PCI_ERR(pcie_dev,
				       vastai_pci_get_die_id(pcie_dev,
							     die_index),
				       "write msg error %d\n", ret);
			break;
		}

		ret = vastai_pci_mem_write(pcie_dev, die_index, int_addr,
					   &msg_addr_dev, 4);
		if (ret) {
			VASTAI_PCI_ERR(pcie_dev,
				       vastai_pci_get_die_id(pcie_dev,
							     die_index),
				       "write int error %d\n", ret);
			break;
		}
	}
#ifdef CONFIG_VASTAI_TOOLS_SUITE
	vatools_profiler_msg_checkpoint(&msg_profiler);
#endif

	return ret;
}

int vastai_pci_send_msg(void *pcie_dev, int die_index,
			union core_bitmap core_id, void *msg_buf, u32 msg_len)
{
	return vastai_pci_send_msg_pid(pcie_dev, die_index,
				core_id, msg_buf, msg_len, -1);

}

int vastai_pci_reset_bus(struct vastai_pci_info *priv)
{
	int ret = 0;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "start to reset pcie bus.\n");
#ifdef CONFIG_VASTAI_PCI_LINK_MONITOR
	vastai_enable_pcie_link_monitor(priv, false);
#endif

	ret = pci_probe_reset_bus(priv->dev->bus);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_probe_reset_bus fail, ret=%d", ret);
		return ret;
	}
	priv->dev->state_saved = true;
#if defined(RHEL_RELEASE_VERSION)
#if RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(8, 0)
	pci_reset_bridge_secondary_bus(priv->dev->bus->self);
#else
	ret = pci_bridge_secondary_bus_reset(priv->dev->bus->self);
#endif
#elif KERNEL_VERSION(4, 19, 0) <= LINUX_VERSION_CODE
	ret = pci_bridge_secondary_bus_reset(priv->dev->bus->self);
#else
	pci_reset_bridge_secondary_bus(priv->dev->bus->self);
#endif
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pci_reset_bus fail, ret=%d",
			       ret);
		return ret;
	}
#ifdef CONFIG_VASTAI_PCI_LINK_MONITOR
	if(atomic_read(&(priv->pci_link_stat.is_reseting))) {
		priv->dev->saved_config_space[8] = 0;
		priv->dev->saved_config_space[9] = 0;
	}
#endif
	pci_restore_state(priv->dev);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "reset pcie bus finish.\n");

	return ret;
}

void vastai_pci_set_vdsp_depth(struct vastai_pci_info *priv, u32 die_id,
			       u32 elem_count, u32 flag)
{
	struct vastai_sv100_die *die = &(priv->dies[die_id]);
	int die_index = die->die_index;
	struct vastai_fifo fifo = { 0 };

	if (elem_count > VASTAI_PCIE_2_VDSP_MSG_DEPTH) {
		VASTAI_PCI_ERR(priv, die_id,
			       "%s elem_cout %d overflow, set to %ld\n",
			       __func__, elem_count,
			       VASTAI_PCIE_2_VDSP_MSG_DEPTH);
		elem_count = VASTAI_PCIE_2_VDSP_MSG_DEPTH;
	}
	spin_lock(&(die->msg_ring_buf_lock));
	fifo.elem_count = elem_count;
	fifo.elem_size = sizeof(union buf_data);
	vastai_pci_mem_write_direct(priv, die_index, MSG_PCIE_TO_VDSP0, &fifo,
			     sizeof(fifo));
	vastai_pci_mem_write_direct(priv, die_index, MSG_PCIE_TO_VDSP1, &fifo,
			     sizeof(fifo));
	vastai_pci_mem_write_direct(priv, die_index, MSG_PCIE_TO_VDSP2, &fifo,
			     sizeof(fifo));
	vastai_pci_mem_write_direct(priv, die_index, MSG_PCIE_TO_VDSP3, &fifo,
			     sizeof(fifo));
	if (flag) {
		VASTAI_PCI_INFO(
			priv, die_index,
			"init vdsp fifo: rd[0x%x] wr[0x%x] elem_count[0x%x] elem_size[0x%x]\n",
			fifo.rd, fifo.wr, fifo.elem_count, fifo.elem_size);
	}
	spin_unlock(&(die->msg_ring_buf_lock));
}

#ifdef CONFIG_VASTAI_CMCU_RINGBUF
void vastai_pci_set_cmcu_depth(struct vastai_pci_info *priv, u32 die_id, u32 elem_count)
{
	struct vastai_sv100_die *die = &(priv->dies[die_id]);
	int die_index = die->die_index;
	struct vastai_fifo fifo = {0};

	if (elem_count > VASTAI_PCIE_2_CMCU_MSG_DEPTH) {
		VASTAI_PCI_ERR(priv, die_id,
				"%s elem_cout %d overflow, set to %ld\n",
				__func__, elem_count, VASTAI_PCIE_2_CMCU_MSG_DEPTH);
		elem_count = VASTAI_PCIE_2_CMCU_MSG_DEPTH;
	}
	spin_lock(&(die->msg_ring_buf_lock));
	fifo.elem_count = elem_count;
	fifo.elem_size = sizeof(union buf_data);
	vastai_pci_mem_write_direct(priv,
			     die_index,
			     MSG_PCIE_TO_CMCU,
			     &fifo,
			     sizeof(fifo));
	VASTAI_PCI_DBG(priv, die_index,
		       "init cmcu fifo: rd[0x%x] wr[0x%x] elem_count[0x%x] elem_size[0x%x]\n",
		       fifo.rd, fifo.wr, fifo.elem_count, fifo.elem_size);
	spin_unlock(&(die->msg_ring_buf_lock));
}
#endif

/* Each function has its own vmsgq 0/1, but they
 * may share and compete the same physical msgq.
 */
int vastai_get_msgq_id(struct vastai_pci_info *priv, int type)
{
	int msgq_id = -1;
	if (type == MSGQ_DMA) {
		msgq_id = MSGQ_ID_0;
	} else if (type == MSGQ_CMD) {
		msgq_id = MSGQ_ID_1;
	} else {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s wrong msgq cmd_type=%d",
					   __func__, type);
	}
	return msgq_id;
}

int vastai_send_pcie_cmd(struct vastai_pci_info *priv, u32 die_index,
			 u32 sub_cmd, u64 paras)
{
	struct pcie_transfer_info trans = {
		.type = VASTAI_PCIE,
		.pcie_config_addr = sub_cmd,
		.data_struct_addr = paras,
	};
	int ret = 0;
	u32 die_id = vastai_pci_get_die_id(priv, die_index);
	u32 msgq_id;

	msgq_id = vastai_get_msgq_id(priv, MSGQ_CMD);
	ret = vastai_pci_vmsgq_wr(&(priv->dies[die_id].vmsgq[msgq_id]),
				 &trans, sizeof(trans));

	if (ret) {
		VASTAI_PCI_ERR(priv, die_index, "%s sub_cmd[%d] error, ret=%d",
			       __func__, sub_cmd, ret);
		return -EINVAL;
	}

	return 0;
}

int vastai_send_pcie_cmd_spe(struct vastai_pci_info *priv, u32 die_index,
					u32 sub_cmd, u64 paras)
{
	struct pcie_transfer_info trans = {
		.type = VASTAI_PCIE,
		.pcie_config_addr = sub_cmd,
		.data_struct_addr = paras,
	};
	int ret = 0;
	u32 die_id = vastai_pci_get_die_id(priv, die_index);
	u32 msgq_id;

	msgq_id = vastai_get_msgq_id(priv, MSGQ_CMD);
	ret = vastai_pci_vmsgq_wr_spe(&(priv->dies[die_id].vmsgq[msgq_id]),
				 &trans, sizeof(trans));

	if (ret) {
		VASTAI_PCI_ERR(priv, die_index, "%s sub_cmd[%d] error, ret=%d",
			       __func__, sub_cmd, ret);
		return -EINVAL;
	}

	return 0;
}

/**
 * @brief send a ctrl cmd to smcu by msgqueue1, smcu will retransmit interrupt to
 * 	other mcu / vdsp.
 *
 * @param pci_info: vastai sv100 pci base struct.
 * @param die_index: the destination die.
 * @param core_bit_map: this is a __BIT_MAP__ smcu will trigger it one by one.
 * @param info: private data. low 8bit is invalied. (for seq_id)
 * @return int: zero is success.
 */
int vastai_send_ctrl_cmd(struct vastai_pci_info *pci_info, u32 die_index,
			 u32 core_bit_map, u64 info)
{
	struct pcie_transfer_info trans = {
		.type = VASTAI_PCIE_CTRL_QUEUE,
		.pcie_config_addr = core_bit_map,
		.data_struct_addr = info,
	};
	int ret = 0;
	u32 msgq_id;
	u32 die_id = vastai_pci_get_die_id(pci_info, die_index);
	struct ctrl_cmd_list *cmd_list = &pci_info->dies[die_id].ctrl_cmd_list;
	u8 seq_id = atomic_inc_return(&cmd_list->ctrl_seq_id) & 0xFF;
	struct ctrl_cmd_entry *cmd_entry =
		vmalloc(sizeof(struct ctrl_cmd_entry));
	unsigned long irq_flag;

	if (!cmd_entry) {
		return -ENOMEM;
	}
	cmd_entry->seq_id = seq_id;
	INIT_LIST_HEAD(&cmd_entry->list);
	init_completion(&cmd_entry->msg_comp);
	cmd_entry->completion_info.smcu_to_host.val = 0;

	spin_lock_irqsave(&cmd_list->ctrl_cmd_list_lock, irq_flag);
	/* we expect the response (completion) to be orderly from fw,
	so used list_add_tail and not performance tuning */
	list_add_tail(&cmd_entry->list, &cmd_list->ctrl_cmd_head.list);
	spin_unlock_irqrestore(&cmd_list->ctrl_cmd_list_lock, irq_flag);

	trans.data_struct_addr &= ~(0xFF);
	trans.data_struct_addr |= (seq_id & (0xFF));

	msgq_id = vastai_get_msgq_id(pci_info, MSGQ_CMD);
	ret = vastai_pci_vmsgq_wr(&(pci_info->dies[die_id].vmsgq[msgq_id]), &trans,
				 sizeof(trans));
	if (ret) {
		goto FREE_CMD_ENTRY;
	}

	ret = wait_for_completion_interruptible_timeout(&cmd_entry->msg_comp,
							1000);
	if (ret == 0) {
		ret = -ETIMEDOUT;
	}

	if (ret > 0) {
		ret = cmd_entry->completion_info.ret_st.ret_val;
	}

FREE_CMD_ENTRY:
	spin_lock_irqsave(&cmd_list->ctrl_cmd_list_lock, irq_flag);
	list_del(&cmd_entry->list);
	spin_unlock_irqrestore(&cmd_list->ctrl_cmd_list_lock, irq_flag);
	vfree(cmd_entry);
	return ret;
}

/**
 * @brief recv a ctrl cmd completion from smcu, checkout ctrl cmd list, and call
 * 	completion done for it.
 *
 * @param pci_info: vastai sv100 pci base struct.
 * @param die_index: the destination die.
 * @param msg: smcu to host msg. will cpy for user thread.
 */
void vastai_ctrl_cmd_compeletion(struct vastai_pci_info *pci_info,
				 u32 die_index, struct smcu_to_host_msg1 *msg)
{
	unsigned long irq_flag;
	u32 die_id = vastai_pci_get_die_id(pci_info, die_index);
	struct ctrl_cmd_list *cmd_list = &pci_info->dies[die_id].ctrl_cmd_list;
	union completion_info_st *info_st = (union completion_info_st *)msg;
	struct ctrl_cmd_entry *entry = NULL;
	struct ctrl_cmd_entry *tmp;
	int cmd_list_len = 0;

	spin_lock_irqsave(&cmd_list->ctrl_cmd_list_lock, irq_flag);
	list_for_each_entry (tmp, &cmd_list->ctrl_cmd_head.list, list) {
		if (tmp->seq_id == info_st->ret_st.seq_id) {
			entry = tmp;
			break;
		}
		cmd_list_len++;
	}
	spin_unlock_irqrestore(&cmd_list->ctrl_cmd_list_lock, irq_flag);

	if (entry) {
		memcpy(&entry->completion_info, msg, sizeof(*msg));
		complete(&entry->msg_comp);
	}

	if (cmd_list_len > 10) {
		VASTAI_PCI_DBG(
			pci_info, die_id,
			"%s: check performance! dies->cmd_list is too long[%d].\n",
			__func__, cmd_list_len);
	}
}

void vastai_update_time(struct vastai_pci_info *priv)
{
	struct vastai_time timeinfo = { 0 };
	struct rtc_time tm;
	int i = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0)
	struct timex txc;
	/*GET UTC time*/
	do_gettimeofday(&(txc.time));
	/*convert UTC to localtime*/
	rtc_time_to_tm(txc.time.tv_sec, &tm);
#else
	time64_t sec = 0;
	sec = ktime_to_ms(ktime_get());
	rtc_time64_to_tm(sec, &tm);
#endif
	timeinfo.year = tm.tm_year + 1900;
	timeinfo.mon = tm.tm_mon + 1;
	timeinfo.day = tm.tm_mday;
	timeinfo.hour = tm.tm_hour;
	timeinfo.min = tm.tm_min;
	timeinfo.sec = tm.tm_sec;

	// u32 die_index = priv->dies[0].die_index;
	/* Update CSRAM periodically by 10 seconds */
	for (i = 0; i < priv->die_num; i++) {
		struct vastai_sv100_die *die = &(priv->dies[i]);
		int die_index = die->die_index;
		vastai_pci_mem_write(priv, die_index, CSRAM_RESEREVE1,
				     &timeinfo, sizeof(timeinfo));
	}
	msleep(10 * 1000);
}

bool is_AI_sys_core(int core)
{
	int i;
	for(i=0; i<sizeof(ai_group)/sizeof(ai_group[0]); i++)
	{
		if(ai_group[i] == core)
			return true;
	}

	return false;
}

static void vastai_heartbeat_work(struct work_struct *work)
{
	struct heartbeat_monitor *heartbeat = NULL;
	struct vastai_pci_info *priv = NULL;
	u32 cur_count[HEARTBEAT_CNT_REG_NUM] = { 0 };
	u32 *last_count = NULL;
	u32 tmp_count = 0;
	int ret = 0, die_index = 0, die_id = 0, core_id = 0;

	priv = container_of(work, struct vastai_pci_info,
			    heartbeat.heartbeat_work.work);
	heartbeat = &priv->heartbeat;

	for (die_id = 0; die_id < priv->die_num; die_id++) {
		die_index = priv->dies[die_id].die_index;
		last_count = heartbeat->dies[die_id].last_counts;

		ret = vastai_pci_mem_read(priv, die_index, HEART_BEAT_CNT,
					  cur_count, sizeof(cur_count));

		if (ret && (VASTAI_NORMAL_STATE == vastai_get_pci_state(priv))) {
			VASTAI_PCI_ERR(priv, die_id,
				       "%s: read heartbeat reg error, %d\n",
				       __func__, ret);
			//TODO: this moment peci error, how to reset dev
			break;
		}

		for (core_id = 0; core_id < HEARTBEAT_CNT_REG_NUM; core_id++) {
			if (priv->dies[die_id].heartbeat_mask & BIT_MASK(core_id)) {
				continue;
			}

			VASTAI_PCI_DBG(
				priv, die_id,
				"%s: core[%d] heartbeat cnt, last=%u cur=%u\n",
				__func__, core_id, last_count[core_id],
				cur_count[core_id]);

			if (cur_count[core_id] != last_count[core_id]) {
				continue;
			}

			/* heartbeat counter don't change, read again */
			ret = vastai_pci_mem_read(priv, die_index,
						  HEART_BEAT_CNT +
							  core_id * sizeof(u32),
						  &tmp_count, sizeof(u32));
			if (ret && (VASTAI_NORMAL_STATE == vastai_get_pci_state(priv))) {
				VASTAI_PCI_ERR(
					priv, die_id,
					"%s: read heartbeat reg error, %d\n",
					__func__, ret);
				//TODO: this moment peci error, how to reset dev
				break;
			}
			if (tmp_count != last_count[core_id]) {
				continue;
			}

			VASTAI_PCI_DBG(priv, die_id,
                                       "%s: core[%d] heartbeat stopped!!!!!\n",
                                       __func__, core_id);
			continue; //do not reset

			/* core is hanged, set status of die and core */
			set_bit(die_id, &heartbeat->dies_status);
			set_bit(core_id, &heartbeat->dies[die_id].cores_status);
		}

		memcpy(last_count, cur_count,
		       sizeof(heartbeat->dies[die_id].last_counts));
	}

	if (!(heartbeat->dies_status))
		goto next;

	/* TODO: dump */

	/**
	 * reset all dies, if one or all of dies hang
	 * TODO: multi dies case
	 */
	if (heartbeat->dies[0].cores_status & BIT_MASK(HEARTBEAT_SMCU)) {
		/* smcu hang: fw will tigger wdg, host wait romcode */
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"%s: smcu hang(dev_wdg_reset and reboot)\n",
				__func__);
		ret = vastai_device_reset(priv, VASTAI_RESET_BOOT_DEV);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s: smcu hang, reset error, %d\n",
				       __func__, ret);
			goto stop;
		}
	} else {
		/**
		 * smcu no hang: host send msg irq to triger fw wdg
		 * and wait romcode
		 */
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"%s: other cores hang(msg_reset and reboot)\n",
				__func__);
		ret = vastai_device_reset(priv, VASTAI_RESET_MSG_TRIGER |
							VASTAI_RESET_BOOT_DEV);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s: smcu not hang, reset error, %d\n",
				       __func__, ret);
			goto stop;
		}
	}

	heartbeat->dies_status = 0;
	for (die_id = 0; die_id < priv->die_num; die_id++) {
		heartbeat->dies[die_id].cores_status = 0;
		memset(heartbeat->dies[die_id].last_counts, 0,
		       sizeof(heartbeat->dies[die_id].last_counts));
	}

	// vastai_pci_resource_recycling(work);
next:
	schedule_delayed_work(&heartbeat->heartbeat_work,
			      heartbeat->timeout_threshold);
	return;

stop:
	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
		       "heartbeat_work stop, dies_status=%lx\n",
		       heartbeat->dies_status);
	for (die_id = 0; die_id < priv->die_num; die_id++) {
		VASTAI_PCI_ERR(priv, die_id, "cores_heartbeat_status=%lx\n",
			       heartbeat->dies[die_id].cores_status);
	}
}

#if 0
static int vastai_dev_reset_irq_handler(void *pci_dev, u32 die_index, int irq,
					void *ctxt, void *out)
{
	struct vastai_pci_info *priv = pci_dev;
	u32 die_irq = irq % VASTAI_PCIE_MSIX_NUM_PER_DIE;
	u32 die_id = irq / VASTAI_PCIE_MSIX_NUM_PER_DIE;

	if (die_irq == SMCU_2_HOST_INT0) {
		/* dma done: wait romcode done */
		VASTAI_PCI_ERR(priv, die_id,
			       "%s: TODO irq[%d]\n", __func__, die_irq);
	} else if (die_irq == SMCU_2_HOST_INT1) {
		/* dma no done: pcie hot reset */
		VASTAI_PCI_ERR(priv, die_id,
			       "%s: TODO irq[%d]\n", __func__, die_irq);
	} else {
		VASTAI_PCI_ERR(priv, die_id,
			       "%s: irq[%d] is invalid\n", __func__, die_irq);
	}

	return 0;
}
#endif

static void vastai_heartbeat_monitor_init(struct vastai_pci_info *priv)
{
	struct heartbeat_monitor *heartbeat = &priv->heartbeat;
	//int ret;
	//int i;
	//int index = 0;
	//int dev_reset_irqs[] = {SMCU_2_HOST_INT0, SMCU_2_HOST_INT1};

	memset(heartbeat, 0, sizeof(*heartbeat));

	heartbeat->timeout_threshold = msecs_to_jiffies(3000); /* 3s */

#if 0
	/**
	 * when fw reset from spi boot, dev re-initail procedure:
	 * 1. check dma done
	 * 2. if dma done send irq SMCU_2_HOST_INI0 to host
	 * 3. if dma no done send irq SMCU_2_HOST_INI1 to host
	 */
	for (i = 0; i < priv->die_num; i++) {
		for (index = 0; index < ARRAY_SIZE(dev_reset_irqs); index++) {
			ret = vastai_irq_register(priv,
						  priv->dies[i].die_index,
						  dev_reset_irqs[index],
						  vastai_dev_reset_irq_handler,
						  NULL);
			if (ret)
				VASTAI_PCI_ERR(priv, i,
					       "%s: irq[%d] register error\n",
					       __func__, dev_reset_irqs[index]);
		}
	}
#endif

	INIT_DELAYED_WORK(&heartbeat->heartbeat_work, vastai_heartbeat_work);
	schedule_delayed_work(&heartbeat->heartbeat_work,
			      heartbeat->timeout_threshold);
}

static void vastai_heartbeart_monitor_deinit(struct vastai_pci_info *priv)
{
	int dev_reset_irqs[] = { SMCU_2_HOST_INT0, SMCU_2_HOST_INT1 };
	int ret = 0, i = 0, index = 0;

	cancel_delayed_work_sync(&priv->heartbeat.heartbeat_work);

	for (i = 0; i < priv->die_num; i++) {
		for (index = 0; index < ARRAY_SIZE(dev_reset_irqs); index++) {
			ret = vastai_irq_unregister(priv,
						    priv->dies[i].die_index,
						    dev_reset_irqs[index]);
			if (ret)
				VASTAI_PCI_ERR(priv, i,
					       "%s: irq[%d] unregister error\n",
					       __func__, dev_reset_irqs[index]);
		}
	}
}

int vastai_bbox_init(struct vastai_pci_info *priv)
{

	spin_lock_init(&(priv->bbox_info.bbox_lock));
	memset(&(priv->bbox_info.val), 0, sizeof(priv->bbox_info.val));

	return 0;
}


int vastai_bbox_set(struct vastai_pci_info *priv, u32 bbox)
{
	unsigned long flags;
	int ret = 0;
	int die_id;

	spin_lock_irqsave(&priv->bbox_info.bbox_lock, flags);
	priv->bbox_info.val = bbox;
	for (die_id = 0; die_id < priv->die_num; die_id++) {
		ret |= vastai_send_pcie_cmd(priv, priv->dies[die_id].die_index,
					     VASTAI_PCIE_SUB_SET_BBOX, (u64)bbox);
	}
	spin_unlock_irqrestore(&priv->bbox_info.bbox_lock, flags);

	return ret;
}

u32 vastai_bbox_get(struct vastai_pci_info *priv)
{
	return priv->bbox_info.val;
}


static void vastai_pci_get_ddr_attr(struct vastai_pci_info *priv)
{
	unsigned int die_index = priv->dies[0].die_index;
	union mcu_dsp_status reg = { 0 };
	u64 reg_addr = VASTAI_MCU_DSP_STATUS_REG;

	vastai_pci_mem_read(priv, die_index, reg_addr, &(reg.val), 4);
#ifdef CONFIG_VASTAI_AI_EMU
	/* fictitious ddr on emulation */
	priv->ddr_attr = 0x200000000;
#else
	if (reg.bit.ddr_attr == 0) {
		/* ddr dram density: 2GB, mc_num: 2, total density: 4GB */
		priv->ddr_attr = 0x100000000;
	} else if (reg.bit.ddr_attr == 1) {
		/* ddr dram density: 2GB, mc_num: 4, total density: 8GB */
		priv->ddr_attr = 0x200000000;
	} else if (reg.bit.ddr_attr == 2) {
		/* ddr dram density: 4GB, mc_num: 2, total density: 8GB */
		priv->ddr_attr = 0x200000000;
	} else if (reg.bit.ddr_attr == 3) {
		/* ddr dram density: 4GB, mc_num: 4, total density: 16GB */
		priv->ddr_attr = 0x400000000;
	} else {
		/* ddr dram density: 2GB, mc_num: 2, total density: 4GB */
		priv->ddr_attr = 0x100000000;
	}
#endif

	/* 1 MB is reserved for buffering because the video ENC may have DDR access overbounds. */
	priv->ddr_attr -= 0x100000;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s=0x%llx (%lldGB), val=0x%x\n",
			__func__, priv->ddr_attr,
			(priv->ddr_attr / 1024 / 1024 / 1024),
			reg.bit.ddr_attr);
}

/* 这个函数用来初始化与fw强相关的变量。它会在fw被reboot时调用 */
void vastai_pci_init_all_die(struct vastai_pci_info *priv)
{
	int i = 0;

	for (i = 0; i < priv->die_num; i++) {
		struct vastai_sv100_die *die = &(priv->dies[i]);

		vastai_die_init(die, priv, i);
	}
}

union core_bitmap core_point_to_core_bitmap(u8 core_point_num)
{
	union core_bitmap _core_bitmap;

	_core_bitmap.val = 0;
	if(core_point_num == CORE_POINT_CMCU)
		_core_bitmap.cmcu = 1;
	else if ((core_point_num >= CORE_POINT_VDSP0) && (core_point_num <= CORE_POINT_VDSP3))
		_core_bitmap.vdsp = 1<<(core_point_num-CORE_POINT_VDSP0);
	else if ((core_point_num >= CORE_POINT_VEMCU0) && (core_point_num <= CORE_POINT_VEMCU3))
		_core_bitmap.encode = 1<<(core_point_num-CORE_POINT_VEMCU0);
	else if ((core_point_num >= CORE_POINT_VDMCU0) && (core_point_num <= CORE_POINT_VDMCU2))
		_core_bitmap.decode = 1<<(core_point_num-CORE_POINT_VDMCU0);

	return _core_bitmap;
}

void vastai_clear_fifo(void *pci_info, int die_index, union core_bitmap core_id)
{
	struct vastai_fifo fifo;
	 struct msg_addr_entry *msg_addr_table = get_msg_addr_entry();

	while (core_id.val) {
		int core_index = ffs(core_id.val) - 1;
		const u64 msg_addr = msg_addr_table[core_index].msg_addr;
		u32 is_use_ring_buf = msg_addr_table[core_index].is_use_ring_buf;

		core_id.val &= ~(BIT(core_index));
		if (is_use_ring_buf) {
			struct vastai_pci_info *priv = pci_info;
			u32 die_id = vastai_pci_get_die_id(pci_info, die_index);
			struct vastai_sv100_die *die = &(priv->dies[die_id]);

			spin_lock(&(die->msg_ring_buf_lock));

			vastai_pci_mem_read(pci_info, die_index, msg_addr, &fifo, sizeof(struct vastai_fifo));
			spin_unlock(&(die->msg_ring_buf_lock));
			if(vastai_fifo_is_empty(&fifo)) {
				return;
			}

			spin_lock(&(die->msg_ring_buf_lock));
			msleep(1500);
			vastai_pci_mem_read(pci_info, die_index, msg_addr, &fifo, sizeof(struct vastai_fifo));
			fifo.wr = fifo.rd;
			vastai_pci_mem_write(pci_info, die_index, msg_addr + offsetof(struct vastai_fifo, wr),
					     &(fifo.wr), sizeof(fifo.wr));

			spin_unlock(&(die->msg_ring_buf_lock));
		}
	}

	return;
}

void exception_proc_work(struct work_struct *work)
{
	struct vastai_sv100_die *dies =
		container_of(work, struct vastai_sv100_die , exception_work);
	struct vastai_pci_info *priv = dies->pci_info;
	u32 die_id = vastai_pci_get_die_id(priv, dies->die_index);
	u64 para = 0;
	union core_bitmap core_id;

	//TODO: lock/unlock
	while (dies->core_status_bitmap) {
		para = ffs(dies->core_status_bitmap) - 1;
		core_id.val = 0;
		core_id = core_point_to_core_bitmap((u8)para);
		VASTAI_PCI_INFO(priv, die_id, "receive core exception index:%u\n", (u32)para);
		if (core_exception_en) {
			vastai_clear_fifo(priv, priv->dies[die_id].die_index, core_id);
			vastai_send_pcie_cmd(priv, priv->dies[die_id].die_index,
					VASTAI_PCIE_SUB_CORE_RESET, para);
		}
		dies->core_status_bitmap &= ~(1 << (u32)para);
	}

}

void vastai_exception_init(struct vastai_pci_info *priv)
{
	int i;

	for (i = 0; i < priv->die_num; i++) {
		priv->dies[i].heartbeat_mask = 0xC0000000;
		if(priv->vadev.dev_cap == VASTAI_HW_TYPE_VIDEO_ONLY) {
			int ai_loop;
			for(ai_loop = 0; ai_loop < sizeof(ai_group)/sizeof(ai_group[0]); ai_loop++)
				priv->dies[i].heartbeat_mask |= (1<<ai_group[ai_loop]);
		}
		if(priv->vadev.dev_cap == VASTAI_HW_TYPE_DEFAULT ||
			priv->vadev.dev_cap == VASTAI_HW_TYPE_AI_ONLY) {
			int video_loop;
			for(video_loop = HEARTBEAT_VDMCU0; video_loop <= HEARTBEAT_VEMCU3; video_loop++)
				priv->dies[i].heartbeat_mask |= (1<<video_loop);
		}

		INIT_WORK(&priv->dies[i].exception_work, exception_proc_work);
	}
}

int vastai_wait_ecc_2_bit(struct vastai_pci_info *pdev,
			  union die_index_data die_index)
{
	struct vastai_sv100_die *die;
	int ret;

	if (!pdev || (pdev->die_num <= die_index.die_id)) {
		return -EINVAL;
	}

	die = &(pdev->dies[die_index.die_id]);

	ret = wait_for_completion_interruptible(&(die->ecc_2bit_err));
	reinit_completion(&(die->ecc_2bit_err));

	return ret;
}

int vastai_get_render_id(struct vastai_pci_info *pdev, u32 die_index)
{
	union die_index_data die_index_u = { .val = die_index };

	u32 die_id = die_index_u.die_id;

	return pdev->dies[die_id].render_id;
}

int vastai_get_vacc_id(struct vastai_pci_info *pdev, u32 die_index)
{
	union die_index_data die_index_u = { .val = die_index };

	return die_index_u.seq_num;
}

/* 1. if vastai_pci_info_list is empty, return 0;
   2. if vastai_pci_info_list contain pdev, return all pre-entry die_num's sum.
   3. if vastai_pci_info_list is no empty, and not contain pdev, return all entry die_num's sum
   **BE CAREFULL!** If we probe with work queue, this function will be a bug.
   */
int vastai_get_seqnum(struct vastai_pci_info *pdev, unsigned int die_id)
{
	struct vastai_pci_info *pos;
	int seq_num = 0;

	if (pdev == NULL || die_id >= pdev->die_num)
		return -EINVAL;

	list_for_each_entry (pos, &vastai_pci_info_list, dev_list) {
		if (pos != pdev) {
			seq_num += pos->die_num;
		} else {
			break;
		}
	}

	return seq_num + die_id;
}
static int vastai_pci_get_dlc_cores_info( struct vastai_pci_info *pcie_info , u32 die_index){
	int ret;
	/* send get dlc cores info at cmd to smcu */
	u32 die_id = vastai_pci_get_die_id(pcie_info , die_index);
	mutex_lock(&(pcie_info->dies[die_id].get_dlc_mutex));
	ret = vastai_send_pcie_cmd(pcie_info, die_index,
		VASTAI_PCIE_SUB_GET_DLC_CORE, 0);
	if (ret < 0){
		VASTAI_PCI_ERR(pcie_info, DUMMY_DIE_ID,
			"%s send cmd fail!\n",
			__func__);
		mutex_unlock(&(pcie_info->dies[die_id].get_dlc_mutex));
		return ret;
	}

	/* wait completion int from smcu */
	if (wait_for_completion_timeout(
		&(pcie_info->dies[die_id].get_dlc_core_done),
		msecs_to_jiffies(
		VASTAI_PCIE_WAIT_SMCU_MSG1_TIME)) <= 0) {

		VASTAI_PCI_ERR(pcie_info, DUMMY_DIE_ID,
			"%s timeout!\n",
			__func__);
		mutex_unlock(&(pcie_info->dies[die_id].get_dlc_mutex));
		return -ETIMEDOUT;
	}
	reinit_completion(&(pcie_info->dies[die_id].get_dlc_core_done));
	mutex_unlock(&(pcie_info->dies[die_id].get_dlc_mutex));

	return ret;
}

static u32 vastai_get_harvest_core_flag(u16 dlc_core){

	u32 harvest_core_flag;
	u32 offset_big_core_bit = 0;
	u32 offset_small_core_bit = 0;

	offset_small_core_bit = (0x01 << ((dlc_core>>9)&0x7)) | (0x01<< ((dlc_core>>6)&0x7));
	offset_big_core_bit = (0x00010000  << ((dlc_core>>3)&0x7)) | (0x00010000  << ((dlc_core)&0x7));
	harvest_core_flag =  offset_big_core_bit + offset_small_core_bit;
	return harvest_core_flag;

}

u32 vastai_get_dlc_cores_info(struct vastai_pci_info *pcie_info, u32 die_index){

	int ret;
	u8 type;
	u16 dlc_core;
	u32 harvest_core_flag;
	u32 die_id;
	/*send get dlc cores info at cmd to smcu */
	die_id = vastai_pci_get_die_id(pcie_info , die_index);
	ret = vastai_pci_get_dlc_cores_info(pcie_info, die_index);
	if( ret < 0 ){
		goto DEFAULT;
	}

	type = pcie_info->harvest_info[die_id].type;
	dlc_core = pcie_info->harvest_info[die_id].dlc_cores;

	if(type != 1 && type != 4){

		VASTAI_PCI_INFO(pcie_info, DUMMY_DIE_ID,
			"%s it is not harvest type\n",__func__);
DEFAULT:
		harvest_core_flag = 0x00FF00FF;
		return harvest_core_flag;
	}

	harvest_core_flag = vastai_get_harvest_core_flag(dlc_core);
	return harvest_core_flag;
}


/*
 * if we pcie goto error, timer will stop.
 */
static void vastai_pci_timer(struct timer_list *timer)
{
	struct vastai_pci_info *priv = from_timer(priv, timer, pci_timer);
	int i;
	int ret = 0;

	/* timer should be trigger in next 10s */
	timer->expires = jiffies + 10 * HZ;

	if (atomic_read(&(priv->pci_state)) != VASTAI_NORMAL_STATE) {
		/* trigger timer */
		add_timer(timer);
		return;
	}

	/* TODO: refactor this into die */
	for (i = 0; i < priv->die_num; i++) {
		ret = vastai_send_pcie_cmd(priv, priv->dies[i].die_index,
					VASTAI_PCIE_SUB_TIME_SYNC,
					vastai_get_host_time_ns()/1000000);
		if (ret) {
#if 0
			atomic_set(&(priv->pci_state), VASTAI_ERROR_STATE);
#else
			// vastai_pci_state_sync_error(priv);
#endif
			break;
		}
	}

	/* trigger timer */
	add_timer(timer);
}

void vastai_pci_timer_init(struct vastai_pci_info *pci_info)
{
	timer_setup(&(pci_info->pci_timer), vastai_pci_timer, 0);
	vastai_pci_timer(&(pci_info->pci_timer));
}

static int vastai_pci_legacy_irq_init(struct vastai_pci_info *priv, struct pci_dev *pdev)
{
	int ret = 0;
	int i = 0;
	const char *pname;
	priv->irq = pdev->irq;

	pname = devm_kasprintf(&(priv->dev->dev), GFP_KERNEL,
			       VASTAI_PCI_DRIVER_PREFIX ":%s %d",
			       VASTAI_GET_PCI_NAME(priv), priv->irq);
	/* legacy irq will not init msix array
	 * Manual initalization */
	priv->msix[0].entry = 0;
	priv->msix[0].vector = 0;
	ret = request_irq(
		priv->irq, (irq_handler_t)(&vastai_pci_legacy_irq),
		IRQF_NO_SUSPEND | IRQF_NO_THREAD | IRQF_PERCPU, pname,
		(void *)&(priv->msix[i]));
	if (ret)
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s request_irq(%d), error %d\n",
			       __func__, priv->irq, ret);
	else
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
			       "%s request_irq(%d) ok\n", __func__,
			       priv->irq);

	return ret;
}

static int vastai_pci_msi_irq_init(struct vastai_pci_info *priv, struct pci_dev *pdev)
{
	int ret = 0;
	int i = 0;
	priv->irq_num = pci_msi_vec_count(pdev);
	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
		       "pci_msi_vec_count irq_num=%d\n",
		       priv->irq_num);

	ret = pci_alloc_irq_vectors(pdev, 1, priv->irq_num,
				    PCI_IRQ_MSI);
	if (ret > 0) {
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
			       "pci_enable_msi_range ret=%d\n", ret);
		priv->irq_num = ret;
	} else {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_enable_msi_range err=%d\n", ret);
		priv->irq_type = VASTAI_INVALID_IRQ;
	}
	priv->irq = pdev->irq;

	for (i = 0; i < priv->irq_num; i++) {
		const char *pname =
			devm_kasprintf(&(priv->dev->dev), GFP_KERNEL,
				       VASTAI_PCI_DRIVER_PREFIX
				       ":%s %d",
				       VASTAI_GET_PCI_NAME(priv), i);
		/* msi irq will not init msix array
		 * Manual initalization */
		priv->msix[i].entry = i;
		priv->msix[i].vector = priv->irq + i;
		ret = request_irq(priv->irq + i,
				  (irq_handler_t)(&vastai_pci_msi_irq),
				  IRQF_SHARED, pname, (void *)&(priv->msix[i]));
		// IRQF_SHARED is better replaced by 0
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s request_irq(%d), error %d\n",
				       __func__, priv->irq + i, ret);
			break;
		}
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
			       "%s request_irq(%d) ok\n", __func__,
			       priv->irq + i);
	}
	// if (i == priv->irq_num)
	//	priv->irq_en = 1;

	return ret;
}

static int vastai_pci_msix_irq_init(struct vastai_pci_info *priv, struct pci_dev *pdev)
{
	int ret = 0;
	int i = 0;

	for (i = 0; i < VASTAI_PCIE_MSIX_NUM; i++) {
		priv->msix[i].entry = i;
		priv->msix[i].vector = 0;
	}

	priv->irq_num = pci_enable_msix_range(pdev, priv->msix, 1,
					      VASTAI_PCIE_MSIX_NUM);
	if (priv->irq_num > 0) {
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
			       "pci_enable_msix_range %d ok\n",
			       priv->irq_num);
	} else {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_enable_msix_range %d err\n",
			       priv->irq_num);
		priv->irq_type = VASTAI_INVALID_IRQ;
	}
	priv->irq = priv->msix[0].vector;

	for (i = 0; i < priv->irq_num; i++) {
		const char *pname =
			devm_kasprintf(&(priv->dev->dev), GFP_KERNEL,
				       VASTAI_PCI_DRIVER_PREFIX
				       ":%s %d",
				       VASTAI_GET_PCI_NAME(priv), i);
		ret = request_irq(priv->msix[i].vector,
				  (irq_handler_t)(&vastai_pci_msi_irq),
				  IRQF_SHARED, pname, (void *)&(priv->msix[i]));
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s request_irq(%d), error %d\n",
				       __func__, priv->msix[i].vector,
				       ret);
			priv->irq_num = i;
			vastai_pci_disable_irq(priv);
			return ret;
		}

		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
			       "%s request_irq(%d) ok\n", __func__,
			       priv->msix[i].vector);
	}

	return ret;
}

static int vastai_pci_irq_init(struct vastai_pci_info *priv, struct pci_dev *pdev)
{
	int ret = 0;

	priv->irq_type = VASTAI_MSIX_IRQ;
	priv->irq = pdev->irq;

	if (priv->irq_type == VASTAI_LEGACY_IRQ) {
		ret = vastai_pci_legacy_irq_init(priv, pdev);
	}

	if (priv->irq_type == VASTAI_MSI_IRQ) {
		ret = vastai_pci_msi_irq_init(priv, pdev);
	}

	if (priv->irq_type == VASTAI_MSIX_IRQ) {
		ret = vastai_pci_msix_irq_init(priv, pdev);
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"irq_type %d, dev->irq %d(+%d)\n",
			priv->irq_type,
			priv->irq, priv->irq_num);

	return ret;
}

void vastai_check_common_include(struct vastai_pci_info *pci_info)
{
	char *find_ci_hash = 0;
	find_ci_hash = strstr(pci_info->dies[0].fw_ver.smcu_ver, "common_include: ");
	if(find_ci_hash == NULL) {
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "old fw, do not have common include hash id\n");
		return;
	}
	find_ci_hash += strlen("common_include: ");

	if(strcmp(find_ci_hash, VASTAI_PCI_COMMON_INCLUDE_COMMIT_HASH))
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[vawarn] FW common [%s] and driver common [%s] mismatch\n",
			find_ci_hash, VASTAI_PCI_COMMON_INCLUDE_COMMIT_HASH);
	else
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "Fw common [%s] match\n",
				VASTAI_PCI_COMMON_INCLUDE_COMMIT_HASH);
}

static int vastai_pci_probe(struct pci_dev *pdev,
			    const struct pci_device_id *id)
{
	struct vastai_pci_info *priv = NULL;
	int ret = 0, i = 0;

	VASTAI_PCI_FUNC_ENTERY;

#if 0
	if (pdev->devfn != 0) {
		/* We only claim function 0 */
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "We only claim func0, now func=%d\n", pdev->devfn);
		return -1;
	}
#endif

	priv = kzalloc(sizeof(struct vastai_pci_info), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	vastai_pci_state_reset_done(priv);

	spin_lock_init(&priv->trans_lock);
	spin_lock_init(&priv->msgq_1_lock);
	init_completion(&priv->set_bar_done);

	init_completion(&priv->activate_vf_done);
	init_completion(&priv->deactivate_vf_done);

	if (id->device == 0x100) {
		priv->is_physfn = 1;
		VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID,
				"This is a PF Device, device id: 0x%x\n",
				id->device);
	} else if (id->device == 0x101) {
		/*
		 * We only install vf driver in vm.
		 * Remove this condition if you want to install vf driver in host
		 */
		if (pdev->is_virtfn) {
			kfree(priv);
			return -ENXIO;
		}
		priv->is_virtfn = 1;
		VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID,
				"This is a VF Device, device id: 0x%x\n",
				id->device);
	} else {
		VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID,
				"This is a Invalid Device, device id: 0x%x\n",
				id->device);
		kfree(priv);
		return -ENXIO;
	}

	for (i = 0; i < ARRAY_SIZE(priv->dies); i++) {
		init_completion(&(priv->dies[i].log_comp));
		init_completion(&(priv->dies[i].ecc_2bit_err));
		init_completion(&(priv->dies[i].rw_xspi));
		init_completion(&(priv->dies[i].smi_ack[0]));
		init_completion(&(priv->dies[i].smi_ack[1]));
		init_completion(&(priv->dies[i].core_time_sync_ack));
	}

	INIT_LIST_HEAD(&(priv->vacc_dev_head));
	if (list_empty(&(vastai_pci_info_list))) {
		/* this is first device */
		priv->dev_id = 0;
	} else {
		struct vastai_pci_info *last = NULL;
		last = list_last_entry(&(vastai_pci_info_list),
				       struct vastai_pci_info, dev_list);
		priv->dev_id = last->dev_id + 1;
	}

	/* we assume only one die in this device first, and get the real die num at boot */
	priv->die_num = 1; /* this member will modify at vastai_pci_boot */

	priv->dev = pdev;
	pci_set_drvdata(pdev, priv);
	/*disable ats capbility*
	*kernel 4.14 not support untrusted property*
	*grub add "intel_iommu=strict" can disable ats*/
	// pdev->untrusted = 1;

	if (pci_enable_device(pdev) < 0) {
		ret = -EIO;
		goto ERROR_FREE_PRIV;
	}

	for (i = 0; i < VASTAI_PCI_BAR_NUM; i++) {
		int flag = pci_resource_flags(pdev, i);
		if (!(flag & IORESOURCE_MEM))
			continue;

		priv->bar[i].mmio_start = pci_resource_start(pdev, i);
		priv->bar[i].mmio_end = pci_resource_end(pdev, i);
		priv->bar[i].mmio_flags = pci_resource_flags(pdev, i);
		priv->bar[i].mmio_len = pci_resource_len(pdev, i);
		priv->bar[i].mem =
			ioremap(priv->bar[i].mmio_start, priv->bar[i].mmio_len);
		priv->bar[i].vmem = priv->bar[i].mem;
		if (priv->bar[i].vmem == NULL) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s:cannot remap bar%d mmio, aborting\n",
				       pci_name(pdev), i);
			ret = -EIO;
			goto ERROR_DISABLE_PCIE;
		}
		vastai_pci_init_bar_at(priv, i);
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"BAR(%d) mmio: 0x%016lx ~ 0x%016lx\n", i,
				(unsigned long)priv->bar[i].mmio_start,
				(unsigned long)priv->bar[i].mmio_end);
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"BAR(%d) vmem: 0x%016lx ~ 0x%016lx\n", i,
				(unsigned long)priv->bar[i].vmem,
				(unsigned long)priv->bar[i].vmem +
					(unsigned long)priv->bar[i].mmio_len -
					1);
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"BAR(%d)  axi: 0x%016lx ~ 0x%016llx\n", i,
				priv->bar[i].at_addr,
				priv->bar[i].at_addr + priv->bar[i].mmio_len -
					1);
		VASTAI_PCI_INFO(
			priv, DUMMY_DIE_ID, "BAR(%d) size: 0x%016lx ~ %17ldM\n",
			i, (unsigned long)priv->bar[i].mmio_len,
			(unsigned long)priv->bar[i].mmio_len / 1024 / 1024);
		priv->bar_num += 1;
	}

	if (priv->bar_num < 4) {
		VASTAI_PCI_ERR(
			priv, DUMMY_DIE_ID,
			"%s bar_num[%d] is less than 4. Pls check bar configuration.\n",
			__FUNCTION__, priv->bar_num);
		ret = -EIO;
		goto ERROR_DISABLE_PCIE;
	}

	ret = vastai_dmabuf_init(priv);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "Failed to init vastai dmabuf. errno %d\n", ret);

		goto ERROR_DISABLE_PCIE;
	}
#ifdef ENABLE_MEMPOOL_EX
	ret = vastai_mempool_ex_init(priv);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "Failed to init mempool. errno %d\n", ret);
	}
#endif
	/*
	 * Enabling the device will enable any BARs,
	 * device memory is now accessible.
	 */
	if (pci_request_regions(pdev, "vastai_pci_basic") < 0) {
		ret = -1;
		goto ERROR_DEINIT_DMABUF;
	}

	ret = vastai_pci_irq_init(priv, pdev);
	if(ret) {
		goto ERROR_RELEASE_REGIONS;
	}

	list_add_tail(&(priv->dev_list), &(vastai_pci_info_list));
#if 0
	/* change to GEN4 */
	ret = vastai_pci_generation_sel(priv, VASTAI_PCI_GEN4);
	if (ret < 0) {
		goto ERROR_RELEASE_REGIONS;
	}
#endif
	/* Enable AER */
	pci_enable_pcie_error_reporting(pdev);
	/* Enable bus mastering (we don't use it though) */
	pci_set_master(pdev);

	/* we create kchar before boot fw,
	 * since we want to keep a debug window to see register that boot is not solid.
	 * For same purpose, we will not destroy pcie_info, after we boot failed
	 */
	vastai_pci_ioctl_init(priv); // NTD

#ifdef CONFIG_VASTAI_HW_TYPE
	priv->vadev.dev_cap = CONFIG_VASTAI_HW_TYPE;
#else
	priv->vadev.dev_cap = VASTAI_HW_TYPE_DEFAULT;
#endif
	pci_save_state(priv->dev);

#ifdef CONFIG_VASTAI_PCI_BOOT
	INIT_DELAYED_WORK(&priv->boot_wq, vastai_pci_boot_wq);
#if 0
	schedule_delayed_work(&priv->boot_wq,
			      msecs_to_jiffies(VASTAI_BOOT_WQ_DELAYED_MS));
#else
	if (priv->is_virtfn) {
		ret = vastai_pci_vf_init(priv);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s:cannot do vf init %d\n",
						pci_name(pdev), ret);
			goto ERROR_BOOT_FAIL;
		}
	} else {
		ret = vastai_pci_boot(priv);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s:cannot do fw boot %d\n",
				       pci_name(pdev), ret);
			goto ERROR_BOOT_FAIL;
		}
	}
#endif
#endif
	/* get ddr attributes from die0 */
	vastai_pci_get_ddr_attr(priv);

	/* get fw version */
	vastai_pci_get_fw_ver(priv, 0xFFF);
	vastai_check_common_include(priv);

	ret = dmi_set_device(priv);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s:flash dmi is fail %d\n",
			       pci_name(pdev), ret);
	}

	if(priv->is_physfn) vastai_pci_timer_init(priv);

	vastai_bbox_init(priv);

	vastai_exception_init(priv);
	/* *********************************
	 * if ai and video register error,
	 * Pcie driver should not return error,
	 * print a msg is enough for debug
	 * *********************************/
#ifdef CONFIG_VASTAI_AI
	for (i = 0; i < priv->die_num; i++) {
		ret = vacc_init(priv, priv->dies[i].die_index);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s:cannot do vacc_init %d\n",
				       pci_name(pdev), ret);
		}
	}
#endif
#ifdef CONFIG_VASTAI_VIDEO
	INIT_LIST_HEAD(&priv->render_head);
	for (i = 0; i < priv->die_num; i++) {
		ret = vastai_video_init(priv, priv->dies[i].die_index,
				  &(priv->dies[i].render_id));
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s:cannot do vastai_video_init %d\n",
				       pci_name(pdev), ret);
		}
	}

#endif

#ifdef CONFIG_VASTAI_LOGSYS
	/* logsys probe init */
	ret = vastai_logsys_probe_init(priv);
	if (ret != 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "vastai_logsys_probe_init fail ret[%d]\n", ret);
	}
#endif

	/**
	 * heartbeat monitor init, init at the end
	 */
#if VASTAI_HEARTBEAT_ENABLE

	if (priv->is_physfn) vastai_heartbeat_monitor_init(priv);
#else
	(void)vastai_heartbeat_monitor_init;
#endif
	/* *********************************
	 * if vatools register error,
	 * Pcie driver should not return error,
	 * print a msg is enough for debug
	 * *********************************/
#ifdef CONFIG_VASTAI_TOOLS_SUITE
#ifdef CONFIG_TOOLS_V2
	mutex_init(&priv->tools_dev_mutex);
#endif
	ret = tools_init(priv);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s:cannot do tools_init %d\n", pci_name(pdev),
			       ret);
	}
#endif

#ifdef CONFIG_VASTAI_PCI_LINK_MONITOR
	if (priv->is_physfn) vastai_pci_link_monitor_init(priv);
#else
	(void)vastai_pci_link_monitor_init;
#endif

	if (priv->is_physfn && vf_num > 0) {
		if (pdev->driver->sriov_configure)
			pdev->driver->sriov_configure(pdev, vf_num);
	}

	VASTAI_PCI_FUNC_EXIT;
	//VASTAI_PCI_FUNC_EXIT;
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "%s done!\n", __func__);

	return 0;

ERROR_BOOT_FAIL:
#ifdef CONFIG_VASTAI_PCI_BOOT
	cancel_delayed_work_sync(&priv->boot_wq);
#endif
	vastai_pci_ioctl_deinit(priv);
	pci_disable_pcie_error_reporting(pdev);
	list_del(&(priv->dev_list));
#if 0
	atomic_set(&(priv->pci_state), VASTAI_EXIT_STATE);
#else
	vastai_pci_state_power_off(priv);
#endif
	vastai_pci_disable_irq(priv);
ERROR_RELEASE_REGIONS:
	pci_release_regions(pdev);
ERROR_DEINIT_DMABUF:
	vastai_dmabuf_deinit(priv);
ERROR_DISABLE_PCIE:
	pci_disable_device(pdev);
ERROR_FREE_PRIV:
	pci_set_drvdata(pdev, NULL);
	kfree(priv);
	return ret;
}

static void vastai_pci_remove(struct pci_dev *pdev)
{
	struct vastai_pci_info *priv = pci_get_drvdata(pdev);
	int i = 0;

	VASTAI_PCI_FUNC_ENTERY;

	if (priv->is_physfn && pci_num_vf(pdev) > 0) {
		if (pdev->driver->sriov_configure)
			pdev->driver->sriov_configure(pdev, 0);
	}

	atomic_set(&(priv->pci_state), VASTAI_EXIT_STATE);
	if (!priv) {
		VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID, "%s priv is NULL\n",
			       __func__);
		return;
	}
	if (priv->is_physfn)
	{
		del_timer_sync(&(priv->pci_timer));

#if 0
		if (pdev->devfn != 0)
			return;
#endif
	}
#ifdef CONFIG_VASTAI_PCI_LINK_MONITOR
	if (priv->is_physfn) vastai_pci_link_monitor_uninit(priv);
#else
	(void)vastai_pci_link_monitor_uninit;
#endif

#if VASTAI_HEARTBEAT_ENABLE
	if (priv->is_physfn) vastai_heartbeart_monitor_deinit(priv);
#else
	(void)vastai_heartbeart_monitor_deinit;
#endif

#ifdef CONFIG_VASTAI_LOGSYS
	vastai_logsys_remove_deinit(priv);
#endif

#ifdef CONFIG_VASTAI_AI
	vacc_exit(priv);
#endif

#ifdef CONFIG_VASTAI_TOOLS_SUITE
	tools_exit(priv);
#endif

#ifdef CONFIG_VASTAI_VIDEO
	vastai_render_file_deinit(priv);
#endif
	vastai_pci_ioctl_deinit(priv);

#if 0
	if (pdev->irq)
		free_irq(pdev->irq, pdev);

	if (msi_enabled)
		pci_disable_msi(pdev);
#endif

	vastai_pci_disable_irq(priv);
	pci_disable_pcie_error_reporting(pdev);

	/* Not care this cmd is success or failed.
	 * Only check it when next insmod */
	if (vastai_pci_reset_out && priv->is_physfn)
		vastai_device_send_reset(priv);
	if (get_vastai_pci_device_number() == 1 && vastai_pci_reset_out)
		ssleep(VASTAI_WAITING_RESET);

#ifdef CONFIG_VASTAI_PCI_BOOT
	cancel_delayed_work_sync(&priv->boot_wq);
#endif
	vastai_dmabuf_deinit(priv);
#ifdef ENABLE_MEMPOOL_EX
	vastai_mempool_ex_deinit(priv);
#endif
	for (i = 0; i < priv->die_num; i++) {
		vastai_die_deinit(&(priv->dies[i]));
	}
	pci_release_regions(pdev);
	pci_disable_device(pdev);

	for (i = 0; i < VASTAI_PCI_BAR_NUM; i++) {
		int flag = pci_resource_flags(pdev, i);
		if (!(flag & IORESOURCE_MEM))
			continue;
		if (priv->bar[i].vmem != NULL) {
			iounmap(priv->bar[i].vmem);
		}
	}

	list_del(&(priv->dev_list));
	kfree(priv);
	//VASTAI_PCI_FUNC_EXIT;
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "%s done!\n", __func__);
}

#if 0
static void vastai_pci_shutdown(struct pci_dev *pdev)
{
	VASTAI_PCI_FUNC_ENTERY;

	pci_disable_device(pdev);

	pci_set_power_state(pdev, pci_choose_state(pdev, PMSG_SUSPEND));
	VASTAI_PCI_FUNC_EXIT;
}
#endif

#if KERNEL_VERSION(5, 9, 0) <= LINUX_VERSION_CODE
/* 2020-08-07 18:48:15 -0700
 * Merge tag 'pci-v5.9-changes' of
 * git://git.kernel.org/pub/scm/linux/kernel/git/helgaas/pci
 */
static pci_ers_result_t
vastai_pci_io_error_detected(struct pci_dev *pdev, pci_channel_state_t state)
#else
static pci_ers_result_t
vastai_pci_io_error_detected(struct pci_dev *pdev, enum pci_channel_state state)
#endif
{
	switch (state) {
	case pci_channel_io_normal:
		dev_info(&pdev->dev, "Can recover vastai basic device\n");
		return PCI_ERS_RESULT_CAN_RECOVER;
	case pci_channel_io_frozen:
		pci_disable_device(pdev);
		dev_info(&pdev->dev, "Request reset for vastai basic device\n");
		return PCI_ERS_RESULT_NEED_RESET;
	case pci_channel_io_perm_failure:
		dev_err(&pdev->dev, "%s: pci_channel_io_perm_failure.\n",
			__func__);
		return PCI_ERS_RESULT_DISCONNECT;
	}

	/* Request a slot reset. */
	return PCI_ERS_RESULT_NEED_RESET;
}

static pci_ers_result_t vastai_pci_io_slot_reset(struct pci_dev *pdev)
{
	pdev->error_state = pci_channel_io_normal;

#ifdef CONFIG_VASTAI_PCI_LINK_MONITOR
	return PCI_ERS_RESULT_RECOVERED;
#endif
	pci_restore_state(pdev);
	if (pci_enable_device(pdev)) {
		printk("can't enable device\n");
		dev_info(&pdev->dev, "Cannot re-enable vastai basic "
				     "device after reset.\n");
		return PCI_ERS_RESULT_DISCONNECT;
	}
	pci_set_master(pdev);

#if defined(RHEL_RELEASE_VERSION) /* centos and red-hat */
#if RHEL_RELEASE_CODE > RHEL_RELEASE_VERSION(8, 1)
	//pci_aer_clear_nonfatal_status(pdev);
#else
	//pci_cleanup_aer_uncorrect_error_status(pdev);
#endif
#else /* ubuntu and other */
#if (KERNEL_VERSION(5, 7, 0) <= LINUX_VERSION_CODE)
	pci_aer_clear_nonfatal_status(pdev);
#else
	pci_cleanup_aer_uncorrect_error_status(pdev);
#endif
#endif

	return PCI_ERS_RESULT_RECOVERED;
}

static void vastai_pci_io_resume(struct pci_dev *pdev)
{
	dev_info(&pdev->dev, "resuming vastai basic device\n");
}

static const struct pci_error_handlers vastai_pci_err_handler = {
	.error_detected = vastai_pci_io_error_detected,
	.slot_reset = vastai_pci_io_slot_reset,
	.resume = vastai_pci_io_resume,
};

static int vastai_pci_sriov_configure(struct pci_dev *pdev, int num_vfs)
{
	int err = 0;
	struct vastai_pci_info *pci_info = pci_get_drvdata(pdev);

	if (!pci_info->is_physfn)
		return -ENODEV;

	if (num_vfs <= 0) {
		pci_disable_sriov(pdev);
		err = vastai_pci_vf_deactivate(pci_info);
		return err;
	}

	if (num_vfs > pci_info->die_num) {
		pci_warn(pdev, "Try to enable %d VFs, but %d actually\n", num_vfs, pci_info->die_num);
		num_vfs = pci_info->die_num;
	}

	err = pci_enable_sriov(pdev, num_vfs);
	if (err != 0) {
		pci_err(pdev, "Enable %d VFs failed: %d\n", num_vfs, err);
		return err;
	}

	err = vastai_pci_vf_activate(pci_info, num_vfs);
	if (err != 0)
		return err;

	return num_vfs;
}

static const struct pci_device_id vastai_pci_ids[] = {
	{ PCI_VDEVICE(VASTAI, 0x0100), 0 },
	{ PCI_VDEVICE(VASTAI, 0x0101), 0 },
	{ PCI_VDEVICE(DAOCLOUD, 0x0100), 0 },
	{ PCI_VDEVICE(DAOCLOUD, 0x0101), 0 },
	{ /* terminate list */ }
};

static struct pci_driver vastai_pci_driver = {
	.name = VASTAI_PCI_DRIVER_PREFIX,
	.id_table = vastai_pci_ids,
	.probe = vastai_pci_probe,
	.remove = vastai_pci_remove,
	.shutdown = vastai_pci_remove,
	.sriov_configure = vastai_pci_sriov_configure,
	.err_handler = &vastai_pci_err_handler
};

extern struct class *myclass;
static char *devnode(struct device *dev, umode_t *mode)
{
	struct vastai_file_info *file_info = dev_get_drvdata(dev);

	if (mode)
		*mode = VASTAI_FILE_MODE;
	return kstrdup(file_info->file_name, GFP_KERNEL);
}

static int __init vastai_init(void)
{
	int rc = 0;

	dmi_info_init();
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "%s\n",
			VASTAI_PCI_DRIVER_COPYRIGHT);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "[driver ver]: %s\n",
			VASTAI_PCI_DRIVER_VERSION);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "[branchinfo]: %s\n",
			VASTAI_PCI_GIT_BRANCH_NAME);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "[commithash]: %s\n",
			VASTAI_PCI_GIT_COMMIT_HASH);
	// VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID,
	// 		"[commit msg]: %s\n", VASTAI_PCI_GIT_COMMIT_MSG);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "[pack  time]: %s\n",
			VASTAI_PCI_MKPACKAGE_TIME);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "[build time]: %s %s\n", __DATE__,
			__TIME__);

	VASTAI_PCI_FUNC_ENTERY;
	myclass = class_create(THIS_MODULE, "char_class");
	myclass->devnode = devnode;

	INIT_LIST_HEAD(&vastai_pci_info_list);

	rc = pci_register_driver(&vastai_pci_driver);
	if (rc)
		return rc;

	//VASTAI_PCI_FUNC_EXIT;
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "%s done!\n", __func__);

	return 0;
}

static void __exit vastai_exit(void)
{
	VASTAI_PCI_FUNC_ENTERY;
	pci_unregister_driver(&vastai_pci_driver);
	class_destroy(myclass);
	//VASTAI_PCI_FUNC_EXIT;
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "%s done!\n", __func__);
}

#ifndef CONFIG_VASTAI_KUNIT
module_init(vastai_init);
module_exit(vastai_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Gang Wang <gang.wang@vastaitech.com>");
MODULE_DEVICE_TABLE(pci, vastai_pci_ids);
#endif
