/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/delay.h>
#include "vastai_udma_engine.h"
#include <linux/dma-buf.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/version.h>
#include <linux/bitops.h>
#include <linux/workqueue.h>
#include <linux/kernel.h>
#include "vastai_dmabuf.h"
#include "vatools.h"

static inline int vastai_list_is_first(const struct list_head *list,
					const struct list_head *head)
{
	return list->prev == head;
}

static struct vastai_pci_info *die_2_pci_info(struct vastai_sv100_die *die)
{
	return die->pci_info;
}

const struct vastai_udma_config vastai_udma_config[VASTAI_DMA_CHN] = {
	{
		.trans_info_fifo_addr = DMA_CHN0_TRANSF_INFO,
		.dma_desc_fifo_addr = DMA_CHN0_DMA_DESC,
	}, {
		.trans_info_fifo_addr = DMA_CHN1_TRANSF_INFO,
		.dma_desc_fifo_addr = DMA_CHN1_DMA_DESC,
	}, {
		.trans_info_fifo_addr = DMA_CHN2_TRANSF_INFO,
		.dma_desc_fifo_addr = DMA_CHN2_DMA_DESC,
	}, {
		.trans_info_fifo_addr = DMA_CHN3_TRANSF_INFO,
		.dma_desc_fifo_addr = DMA_CHN3_DMA_DESC,
	},
};
static inline int vastai_get_udma_chn(struct pcie_transfer_info *trans)
{
	return GET_UDMA_CHN(trans->dir, trans->is_rc);
}
/* udma_engine:
	-. pcie dma descriptor create\fill
	-. push pcie dma descriptor to device csram
	-. handle dma complete msix
*/
u64 vastai_udma_push_desc(struct vastai_pci_info *pci_info,
			  unsigned int die_index, unsigned int dma_chn,
			  u64 sys_addr, u64 ext_addr, u32 length, u8 ctr_byte)
{
	PCIE_xd_desc temp = { 0 };
	u32 die_id = ((union die_index_data)die_index).die_id;
	u64 fifo_addr = vastai_udma_config[dma_chn].dma_desc_fifo_addr;
	struct vastai_fifo *fifo;
	int ret;

	int vf_id;
	int vf_bdf;

	if (pci_info->is_virtfn) {
		fifo_addr += VASTAI_DMA_DESC_VF_OFFSET;
		vf_id = pci_info->vf_info.vf_id;
		vf_bdf = pci_info->vf_info.bdf;
		switch (vf_id) {
		case VASTAI_PCI_VF_ID_1:
			/* Configure udma request id for VF1 */
			temp.ext_attr |= vf_bdf << 10;
			break;
		case VASTAI_PCI_VF_ID_2:
		case VASTAI_PCI_VF_ID_3:
		case VASTAI_PCI_VF_ID_4:
			/* VF-2/3/4 go through Die0 EP OB region */
			ext_addr |= 0xb000000000 + (vf_id - 1) * VASTAI_DIE1_BASE;
			break;
		default: break;
		}
	}

	fifo = &(pci_info->dies[die_id].udma_chn[dma_chn].dma_desc_fifo);
	temp.sys_lo_addr = ((uint32_t)(sys_addr & VASTAI_VAL_DW_MASK));
	temp.sys_hi_addr = ((uint32_t)((sys_addr >> 32) & VASTAI_VAL_DW_MASK));

	temp.ext_lo_addr = ((uint32_t)(ext_addr & VASTAI_VAL_DW_MASK));
	temp.ext_hi_addr = ((uint32_t)((ext_addr >> 32) & VASTAI_VAL_DW_MASK));

	temp.size_and_ctrl.size = (length & VASTAI_UDMA_LENGTH_MASK);
	temp.size_and_ctrl.ctrl_bits.control_bits = (ctr_byte);

	ret = vastai_fifo_push_elem(pci_info, die_index, fifo_addr, &temp,
				    fifo);/*TODO:create desc_fifo*/
	/* TODO: handle fifo is full */
	if (ret) {
		VASTAI_PCI_ERR(pci_info, die_id, "push dma_desc is error %d\n",
			       ret);
		return U64_MAX;
	}

	return vastai_fifo_wr_get_front(fifo, fifo_addr);
}

void vastai_set_desc_link(struct vastai_pci_info *pci_info,
			  unsigned int die_index, u64 *desc_tab, int num)
{
	int i;
	int ret;
	u32 die_id = ((union die_index_data)die_index).die_id;

	/* set desc_tab list */
	for (i = 0; i < num - 1; i++) {
		u32 next_desc = desc_tab[i + 1] & VASTAI_VAL_DW_MASK;
		u64 next_desc_hi = (desc_tab[i + 1] >> 32) & VASTAI_VAL_DW_MASK;
		u32 size_and_ctrl;
		ret = vastai_pci_mem_write(pci_info, die_index,
					   desc_tab[i] +
						   offsetof(PCIE_xd_desc, next),
					   &next_desc, sizeof(u32));
		if (ret)
			VASTAI_PCI_ERR(pci_info, die_id,
				       "%s write low error %d\n", __FUNCTION__,
				       ret);
		ret = vastai_pci_mem_write(pci_info, die_index,
					   desc_tab[i] + offsetof(PCIE_xd_desc,
								  next_hi_addr),
					   &next_desc_hi, sizeof(u32));
		if (ret)
			VASTAI_PCI_ERR(pci_info, die_id,
				       "%s write low error %d\n", __FUNCTION__,
				       ret);
		ret = vastai_pci_mem_read(
			pci_info, die_index,
			desc_tab[i] + offsetof(PCIE_xd_desc, size_and_ctrl),
			&size_and_ctrl, sizeof(size_and_ctrl));
		if (ret)
			VASTAI_PCI_ERR(pci_info, die_id,
				       "%s read size_ctrl error %d\n",
				       __FUNCTION__, ret);
		size_and_ctrl |= (BIT(5) << 24); /* set continue */
		ret = vastai_pci_mem_write(
			pci_info, die_index,
			desc_tab[i] + offsetof(PCIE_xd_desc, size_and_ctrl),
			&size_and_ctrl, sizeof(size_and_ctrl));
		if (ret)
			VASTAI_PCI_ERR(pci_info, die_id,
				       "%s write size_ctrl error %d\n",
				       __FUNCTION__, ret);
	}
}

/* for tirgger dma, we need fill transfer info and fill msg_queue */
/* TODO: refactor relastionship between buf_desc and pcie_transfer_info */
static u64 vastai_pci_push_trans_info(struct vastai_pci_info *pci_info,
				      unsigned int die_index,
				      struct pcie_transfer_info *trans,
				      u64 *desc)
{
	u32 trans_cur_addr_u32 = 0;
	u64 trans_cur_addr_u64 = 0;
	struct vastai_fifo *fifo;
	u64 fifo_addr = 0;
	int ret;
	int die_id = vastai_pci_get_die_id(pci_info, die_index);
	int udma_chn = vastai_get_udma_chn(trans);
	struct vastai_sv100_die *die = &(pci_info->dies[die_id]);

	fifo = &(die->udma_chn[udma_chn].trans_info_fifo);
	fifo_addr = vastai_udma_config[udma_chn].trans_info_fifo_addr;

	if (pci_info->is_virtfn)
		fifo_addr += VASTAI_DMA_TRANS_INFO_VF_OFFSET;


	ret = vastai_fifo_push_elem(pci_info, die_index, fifo_addr, trans, fifo);/*TODO：create trans_fifo*/
	if (ret) {
		VASTAI_PCI_ERR(pci_info, die_id,
			       "push trans_info is error %d\n", ret);
		return U64_MAX;
	}
	/* set last desc point to this trans_cur_addr_u32 */
	trans_cur_addr_u64 = vastai_fifo_wr_get_front(fifo, fifo_addr);
	trans_cur_addr_u32 = trans_cur_addr_u64 & UINT_MAX;
	vastai_pci_mem_write(pci_info, die_index,
			     desc[trans->num] +
				     offsetof(PCIE_xd_desc, trans_info_addr),
			     &trans_cur_addr_u32, sizeof(u32));
	return trans_cur_addr_u64;
}

#define work_struct_2_die(work) (container_of(work, struct vastai_sv100_die, dma_done))
static void dma_done_work_func(struct work_struct *work)
{
	struct vastai_sv100_die *die =  work_struct_2_die(work);
	struct vastai_pci_info *pci_info = die_2_pci_info(die);
	struct vastai_fifo fifo_cache = {0};
	struct vastai_udma_chn *chn;
	u32 trans_info_addr = 0;
	u32 die_index = die->die_index;
	u32 die_id = vastai_pci_get_die_id(pci_info, die_index);
	u64 msg0_addr = MSG_SMCUQ0_TO_PCIE;
	int ret = 0;

	VASTAI_PCI_DBG(pci_info, die_id, "%s %d\n", __func__, __LINE__);
	if (!pci_info || !die)
		return;

	if (pci_info->is_virtfn)
		msg0_addr += VASTAI_PCIE_SMCU_0_VF_OFFSET;

	mutex_lock(&die->done_lock);
	while (!vastai_fifo_pop_elem(pci_info, die_index, msg0_addr,
				     &trans_info_addr, &fifo_cache)) {
		struct pcie_transfer_info trans;
		u64 cur_addr;
		u32 dma_chn;
		u64 fifo_addr;
		struct vastai_fifo *fifo;

		vastai_pci_mem_read(pci_info, die_index, trans_info_addr, &trans,
				    sizeof(trans));
		if (trans.dir == 0xff || trans.is_rc == 0xff || trans.type == 0xff ||
			trans.chn == 0xffffffff || trans.pcie_config_addr == 0xffffffff) {
			VASTAI_PCI_ERR(
				pci_info, die_id,
				"read transfer info error\n");
			return;
		}

		dma_chn = GET_UDMA_CHN(trans.dir, trans.is_rc);
		fifo_addr = vastai_udma_config[dma_chn].trans_info_fifo_addr;

		if (pci_info->is_virtfn)
			fifo_addr += VASTAI_DMA_TRANS_INFO_VF_OFFSET;

		fifo = &(pci_info->dies[die_id].udma_chn[dma_chn].trans_info_fifo);

		ret = vastai_fifo_pop_elem(pci_info, die_index,
				     fifo_addr, NULL,
				     fifo);
		if (ret) {
			VASTAI_PCI_ERR(
				pci_info, die_id,
				"pop transfer info fifo err %d\n",
				ret);
			return;
		}
		cur_addr = vastai_fifo_rd_get_front(fifo, fifo_addr);

		if (cur_addr != trans_info_addr) {
			VASTAI_PCI_ERR(
				pci_info, die_id,
				"ERROR transfer info addr, expect 0x%llx get 0x%x\n",
				cur_addr, trans_info_addr);
			mutex_unlock(&die->done_lock);
			return;
		}
		trans.data_struct->time_of_dma_run = trans.time_of_dma_run;
		trans.data_struct->time_of_dma_finish = trans.time_of_dma_finish;

		chn = &(pci_info->dies[die_id].udma_chn[dma_chn]);
		spin_lock(&chn->lock);
		list_del(&trans.data_struct->list);
		spin_unlock(&chn->lock);
		complete(&(trans.data_struct->dma_comp));
		VASTAI_PCI_DBG(pci_info, die_id, "%s get data_struct 0x%p\n",
			       __func__, trans.data_struct);
	}
	mutex_unlock(&die->done_lock);
	return;

}

static int inline vastai_record_vdmcu_id(struct vastai_pci_info *pci_info,
					  int die_index,
					  struct data_struct *data)
{
	u64 sys_addr;
	int die_id = vastai_pci_get_die_id(pci_info, die_index);
	int vdmcu_id = 0xff;

	switch (data->chn) {
		case 0x1000000:
			vdmcu_id = 0;
			break;
		case 0x2000000:
			vdmcu_id = 1;
			break;
		case 0x4000000:
			vdmcu_id = 2;
			break;
	};

	if (vdmcu_id < 0 || vdmcu_id >= 3)
		return 0;

	sys_addr = data->vdmcu_dma_msg_dst;
	if (pci_info->dies[die_id].vdmcu_pre_addr_table[vdmcu_id] == sys_addr) {
		VASTAI_PCI_INFO(pci_info,
			       vastai_pci_get_die_id(pci_info, die_index),
			       "%s vdmcu:%d get same addr=0x%llx\n",
			       __FUNCTION__, vdmcu_id, sys_addr);
		return -EBUSY;
	} else {
		pci_info->dies[die_id].vdmcu_pre_addr_table[vdmcu_id] = sys_addr;
	}
	return 0;
}

static int vastai_pci_start_tx_transfer_info(struct vastai_pci_info *pci_info,
					     unsigned int die_index,
					     struct data_struct *data)
{
	struct pcie_transfer_info trans = { 0 };
	struct buf_desc *buf_desc = NULL;
	u32 num = 0;
	int die_id = vastai_pci_get_die_id(pci_info, die_index);
	u64 *desc;
	u64 trans_cur_addr_u64 = 0;
	int ret;
	u32 msgq_id;

	ret = vastai_record_vdmcu_id(pci_info, die_index, data);
	if (ret)
		return ret;
	/* fill transfer info */
	trans.type = data->type;
	trans.chn = data->chn;
	trans.dir = data->dir;

	trans.is_vf = (pci_info->is_virtfn) ? 1 : 0;

	trans.num = data->num;
	trans.is_rc = data->type;
	trans.data_struct = data;
	trans.vdmcu_dma_msg_dst = data->vdmcu_dma_msg_dst;
	trans.vdmcu_dma_msg_len = data->vdmcu_dma_msg_len;
	VASTAI_PCI_DBG(pci_info, die_id,
		       "%s, type=%d, chn=0x%x, dir=%d, num=%d\n", __func__,
		       trans.type, trans.chn, trans.dir, trans.num);

	desc = kmalloc(sizeof(u64) * (trans.num + 1), GFP_KERNEL);
	if (!desc) {
		VASTAI_PCI_ERR(pci_info, die_id,
			       "%s kmalloc is failed! need buf size is 0x%lx\n",
			       __FUNCTION__, sizeof(u64) * trans.num);
		return -EINVAL;
	}

	list_for_each_entry (buf_desc, &(data->head), node) {
		u64 sys_addr;
		u64 ext_addr;
		u8 ctrl_byte = 0;

		sys_addr = buf_desc->data.data_desc.sys_addr;
		ext_addr = buf_desc->data.data_desc.ext_addr;
		if (die_id != 0)
			ext_addr |= 0xb000000000;
		if (list_is_last(&(buf_desc->node), &(data->head)))
			ctrl_byte |= BIT(0);

		desc[num + 1] = vastai_udma_push_desc(
			pci_info, die_index,
			vastai_get_udma_chn(&trans),
			sys_addr, ext_addr, buf_desc->data.data_desc.len,
			ctrl_byte);
		if (desc[num + 1] == U64_MAX) {
			ret =  -EIO;
			goto ERROR;
		}
		VASTAI_PCI_DBG(
			pci_info, die_id,
			"%s, len=%d, sys=0x%llx, ext=0x%llx, desc[0x%llx]\n",
			__func__, buf_desc->data.data_desc.len,
			buf_desc->data.data_desc.sys_addr,
			buf_desc->data.data_desc.ext_addr, desc[1]);
		if (++num > data->num) /* this is a bug! */ {
			VASTAI_PCI_ERR(pci_info, die_id,
				       "cur num:%d not equal data_num %d\n",
				       num, data->num);
			ret = -EINVAL;
			goto ERROR;
		}
	}
	vastai_set_desc_link(pci_info, die_index, desc + 1, data->num);/* set desc_tab list */
	trans.pcie_config_addr = desc[1];

#if 0
	/* set append list */
	if (pci_info->dies[die_id].dma_last_desc[trans.dir]) {
		desc[0] = pci_info->dies[die_id].dma_last_desc[trans.dir];
		vastai_set_desc_link(pci_info, die_index, desc, 2);
	}
	pci_info->dies[die_id].dma_last_desc[trans.dir] = desc[data->num];
#endif
	trans_cur_addr_u64 = vastai_pci_push_trans_info(pci_info, die_index,
							&trans, desc);
	if (trans_cur_addr_u64 == U64_MAX) {
		ret =  -EIO;
		goto ERROR;
	}
	data->time_of_host_trigger_dma = vastai_get_host_time_ns();

	msgq_id = vastai_get_msgq_id(pci_info, MSGQ_DMA);
	vastai_pci_vmsgq_wr(&(pci_info->dies[die_id].vmsgq[msgq_id]),
						(u32 *)&trans_cur_addr_u64, sizeof(u32));

	kfree(desc);
	return 0;
ERROR:
	kfree(desc);
	return ret;
}

static void vastai_get_dma_transinfo(void *priv, u32 die_index,
				     struct data_struct *data,
				     profiler_dma_checkpoint_t *dma_profiler)
{
	struct buf_desc *buf_desc = NULL;

	dma_profiler->pid = data->pid;
	dma_profiler->dev_id = vastai_pci_info_get_dev_id(priv);
	dma_profiler->die_id = vastai_pci_get_die_id(priv, die_index);
	dma_profiler->die_index = die_index;
	dma_profiler->time_of_driver_notify_host_dma_done =
		vastai_get_host_time_ns();
	dma_profiler->time_of_host_trigger_dma = data->time_of_host_trigger_dma;
	dma_profiler->time_of_dma_run = data->time_of_dma_run;
	dma_profiler->time_of_dma_finish = data->time_of_dma_finish;
	dma_profiler->dma_dir = data->dir;

	list_for_each_entry (buf_desc, &(data->head), node) {
		if (vastai_list_is_first(&(buf_desc->node), &(data->head))) {
			dma_profiler->dma_axi_addr =
				buf_desc->data.data_desc.sys_addr;
		}
		dma_profiler->dma_data_size += buf_desc->data.data_desc.len;
	}

	return;
}

static int vastai_pci_transfer_data_sync(struct vastai_pci_info *pci_info,
					 unsigned int die_index,
					 struct data_struct *data)
{
	int ret = 0;
	int die_id = vastai_pci_get_die_id(pci_info, die_index);
	struct vastai_sv100_die *die = &(pci_info->dies[die_id]);
	profiler_dma_checkpoint_t dma_profiler;
	struct vastai_udma_chn* chn;

	if (list_empty(&data->head)) {
		return -EINVAL;
	}

	if (pci_info->disable == true)
		return -EIO;

	init_completion(&data->dma_comp);
	memset(&dma_profiler, 0, sizeof(profiler_dma_checkpoint_t));
	dma_profiler.time_of_driver_recv_dma_req = vastai_get_host_time_ns();

	chn = &die->udma_chn[GET_UDMA_CHN(data->dir, data->type)];
	mutex_lock(&chn->trans_mutex);
	spin_lock(&chn->lock);
	list_add_tail(&data->list, &chn->head);
	spin_unlock(&chn->lock);
	ret = vastai_pci_start_tx_transfer_info(pci_info, die_index, data);
	mutex_unlock(&chn->trans_mutex);
	if (ret) {
		spin_lock(&chn->lock);
		list_del(&data->list);
		spin_unlock(&chn->lock);
		VASTAI_PCI_ERR(pci_info,
			       vastai_pci_get_die_id(pci_info, die_index),
			       "%s start transfer data failed %d\n",
			       __FUNCTION__,
			       ret);
		return ret;
	}

	ret = wait_for_completion_interruptible(&data->dma_comp);
	if (ret) {
		VASTAI_PCI_ERR(pci_info,
			       vastai_pci_get_die_id(pci_info, die_index),
			       "%s wait_for_completion error %d\n",
			       __FUNCTION__, ret);
	}
	/*record dma profiler info*/
	vastai_get_dma_transinfo(pci_info, die_index, data, &dma_profiler);
#ifdef CONFIG_VASTAI_TOOLS_SUITE
	vatools_profiler_dma_checkpoint(&dma_profiler);
#endif
	return ret;
}

int vastai_pci_data_movement(void *pci_info, int die_index, u64 local_addr,
			     u64 remote_addr, u32 len, bool is_local_2_remote)
{
	struct data_struct dma_data_struct = { 0 };
	struct buf_desc desc;
	int ret = 0;

	memset(&desc, 0, sizeof(desc));
	INIT_LIST_HEAD(&dma_data_struct.head);

	desc.data.data_desc.sys_addr = local_addr;
	desc.data.data_desc.ext_addr = remote_addr;
	desc.data.data_desc.len = len;

	list_add_tail(&(desc.node), &(dma_data_struct.head));

	dma_data_struct.num = 1;
	dma_data_struct.chn = 0;
	dma_data_struct.dir = is_local_2_remote;
	dma_data_struct.type = 1;

	ret = vastai_pci_transfer_data_sync(pci_info, die_index,
					    &dma_data_struct);
	if (ret) {
		/* there are information print at vastai_pci_transfer_data_sync is enough,
		   we don't need print again */
		ssleep(1);
	}

	return ret;
}

struct buf_desc_buf {
	struct buf_desc desc;
	struct vastai_dma_buf *buf;
};

static int fill_dma_buf(void *pcie_dev, struct buf_desc_buf *dma_buf,
			const struct vastai_dmadesc *desc)
{
	if (!desc->is_src_dma_addr) {
		if (iommu_is_enable(pcie_dev)) {
			dma_buf->buf = vastai_dma_buf_sg_get(pcie_dev,
						desc->dma_lenth);
		} else {
			dma_buf->buf = vastai_mempool_get(pcie_dev,
						desc->dma_lenth);
		}

		if (!dma_buf->buf) {
			VASTAI_PCI_ERR(pcie_dev, DUMMY_DIE_ID,
				       "%s udma alloc failed\n", __FUNCTION__);
			return -ENOMEM;
		}
		dma_buf->desc.data.data_desc.ext_addr =
			dma_buf->buf->dma_bus_addr;
	} else {
		dma_buf->desc.data.data_desc.ext_addr =
			desc->host_addr.dma_addr;
	}

	memcpy(dma_buf->desc.data.data_desc.user_data, desc->user_prive,
	       sizeof(u32) * 4);
	dma_buf->desc.data.data_desc.len = desc->dma_lenth;

	dma_buf->desc.data.data_desc.sys_addr = desc->dev_addr;
	return 0;
}

static void clean_dma_buf(void *pcie_dev, struct buf_desc_buf *dma_buf,
			  u32 is_src_dma_addr)
{
	if (!is_src_dma_addr) {
		if (iommu_is_enable(pcie_dev))
			vastai_dma_buf_sg_put(pcie_dev, dma_buf->buf);
		else
			vastai_mempool_put(pcie_dev, dma_buf->buf);
	}
	memset(dma_buf, 0, sizeof(struct buf_desc_buf));
}

int vastai_pci_dma_transfer_sync(void *pci_info, int die_index,
				 union core_bitmap core_id,
				 struct vastai_dmadesc desc[], int desc_num)
{
	return vastai_pci_dma_transfer_sync_pid(pci_info, die_index, core_id,
						desc, desc_num, -1);
}

static struct vastai_pid_entry *
vastai_dma_create_credit(struct vastai_pci_info *pci_info, int die_index,
			 int pid, int *ret)
{
	int die_id = vastai_pci_get_die_id(pci_info, die_index);
	struct vastai_credit *credit = pci_info->dies[die_id].credit;
	struct vastai_pid_entry *pid_entry = NULL;

	*ret = down_interruptible(&credit->credit_sem);
	if (*ret != 0) {
		VASTAI_PCI_ERR(pci_info, die_id,
			       "%s Is user send kill signal? \n", __FUNCTION__);
		return NULL;
	}

	if (pid < 0) {
		/* invalid pid, use default credit */
		return NULL;
	}

	pid_entry = kvmalloc(sizeof(struct vastai_pid_entry), GFP_KERNEL);
	if (!pid_entry) {
		VASTAI_PCI_ERR(
			pci_info, DUMMY_DIE_ID,
			"%s pid_entry alloc failed, use default credit\n",
			__FUNCTION__);
		return NULL;
	}

	INIT_LIST_HEAD(&pid_entry->list);
	pid_entry->pid = pid;
	spin_lock(&credit->lock);
	list_add_tail(&pid_entry->list, &credit->pid_entry_head.list);
	spin_unlock(&credit->lock);

	return pid_entry;
}

static int vastai_dma_get_credit(struct vastai_pci_info *pci_info,
				 int die_index,
				 struct vastai_pid_entry *pid_entry,
				 int expect_num)
{
	int die_id = vastai_pci_get_die_id(pci_info, die_index);
	struct vastai_credit *credit = pci_info->dies[die_id].credit;
	int first_entry_pid = 0;
	int credit_val = 0;

	if (!is_vastai_mempool_rich(pci_info))
		return 1;

	if (pid_entry == NULL) {
		credit_val = VASTAI_DMA_DEFAULT_CREDIT;
		goto OUT;
	}

	/* lock */
	spin_lock(&credit->lock);
	/* get first entry pid */
	first_entry_pid = list_first_entry(&credit->pid_entry_head.list,
					   struct vastai_pid_entry, list)
				  ->pid;
	/* unlock */
	spin_unlock(&credit->lock);

	/* only first entry get high priority */
	if (first_entry_pid == pid_entry->pid) {
		credit_val = VASTAI_DMA_HIGH_PRIORITY;
	} else {
		credit_val = VASTAI_DMA_LOW_PRIORITY;
	}

OUT:
	return min(credit_val, expect_num);
}

static void vastai_dma_destory_credit(struct vastai_pci_info *pci_info,
				      int die_index,
				      struct vastai_pid_entry *pid_entry)
{
	int die_id = vastai_pci_get_die_id(pci_info, die_index);
	struct vastai_credit *credit = pci_info->dies[die_id].credit;

	if (!pid_entry) {
		up(&credit->credit_sem);
		return;
	}

	/* lock */
	spin_lock(&credit->lock);
	/* del_list */
	list_del(&pid_entry->list);
	/* unlock */
	spin_unlock(&credit->lock);
	/* free */
	kvfree(pid_entry);
	/* up */
	up(&credit->credit_sem);
}

int vastai_pci_dma_common_mem_copy(bool is_host_to_dev, struct vastai_dmadesc *cur_desc,
						  struct buf_desc_buf *dma_buf,
						  void *pci_info,
						  int i, int die_index)
{
	int ret = 0;

	if (!(cur_desc[i].is_src_dma_addr)) {
		if (cur_desc[i].is_src_not_user_mem) {
			if(is_host_to_dev)
				memcpy(dma_buf[i].buf->vir, cur_desc[i].host_addr.vir_addr, cur_desc[i].dma_lenth);
			else
				memcpy(cur_desc[i].host_addr.vir_addr, dma_buf[i].buf->vir, cur_desc[i].dma_lenth);
		} else {
			if(is_host_to_dev)
				ret = copy_from_user(dma_buf[i].buf->vir, cur_desc[i].host_addr.vir_addr, cur_desc[i].dma_lenth);
			else
				ret = copy_to_user(cur_desc[i].host_addr.vir_addr, dma_buf[i].buf->vir, cur_desc[i].dma_lenth);
			if (ret) {
				VASTAI_PCI_ERR(
					pci_info,
					vastai_pci_get_die_id(
						pci_info, die_index),
					"%s copy %s user failed dma[0x%p] user[0x%p]\n",
					__FUNCTION__,
					is_host_to_dev ? "from" : "to",
					dma_buf[i].buf->vir,
					(void *)cur_desc[i].host_addr.vir_addr);
				return -EFAULT;
			}
		}
	}

	return ret;
}

int vastai_pci_dma_transfer_sync_pid(void *pci_info, int die_index,
				     union core_bitmap core_id,
				     struct vastai_dmadesc desc[], int desc_num,
				     int pid)
{
	struct data_struct dma_data_struct = { 0 };
	int i = 0;
	int ret = 0;
	bool host_to_dev = desc[0].is_host_to_dev;
	struct buf_desc_buf *dma_buf = NULL;
	struct vastai_dmadesc *cur_desc = NULL;
	int remain_num = desc_num;
	int cur_num = 0;
	struct vastai_pid_entry *pid_entry = NULL;
	if (desc_num < 1) {
		VASTAI_PCI_ERR(pci_info,
			       vastai_pci_get_die_id(pci_info, die_index),
			       "%s desc_num is 0\n", __FUNCTION__);
		return 0;
	}

	pid_entry = vastai_dma_create_credit(pci_info, die_index, pid, &ret);
	if (ret != 0) {
		return ret;
	}

	/* check list is vaild */
	for (i = 0; i < desc_num; i++) {
		if (host_to_dev != desc[i].is_host_to_dev) {
			VASTAI_PCI_ERR(pci_info,
				       vastai_pci_get_die_id(pci_info,
							     die_index),
				       "%s dir not expect, desc[%d : %d] != %d",
				       __FUNCTION__, i, desc[i].is_host_to_dev,
				       host_to_dev);
			ret = -EAGAIN;
			goto UP_DMA_SEM;
		}
	}
LOOP:
	cur_num = vastai_dma_get_credit(pci_info, die_index, pid_entry,
					remain_num);/*TODO:Up to 8 at a time*/
	memset(&dma_data_struct, 0, sizeof(struct data_struct));
	INIT_LIST_HEAD(&dma_data_struct.head);

	dma_buf = kvmalloc(sizeof(*dma_buf) * cur_num, GFP_KERNEL);
	if (!dma_buf) {
		ret = -ENOMEM;
		goto UP_DMA_SEM;
	}
	cur_desc = &desc[desc_num - remain_num];
	remain_num -= cur_num;

	for (i = 0; i < cur_num; i++) {/*TODO：Start filling dma_desc*/
		ret = fill_dma_buf(pci_info, &(dma_buf[i]), &(cur_desc[i]));
		if (ret) {
			desc_num = i;
			goto CLEAN_DMA_BUF;
		}

		if (host_to_dev){
			ret = vastai_pci_dma_common_mem_copy(host_to_dev,
				cur_desc, dma_buf, pci_info, i, die_index);
			if(ret) {
				desc_num = i;
				goto CLEAN_DMA_BUF;
			}
		}

		list_add_tail(&(dma_buf[i].desc.node), &(dma_data_struct.head));
	}

	dma_data_struct.num = cur_num;
	if (remain_num <= 0) {
		dma_data_struct.chn = core_id.val;
		dma_data_struct.vdmcu_dma_msg_len = desc[0].dma_lenth;
		dma_data_struct.vdmcu_dma_msg_dst = desc[0].dev_addr;
	}
	dma_data_struct.dir = host_to_dev ? 0 : 1;
	dma_data_struct.type = 0;
	dma_data_struct.pid = pid;
	ret = vastai_pci_transfer_data_sync(pci_info, die_index,
					    &dma_data_struct);
	if (ret && ret != -EBUSY) {
		ssleep(1);
		goto CLEAN_DMA_BUF;
	}

	for (i = 0; i < cur_num; i++) {
		if (!host_to_dev){
			ret = vastai_pci_dma_common_mem_copy(host_to_dev,
				cur_desc, dma_buf, pci_info, i, die_index);
			if(ret) {
				desc_num = i;
				goto CLEAN_DMA_BUF;
			}
		}
		clean_dma_buf(pci_info, &(dma_buf[i]),
			      cur_desc->is_src_dma_addr);
	}

	kvfree(dma_buf);

	if (remain_num > 0)
		goto LOOP;

	vastai_dma_destory_credit(pci_info, die_index, pid_entry);
	return ret;
CLEAN_DMA_BUF:
	for (i = 0; i < cur_num; i++) {
		clean_dma_buf(pci_info, &(dma_buf[i]),
			      cur_desc->is_src_dma_addr);
	}
	kvfree(dma_buf);
UP_DMA_SEM:
	vastai_dma_destory_credit(pci_info, die_index, pid_entry);
	return ret;
}
static dma_addr_t sg_addr(struct vastai_pci_info *pci_info,
				struct scatterlist *sg)
{
	return iommu_is_enable(pci_info) ? sg_dma_address(sg)
					: sg_phys(sg);

}
static u32 sg_len(struct vastai_pci_info *pci_info,
				struct scatterlist *sg)
{
	return iommu_is_enable(pci_info) ? sg_dma_len(sg)
					: sg->length;

}

/* sg_gather:
	find max size whicth can be gathered.
	return dma_addr and size.
*/

static int sg_gather(struct vastai_pci_info *pci_info ,
			struct vastai_sg_gather_info *sg_info,
			u32 max_gather_size)
{
	u32 len, len_next, cur_sg_remain_len;
	dma_addr_t addr, addr_next;
	struct scatterlist *sg, *sg_next_p;
	u32 i = sg_info->cur_sg_index;

	sg = sg_info->cur_sg;
	if (sg_info->cur_sg_remain_len == 0) {
		len = sg_len(pci_info, sg);
		sg_info->gather_addr = sg_addr(pci_info, sg);
		sg_info->gather_len = 0;
		sg_info->cur_sg_remain_len = len;
	}

	if (sg_info->cur_sg_remain_len > max_gather_size) {
		len = max_gather_size;
		addr = sg_info->gather_addr + sg_info->gather_len;
		cur_sg_remain_len = sg_info->cur_sg_remain_len - len;
	} else {
		len = sg_info->cur_sg_remain_len;
		addr = sg_info->gather_addr + sg_info->gather_len;
		while ((i + 1) < sg_info->sg_len) {
			sg_next_p = sg_next(sg);
			addr_next = sg_addr(pci_info, sg_next_p);
			len_next = sg_len(pci_info, sg_next_p);
			if (len_next == 0)
				break;
			if (addr + len == addr_next) {
				i++;
				sg = sg_next_p;
				if (len + len_next > max_gather_size) {
					cur_sg_remain_len = len + len_next -
							max_gather_size;
					len = max_gather_size;
					break;
				} else {
					len += len_next;
					cur_sg_remain_len = 0;
				}
			} else {
				i++;
				sg = sg_next_p;
				cur_sg_remain_len = 0;
				break;
			}
		}
	}
	sg_info->cur_sg = sg;
	sg_info->cur_sg_index = i;
	sg_info->gather_addr = addr;
	sg_info->gather_len = len;
	sg_info->cur_sg_remain_len = cur_sg_remain_len;

	return 0;
}
int vastai_dma_set_desc_memset(struct vastai_dmadesc* desc,
            dma_addr_t dma_bus_addr,
            int i,
            u64 axi_addr,
            u32 done_size,
            u32 is_host_to_dev,
            u32 is_src_not_user_mem,
            u32 is_src_dma_addr,
            u32 size)
{
	int ret = 0;
	desc[i].is_host_to_dev = is_host_to_dev;
	desc[i].is_src_not_user_mem = is_src_not_user_mem;
	desc[i].is_src_dma_addr = is_src_dma_addr;
	desc[i].dev_addr = axi_addr+ done_size;

    if(size < VASTAI_MAX_MEMSET_BUF){

        desc[i].dma_lenth = size;
    }else{

        desc[i].dma_lenth = VASTAI_MAX_MEMSET_BUF;
    }

	desc[i].host_addr.dma_addr = dma_bus_addr;
	return ret;
}
int vastai_dma_sg_set_desc(struct vastai_dmadesc* desc,
            struct vastai_sg_gather_info sg_info,
            int i,
            u64 axi_addr,
            u32 done_size,
            u32 is_host_to_dev,
            u32 is_src_not_user_mem,
            u32 is_src_dma_addr)
{
	int ret=0;
	desc[i].is_host_to_dev = is_host_to_dev;
	desc[i].is_src_not_user_mem = is_src_not_user_mem;
	desc[i].is_src_dma_addr = is_src_dma_addr;
	desc[i].dev_addr = axi_addr+ done_size;
	desc[i].dma_lenth = sg_info.gather_len;
	desc[i].host_addr.dma_addr = sg_info.gather_addr;
	return ret;
}

int vastai_pci_dma_transfer_memset(void *pci_info,
			int die_index,
			union core_bitmap core_id,
			dma_addr_t dma_bus_addr,
			u32 is_src_dma_addr,
			u32 is_host_to_dev,
			u32 is_src_not_user_mem,
			u64 axi_addr,
			u32 size)
{
	u32 done_size = 0;
	int ret = 0;
	int i = 0;
	struct vastai_dmadesc *desc;
	u32 entry_count = DIV_ROUND_UP(size, VASTAI_MAX_MEMSET_BUF);

	desc = kvmalloc(sizeof(struct vastai_dmadesc) * entry_count,
					GFP_KERNEL);
	while (done_size < size) {
		if (ret)
			break;
        ret = vastai_dma_set_desc_memset(desc,
                                         dma_bus_addr,
                                         i,
                                         axi_addr,
                                         done_size,
                                         is_host_to_dev,
                                         is_src_not_user_mem,
                                         is_src_dma_addr,
                                         size);

		if (ret)
		  break;
		++i;
        if(size < VASTAI_MAX_MEMSET_BUF){

            done_size += size;
        }else{

            done_size += VASTAI_MAX_MEMSET_BUF;
        }
	}
	ret = vastai_pci_dma_transfer_sync(
                pci_info, die_index, core_id,
				desc, i);
	kvfree(desc);
	return ret;
}

int vastai_pci_dma_transfer_sg(void *pci_info,
			int die_index,
			union core_bitmap core_id,
			struct vastai_dma_buf *dm,
			u32 is_src_dma_addr,
			u32 is_host_to_dev,
			u32 is_src_not_user_mem,
			u64 axi_addr,
			u32 size)
{
	u32 max_gather_size;
	u32 done_size = 0;
	int ret = 0;
	int i=0;
	struct scatterlist *sg;
	struct vastai_sg_gather_info sg_info;
	struct vastai_dmadesc *desc;
	u32 entry_count = DIV_ROUND_UP(size, VASTAI_MAX_DMA_PAYLOAD);
	/*...................Optimize desc to become an array and record its size..............*/
	memset(&sg_info, 0, sizeof(struct vastai_sg_gather_info));
	sg_info.cur_sg = dm->sg_table.sgl;
	sg_info.sg_len = dm->sg_table.nents;
	sg = sg_info.cur_sg;

	desc = kvmalloc(sizeof(struct vastai_dmadesc) * entry_count,
					GFP_KERNEL);
	while (done_size < size) {
		max_gather_size = min((u32)VASTAI_MAX_DMA_PAYLOAD,
							size - done_size);
		ret = sg_gather(pci_info, &sg_info, max_gather_size);
		if (ret)
			break;
        ret=vastai_dma_sg_set_desc(desc,
                                sg_info,
                                i,
                                axi_addr,
                                done_size,
                                is_host_to_dev,
                                is_src_not_user_mem,
                                is_src_dma_addr);

		if (ret)
		  break;
		++i;
		done_size += sg_info.gather_len;
	}
	ret = vastai_pci_dma_transfer_sync(
                pci_info, die_index, core_id,
				desc, i);
	kvfree(desc);
	return ret;
}

int vastai_pci_dma_transfer_by_uaddr(struct vastai_pci_info *pci_info,
			int die_index,
			union core_bitmap core_id,
			u32 is_host_to_dev,
			u64 axi_addr,
			u64 user_addr,
			u32 size)
{
	struct device *dev = &(pci_info->dev->dev);
	struct sg_table *sgt;
	int num_pinned_pages = 0;
	unsigned long offset = user_addr & ((1 << PAGE_SHIFT) - 1);
	unsigned int num_pages = (size + offset + PAGE_SIZE - 1) >> PAGE_SHIFT;
	unsigned int gup_flags = 0;
	int i;
	int ret;
	struct vastai_dma_buf *dm =
		kzalloc(sizeof(struct vastai_dma_buf), GFP_KERNEL);
	if (!dm)
		return -ENOMEM;

	dm->pages = kvmalloc_array(num_pages, sizeof(struct page *),
			GFP_KERNEL | __GFP_ZERO);
	if (!dm->pages) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"kvmalloc_array failed\n");

		ret = -ENOMEM;
		goto ERROR_FREE_DM;
	}

	dm->pci_info = pci_info;
	dm->num_pages = num_pages;
	gup_flags = is_host_to_dev ? 0 : FOLL_WRITE;
	num_pinned_pages = get_user_pages_fast(
			user_addr,
			(int)num_pages,
			gup_flags,
			dm->pages);
	if (num_pinned_pages != num_pages)
	{
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"get_user_pages_fast failed: (%d - %u)\n",
			num_pinned_pages, num_pages);

		ret = -EINVAL;
		goto ERROR_FREE_PAGE;
	}

	ret = sg_alloc_table_from_pages(&dm->sg_table, dm->pages,
				dm->num_pages, 0, size, GFP_KERNEL);
	if (ret) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"sg_alloc_table_from_pages failed:%d \n", ret);

		goto ERROR_FREE_PIN_PAGE;
	}
	sgt = &dm->sg_table;
	sgt->nents = dma_map_sg(dev, sgt->sgl, sgt->orig_nents, 0);
	if (!sgt->nents) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"dma_map_sg failed:%d \n", ret);

		goto ERROR_FREE_TABLE;
	}
#if 1
	ret = vastai_pci_dma_transfer_sg(pci_info,
				  die_index,
				  core_id, dm,
				  1,
				  is_host_to_dev,
				  1,
				  axi_addr,
				  size);
	if (ret) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"vastai_pci_dma_transfer_sg failed:%d \n", ret);
	}
#endif
//	printk("transfer ok\n");

	dma_unmap_sg(dev, sgt->sgl, sgt->orig_nents, 0);
ERROR_FREE_TABLE:
	sg_free_table(&dm->sg_table);
ERROR_FREE_PIN_PAGE:
	for (i = 0; i < num_pinned_pages; i++)
		put_page(dm->pages[i]);
ERROR_FREE_PAGE:
	kvfree(dm->pages);
ERROR_FREE_DM:
	kfree(dm);

	return ret;
}

int vastai_udma_init(struct vastai_sv100_die *die)
{
	int chn;
	u32 dma_desc_fifo_size = VASTAI_DMA_DESC_FIFO_SIZE;
	u32 transfer_info_fifo_size = VASTAI_DMA_TRANS_INFO_FIFO_SIZE;

	if (die->pci_info->is_virtfn) {
		dma_desc_fifo_size -= VASTAI_DMA_DESC_VF_OFFSET;
		transfer_info_fifo_size -= VASTAI_DMA_TRANS_INFO_VF_OFFSET;
	}

	die->credit = kvmalloc(sizeof(struct vastai_credit),
			       GFP_KERNEL);
	if (!die->credit) {
		VASTAI_PCI_ERR(die->pci_info, DUMMY_DIE_ID,
			"%s alloc credit failed! pls check pc's mem\n",
			__func__);
		goto CREDIT_ALLOC_FAILED;
	}
	memset(die->credit, 0x0, sizeof(struct vastai_credit));
	INIT_LIST_HEAD(&die->credit->pid_entry_head.list);
	spin_lock_init(&die->credit->lock);
	sema_init(&die->credit->credit_sem, VASTAI_DMA_CREDIT_LIMIT);


	die->udma_chn = kmalloc_array(VASTAI_DMA_CHN,
				      sizeof(struct vastai_udma_chn),
				      GFP_KERNEL);
	memset(die->udma_chn, 0x0,
		sizeof(struct vastai_udma_chn) * VASTAI_DMA_CHN);
	if (!die->udma_chn) {
		VASTAI_PCI_ERR(die->pci_info, DUMMY_DIE_ID,
			       "%s alloc udma chn failed!\n", __func__);
		goto UDMA_CHN_ALLOC_FAILED;
	}

	die->udma_wq = alloc_workqueue("vastai_dma_wq",
			WQ_MEM_RECLAIM|WQ_UNBOUND|WQ_CPU_INTENSIVE, 1);
	if (!die->udma_wq) {
		VASTAI_PCI_ERR(die->pci_info, DUMMY_DIE_ID,
			       "%s alloc dma_wq failed!\n", __func__);
		goto UDMA_WQ_FAILED;
	}

	for (chn = 0; chn < VASTAI_DMA_CHN; chn ++) {
		struct vastai_udma_chn *udma_chn = &(die->udma_chn[chn]);
		INIT_LIST_HEAD(&udma_chn->head);
		spin_lock_init(&udma_chn->lock);
		mutex_init(&(udma_chn->trans_mutex));
		vastai_fifo_init_arg(&(udma_chn->dma_desc_fifo),
				     ALIGN(sizeof(PCIE_xd_desc), 4),
				     dma_desc_fifo_size);
		vastai_fifo_init_arg(&(udma_chn->trans_info_fifo),
				ALIGN(sizeof(struct pcie_transfer_info), 4),
				transfer_info_fifo_size);
	}
	INIT_WORK(&(die->dma_done), dma_done_work_func);
	mutex_init(&(die->done_lock));
	return 0;

UDMA_WQ_FAILED:
	kfree(die->udma_wq);

UDMA_CHN_ALLOC_FAILED:
	kfree(die->udma_chn);

CREDIT_ALLOC_FAILED:
	kfree(die->credit);
	return -ENOMEM;
}

void vastai_udma_deinit(struct vastai_sv100_die *die)
{
	flush_workqueue(die->udma_wq);
	destroy_workqueue(die->udma_wq);
	die->udma_wq = NULL;
	kfree(die->udma_chn);
	die->udma_chn = NULL;
	kvfree(die->credit);
	die->credit = NULL;
	VASTAI_PCI_INFO(die->pci_info, DUMMY_DIE_ID, "%s done!\n", __func__);
}

struct vastai_dma_pos_info {
	u32 current_channel;
	u32 current_line;
	u32 current_line_pos;
	u32 next_dst_offset;
	int complete;
};

/**
 * 将channel buf 里的数据搬运到一个 dma buf中。它通过ret val， 使自己对上一次的结果有记忆。
 */
static int vastai_dma_copy_from_user(struct vastai_channel_buf *channel_buf,
				     int channel_num,
				     struct vastai_dma_buf *buf,
				     struct vastai_dma_pos_info *meta)
{
	int channel_i = meta->current_channel;
	u32 line_i = meta->current_line;
	int ret = 0;
	u32 dst_done_size = meta->next_dst_offset;
	u32 src_done_size =
		(channel_buf[meta->current_channel].src_width_offset +
		 channel_buf[meta->current_channel].width) *
			meta->current_line +
		meta->current_line_pos;

	if (meta->complete == 1)
		return 0;

	if (meta->current_line_pos) {
		u32 real_size;
		if (channel_buf[channel_i].width < meta->current_line_pos)
			return -EINVAL;
		real_size =
			channel_buf[channel_i].width - meta->current_line_pos;
		ret = copy_from_user(buf->vir + dst_done_size,
				     channel_buf[channel_i].user_buf +
					     src_done_size,
				     real_size);
		if (ret)
			return ret;

		dst_done_size +=
			(real_size + channel_buf[channel_i].dst_width_offset);
		src_done_size +=
			(real_size + channel_buf[channel_i].src_width_offset);
		line_i += 1;
	}

	for (; channel_i < channel_num; channel_i++) {
		for (; line_i < channel_buf[channel_i].high; line_i++) {
			/* 如果 当前剩余的dma buf 不够 */
			if ((channel_buf[channel_i].width +
			     channel_buf[channel_i].dst_width_offset) >
			    (buf->size - dst_done_size)) {
				u32 real_size =
					min(channel_buf[channel_i].width,
					    buf->size - dst_done_size);
				u32 line_remain = channel_buf[channel_i].width -
						  real_size;

				ret = copy_from_user(
					buf->vir + dst_done_size,
					channel_buf[channel_i].user_buf +
						src_done_size,
					real_size);
				if (ret)
					return ret;

				dst_done_size += real_size;
				meta->current_channel = channel_i;
				meta->current_line_pos =
					line_remain ? real_size : 0;
				meta->next_dst_offset =
					line_remain ?
						0 :
						(channel_buf[channel_i]
							 .dst_width_offset -
						 (buf->size - dst_done_size));
				meta->current_line = meta->next_dst_offset ?
							     ++line_i :
							     line_i;
				return 0;
			}
			/* 正常逻辑 */
			ret = copy_from_user(buf->vir + dst_done_size,
					     channel_buf[channel_i].user_buf +
						     src_done_size,
					     channel_buf[channel_i].width);
			if (ret)
				return ret;

			dst_done_size +=
				(channel_buf[channel_i].width +
				 channel_buf[channel_i].dst_width_offset);
			src_done_size +=
				(channel_buf[channel_i].width +
				 channel_buf[channel_i].src_width_offset);
		}
		line_i = 0;
		src_done_size = 0;
		dst_done_size += (channel_buf[channel_i].width +
				  channel_buf[channel_i].dst_width_offset) *
				 channel_buf[channel_i].dst_high_offset;
		if (dst_done_size > buf->size && channel_i < channel_num) {
			meta->current_channel = ++channel_i;
			meta->current_line = 0;
			meta->current_line_pos = 0;
			meta->next_dst_offset = dst_done_size - buf->size;
			return 0;
		}
	}
	/* all is done */
	meta->current_channel = 0;
	meta->current_line = 0;
	meta->current_line_pos = 0;
	meta->next_dst_offset = 0;
	meta->complete = 1;
	return 0;
}

int vastai_dma_video(void *pci_info, int die_index, union core_bitmap core_id,
		     struct vastai_channel_buf *channel_buf, int channel_num,
		     u64 dev_addr, int pid)
{
	unsigned long channel_buf_total_size = 0;
	int i = 0;
	struct vastai_pid_entry *pid_entry = NULL;
	int ret;
	struct data_struct dma_data_struct = { 0 };
	unsigned long dma_transfer_total_len = 0;
	struct vastai_dma_pos_info meta = { 0 };
	int dma_num;
	struct buf_desc_buf *dma_buf = NULL;

	memset(&meta, 0, sizeof(struct vastai_dma_pos_info));
	for (i = 0; i < channel_num; i++) {
		channel_buf_total_size +=
			(channel_buf[i].width +
			 channel_buf[i].dst_width_offset) *
			(channel_buf[i].high + channel_buf[i].dst_high_offset);
	}
	pid_entry = vastai_dma_create_credit(pci_info, die_index, pid, &ret);
	if (ret != 0)
		return ret;

	while (channel_buf_total_size - dma_transfer_total_len) {
		unsigned long remain_size =
			channel_buf_total_size - dma_transfer_total_len;
		dma_num = DIV_ROUND_UP(remain_size, VASTAI_MAX_DMA_BUF);

		memset(&dma_data_struct, 0, sizeof(struct data_struct));
		INIT_LIST_HEAD(&dma_data_struct.head);

		dma_num = vastai_dma_get_credit(pci_info, die_index, pid_entry,
						dma_num);
		dma_buf = kvmalloc(sizeof(struct buf_desc_buf) * dma_num,
				   GFP_KERNEL);
		if (!dma_buf) {
			/* TODO ERROR HANDLE */
			VASTAI_PCI_ERR(
				pci_info,
				vastai_pci_get_die_id(pci_info, die_index),
				"%s dma_buf alloc failed\n", __FUNCTION__);
			ret = -ENOMEM;
			goto UP_DMA_SEM;
		}

		for (i = 0; i < dma_num; i++) {
			u32 dma_len = min(remain_size,
					(unsigned long)VASTAI_MAX_DMA_BUF);
			if (iommu_is_enable(pci_info)) {
				dma_buf[i].buf = vastai_dma_buf_sg_get(pci_info,
								dma_len);
			} else {
				dma_buf[i].buf = vastai_mempool_get(pci_info,
								dma_len);
			}
			if (!dma_buf[i].buf) {
				dma_num = i;
				ret = -ENOMEM;
				goto CLEAN_DMA_BUF;
			}
			ret = vastai_dma_copy_from_user(channel_buf,
							channel_num,
							dma_buf[i].buf, &meta);
			if (ret) {
				/* TODO ERROR HANDLE */
				VASTAI_PCI_ERR(
					pci_info,
					vastai_pci_get_die_id(pci_info,
							      die_index),
					"%s vastai_dma_copy_from_user failed, err:%d\n",
					__FUNCTION__, ret);
				dma_num = i;
				goto CLEAN_DMA_BUF;
			}
			dma_buf[i].desc.data.data_desc.ext_addr =
				dma_buf[i].buf->dma_bus_addr;
			dma_buf[i].desc.data.data_desc.len = dma_len;
			dma_buf[i].desc.data.data_desc.sys_addr =
				dev_addr + dma_transfer_total_len;

			list_add_tail(&(dma_buf[i].desc.node),
				      &(dma_data_struct.head));
			remain_size -= dma_len;
			dma_transfer_total_len += dma_len;
		}

		dma_data_struct.num = dma_num;
		dma_data_struct.dir = 0;
		dma_data_struct.type = 0;
		ret = vastai_pci_transfer_data_sync(pci_info, die_index,
						    &dma_data_struct);
		if (ret) {
			/* TODO ERROR HANDLE */
			ssleep(1);
			VASTAI_PCI_ERR(
				pci_info,
				vastai_pci_get_die_id(pci_info, die_index),
				"%s vastai_pci_transfer_data_sync failed, err:%d\n",
				__FUNCTION__, ret);
			goto CLEAN_DMA_BUF;
		}
		for (i = 0; i < dma_num; i++) {
			clean_dma_buf(pci_info, &(dma_buf[i]), 0);
		}
		kvfree(dma_buf);
	}
	vastai_dma_destory_credit(pci_info, die_index, pid_entry);

	return 0;
CLEAN_DMA_BUF:
	for (i = 0; i < dma_num; i++) {
		clean_dma_buf(pci_info, &(dma_buf[i]), 0);
	}
	kvfree(dma_buf);
UP_DMA_SEM:
	vastai_dma_destory_credit(pci_info, die_index, pid_entry);
	return ret;
}
