/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/platform_device.h>
#include <linux/uaccess.h>
#include <linux/version.h>
#include <linux/random.h>
#include <linux/dma-buf.h>

#include "vastai_pci.h"
#include "vastai_pci_api.h"
#include "vastai_pci_test.h"
#include "vastai_udma_test.h"
#include "vastai_udma_engine.h"
#include "vastai_dmabuf.h"
#include "vastai_fifo.h"
#include "vastai_pci_vf.h"
#include "vastai_state.h"

#ifdef CONFIG_VASTAI_KUNIT
#include "stub.h"
#endif

#ifdef EMULATOR_ENV
#include "vastai.h"
#endif
#define TCGETS	       0x5401
#define TCSETS	       0x5402
#define TCSETSW	       0x5403
#define TCSETSF	       0x5404
#define TIOCSERGSTRUCT 0x5458 /* For debugging only */
#define TIOCSERGETLSR  0x5459 /* Get line status register */
#define TIOCSERCONFIG  0x545
struct vastai_dma_buf dmb[4];

static struct vastai_pci_arg_t vastai_pci_args_default[VASTAI_PCI_ARG_MAX] = {
	{ VASTAI_PCI_ARG_ADDR, "addr", 0 },
	{ VASTAI_PCI_ARG_VAL, "val", 0 },
	{ VASTAI_PCI_ARG_LEN, "len", 4 },
	{ VASTAI_PCI_ARG_BAR, "bar", 0 },
	{ VASTAI_PCI_ARG_RUN, "run", 0 },
	{ VASTAI_PCI_ARG_OFFSET, "offset", 0 },
	{ VASTAI_PCI_ARG_VIR, "vir", 0 },
	{ VASTAI_PCI_ARG_REGION, "region", 0 },
	{ VASTAI_PCI_ARG_MODE, "mode", 0 },
	{ VASTAI_PCI_ARG_CASE, "case", 0 },
	{ VASTAI_PCI_ARG_USER_ADDR, "user_addr", 0 },
};

static void vastai_pci_args_init(struct vastai_pci_arg_t *args)
{
	unsigned int i;

	for (i = 0; i < VASTAI_PCI_ARG_MAX; i++)
		*(args + i) = vastai_pci_args_default[i];
}

static inline unsigned long vastai_pci_args_value(struct vastai_pci_arg_t *args,
						  unsigned char index)
{
	unsigned char i;

	for (i = 0; i < VASTAI_PCI_ARG_MAX; i++) {
		if (index == args[i].id)
			return args[i].def;
	}

	return -1;
}

int vastai_pci_malloc_udma_buf(struct vastai_pci_info *pcie_dev,
			       struct vastai_dma_buf *dm, u32 len)
{
	int ret;

	if (dm->vir || dm->dma_bus_addr) {
		if (dm->size >= len) {
			VASTAI_PCI_INFO(pcie_dev, DUMMY_DIE_ID,
					"%s not malloc again! "
					"dm_dma_bus_addr:0x%lx\n",
					__func__, dm->dma_bus_addr);
			return 0;
		} else
			vastai_udma_free(pcie_dev, dm);
	}
	ret = vastai_udma_malloc(pcie_dev, dm, len);
	if (ret != 0)
		return ret;
	memset((unsigned char *)(dm->vir), VASTAI_UDMA_TEST_VAL, len);
	VASTAI_PCI_DBG(pcie_dev, DUMMY_DIE_ID, "%s dm_phy:0x%lx\n", __func__,
		       dm->dma_bus_addr);
	return 0;
}

void vastai_pci_set_udma_buf(struct vastai_dma_buf *dm, int val)
{
	if (!dm->vir || !dm->dma_bus_addr || !dm->size) {
		VASTAI_PCI_INFO(
			NULL, DUMMY_DIE_ID,
			"%s please malloc first! dm_dma_bus_addr:0x%lx\n",
			__func__, dm->dma_bus_addr);
		return;
	}
	memset((unsigned char *)(dm->vir), val, dm->size);
	VASTAI_PCI_HEX_DUMP(KERN_INFO, "dma buf: ", DUMP_PREFIX_NONE, 32, 4,
			    (void *)(dm->vir),
			    ((dm->size > VASTAI_UDMA_TEST_LEN) ?
				     VASTAI_UDMA_TEST_LEN :
				     dm->size),
			    true);
}

void vastai_pci_print_udma_buf(struct vastai_dma_buf *dm)
{
	if (!dm->vir || !dm->dma_bus_addr || !dm->size) {
		VASTAI_PCI_INFO(
			NULL, DUMMY_DIE_ID,
			"%s please malloc first! dm_dma_bus_addr:0x%lx\n",
			__func__, dm->dma_bus_addr);
		return;
	}
	VASTAI_PCI_HEX_DUMP(KERN_INFO, "dma buf: ", DUMP_PREFIX_NONE, 32, 4,
			    (void *)(dm->vir),
			    ((dm->size > VASTAI_UDMA_TEST_LEN) ?
				     VASTAI_UDMA_TEST_LEN :
				     dm->size),
			    true);
}

int vastai_pci_free_udma_buf(struct vastai_pci_info *pcie_dev,
			     struct vastai_dma_buf *dm)
{
	if (!dm->vir || !dm->dma_bus_addr || !dm->size) {
		VASTAI_PCI_INFO(
			pcie_dev, DUMMY_DIE_ID,
			"%s no malloc, no free! dm_dma_bus_addr:0x%lx\n",
			__func__, dm->dma_bus_addr);
		return 0;
	}
	return vastai_udma_free(pcie_dev, dm);
}

/* TC21: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer to SRAM, as inbound transfer
 * 		(from remote memory to local memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc21(struct vastai_pci_info *priv)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE0, VASTAI_DIR_IN, 0x100,
					0x840000000, dmb[0].dma_bus_addr);
	return VASTAI_TEST_PASS;
}

/* TC22: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from SRAM, as outbound transfer
 * 		(from local memory to remote memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc22(struct vastai_pci_info *priv)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE0, VASTAI_DIR_OUT,
					0x100, 0x840000000,
					dmb[0].dma_bus_addr);
	vastai_pci_print_udma_buf(&dmb[0]);
	return VASTAI_TEST_PASS;
}

/* TC23: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer to SRAM, as inbound transfer
 * 		(from remote memory to local memory), specified length.
 * param in len The length of transfer in bytes.
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc23(struct vastai_pci_info *priv, u32 len)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], len);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE0, VASTAI_DIR_IN, len,
					VASTAI_CSRAM_ADDR(VASTAI_DIE0),
					dmb[0].dma_bus_addr);
	return VASTAI_TEST_PASS;
}

/* TC24: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from SRAM, as outbound transfer
 * 		(from local memory to remote memory), specified length.
 * param in len The length of transfer in bytes.
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc24(struct vastai_pci_info *priv, u32 len)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], len);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE0, VASTAI_DIR_OUT, len,
					VASTAI_CSRAM_ADDR(VASTAI_DIE0),
					dmb[0].dma_bus_addr);
	vastai_pci_print_udma_buf(&dmb[0]);
	return VASTAI_TEST_PASS;
}

/* TC25: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer to DDR, as inbound transfer
 * 		(from remote memory to local memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc25(struct vastai_pci_info *priv)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE0, VASTAI_DIR_IN, 0x100,
					VASTAI_DDR_SPACE_ADDR(VASTAI_DIE0),
					dmb[0].dma_bus_addr);
	return VASTAI_TEST_PASS;
}

/* TC26: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from DDR, as outbound transfer
 * 		(from local memory to remote memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc26(struct vastai_pci_info *priv)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE0, VASTAI_DIR_OUT,
					0x100,
					VASTAI_DDR_SPACE_ADDR(VASTAI_DIE0),
					dmb[0].dma_bus_addr);
	vastai_pci_print_udma_buf(&dmb[0]);
	return VASTAI_TEST_PASS;
}

/* TC27: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer to DDR, as inbound transfer
 * 		(from remote memory to local memory), 0x1000000 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc27(struct vastai_pci_info *priv)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_MALLOC_MAX);
	vastai_pci_malloc_udma_buf(priv, &dmb[1], VASTAI_UDMA_MALLOC_MAX);
	vastai_pci_malloc_udma_buf(priv, &dmb[2], VASTAI_UDMA_MALLOC_MAX);
	vastai_pci_malloc_udma_buf(priv, &dmb[3], VASTAI_UDMA_MALLOC_MAX);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE0, VASTAI_DIR_IN,
					VASTAI_UDMA_LENGTH_MAX,
					VASTAI_DDR_SPACE_ADDR(VASTAI_DIE0),
					dmb[3].dma_bus_addr);
	return VASTAI_TEST_PASS;
}

/* TC28: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from DDR, as outbound transfer
 * 		(from local memory to remote memory), 0x1000000 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc28(struct vastai_pci_info *priv)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_LENGTH_MAX);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE0, VASTAI_DIR_OUT,
					VASTAI_UDMA_LENGTH_MAX,
					VASTAI_DDR_SPACE_ADDR(VASTAI_DIE0),
					dmb[0].dma_bus_addr);
	vastai_pci_print_udma_buf(&dmb[0]);
	return VASTAI_TEST_PASS;
}

/* TC30: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from DDR, as ioutbound transfer
 * 		(from local memory to remote memory), specified length.
 * param in len The length of transfer in bytes.
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc30(struct vastai_pci_info *priv, u32 len)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], len);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE0, VASTAI_DIR_OUT, len,
					VASTAI_DDR_SPACE_ADDR(VASTAI_DIE0),
					dmb[0].dma_bus_addr);
	vastai_pci_print_udma_buf(&dmb[0]);
	return VASTAI_TEST_PASS;
}

/* TC52: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from SRAM, as outbound transfer
 * 		(from local memory to remote memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc52(struct vastai_pci_info *priv)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE1, VASTAI_DIR_OUT,
					0x100, VASTAI_CSRAM_ADDR(VASTAI_DIE1),
					dmb[0].dma_bus_addr);
	vastai_pci_print_udma_buf(&dmb[0]);
	return VASTAI_TEST_PASS;
}

/* TC53: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer to DDR, as inbound transfer
 * 		(from remote memory to local memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc53(struct vastai_pci_info *priv)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE1, VASTAI_DIR_IN, 0x100,
					VASTAI_DDR_SPACE_ADDR(VASTAI_DIE1),
					dmb[0].dma_bus_addr);
	return VASTAI_TEST_PASS;
}

/* TC54: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from DDR, as outbound transfer
 * 		(from local memory to remote memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc54(struct vastai_pci_info *priv)
{
	vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
	vastai_udma_start_bulk_transfer(priv, VASTAI_DIE1, VASTAI_DIR_OUT,
					0x100,
					VASTAI_DDR_SPACE_ADDR(VASTAI_DIE1),
					dmb[0].dma_bus_addr);
	vastai_pci_print_udma_buf(&dmb[0]);
	return VASTAI_TEST_PASS;
}

/* TC101: reset test via pcie driver
 * keyword: reset
 * description: Perform a reset
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc101(struct vastai_pci_info *priv)
{
	vastai_pci_reset_bus(priv);
	vastai_pci_tc22(priv);
	return VASTAI_TEST_PASS;
}

/* TC102: reset when UDMA transfer test via pcie driver
 * keyword: reset
 * description: Perform a reset
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
u8 vastai_test_udma_reset_flag;
static int vastai_pci_tc102(struct vastai_pci_info *priv)
{
	vastai_test_udma_reset_flag = 1;
	vastai_pci_tc22(priv);
	vastai_test_udma_reset_flag = 0;
	vastai_pci_tc22(priv);
	return VASTAI_TEST_PASS;
}

/* TC103: reset test via pcie driver
 * keyword: reset
 * description: Perform a reset
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc103(struct vastai_pci_info *priv)
{
	vastai_pci_reset_bus(priv);
	vastai_pci_tc52(priv);
	return VASTAI_TEST_PASS;
}

/* TC104: reset when UDMA transfer test via pcie driver
 * keyword: reset
 * description: Perform a reset
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
u8 vastai_test_udma_reset_flag;
static int vastai_pci_tc104(struct vastai_pci_info *priv)
{
	vastai_test_udma_reset_flag = 1;
	vastai_pci_tc52(priv);
	vastai_test_udma_reset_flag = 0;
	vastai_pci_tc52(priv);
	return VASTAI_TEST_PASS;
}

/* Host to fw dma sync unit test with src is virtual buf */
static int vastai_pci_tc155(struct vastai_pci_info *priv)
{
	struct vastai_dmadesc desc[2];
	int die_index = priv->dies[0].die_index;
	union core_bitmap core_id = { .val = 0 }; /* no core need be tirgger */

	desc[0].dev_addr = 0x840000000;
	desc[0].is_src_dma_addr = 0;
	desc[0].is_src_not_user_mem = 1;
	desc[0].host_addr.vir_addr = vmalloc(0x100);
	desc[0].is_host_to_dev = 1;
	desc[0].dma_lenth = 0x100;
	memset(desc[0].host_addr.vir_addr, '1', 0x100);

	desc[1].dev_addr = 0x840000100;
	desc[1].is_src_dma_addr = 0;
	desc[1].is_src_not_user_mem = 1;
	desc[1].host_addr.vir_addr = vmalloc(0x100);
	desc[1].is_host_to_dev = 1;
	desc[1].dma_lenth = 0x100;
	memset(desc[1].host_addr.vir_addr, '2', 0x100);

	core_id.decode = 0x1;
	vastai_pci_dma_transfer_sync(priv, die_index, core_id, desc, 2);
	VASTAI_PCI_HEX_DUMP(KERN_INFO, "host buf: ", DUMP_PREFIX_NONE, 32, 4,
			    desc[0].host_addr.vir_addr, 0x100, true);
	VASTAI_PCI_HEX_DUMP(KERN_INFO, "host buf: ", DUMP_PREFIX_NONE, 32, 4,
			    desc[1].host_addr.vir_addr, 0x100, true);

	vfree(desc[0].host_addr.vir_addr);
	vfree(desc[1].host_addr.vir_addr);
	return VASTAI_TEST_PASS;
}

/* Fw to Host dma sync unit test with src is virtual buf */
static int vastai_pci_tc156(struct vastai_pci_info *priv)
{
	struct vastai_dmadesc desc[2];
	int die_index = priv->dies[0].die_index;
	union core_bitmap core_id = { .val = 0 }; /* no core need be tirgger */

	desc[0].dev_addr = 0x840000000;
	desc[0].is_src_dma_addr = 0;
	desc[0].is_src_not_user_mem = 1;
	desc[0].host_addr.vir_addr = vmalloc(0x100);
	desc[0].is_host_to_dev = 0;
	desc[0].dma_lenth = 0x100;

	desc[1].dev_addr = 0x840000100;
	desc[1].is_src_dma_addr = 0;
	desc[1].is_src_not_user_mem = 1;
	desc[1].host_addr.vir_addr = vmalloc(0x100);
	desc[1].is_host_to_dev = 0;
	desc[1].dma_lenth = 0x100;

	core_id.decode = 0x1;
	vastai_pci_dma_transfer_sync(priv, die_index, core_id, desc, 2);
	VASTAI_PCI_HEX_DUMP(KERN_INFO, "host buf: ", DUMP_PREFIX_NONE, 32, 4,
			    desc[0].host_addr.vir_addr, 0x100, true);
	VASTAI_PCI_HEX_DUMP(KERN_INFO, "host buf: ", DUMP_PREFIX_NONE, 32, 4,
			    desc[1].host_addr.vir_addr, 0x100, true);
	vfree(desc[0].host_addr.vir_addr);
	vfree(desc[1].host_addr.vir_addr);
	return VASTAI_TEST_PASS;
}

/* Host to fw dma sync , long length */
/* test Die 0 */
static int vastai_pci_tc157(struct vastai_pci_info *priv)
{
	struct vastai_dmadesc desc[2];
	int die_index = priv->dies[0].die_index;
	union core_bitmap core_id = { .val = 0 }; /* no core need be tirgger */
	struct vastai_dma_buf dm[2];
	int i;
	u32 desc_cnt = (get_random_u32() % 2) + 1;
	u32 offset = 0;
	u32 is_host_to_dev = get_random_u32() % 2;

	for (i = 0; i < desc_cnt; i++) {
		u32 test_len = (get_random_u32() % 0x200000) + 900000;
		desc[i].dev_addr = ALIGN(0x840000000 + offset, 0x1000);
		desc[i].is_src_dma_addr = 1;
		desc[i].is_src_not_user_mem = 1;
		vastai_udma_malloc(priv, &dm[i], test_len);
		desc[i].host_addr.dma_addr = dm[i].dma_bus_addr;
		desc[i].host_addr.vir_addr = dm[i].vir;
		desc[i].is_host_to_dev = is_host_to_dev;
		desc[i].dma_lenth = test_len;
		memset(desc[i].host_addr.vir_addr, '2', test_len);
		offset += test_len;
	}

	core_id.decode = 0x1;
	vastai_pci_dma_transfer_sync(priv, die_index, core_id, desc, desc_cnt);

	for (i = 0; i < desc_cnt; i++) {
		vastai_udma_free(priv, &dm[i]);
	}
	return VASTAI_TEST_PASS;
}

static int vastai_pci_tc158(struct vastai_pci_info *priv)
{
	union core_bitmap core_id = { .vdsp = BIT(0) };
	int i = 0;
	static int index = 0;

	for (i = 0; i < 10; i++) {
		u32 msg_buf = 0x80000000 + index;
		index += 1;
		vastai_pci_send_msg(priv, 0, core_id, &msg_buf,
				    sizeof(msg_buf));
	}
	return 0;
}

/* test Die 1 */
static int vastai_pci_tc159(struct vastai_pci_info *priv)
{
	struct vastai_dmadesc desc[2];
	int die_index = priv->dies[1].die_index;
	union core_bitmap core_id = { .val = 0 }; /* no core need be tirgger */
	struct vastai_dma_buf dm[2];
	int i;
	u32 desc_cnt = (get_random_u32() % 2) + 1;
	u32 offset = 0;
	u32 is_host_to_dev = get_random_u32() % 2;

	for (i = 0; i < desc_cnt; i++) {
		u32 test_len = (get_random_u32() % 0x200000) + 900000;
		desc[i].dev_addr = ALIGN(0x840000000 + offset, 0x1000);
		desc[i].is_src_dma_addr = 1;
		desc[i].is_src_not_user_mem = 1;
		vastai_udma_malloc(priv, &dm[i], test_len);
		desc[i].host_addr.dma_addr = dm[i].dma_bus_addr;
		desc[i].host_addr.vir_addr = dm[i].vir;
		desc[i].is_host_to_dev = is_host_to_dev;
		desc[i].dma_lenth = test_len;
		memset(desc[i].host_addr.vir_addr, '2', test_len);
		offset += test_len;
	}

	core_id.decode = 0x1;
	vastai_pci_dma_transfer_sync(priv, die_index, core_id, desc, desc_cnt);

	for (i = 0; i < desc_cnt; i++) {
		vastai_udma_free(priv, &dm[i]);
	}
	return VASTAI_TEST_PASS;
}

/*
   1. send buf to die0.
   2. move buf from die0 to die1
   3. recv buf from die1
   4. check buf
   */
static int vastai_pci_tc160(struct vastai_pci_info *priv)
{
	struct vastai_dmadesc desc[2];
	int die_index = priv->dies[0].die_index;
	union core_bitmap core_id = { .val = 0 }; /* no core need be tirgger */
	int i;

	desc[0].dev_addr = 0x840000000;
	desc[0].is_src_dma_addr = 0;
	desc[0].is_src_not_user_mem = 1;
	desc[0].host_addr.vir_addr = vmalloc(0x100);
	desc[0].is_host_to_dev = 1;
	desc[0].dma_lenth = 0x100;

	memset(desc[0].host_addr.vir_addr, 'a', 0x100);
	vastai_pci_dma_transfer_sync(priv, die_index, core_id, desc, 1);

	vastai_pci_data_movement(priv, die_index, 0x840000000, 0x840000000,
				 0x100, 1);

	die_index = priv->dies[1].die_index;
	desc[1].dev_addr = 0x840000000;
	desc[1].is_src_dma_addr = 0;
	desc[1].is_src_not_user_mem = 1;
	desc[1].host_addr.vir_addr = vmalloc(0x100);
	desc[1].is_host_to_dev = 0;
	desc[1].dma_lenth = 0x100;

	vastai_pci_dma_transfer_sync(priv, die_index, core_id, &(desc[1]), 1);

	for (i = 0; i < 0x100; i++) {
		if (((char *)desc[1].host_addr.vir_addr)[i] !=
		    ((char *)desc[0].host_addr.vir_addr)[i])
			printk("%s check error\n", __FUNCTION__);
	}

	return VASTAI_TEST_PASS;
}

/*
   1. send buf to die0.
   2. move buf from die1 to die0
   3. recv buf from die0
   4. check buf
   */
static int vastai_pci_tc161(struct vastai_pci_info *priv)
{
	struct vastai_dmadesc desc[2];
	int die_index = priv->dies[1].die_index;
	union core_bitmap core_id = { .val = 0 }; /* no core need be tirgger */
	int i;

	desc[0].dev_addr = 0x840000000;
	desc[0].is_src_dma_addr = 0;
	desc[0].is_src_not_user_mem = 1;
	desc[0].host_addr.vir_addr = vmalloc(0x100);
	desc[0].is_host_to_dev = 1;
	desc[0].dma_lenth = 0x100;

	memset(desc[0].host_addr.vir_addr, 'a', 0x100);
	vastai_pci_dma_transfer_sync(priv, die_index, core_id, desc, 1);

	die_index = priv->dies[0].die_index;
	vastai_pci_data_movement(priv, die_index, 0x840000000, 0x840000000,
				 0x100, 0);

	desc[1].dev_addr = 0x840000000;
	desc[1].is_src_dma_addr = 0;
	desc[1].is_src_not_user_mem = 1;
	desc[1].host_addr.vir_addr = vmalloc(0x100);
	desc[1].is_host_to_dev = 0;
	desc[1].dma_lenth = 0x100;

	vastai_pci_dma_transfer_sync(priv, die_index, core_id, &(desc[1]), 1);

	for (i = 0; i < 0x100; i++) {
		if (((char *)desc[1].host_addr.vir_addr)[i] !=
		    ((char *)desc[0].host_addr.vir_addr)[i])
			printk("%s check error\n", __FUNCTION__);
	}

	return VASTAI_TEST_PASS;
}
int vastai_send_ctrl_cmd(struct vastai_pci_info *pci_info, u32 die_index,
			 u32 core_bit_map, u64 info);
static int vastai_pci_tc162(struct vastai_pci_info *priv)
{
	int ret;
	/* cmcu */
	ret = vastai_send_ctrl_cmd(priv, 0, 0x1 << 12, 0);
	if (ret) {
		printk("ERROR: test failed, errno %d\n", ret);
	} else {
		printk("SUCCESS\n");
	}
	return 0;
}

static int vastai_pci_tc163(struct vastai_pci_info *priv)
{
	int ret;
	/* vdsps */
	ret = vastai_send_ctrl_cmd(priv, 0, 0xf << 8, 0);
	if (ret) {
		printk("ERROR: test failed, errno %d\n", ret);
	} else {
		printk("SUCCESS\n");
	}
	return 0;
}

static int vastai_pci_tc164(struct vastai_pci_info *priv)
{
	int ret;
	/* null */
	ret = vastai_send_ctrl_cmd(priv, 0, 0x0, 0);
	if (ret) {
		printk("ERROR: test failed, errno %d\n", ret);
	} else {
		printk("SUCCESS\n");
	}
	return 0;
}

static int vastai_pci_test_case(struct vastai_pci_info *priv,
				struct char_drv_info *dev, u32 case_no)
{
	int ret = 0;
	u32 len;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"******** VASTAI PCI TEST[%d] start ********\n",
			case_no);
	len = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
	switch (case_no) {
	case 21:
		ret = vastai_pci_tc21(priv);
		break;
	case 22:
		ret = vastai_pci_tc22(priv);
		break;
	case 23:
		ret = vastai_pci_tc23(priv, len);
		break;
	case 24:
		ret = vastai_pci_tc24(priv, len);
		break;
	case 25:
		ret = vastai_pci_tc25(priv);
		break;
	case 26:
		ret = vastai_pci_tc26(priv);
		break;
	case 27:
		ret = vastai_pci_tc27(priv);
		break;
	case 28:
		ret = vastai_pci_tc28(priv);
		break;
	case 30:
		ret = vastai_pci_tc30(priv, len);
		break;
	case 52:
		ret = vastai_pci_tc52(priv);
		break;
	case 53:
		ret = vastai_pci_tc53(priv);
		break;
	case 54:
		ret = vastai_pci_tc54(priv);
		break;
	case 101:
		ret = vastai_pci_tc101(priv);
		break;
	case 102:
		ret = vastai_pci_tc102(priv);
		break;
	case 103:
		ret = vastai_pci_tc103(priv);
		break;
	case 104:
		ret = vastai_pci_tc104(priv);
		break;
	case 155:
		ret = vastai_pci_tc155(priv);
		break;
	case 156:
		ret = vastai_pci_tc156(priv);
		break;
	case 157:
		ret = vastai_pci_tc157(priv);
		break;
	case 158:
		ret = vastai_pci_tc158(priv);
		break;
	case 159:
		ret = vastai_pci_tc159(priv);
		break;
	case 160:
		ret = vastai_pci_tc160(priv);
		break;
	case 161:
		ret = vastai_pci_tc161(priv);
		break;
	case 162:
		ret = vastai_pci_tc162(priv);
		break;
	case 163:
		ret = vastai_pci_tc163(priv);
		break;
	case 164:
		ret = vastai_pci_tc164(priv);
		break;
	default:
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "Unsupported test "
			       "case_no = %d\n",
			       case_no);
		break;
	}
	if (ret == VASTAI_TEST_PASS)
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"******** VASTAI PCI TEST[%d] done "
				"********\n",
				case_no);
	else
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "******** VASTAI PCI TEST[%d] fail "
			       "ret = %d ********\n",
			       case_no, ret);
	return ret;
}

static int vastai_pci_cmdline_args(char *cmdline, char *cmd,
				   struct vastai_pci_arg_t *args, int argc)
{
	int status = 0, DONE, FOUND = 0;
	char *end, *argname, *cp;
	unsigned short base;
	unsigned long result, value, val, i, j;

	for (i = 0; i < argc; i++)
		args[i].def = VASTAI_VAL_DW_MASK;

	/* get cmd */
	while (*cmdline == ' ' || *cmdline == '\t')
		cmdline++;

	while (*cmdline != ' ' && *cmdline != '\t' && *cmdline != '\0') {
		*cmd = *cmdline;
		cmd++;
		cmdline++;
	}
	*cmd = '\0';
	if (*cmdline == '\0')
		goto WEDONE;

	*cmdline = '\0';
	cmdline++;
	while (*cmdline == ' ' || *cmdline == '\t')
		cmdline++;

	end = cmdline;
	while (*end == ' ' || *end == '\t')
		end++;

	/*
	 * Parse cmdline
	 */
	DONE = (*end == '\0') ? 1 : 0;
	while (!DONE) {
		/* get the register name */
		while (*end != '=' && *end != '\0')
			end++;
		if (*end == '\0') {
			status = 1;
			goto WEDONE;
		}
		*end = '\0';
		argname = cmdline;
		/* now get value to write to register */
		cmdline = ++end;
		/* if there's whitespace after the '=', exit with an error */
		if (*end == ' ' || *end == '\t' || *end == '\n') {
			status = 1;
			goto WEDONE;
		}
		while (*end != ' ' && *end != '\t' && *end != '\n' &&
		       *end != '\0')
			end++;
		if (*end == '\0')
			DONE = 1;
		else
			*end = '\0';

		if (!strcmp(argname, "file") || !strcmp(argname, "filec")) {
			val = 1;
		} else {
			/* get the base, convert value to base-10 if necessary
			 */
			val = 0;
			result = 0;
			cp = cmdline;
			if (cp[0] == '0' && (cp[1] == 'x' || cp[1] == 'X')) {
				base = 16;
				cp += 2;

			} else {
				base = 10;
			}
			while (isxdigit(*cp)) {
				value = isdigit(*cp) ?
						(*cp - '0') :
						((islower(*cp) ? toupper(*cp) :
								 *cp) -
						 'A' + 10);

				result = result * base + value;
				cp++;
			}

			val = result;
		}

		FOUND = 0;
		/*
		 * verify the register arg is valid, and if the value is not
		 * too big, write it to the corresponding location in arg_vals
		 */
		for (j = 0; j < argc && !FOUND; j++) {
			if (!strcmp(argname, args[j].name)) {
				args[j].def = val;
				FOUND = 1;
			}
		}
		if (!FOUND) {
			VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID, "arg %s err\n",
				       argname);
			status = 0;
			goto WEDONE;
		}

		/*
		 * point cmdline and end to next non-whitespace
		 * (next argument)
		 */
		cmdline = ++end;
		while (*cmdline == ' ' || *cmdline == '\t' || *cmdline == '\n')
			cmdline++;
		end = cmdline;
		/*
		 * if, after skipping whitespace, we hit end of line or EOF,
		 * we're done
		 */
		if (*end == '\0')
			DONE = 1;
	}

WEDONE:
	return status;
}

static int vastai_xspiflash_read(struct vastai_pci_info *priv, u32 die_id,
				unsigned long addr, struct char_drv_info *dev,
				bool print_flag)
{
	int ret = 0;
	int i = 0;
	unsigned char print_str[64];

	ret = vastai_pci_read_xspi(priv, die_id, addr, dev->dm->vir,
			dev->valid_data_length, 0);
	if (ret) {
		return ret;
	}

	if (print_flag == 1) {
		for (i = 0; i < dev->valid_data_length; i += 16) {
			sprintf(print_str,
				"%s [read] 0x%010lx: ", VASTAI_PCI_DEBUG_PREFIX,
				(addr + i + XSPI_READ_BASE));
			VASTAI_PCI_HEX_DUMP(KERN_WARNING, print_str,
					    DUMP_PREFIX_NONE, 16, 4,
					    dev->dm->vir + i, 16, true);
		}
	}
	return 0;
}

void vastai_pci_test_vftest(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	vastai_send_msg_to_peer_test(priv, val);
}

void vastai_pci_test_vfon(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long addr, val;
	int err = 0;
	int ret = 0;
	struct vf_info_t tmp;

	addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

	//check old value
	ret = vastai_pci_mem_read(priv, priv->dies[0].die_index,
		(CSRAM_BASE_ADDR + VF_INFO_OFFSET), &priv->vf_info,
		sizeof (struct vf_info_t));
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s vfon [%d -> %ld]\n",__func__, priv->vf_info.vf_activated, val);
	err |= ret;

	//write new value
	priv->vf_info.vf_activated = (u32)val;
	ret = vastai_pci_mem_write(priv, priv->dies[0].die_index,
					 (CSRAM_BASE_ADDR + VF_INFO_OFFSET), &priv->vf_info, sizeof (struct vf_info_t));
	err |= ret;

	//check value is correctly written
	ret = vastai_pci_mem_read(priv, priv->dies[0].die_index,
					 (CSRAM_BASE_ADDR + VF_INFO_OFFSET), &tmp, sizeof (struct vf_info_t));

	err |= ret;
	if (tmp.vf_activated != val) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s vfon expect %ld, but %d\n",__func__, val, tmp.vf_activated);
	}

	if (err < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,	"%s error:%d\n", __func__, err);
	}
}

/* read configuration space */
void vastai_pci_test_rcfg(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long addr;
	int ret = 0;
	int i = 0;
	unsigned char print_str[64];

	addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
	dev->valid_data_length =
		vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"%s [read config] len=%d bytes\n", __func__,
			dev->valid_data_length);
	if (dev->valid_data_length)
		ret = vastai_pci_config_read(priv, addr, dev->dm->vir,
					     dev->valid_data_length);
	if (!ret) {
		for (i = 0; i < dev->valid_data_length; i += 16) {
			sprintf(print_str,
				"%s config "
				"space "
				"0x%03lx: ",
				VASTAI_PCI_DEBUG_PREFIX, (i + addr));
			VASTAI_PCI_HEX_DUMP(KERN_WARNING, print_str,
					    DUMP_PREFIX_NONE, 16, 4,
					    dev->dm->vir + i, 16, true);
		}
	}
}

void vastai_pci_test_wcfg(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long addr, val;
	int ret = 0;

	addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	dev->valid_data_length =
		vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"%s [write config] addr=0x%lx, "
			"len=%d, val=0x%lx\n",
			__func__, addr, dev->valid_data_length, val);
	if (dev->valid_data_length)
		ret = vastai_pci_config_write(priv, addr, (char *)&val,
					      dev->valid_data_length);
}

/* read reg:
* cmd format e.g. "read addr=0x4a000000 len=100"
*/
void vastai_pci_test_read(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long addr;
	int ret = 0;
	unsigned char print_str[64];
	u32 i;

	addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
	dev->valid_data_length =
		vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
	if (dev->valid_data_length <= 0) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"%s read len uses default 4 bytes\n",
				__func__);
		dev->valid_data_length = 4;
	} else if (dev->valid_data_length > dev->dm->size) {
		dev->valid_data_length = dev->dm->size;
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s read len overflow, "
			       "change to %d bytes\n",
			       __func__, dev->valid_data_length);
	}

	ret = vastai_pci_mem_read(priv, priv->dies[0].die_index, addr,
				  dev->dm->vir, dev->valid_data_length);

	if (ret < 0)
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s [read] 0x%lx, len=%d, error:%d\n",
			       __func__, addr, dev->valid_data_length,
			       ret);
	else {
		for (i = 0; i < dev->valid_data_length; i += 16) {
			sprintf(print_str, "%s [read] 0x%010lx: ",
				VASTAI_PCI_DEBUG_PREFIX, (addr + i));
			VASTAI_PCI_HEX_DUMP(KERN_WARNING, print_str,
					    DUMP_PREFIX_NONE, 16, 4,
					    dev->dm->vir + i, 16, true);
		}
	}

}

void vastai_pci_test_read_buf(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long addr;
	int ret = 0;

	addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
	dev->valid_data_length =
		vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
	if (dev->valid_data_length <= 0) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"%s read len uses default 4 bytes\n",
				__func__);
		dev->valid_data_length = 4;
	} else if (dev->valid_data_length > dev->dm->size) {
		dev->valid_data_length = dev->dm->size;
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s read len overflow, "
			       "change to %d bytes\n",
			       __func__, dev->valid_data_length);
	}
	ret = vastai_pci_mem_read(priv, priv->dies[0].die_index, addr,
				  dev->dm->vir, dev->valid_data_length);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s [read] 0x%lx, len=%d, error:%d\n",
			       __func__, addr, dev->valid_data_length,
			       ret);
}


/* write reg:
* cmd format e.g. "write addr=0x4a000000 val=0x0 len=100"
*/
void vastai_pci_test_write(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long addr, val;
	u32 i;
	int ret = 0;

	addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	dev->valid_data_length =
		vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
	if (dev->valid_data_length <= 0) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"%s write len uses default 4 bytes\n",
				__func__);
		dev->valid_data_length = 4;
	} else if (dev->valid_data_length > dev->dm->size) {
		dev->valid_data_length = dev->dm->size;
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s write len overflow, "
			       "change to %d bytes\n",
			       __func__, dev->valid_data_length);
	}
	for (i = 0; i < dev->valid_data_length; i += 4) {
		((char *)dev->dm->vir)[i] = val & 0xFF;
		if ((i + 1) == dev->valid_data_length)
			break;
		((char *)dev->dm->vir)[i + 1] = (val >> 8) & 0xFF;
		if ((i + 2) == dev->valid_data_length)
			break;
		((char *)dev->dm->vir)[i + 2] = (val >> 16) & 0xFF;
		if ((i + 3) == dev->valid_data_length)
			break;
		((char *)dev->dm->vir)[i + 3] = (val >> 24) & 0xFF;
	}
	ret = vastai_pci_mem_write(priv, priv->dies[0].die_index, addr,
				   dev->dm->vir,
				   dev->valid_data_length);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s [write] 0x%lx, len=%d, error:%d\n",
			       __func__, addr, dev->valid_data_length,
			       ret);
	else
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"%s [write] addr=0x%lx, "
				"len=%d, val=0x%lx\n",
				__func__, addr, dev->valid_data_length,
				val);

}

void vastai_pci_test_write_buf(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long addr;
	int ret = 0;

	unsigned long user_addr;

	addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
	user_addr = vastai_pci_args_value(dev->args,
					  VASTAI_PCI_ARG_USER_ADDR);
	dev->valid_data_length =
		vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
	if (dev->valid_data_length > dev->dm->size) {
		dev->valid_data_length = dev->dm->size;
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s write_buf len overflow, "
			       "change to %d bytes\n",
			       __func__, dev->valid_data_length);
	}

	ret = copy_from_user(dev->dm->vir, (char __user *)user_addr,
			     dev->valid_data_length);
	if (ret)
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s copy_from_user is error\n",
			       __func__);
	ret = vastai_pci_mem_write(priv, priv->dies[0].die_index, addr,
				   dev->dm->vir,
				   dev->valid_data_length);
	if (ret < 0)
		VASTAI_PCI_ERR(
			priv, DUMMY_DIE_ID,
			"%s [write_buf] 0x%lx, len=%d, error:%d\n",
			__func__, addr, dev->valid_data_length, ret);
	else
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
			       "%s [write_buf] addr=0x%lx, "
			       "len=%d\n",
			       __func__, addr, dev->valid_data_length);
}

void vastai_pci_test_getver(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_pci_show_fw_ver(priv, 0x3FF);
}

void vastai_pci_test_boot(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_device_reset(priv, VASTAI_RESET_BOOT_DEV);
}

/* reset pcie bus */
void vastai_pci_test_hotreset(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	int ret;

	ret = vastai_pci_state_change(priv, VASTAI_NORMAL_STATE, VASTAI_RESET_STATE);
	if (ret)
		return;
	vastai_pci_reset_bus(priv);
	ret = vastai_pci_state_change(priv, VASTAI_RESET_STATE, VASTAI_NO_RUN_STATE);
}

/* reset and boot device */
void vastai_pci_test_hotreboot(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	/* [mod] gang 2022-5-18
	 * Temp for diags driver:
	 * Some diags cases will hang smcu, then re-initial reset fail.
	 * We need to use hot-reset to reset device.
	 */
	int ret;

	ret = vastai_pci_state_change(priv, VASTAI_NORMAL_STATE, VASTAI_RESET_STATE);
	if (ret)
		return;
	vastai_pci_reset_bus(priv);
	ret = vastai_pci_state_change(priv, VASTAI_RESET_STATE, VASTAI_NO_RUN_STATE);
	if (ret)
		return;
	vastai_device_reset(priv, VASTAI_RESET_BOOT_DEV);
}

/* reset mcu and dsp */
void vastai_pci_test_reset(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	/* send msg_irq trigger a re-initial reset. */
	vastai_device_reset(priv, VASTAI_RESET_MSG_TRIGER);
}

/* reset and boot device, re-download ddr bandwidth fw */
void vastai_pci_test_reboot_ddrbw(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	/* trigger a re-initial reset and reboot dev . */
	priv->ddr_bw_test = 1;
	vastai_device_reset(priv, VASTAI_RESET_MSG_TRIGER |
					  VASTAI_RESET_BOOT_DEV);
	priv->ddr_bw_test = 0;
}

/* reset and boot device */
void vastai_pci_test_reboot(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;

	/* trigger a re-initial reset and reboot dev . */
	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	if (val != VASTAI_VAL_DW_MASK && val != priv->vadev.dev_cap) {
		priv->vadev.dev_cap = val;
		vastai_send_pcie_cmd(priv, priv->dies[0].die_index,
				     VASTAI_PCIE_SUB_POWERON, 0);
		msleep(100);
	}

	vastai_device_reset(priv, VASTAI_RESET_MSG_TRIGER |
					  VASTAI_RESET_BOOT_DEV);
}

/* power on video module */
void vastai_pci_test_poweron_video(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_send_pcie_cmd(priv, priv->dies[0].die_index,
					     VASTAI_PCIE_SUB_POWERON, 0);
			msleep(100);
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
					"poweron video finished!\n");
}

void vastai_pci_test_poweroff_video(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_send_pcie_cmd(priv, priv->dies[0].die_index,
					     VASTAI_PCIE_SUB_POWEROFF, 0);
			msleep(100);
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
					"poweron video finished!\n");

}

void vastai_pci_test_poweron(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_send_pcie_cmd(priv, priv->dies[0].die_index,
				     VASTAI_PCIE_SUB_POWERON, 0);
		msleep(100);
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"poweron video finished!\n");
}

void vastai_pci_test_bmcu_upgrade(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_send_pcie_cmd(priv, priv->dies[0].die_index,
			     VASTAI_PCIE_SUB_BMCU_UPGRADE, 0);
	ssleep(16);
	vastai_pci_get_fw_ver(priv, 0x4);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "upgrade BMCU finished!\n");

}

void vastai_pci_test_bmcu_degrade(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_send_pcie_cmd(priv, priv->dies[0].die_index,
					     VASTAI_PCIE_SUB_BMCU_DEGRADE, 0);
			ssleep(16);
			vastai_pci_get_fw_ver(priv, 0x4);
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "degrade BMCU finished!\n");
}

/* upgrade BMCU by pcie*/
void vastai_pci_test_flashbmcu(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_pci_flash_bmcu(priv, 0);
}

/* select pcie generation between host and master die */
void vastai_pci_test_selgen(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	vastai_pci_generation_sel(priv, val);
}

/* select pcie generation between die0 and die1 */
void vastai_pci_test_seld0gen(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	vastai_pci_generation_sel_slave_die(priv, val, VASTAI_DIE0);
}

/* select pcie generation between die1 and die2 */
void vastai_pci_test_seld1gen(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	vastai_pci_generation_sel_slave_die(priv, val, VASTAI_DIE1);
}

/* select pcie generation between die2 and die3 */
void vastai_pci_test_seld2gen(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	vastai_pci_generation_sel_slave_die(priv, val, VASTAI_DIE2);
}

void vastai_pci_test_getloglevel(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "current loglevel=[0x%x]\n",
			vastai_pci_log_level);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"(bits correspond to levels in turn: perf, "
			"data, debug, normal)\n");
}

/* set log level, "setloglevel val=xxx" */
void vastai_pci_test_setloglevel(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	vastai_pci_log_level = val;
	if (vastai_pci_log_level == 0)
		VASTAI_PCI_ERR(
			priv, DUMMY_DIE_ID,
			"current loglevel=[0x%x], printing is closed\n",
			vastai_pci_log_level);
	else
		VASTAI_PCI_INFO(
			priv, DUMMY_DIE_ID,
			"set level=0x%lx, current loglevel=[0x%x]\n",
			val, vastai_pci_log_level);

}

/* set log level, "setvdspdep val=xxx" */
void vastai_pci_test_setvdspdep(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;
	u32 i;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	for (i = 0; i < priv->die_num; i++) {
		vastai_pci_set_vdsp_depth(priv, i, val, 1);
	}
}

/* malloc a dma buf for testing */
void vastai_pci_test_mallocbuf(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	dev->valid_data_length =
		vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
	if (dev->valid_data_length <= 0) {
		dev->valid_data_length = VASTAI_UDMA_TEST_LEN;
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"%s malloc default len=%d\n", __func__,
				dev->valid_data_length);
	}
	vastai_pci_malloc_udma_buf(priv, &dmb[0],
				   dev->valid_data_length);
}

/* set a dma buf for testing */
void vastai_pci_test_setbuf(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	vastai_pci_set_udma_buf(&dmb[0], val);
}

/* print a dma buf for testing */
void vastai_pci_test_printbuf(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_pci_print_udma_buf(&dmb[0]);
}

/* free a dma buf for testing */
void vastai_pci_test_freebuf(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_pci_free_udma_buf(priv, &dmb[0]);
}

/* flash xspi bootcode, "flashxspi_app val=xxx" */
void vastai_pci_test_flashxspi_app(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;
	u32 i;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	if ((val >= 0) && (val < priv->die_num)) {
		vastai_pci_flash_xspi(priv, val, dev, 0);
	} else if (val == VASTAI_VAL_DW_MASK) {
		for (i = 0; i < priv->die_num; i++)
			vastai_pci_flash_xspi(priv, i, dev, 0);
	} else {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s input error val=0x%lx\n", __func__,
			       val);
	}
	vastai_pci_get_fw_ver(priv, 0xC00);
}

/* flash xspi bootcode, "flashxspi val=xxx" */
void vastai_pci_test_flashxspi(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;
	u32 i;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	if ((val >= 0) && (val < priv->die_num)) {
		vastai_pci_flash_xspi(priv, val, dev, 1);
	} else if (val == VASTAI_VAL_DW_MASK) {
		for (i = 0; i < priv->die_num; i++)
			vastai_pci_flash_xspi(priv, i, dev, 1);
	} else {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s input error val=0x%lx\n", __func__,
			       val);
	}
	vastai_pci_get_fw_ver(priv, 0xC00);

}

/* set flash write protect enable */
void vastai_pci_test_wpen(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;
	u32 i;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	if ((val >= 0) && (val < priv->die_num)) {
		vastai_pci_set_flash_wp(priv, val, dev, 1);
	} else if (val == VASTAI_VAL_DW_MASK) {
		for (i = 0; i < priv->die_num; i++)
			vastai_pci_set_flash_wp(priv, i, dev, 1);
	} else
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s input error val=0x%lx\n", __func__,
			       val);
}

/* set flash write protect disable */
void vastai_pci_test_wpdis(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;
	u32 i;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	if ((val >= 0) && (val < priv->die_num)) {
		vastai_pci_set_flash_wp(priv, val, dev, 0);
	} else if (val == VASTAI_VAL_DW_MASK) {
		for (i = 0; i < priv->die_num; i++)
			vastai_pci_set_flash_wp(priv, i, dev, 0);
	} else
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s input error val=0x%lx\n", __func__,
			       val);

}

/* read xspi bootcode, "readxspi addr=xxx val=xxx len=xxx" */
void vastai_pci_test_readxspi(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long addr, val;
	u32 i = 0;

	addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	dev->valid_data_length =
		vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
	if ((val >= 0) && (val < priv->die_num))
		vastai_xspiflash_read(priv, val, addr, dev, 1);
	else if (val == VASTAI_VAL_DW_MASK) {
		for (i = 0; i < priv->die_num; i++)
			vastai_xspiflash_read(priv, i, addr, dev, 1);
	} else
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s input error val=0x%lx\n", __func__,
			       val);
}

void vastai_pci_test_test(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;
	int ret;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	ret = vastai_pci_test_case(priv, dev, val);
}

void vastai_pci_test_cedar_sym(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	int ret;
	u32 i;
	unsigned int reg_val = 0;

	u64 zero = 0;
	dev->valid_data_length = 0;
	for (i = 0; i < priv->die_num; i++) {
		ret = vastai_send_pcie_cmd(
			priv, priv->dies[i].die_index,
			VASTAI_PCIE_SUB_CEDAR_SYM_ALG, 0);
		do {
			vastai_pci_mem_read(priv,
					    priv->dies[i].die_index,
					    SLT_TEST_CEDAR_SYM_DONE,
					    &reg_val, 4);
		} while (reg_val != VASTAI_VAL_DW_MASK);
		vastai_pci_mem_read(priv, priv->dies[i].die_index,
				    SLT_TEST_CEDAR_SYM_RES, &reg_val,
				    4);
		if (0x0 == reg_val)
			VASTAI_PCI_INFO(priv, priv->dies[i].die_index,
					"cedar sym slt test pass!\n");
		else
			VASTAI_PCI_ERR(priv, priv->dies[i].die_index,
				       "cedar sym slt test fail!\n");
		memcpy((dev->dm->vir + 4 * i), &reg_val, 4);
		dev->valid_data_length += 4;
		vastai_pci_mem_write(priv, priv->dies[i].die_index,
				     SLT_TEST_CEDAR_SYM_RES, &zero,
				     sizeof(zero));
	}
}

void vastai_pci_test_cedar_asy(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	int ret;
	u32 i;
	unsigned int reg_val = 0;

	u64 zero = 0;
	dev->valid_data_length = 0;
	for (i = 0; i < priv->die_num; i++) {
		ret = vastai_send_pcie_cmd(
			priv, priv->dies[i].die_index,
			VASTAI_PCIE_SUB_CEDAR_ASYM_ALG, 0);
		do {
			vastai_pci_mem_read(priv,
					    priv->dies[i].die_index,
					    SLT_TEST_CEDAR_ASY_DONE,
					    &reg_val, 4);
		} while (reg_val != 0xF);
		vastai_pci_mem_read(priv, priv->dies[i].die_index,
				    SLT_TEST_CEDAR_ASY_RES, &reg_val,
				    4);
		if (0x0 == reg_val)
			VASTAI_PCI_INFO(priv, priv->dies[i].die_index,
					"cedar asy slt test pass!\n");
		else
			VASTAI_PCI_ERR(priv, priv->dies[i].die_index,
				       "cedar asy slt test fail!\n");
		memcpy((dev->dm->vir + 4 * i), &reg_val, 4);
		dev->valid_data_length += 4;
		vastai_pci_mem_write(priv, priv->dies[i].die_index,
				     SLT_TEST_CEDAR_ASY_RES, &zero,
				     sizeof(zero));
	}
}

/* cedar svc for customer*/
void vastai_pci_test_cedar_svc(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	int ret;
	unsigned int reg_val = 0;

	u64 zero = 0;
	int die_id =
		vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	dev->valid_data_length = 0;
	vastai_pci_mem_write(priv, priv->dies[die_id].die_index,
			     CEDAR_SVC_STS_RES, &zero,
			     sizeof(zero)); //clear sts and done_flag
	ret = vastai_send_pcie_cmd(priv, priv->dies[die_id].die_index,
				   VASTAI_PCIE_SUB_CEDAR_SVC, 0);
	VASTAI_PCI_INFO(priv, priv->dies[die_id].die_index,
			"cedar svc cmd send to smcu!\n");
	do {
		vastai_pci_mem_read(priv, priv->dies[die_id].die_index,
				    CEDAR_SVC_DONE_FLAG, &reg_val, 4);
	} while (reg_val != 0xFF);
	VASTAI_PCI_INFO(priv, priv->dies[die_id].die_index,
			"cedar svc cmd done!\n");
	vastai_pci_mem_read(priv, priv->dies[die_id].die_index,
			    CEDAR_SVC_STS_RES, &reg_val, 4);
	if (0x0 == reg_val)
		VASTAI_PCI_INFO(priv, priv->dies[die_id].die_index,
				"cedar svc pass!\n");
	else if (0xA5 == reg_val)
		VASTAI_PCI_ERR(priv, priv->dies[die_id].die_index,
			       "cedar svc no entry map!\n");
	else
		VASTAI_PCI_ERR(priv, priv->dies[die_id].die_index,
			       "cedar svc fail!\n");
	memcpy(dev->dm->vir, &reg_val, 4);
	dev->valid_data_length = 4;
}

/**
 * disable smcu heartbeat counter for testing heartbeat
 * 0 - hang timer counter of smcu
 * 1 - hang thread counter of smcu
 */
void vastai_pci_test_hang_smcu(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	if (val == 0)
		vastai_send_pcie_cmd(priv, priv->dies[0].die_index,
				     VASTAI_PCIE_SUB_HANG_SMCU, 0);
	else if (val == 1)
		vastai_send_pcie_cmd(priv, priv->dies[0].die_index,
				     VASTAI_PCIE_SUB_HANG_SMCU_THREAD,
				     0);
}

void vastai_pci_test_toggle_dma_callback(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;

	int i = 0;
	u32 flage = 0;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	if (val == 0) {
		flage = VASTAI_PCIE_SUB_CLOSE_DMA_CALLBACK;
	} else {
		flage = VASTAI_PCIE_SUB_OPEN_DMA_CALLBACK;
	}
	for (i = 0; i < priv->die_num; i++) {
		vastai_send_pcie_cmd(priv, priv->dies[i].die_index,
				     flage, 0);
	}
}

void vastai_pci_test_iommu(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
		"iommu is %s\n", iommu_is_enable(priv) ? "enable" :"disable");
}

void vastai_pci_test_core_reset(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	u32 core_id;
	u32 die_id = 0;

	core_id = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	vastai_send_pcie_cmd(priv, priv->dies[die_id].die_index,
				VASTAI_PCIE_SUB_CORE_RESET, core_id);

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"reset core id %d\n", core_id);
}

void vastai_pci_test_exp_tri(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	u32 core_id;
	u32 die_id = 0;
	core_id = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

	printk("core_id:%x\n", core_id);
	vastai_send_pcie_cmd(priv, priv->dies[die_id].die_index,
				VASTAI_PCIE_SUB_TRI_EXP_TEST, core_id);

}

void vastai_pci_test_switchbbox(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;
	int ret;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	ret = vastai_bbox_set(priv, (u32)val);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"switchbbox:0x%x\n", (u32)val);
}

void vastai_pci_test_set_state(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"origin state:0x%x\n", (u32)atomic_read(&priv->pci_state));
	atomic_set(&(priv->pci_state), val);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"set_state:0x%x\n", (u32)val);
}

/* read harvest ai core */
void vastai_pci_test_read_harvest(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	unsigned long val;
	u32 i;

	val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
	if ((val >= 0) && (val < priv->die_num)) {
		/* send upgrade cmd to smcu */
		vastai_send_pcie_cmd(priv, val, VASTAI_PCIE_SUB_GET_DLC_CORE, 0);
	} else if (val == VASTAI_VAL_DW_MASK) {
		for (i = 0; i < priv->die_num; i++)
			vastai_send_pcie_cmd(priv, i, VASTAI_PCIE_SUB_GET_DLC_CORE, 0);
	} else {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s input error val=0x%lx\n", __func__,
			       val);
	}
}


void vastai_pci_test_show_mempool(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_show_mempool_ex(priv);
}

void vastai_pci_test_re_mempool(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	vastai_reset_mempool_ex(priv);
}

typedef void (*cmd_function_t) (struct char_drv_info*, struct vastai_pci_info*);

struct vastai_pci_test_info {
	char* cmd;
	cmd_function_t pFunc;
	//TODO: help cmd for the every cmd to point out how to use it
};


static int vastai_pci_proc_cmd(unsigned char *input, int input_len,
			       struct char_drv_info *dev)
{
	int ret, i;
	unsigned char cmd[64];
	struct vastai_pci_info *priv = dev->pcie_dev_info;
	struct vastai_pci_test_info test_info[] = {
			{"vftest", vastai_pci_test_vftest},
			{"vfon",   vastai_pci_test_vfon},
			{"rcfg",   vastai_pci_test_rcfg},
			{"config",   vastai_pci_test_rcfg},
			{"wcfg",   vastai_pci_test_wcfg},
			{"wconfig",   vastai_pci_test_wcfg},
			{"read",   vastai_pci_test_read},
			{"read_buf",   vastai_pci_test_read_buf},
			{"write",   vastai_pci_test_write},
			{"write_buf",   vastai_pci_test_write_buf},
			{"getver",   vastai_pci_test_getver},
#ifdef CONFIG_VASTAI_PCI_BOOT
			{"boot",   vastai_pci_test_boot},
#endif
			{"hotreset",   vastai_pci_test_hotreset},
			{"hotreboot",   vastai_pci_test_hotreboot},
			{"reset",   vastai_pci_test_reset},
			{"reboot_ddrbw",   vastai_pci_test_reboot_ddrbw},
			{"reboot",   vastai_pci_test_reboot},
			{"poweron_video",   vastai_pci_test_poweron_video},
			{"poweroff_video",   vastai_pci_test_poweroff_video},
			{"poweron",   vastai_pci_test_poweron},
			{"bmcu_upgrade",   vastai_pci_test_bmcu_upgrade},
			{"bmcu_degrade",   vastai_pci_test_bmcu_degrade},
			{"flashbmcu",   vastai_pci_test_flashbmcu},
			{"selgen",   vastai_pci_test_selgen},
			{"seld0gen",   vastai_pci_test_seld0gen},
			{"seld1gen",   vastai_pci_test_seld1gen},
			{"seld2gen",   vastai_pci_test_seld2gen},
			{"getloglevel",   vastai_pci_test_getloglevel},
			{"setloglevel",   vastai_pci_test_setloglevel},
			{"setvdspdep",   vastai_pci_test_setvdspdep},
			{"mallocbuf",   vastai_pci_test_mallocbuf},
			{"setbuf",   vastai_pci_test_setbuf},
			{"printbuf",   vastai_pci_test_printbuf},
			{"freebuf",   vastai_pci_test_freebuf},
			{"flashxspi_app",   vastai_pci_test_flashxspi_app},
			{"flashxspi",   vastai_pci_test_flashxspi},
			{"wpen",   vastai_pci_test_wpen},
			{"wpdis",   vastai_pci_test_wpdis},
			{"readxspi",   vastai_pci_test_readxspi},
			{"test",   vastai_pci_test_test},
			{"cedar_sym",   vastai_pci_test_cedar_sym},
			{"cedar_asy",   vastai_pci_test_cedar_asy},
			{"hang_smcu",   vastai_pci_test_hang_smcu},
			{"toggle_dma_callback",   vastai_pci_test_toggle_dma_callback},
			{"iommu",   vastai_pci_test_iommu},
			{"core_reset",   vastai_pci_test_core_reset},
			{"exp_tri",   vastai_pci_test_exp_tri},
			{"switchbbox",   vastai_pci_test_switchbbox},
			{"set_state",   vastai_pci_test_set_state},
			{"harvest", vastai_pci_test_read_harvest},
#ifdef ENABLE_MEMPOOL_EX
			{"show_mempool",   vastai_pci_test_show_mempool},
			{"re_mempool",   vastai_pci_test_re_mempool},
#endif
		};


	ret = vastai_pci_cmdline_args(input, cmd, dev->args, VASTAI_PCI_ARG_MAX);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s cmdline_args err=%d\n",
			       __func__, ret);
		return -1;
	}
	/* NTD: printing help information */
	else if (!strncmp("help", cmd, strlen("help"))) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s cmd list:\n", __func__);
		for(i=0; i<sizeof(test_info)/sizeof(test_info[0]); i++) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s \n", test_info[i].cmd);
		}
	} else {
		for(i=0; i<sizeof(test_info)/sizeof(test_info[0]); i++) {
			if(!strcmp(test_info[i].cmd, cmd)) {
				test_info[i].pFunc(dev, priv);
				break;
			}
		}

		/* unknown cmd */
		if(i==sizeof(test_info)/sizeof(test_info[0]))
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s unknown cmd %s\n",
					__func__, cmd);
	}

	/* return error code for user */
	return ret;
}

static int vastai_pci_char_open(struct inode *inode, struct file *filp)
{
	struct char_drv_info *dev;
	struct char_drv_info *
		anonymous_dev; /* we need a anonymous_dev for multi thread open kchar */
	struct vastai_file_info *file_info;

	/* VASTAI_PCI_FUNC_ENTERY; */
	file_info = to_vastai_file_info(inode->i_cdev);
	dev = container_of(file_info, struct char_drv_info, file_info);
	anonymous_dev = kmalloc(sizeof(*dev), GFP_KERNEL);
	if (!anonymous_dev) {
		return -ENOMEM;
	}
	memcpy(anonymous_dev, dev, sizeof(*dev));
	anonymous_dev->valid_data_length = 0;
	if (iommu_is_enable(dev->pcie_dev_info)) {
		anonymous_dev->dm  = vastai_dma_buf_sg_get(
				dev->pcie_dev_info,
				SZ_4M);
	}
	else {
		anonymous_dev->dm  = vastai_mempool_get(
				anonymous_dev->pcie_dev_info,
				SZ_4M);
	}
	if (!anonymous_dev->dm) {
		VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
			       "malloc buf failed\n");
		return -ENOMEM;
	}

	vastai_pci_args_init(anonymous_dev->args);
	filp->private_data = anonymous_dev;

	return 0;
}

static ssize_t vastai_pci_char_read(struct file *filp, char __user *buffer,
				    size_t count, loff_t *offset)
{
	int ret;
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	char *read_buf = (char *)(dev->dm->vir) + *offset;

	VASTAI_PCI_FUNC_ENTERY;
	VASTAI_PCI_DBG(dev->pcie_dev_info, DUMMY_DIE_ID,
		       "%s count[%ld] offset[%lld] dev->l[%d]!\n", __FUNCTION__,
		       count, *offset, dev->valid_data_length);

	/* ***********************************
	 * arg check:
	 * 1. count != 0 && count <= remain_size (remain_size = dev->l - *offset)
	 * 2. dev->l < dm.size
	 * 3. remain_size >= 0
	 * ***********************************/
	if ((count == 0) || ((*offset) >= (dev->valid_data_length)) ||
	    (dev->valid_data_length > dev->dm->size)) {
		return 0;
	}
	if (count > (dev->valid_data_length - *offset)) {
		count = dev->valid_data_length - *offset;
	}

	ret = copy_to_user((void *)buffer, read_buf, count);
	if (ret != 0) {
		VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
			       "%s copy_to_user get error!\n", __FUNCTION__);
		return -ret;
	}

	*offset += count;

	return count;
}

static int vastai_cmd_dma(struct file *filp, const char __user *buffer,
			  size_t count, loff_t *offset);
static int vastai_cmd_rc(struct file *filp, const char __user *buffer,
			 size_t count, loff_t *offset);
static int vastai_diag_get_die_info(struct file *filp,
				    const char __user *buffer, size_t count,
				    loff_t *offset);
static int vastai_send_smi(struct file *filp, const char __user *buffer,
			   size_t count, loff_t *offset);
static int vastai_get_fw_version(struct file *filp, const char __user *buffer,
				 size_t count, loff_t *offset);
static int vastai_power_off_soc(struct file *filp,
				const char __user *buffer,
				size_t count,
				loff_t *offset);
static int vastai_set_machine_state(struct file *filp,
				    const char __user *buffer,
				    size_t count,
				    loff_t *offset);
struct {
	char *cmd_name;
	char *usage;
	int (*cmd_fn)(struct file *filp, const char __user *buffer,
		      size_t count, loff_t *offset);
} cmd_table[] = {
	{ .cmd_name = "DMA", .cmd_fn = vastai_cmd_dma, .usage = NULL },
	{ .cmd_name = "RC", .cmd_fn = vastai_cmd_rc, .usage = NULL },
	{ .cmd_name = "get_die_info",
	  .cmd_fn = vastai_diag_get_die_info,
	  .usage = NULL },
	{ .cmd_name = "smi", .cmd_fn = vastai_send_smi, .usage = NULL },
	{ .cmd_name = "get_fw_version",
	  .cmd_fn = vastai_get_fw_version,
	  .usage = NULL },
	{ .cmd_name = "power off soc", .cmd_fn = vastai_power_off_soc,
	  .usage = NULL },
	{ .cmd_name = "set m_state", .cmd_fn = vastai_set_machine_state,
	  .usage = NULL },
};

static int vastai_set_machine_state(struct file *filp,
				    const char __user *buffer,
				    size_t count,
				    loff_t *offset)
{
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	struct vastai_pci_info *priv = dev->pcie_dev_info;
	int ret;
	int m_state;
	char temp_buf[64] = { 0 };

	ret = copy_from_user(temp_buf, buffer, 32);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s copy from user error\n",
			       __func__);
		return -EINVAL;
	}
	sscanf(temp_buf, "set m_state %d", &m_state);
	atomic_set(&priv->pci_state, m_state);

	return 0;
}

static int vastai_power_off_soc(struct file *filp,
				const char __user *buffer,
				size_t count,
				loff_t *offset)
{
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	struct vastai_pci_info *priv = dev->pcie_dev_info;
	int ret;

	ret = vastai_send_pcie_cmd(priv, priv->dies[0].die_index,
				VASTAI_PCIE_SUB_REQUEST_BMCU_POWEROFF,
				0);
	return ret;
}

/* TODO: use sys/bus/pci/ dir struction to characterize die information */
static int vastai_diag_get_die_info(struct file *filp,
				    const char __user *buffer, size_t count,
				    loff_t *offset)
{
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	struct vastai_pci_info *priv = dev->pcie_dev_info;
	int ret;
	int die_index[4] = { 0 };
	int die_num;
	int i = 0;

	ret = get_vastai_die_info(priv, die_index, &die_num);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s get_vastai_die_info error %d\n",
			       __FUNCTION__, ret);
		return ret;
	}

	((int *)dev->dm->vir)[0] = die_num;
	dev->valid_data_length += 4;
	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
		       "vastai device_id [%d] die_num[%d]\n", priv->dev_id,
		       die_num);
	for (i = 0; i < die_num; i++) {
		((int *)dev->dm->vir)[i + 1] = die_index[i];
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "[%02d] ", die_index[i]);
		dev->valid_data_length += 4;
	}

	return 0;
}

static int vastai_get_fw_version(struct file *filp, const char __user *buffer,
				 size_t count, loff_t *offset)
{
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	struct vastai_pci_info *priv = dev->pcie_dev_info;

	memcpy(dev->dm->vir, &priv->dies[0].fw_ver, sizeof(struct vastai_fw_version));
	dev->valid_data_length += sizeof(struct vastai_fw_version);
	return 0;
}

static int vastai_send_smi(struct file *filp, const char __user *buffer,
			   size_t count, loff_t *offset)
{
	struct pcie_transfer_info trans = { .type = VASTAI_SMI };
	int ret;
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	struct vastai_pci_info *priv = dev->pcie_dev_info;
	char temp_buf[64] = { 0 };
	int die_index;
	u32 die_id;
	u32 msgq_id;

	ret = copy_from_user(temp_buf, buffer, 32);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s copy from user error\n",
			       __func__);
		return -EINVAL;
	}
	sscanf(temp_buf, "smi %d", &die_index);

	die_id = vastai_pci_get_die_id(priv, die_index);
	msgq_id = vastai_get_msgq_id(priv, MSGQ_CMD);
	ret = vastai_pci_vmsgq_wr(&(priv->dies[die_id].vmsgq[msgq_id]),
				 &trans, sizeof(trans));

	if (ret)
		return -EINVAL;
	return 0;
}

static size_t get_user_dma_desc_size(u32 desc_num)
{
	struct user_dma_desc temp;
	return sizeof(temp) + sizeof(temp.desc_tab[0]) * desc_num;
}

static int vastai_cmd_rc(struct file *filp, const char __user *buffer,
			 size_t count, loff_t *offset)
{
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	struct vastai_pci_info *priv = dev->pcie_dev_info;
	struct user_dma_desc *dma_desc = NULL;
	u32 desc_num = 1;
	int i;
	int ret;
	u64 desc_tab[64];
	int die_index;

	dma_desc = kzalloc(get_user_dma_desc_size(desc_num), GFP_KERNEL);

REFILL_DMA_DESC:
	if (!dma_desc) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s kmalloc failed size %ld\n", __FUNCTION__,
			       get_user_dma_desc_size(desc_num));
		return -ENOMEM;
	}

	ret = copy_from_user(dma_desc, buffer,
			     get_user_dma_desc_size(desc_num));
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s copy from user error\n",
			       __func__);
		goto FREE_DMA_DESC;
	}

	if (strncmp(dma_desc->cmd_str, "RC ", sizeof(dma_desc->cmd_str))) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s get err cmd line %s\n",
			       __FUNCTION__, dma_desc->cmd_str);
		/* continue */
	}

	if (dma_desc->num != desc_num && desc_num == 1) {
		desc_num = dma_desc->num;
		if (desc_num == 0) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "desc_num less 1!\n");
			ret = -EINVAL;
			goto FREE_DMA_DESC;
		}
		kfree(dma_desc);
		dma_desc =
			kzalloc(get_user_dma_desc_size(desc_num), GFP_KERNEL);
		goto REFILL_DMA_DESC;
	}
	die_index = priv->dies[dma_desc->die_id].die_index;

	for (i = 0; i < desc_num; i++) {
		u8 ctr_byte = vastai_udma_ctr_byte(i, desc_num, dma_desc->mode);

		desc_tab[i] =
			vastai_udma_push_desc(priv, die_index, dma_desc->ch % 4,
					      dma_desc->desc_tab[i].dev_addr,
					      dma_desc->desc_tab[i].host_addr,
					      dma_desc->desc_tab[i].length,
					      ctr_byte);
	}
	vastai_udma_trigger(priv, die_index, dma_desc->ch,
			    dma_desc->is_host_to_dev, desc_tab, desc_num);
	ret = vastai_udma_polling_done(priv, die_index, dma_desc->ch, desc_tab,
				       desc_num);
FREE_DMA_DESC:
	kfree(dma_desc);
	return ret;
}

static int vastai_cmd_dma(struct file *filp, const char __user *buffer,
			  size_t count, loff_t *offset)
{
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	struct vastai_pci_info *priv = dev->pcie_dev_info;
	struct user_dma_desc *dma_desc = NULL;
	u32 desc_num = 1;
	int i;
	int ret;
	u64 desc_tab[64];
	u64 dma_length = 0;
	int die_index;

	dma_desc = kzalloc(get_user_dma_desc_size(desc_num), GFP_KERNEL);

REFILL_DMA_DESC:
	if (!dma_desc) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s kmalloc failed size %ld\n", __FUNCTION__,
			       get_user_dma_desc_size(desc_num));
		return -ENOMEM;
	}

	ret = copy_from_user(dma_desc, buffer,
			     get_user_dma_desc_size(desc_num));
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s copy from user error\n",
			       __func__);
		goto FREE_DMA_DESC;
	}

	if (strncmp(dma_desc->cmd_str, "DMA ", sizeof(dma_desc->cmd_str))) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s get err cmd line %s\n",
			       __FUNCTION__, dma_desc->cmd_str);
		/* continue */
	}

	if (dma_desc->num != desc_num && desc_num == 1) {
		desc_num = dma_desc->num;
		if (desc_num == 0) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "desc_num less 1!\n");
			ret = -EINVAL;
			goto FREE_DMA_DESC;
		}
		kfree(dma_desc);
		dma_desc =
			kzalloc(get_user_dma_desc_size(desc_num), GFP_KERNEL);
		goto REFILL_DMA_DESC;
	}
	die_index = priv->dies[dma_desc->die_id].die_index;

	for (i = 0; i < desc_num; i++) {
		u8 ctr_byte = vastai_udma_ctr_byte(i, desc_num, dma_desc->mode);
		bool is_post_write = ctr_byte & BIT(2);
		bool is_pre_fetch = ctr_byte & BIT(1);

		if (dma_length + dma_desc->desc_tab[i].length > dev->dm->size) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "transfer too long 0x%llx!\n",
				       dma_length);
			return -EINVAL;
		}

		if (dma_desc->is_host_to_dev && !is_post_write) {
			ret = copy_from_user(
				dev->dm->vir + dma_length,
				(void *)dma_desc->desc_tab[i].host_addr,
				dma_desc->desc_tab[i].length);
			if (ret != 0) {
				VASTAI_PCI_ERR(
					priv, DUMMY_DIE_ID,
					"%s %d copy_from_user get error!\n",
					__FUNCTION__, __LINE__);
				return ret;
			}
		}
		desc_tab[i] = vastai_udma_push_desc(
			priv, die_index, dma_desc->ch,
			dma_desc->desc_tab[i].dev_addr,
			(dev->dm->dma_bus_addr + dma_length) |
				((dma_desc->die_id) ? 0xb000000000 : 0x0),
			dma_desc->desc_tab[i].length, ctr_byte);
		if (!(dma_desc->is_host_to_dev && is_post_write) &&
		    !(!(dma_desc->is_host_to_dev) && is_pre_fetch))
			dma_length += dma_desc->desc_tab[i].length;
	}
	vastai_udma_trigger(priv, die_index, dma_desc->ch,
			    dma_desc->is_host_to_dev, desc_tab, desc_num);
	ret = vastai_udma_polling_done(priv, die_index, dma_desc->ch, desc_tab,
				       desc_num);
	dev->valid_data_length = dma_length;
FREE_DMA_DESC:
	kfree(dma_desc);
	return ret;
}

static ssize_t vastai_pci_char_write(struct file *filp,
				     const char __user *buffer, size_t count,
				     loff_t *offset)
{
	int ret;
	unsigned char *input_kernel_buf;
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	int i;
	size_t real_count = min(count, (size_t)VASTAI_CMD_READ_LEN);

	/* VASTAI_PCI_FUNC_ENTERY; */
	if (count == 0)
		return count;

	if (!dev->dm->vir) {
		VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
			       "%s dev->replay is Empty!\n", __FUNCTION__);
		return count;
	}

	input_kernel_buf = kzalloc(VASTAI_CMD_READ_LEN, GFP_KERNEL);
	if (!input_kernel_buf) {
		return -ENOMEM;
	}
	ret = copy_from_user(input_kernel_buf, (char __user *)buffer,
			     real_count);
	if (ret != 0) {
		VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
			       "%s copy_from_user get error!\n", __FUNCTION__);
		goto FREE_INPUT_BUF;
	}

	ret = -1;
	for (i = 0; i < ARRAY_SIZE(cmd_table); i++) {
		char *cmd = cmd_table[i].cmd_name;
		if (strncmp(cmd, input_kernel_buf, strlen(cmd)) == 0) {
			ret = cmd_table[i].cmd_fn(filp, buffer, count, offset);
			if (ret < 0) {
				VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
					       "get error code %d\n", ret);
				goto FREE_INPUT_BUF;
			}
		}
	}
	input_kernel_buf[real_count - 1] = 0; /* set \0 for string */

	if (ret != 0) { /* Is this cmd handle by cmd_table ? */
		VASTAI_PCI_DBG(dev->pcie_dev_info, DUMMY_DIE_ID, "write:%s\n",
			       input_kernel_buf);
		ret = vastai_pci_proc_cmd(input_kernel_buf, VASTAI_CMD_READ_LEN,
					  dev);
	}

FREE_INPUT_BUF:
	kfree(input_kernel_buf);

	if (ret != 0) {
		VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
			       "%s we get error code %d!\n", __FUNCTION__, ret);
		return ret;
	}
	return count;
}

static long vastai_pci_char_ioctl(struct file *filp, unsigned int cmd,
				  unsigned long arg)
{
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	struct vastai_pci_info *priv = dev->pcie_dev_info;
	struct kchar_cmd kcmd;
	int ret = 0;

	if ((atomic_read(&priv->pci_state) != VASTAI_NORMAL_STATE) &&
			(atomic_read(&priv->pci_state) != VASTAI_DEBUG_STATE))
		return -ENODEV;
	ret = copy_from_user(&kcmd, (void *)arg, sizeof(kcmd));
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s copy_from_user is error\n", __func__);
		return -EIO;
	}

	switch (cmd) {

	case VASTAI_PCI_IOCTL_ALLOC: {
		struct dma_buf *dmabuf;

		if (iommu_is_enable(priv))
			dmabuf = vastai_alloc_dma_buf_sg(
					priv,
					kcmd.alloc_cmd.size);
		else
			dmabuf = vastai_alloc_dma_buf(
					priv,
					kcmd.alloc_cmd.size);
		if (dmabuf) {
			int fd = dma_buf_fd(dmabuf, O_CLOEXEC);
			struct vastai_dma_buf *dm = get_vastai_dma_buf(dmabuf, 0);
			kcmd.alloc_cmd.dma_buf_fd = fd;
			kcmd.alloc_cmd.dma_addr_t = dm->dma_bus_addr;
		} else {
			kcmd.alloc_cmd.dma_buf_fd = -ENOMEM;
		}
		break;
	}

	case VASTAI_PCI_IOCTL_DMA_START: {
		struct dma_buf *dmabuf =
			dma_buf_get(kcmd.dma_start_cmd.dma_buf_fd);
		struct vastai_dmadesc desc;
		union core_bitmap core_id = { .val = 0 };
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
		struct timespec tm_begin, tm_end;
#else
		u64 tm_begin, tm_end;
#endif
		unsigned long time_total_ns;
		u32 done_size = 0;
		u32 size = kcmd.dma_start_cmd.size;
		u32 i = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
		/* getnstimeofday function is removed from kernel5.6 */
		getnstimeofday(&tm_begin);
#else
		tm_begin = ktime_get_ns();
#endif
		if (IS_ERR(dmabuf)) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s dma buf fd is invalid\n", __func__);
			return PTR_ERR(dmabuf);
		}

		if (iommu_is_enable(priv)) {
			union core_bitmap core_id = { .val = 0 };
			struct vastai_dma_buf *dm =
				get_vastai_dma_buf(dmabuf, 0);
			u32 is_src_dma_addr = 1;
			u32 is_src_not_user_mem = 1;
			if (kcmd.dma_start_cmd.size > dm->size) {
				ret = -EINVAL;
				goto free_buf;
			}
			ret = vastai_pci_dma_transfer_sg(
				priv,
				kcmd.dma_start_cmd.die_index,
				core_id,
				dm,
				is_src_dma_addr,
				!kcmd.dma_start_cmd.is_dev_to_host,
				is_src_not_user_mem,
				kcmd.dma_start_cmd.axi_addr,
				kcmd.dma_start_cmd.size);
		}else {
			desc.is_host_to_dev = !kcmd.dma_start_cmd.is_dev_to_host;
			desc.is_src_not_user_mem = 1;
			desc.is_src_dma_addr = 1;

			while (done_size < kcmd.dma_start_cmd.size) {
				struct vastai_dma_buf *dm =
					get_vastai_dma_buf(dmabuf, i);
				u32 temp_size =
					(size - done_size) < VASTAI_MAX_DMA_BUF ?
						(size - done_size) :
						VASTAI_MAX_DMA_BUF;

				desc.dev_addr = kcmd.dma_start_cmd.axi_addr + done_size;
				desc.dma_lenth = temp_size;
				desc.host_addr.dma_addr = dm->dma_bus_addr;

				ret = vastai_pci_dma_transfer_sync(
					priv, kcmd.dma_start_cmd.die_index, core_id,
					&desc, 1);
				if (ret)
					break;

				i++;
				done_size += temp_size;
			}
		}
free_buf:
		dma_buf_put(dmabuf);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
		getnstimeofday(&tm_end);
		time_total_ns =
			timespec_to_ns(&tm_end) - timespec_to_ns(&tm_begin);
#else
		tm_end = ktime_get_ns();
		time_total_ns = tm_end - tm_begin;
#endif

		if (time_total_ns > 30000 && 0) {
			struct vastai_fifo fifo;
			printk("%s die_index 0x%x time %ld\n", __FUNCTION__,
			       kcmd.dma_start_cmd.die_index, time_total_ns);

			vastai_pci_mem_read(priv, kcmd.dma_start_cmd.die_index,
					    0x8c14000, &fifo,
					    sizeof(struct vastai_fifo));

			VASTAI_PCI_DBG(
				priv, DUMMY_DIE_ID,
				"%s rd[0x%x] wr[0x%x] elem_count[0x%x] elem_size[0x%x]\n",
				"0x8c14000", fifo.rd, fifo.wr, fifo.elem_count,
				fifo.elem_size);

			vastai_pci_mem_read(priv, kcmd.dma_start_cmd.die_index,
					    0x8c1e000, &fifo,
					    sizeof(struct vastai_fifo));

			VASTAI_PCI_DBG(
				priv, DUMMY_DIE_ID,
				"%s rd[0x%x] wr[0x%x] elem_count[0x%x] elem_size[0x%x]\n",
				"0x8c1e000", fifo.rd, fifo.wr, fifo.elem_count,
				fifo.elem_size);
		}
		break;
	}
	case VASTAI_PCI_IOCTL_DMA_TRANS: {
		struct vastai_dmadesc *desc;
		union core_bitmap core_id = { .val = 0 };
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
		struct timespec tm_begin, tm_end;
#else
		u64 tm_begin, tm_end;
#endif
		unsigned long time_total_ns;
		u32 done_size = 0;
		u32 size = kcmd.dma_trans_cmd.length;
		u32 i = 0;
		u32 entry_count;
		u32 temp_size;

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
		/* getnstimeofday function is removed from kernel5.6 */
		getnstimeofday(&tm_begin);
#else
		tm_begin = ktime_get_ns();
#endif
		entry_count = DIV_ROUND_UP(size, VASTAI_MAX_DMA_BUF);
		desc = kvmalloc(sizeof(struct vastai_dmadesc) * entry_count,
				GFP_KERNEL);
		for (i = 0; i < entry_count; i++) {
			temp_size = (size - done_size) < VASTAI_MAX_DMA_BUF ?
					    (size - done_size) :
					    VASTAI_MAX_DMA_BUF;
			desc[i].is_host_to_dev =
				!kcmd.dma_trans_cmd.is_dev_to_host;
			desc[i].is_src_not_user_mem = 0;
			desc[i].is_src_dma_addr = 0;
			desc[i].dev_addr =
				kcmd.dma_trans_cmd.axi_addr + done_size;
			desc[i].dma_lenth = temp_size;
			desc[i].host_addr.vir_addr =
				kcmd.dma_trans_cmd.vir_addr + done_size;
			done_size += temp_size;
		}
		ret = vastai_pci_dma_transfer_sync(priv,
						   kcmd.dma_trans_cmd.die_index,
						   core_id, desc, entry_count);

		kvfree(desc);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
		getnstimeofday(&tm_end);
		time_total_ns =
			timespec_to_ns(&tm_end) - timespec_to_ns(&tm_begin);
#else
		tm_end = ktime_get_ns();
		time_total_ns = tm_end - tm_begin;
#endif

		if (time_total_ns > 30000 && 0) {
			struct vastai_fifo fifo;
			printk("%s die_index 0x%x time %ld\n", __FUNCTION__,
			       kcmd.dma_trans_cmd.die_index, time_total_ns);

			vastai_pci_mem_read(priv, kcmd.dma_trans_cmd.die_index,
					    0x8c14000, &fifo,
					    sizeof(struct vastai_fifo));

			VASTAI_PCI_DBG(
				priv, DUMMY_DIE_ID,
				"%s rd[0x%x] wr[0x%x] elem_count[0x%x] elem_size[0x%x]\n",
				"0x8c14000", fifo.rd, fifo.wr, fifo.elem_count,
				fifo.elem_size);

			vastai_pci_mem_read(priv, kcmd.dma_trans_cmd.die_index,
					    0x8c1e000, &fifo,
					    sizeof(struct vastai_fifo));

			VASTAI_PCI_DBG(
				priv, DUMMY_DIE_ID,
				"%s rd[0x%x] wr[0x%x] elem_count[0x%x] elem_size[0x%x]\n",
				"0x8c1e000", fifo.rd, fifo.wr, fifo.elem_count,
				fifo.elem_size);
		}
		break;
	}
	case VASTAI_PCI_IOCTL_DMA_TRANS_SG: {
		union core_bitmap core_id = { .val = 0 };
		struct vastai_channel_buf *channel_buf;
		int i;
		int channel_num = kcmd.dma_trans_sg_cmd.channel_num;
		channel_buf = kvmalloc(sizeof(struct vastai_channel_buf) *
					       channel_num,
				       GFP_KERNEL);
		if (!channel_buf) {
			return -ENOMEM;
		}
		ret = copy_from_user(
			channel_buf, kcmd.dma_trans_sg_cmd.channel_buf,
			sizeof(struct vastai_channel_buf) * channel_num);
		if (ret) {
			kvfree(channel_buf);
			return -1;
		}

		for (i = 0; i < channel_num; i++) {
			VASTAI_PCI_DBG(
				priv, kcmd.dma_trans_sg_cmd.die_index,
				"channel[%d], w:%d, h:%d, dst_w_off:%d, dst_h_off:%d, src_w_off:%d, src_h_off:%d\n",
				i, channel_buf[i].width, channel_buf[i].high,
				channel_buf[i].dst_width_offset,
				channel_buf[i].dst_high_offset,
				channel_buf[i].src_width_offset,
				channel_buf[i].src_high_offset);
		}

		vastai_dma_video(priv, kcmd.dma_trans_sg_cmd.die_index, core_id,
				 channel_buf, channel_num,
				 kcmd.dma_trans_sg_cmd.axi_addr,
				 kcmd.dma_trans_sg_cmd.pid);

		kvfree(channel_buf);
		break;
	}
    case VASTAI_PCI_IOCTL_MEMSET: {
        u32 size = kcmd.dma_memset_cmd.length;
        struct device *dev = &(priv->dev->dev);
        dma_addr_t dma_bus_addr;
        u32 is_src_dma_addr = 1;
        u32 is_src_not_user_mem = 1;
        union core_bitmap core_id = { .val = 0 };
        struct vastai_dma_buf *dm = kmalloc(sizeof(struct vastai_dma_buf),  GFP_KERNEL);

        if(!dm){
            return -ENOMEM;
        }

        if(size < VASTAI_MAX_MEMSET_BUF){

            dm->vir = dma_alloc_coherent(dev, size, &dma_bus_addr,  GFP_KERNEL);
        }else{

            dm->vir = dma_alloc_coherent(dev, VASTAI_MAX_MEMSET_BUF, &dma_bus_addr,  GFP_KERNEL);
        }

        if(!dm->vir || !dma_bus_addr){
         VASTAI_PCI_ERR(priv,DUMMY_DIE_ID,
                        "%s: dma_alloc_coherent fail\n",__func__);
         goto ERROR_FREE_DM;
        }



        memset(dm->vir, kcmd.dma_memset_cmd.val, size);

        ret = vastai_pci_dma_transfer_memset(
				priv,
				kcmd.dma_memset_cmd.die_index,
				core_id,
				dma_bus_addr,
				is_src_dma_addr,
				!kcmd.dma_memset_cmd.is_dev_to_host,
				is_src_not_user_mem,
				kcmd.dma_memset_cmd.axi_addr,
				kcmd.dma_memset_cmd.length);

        if(size < VASTAI_MAX_MEMSET_BUF){

            dma_free_coherent(dev, size, (void *)dm->vir, dma_bus_addr);

        }else{

            dma_free_coherent(dev, VASTAI_MAX_MEMSET_BUF, (void *)dm->vir, dma_bus_addr);
        }
ERROR_FREE_DM:
        kvfree(dm);

        break;
    }
	case VASTAI_PCI_IOCTL_POLL_LOG: {
		/* trigger smcu to poll log buf */
		u32 flag = 0;
		int ret;
		u32 timeout_water = kcmd.log_poll_cmd.timeout_water != 0x0 ?
					    kcmd.log_poll_cmd.timeout_water :
					    500;
		u32 die_index = kcmd.log_poll_cmd.die_index;
		struct vastai_sv100_die *die = &(
			((struct vastai_pci_info *)priv)
				->dies[vastai_pci_get_die_id(priv, die_index)]);

		vastai_pci_mem_write(priv, die_index, LOG_MSG_FLAG, &flag,
				     sizeof(flag));
		/* wait complete */
		ret = wait_for_completion_interruptible_timeout(
			&(die->log_comp), HZ * timeout_water / 1000);
		if (ret == 0) {
			kcmd.log_poll_cmd.is_timeout = 1;
		}

		reinit_completion(&(die->log_comp));

		/* get log ring ctl */
		/* we need size == 0x1D0 buf */
		vastai_pci_mem_read(priv, kcmd.log_poll_cmd.die_index,
				    LOG_CTRL_INFO, dev->dm->vir, 0x1D0);
		ret = copy_to_user(kcmd.log_poll_cmd.log_ring_ctl, dev->dm->vir,
				   0x1D0);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "%s copy_from_user is error\n",
				       __func__);
			return -EIO;
		}
		break;
	}
	case VASTAI_PCI_IOCTL_RENDER: {
		u32 die_index = kcmd.get_render_name.die_index;
		snprintf(kcmd.get_render_name.file_name,
			 sizeof(kcmd.get_render_name.file_name), "renderD%d",
			 priv->dies[vastai_pci_get_die_id(priv, die_index)]
				 .render_id);
		printk("%s\n", kcmd.get_render_name.file_name);
		break;
	}
	case VASTAI_PCI_IOCTL_PORT: {
		strcpy(kcmd.port_name, pci_name(priv->dev));
		printk("%s\n", kcmd.port_name);
		break;
	}
	case VASTAI_PCI_IOCTL_GET_PERFORMANCE: {
		u32 die_index = kcmd.get_die_performance.die_index;
		memcpy(kcmd.get_die_performance.pcie_tx_vdsp_count,
		       priv->dies[vastai_pci_get_die_id(priv, die_index)]
			       .pcie_tx_vdsp_count,
		       sizeof(kcmd.get_die_performance.pcie_tx_vdsp_count));

		memcpy(kcmd.get_die_performance.pcie_rx_vdsp_count,
		       priv->dies[vastai_pci_get_die_id(priv, die_index)]
			       .pcie_rx_vdsp_count,
		       sizeof(kcmd.get_die_performance.pcie_rx_vdsp_count));
		break;
	}
	case VASTAI_PCI_IOCTL_PCI_ID:
                kcmd.get_pci_id.pci_id = priv->dev->vendor | priv->dev->device << 16;
		break;
	case VASTAI_PCI_IOCTL_PCI_SUB_ID:
		kcmd.get_pci_sub_id.pci_sub_id = 0xFFFF;
		break;
	case VASTAI_PCI_IOCTL_MAJOR_MINOR:
		kcmd.get_major_minor.major_minor =
			vastai_file_get_major_minor(&dev->file_info);
		break;
	case TCGETS:
		return -ENOIOCTLCMD;

	case TCSETS:
		return -ENOIOCTLCMD;

	case TIOCSERGETLSR:
		return -ENOIOCTLCMD;

	case TIOCSERCONFIG:
		return -ENOIOCTLCMD;

	default:
		VASTAI_PCI_INFO(dev->pcie_dev_info, DUMMY_DIE_ID,
				"Unsupport cmd 0x%x\n", cmd);
		break;
	}

	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "ioctl error %d\n", ret);
		return ret;
	}

	ret = copy_to_user((void *)arg, &kcmd, sizeof(kcmd));
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s copy_to_user is error\n",
			       __func__);
		return -EIO;
	}

	return 0;
}

static int vastai_pci_char_release(struct inode *inode, struct file *filp)
{
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;

	/* VASTAI_PCI_FUNC_ENTERY; */
	if (iommu_is_enable(dev->pcie_dev_info)) {
		vastai_dma_buf_sg_put(dev->pcie_dev_info, dev->dm);
	}
	else
		vastai_mempool_put(dev->pcie_dev_info, dev->dm);

	/* free anonymous_dev */
	kfree(dev);

	return 0;
}

static const struct file_operations fop = {
	.owner = THIS_MODULE,
	.open = vastai_pci_char_open,
	.release = vastai_pci_char_release,
	.write = vastai_pci_char_write,
	.read = vastai_pci_char_read,
	.llseek = default_llseek,
	.unlocked_ioctl = vastai_pci_char_ioctl,
};
struct class *myclass;

static int vastai_pci_ioctl_cmd_init(struct vastai_pci_info *bus)
{
	int ret;
	struct char_drv_info *drv;

	drv = kmalloc(sizeof(struct char_drv_info), GFP_KERNEL);
	if (!drv) {
		ret = -ENOMEM;
		return ret;
	}
	bus->p_char = drv;

	drv->pcie_dev_info = bus;
	ret = vastai_file_create(&drv->file_info, bus, &fop, "kchar:%d",
				 bus->dev_id);
	VASTAI_PCI_INFO(bus, DUMMY_DIE_ID,
			"init ioctl node /dev/kchar:%d ok.\n", bus->dev_id);

	return ret;
}

int vastai_pci_ioctl_cmd_deinit(struct vastai_pci_info *bus)
{
	struct char_drv_info *drv = bus->p_char;

	vastai_pci_free_udma_buf(bus, &dmb[0]);
	vastai_file_destroy(&drv->file_info);
	VASTAI_PCI_INFO(bus, DUMMY_DIE_ID,
			"deinit ioctl node /dev/kchar:%d ok.\n", bus->dev_id);
	kfree(drv);
	bus->p_char = NULL;

	return 0;
}

int vastai_version_file_init(struct vastai_pci_info *pci_info);
int vastai_ctl_file_init(struct vastai_pci_info *pci_info);
int vastai_pci_ioctl_init(struct vastai_pci_info *bus)
{
	vastai_version_file_init(bus);
	vastai_ctl_file_init(bus);
	return vastai_pci_ioctl_cmd_init(bus);
}

void vastai_version_file_deinit(struct vastai_pci_info *pci_info);
void vastai_ctl_file_deinit(struct vastai_pci_info *pci_info);
int vastai_pci_ioctl_deinit(struct vastai_pci_info *bus)
{
	vastai_version_file_deinit(bus);
	vastai_ctl_file_deinit(bus);
	return vastai_pci_ioctl_cmd_deinit(bus);
}

unsigned int major = 321;
atomic_t minor = ATOMIC_INIT(0);

int vastai_file_create(struct vastai_file_info *file_info,
		       struct vastai_pci_info *pci_info,
		       const struct file_operations *fop, const char *fmt, ...)
{
	va_list vargs;
	int ret = 0;

	file_info->dev_num = MKDEV(major, atomic_inc_return(&minor));
	ret = register_chrdev_region(file_info->dev_num, 1, "vastai");
	if (ret) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
			       "register device num failed! alloc one [%d]\n",
			       file_info->dev_num);
		alloc_chrdev_region(&file_info->dev_num, 0, 1, "char");
	}
	file_info->dev.owner = THIS_MODULE;
	cdev_init(&file_info->dev, fop);
	ret = cdev_add(&file_info->dev, file_info->dev_num, 1);
	if (ret < 0) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID, "cdev_add failed %d\n",
			       ret);
		return ret;
	}

	va_start(vargs, fmt);
	vsnprintf(file_info->file_name, 32, fmt, vargs);
	va_end(vargs);
	file_info->device = device_create(myclass, NULL, file_info->dev_num,
					  file_info, file_info->file_name);

	if (!file_info->device) {
		ret = PTR_ERR(file_info->device);
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
			       "device_create_vargs failed %d\n", ret);
		return ret;
	}
	return ret;
}

void vastai_file_destroy(struct vastai_file_info *file_info)
{
	device_destroy(myclass, file_info->dev_num);
	cdev_del(&file_info->dev);
	unregister_chrdev_region(file_info->dev_num, 1);
	atomic_dec_return(&minor);
}

dev_t vastai_file_get_major_minor(struct vastai_file_info *file_info)
{
	return file_info->dev_num;
}
