/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/platform_device.h>
#include <linux/uaccess.h>
#include <linux/random.h>
#include <linux/dma-buf.h>

#include "vastai_pci.h"
#define TEMP_BUF_SIZE 512

static int version_open(struct inode *inode, struct file *filp)
{
	struct vastai_file_info *file_info;
	struct vastai_version_file *version_file;

	file_info = to_vastai_file_info(inode->i_cdev);
	version_file =
		container_of(file_info, struct vastai_version_file, file_info);

	filp->private_data = version_file;
	return 0;
}

static int version_release(struct inode *inode, struct file *filp)
{
	return 0;
}

static int version_sprint(void *buf, struct vastai_pci_info *pci_info)
{
	int ret = 0;
	int i = 0;

	ret += sprintf(buf + ret, "[%s]\n", pci_info->dies[0].fw_ver.ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->dies[0].fw_ver.smcu_ver);
	ret += sprintf(buf + ret, "[%s] [active]\n",
		       pci_info->dies[0].fw_ver.bmcu_ver);
	ret += sprintf(buf + ret, "[%s] [backup]\n",
		       pci_info->dies[0].fw_ver.bmcu_backup_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->dies[0].fw_ver.vdmcu_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->dies[0].fw_ver.vemcu_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->dies[0].fw_ver.vdsp_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->dies[0].fw_ver.cmcu_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->dies[0].fw_ver.lmcu_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->dies[0].fw_ver.odsp_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->dies[0].fw_ver.driver_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->dies[0].fw_ver.phy_ver);

	for (i = 0; i < pci_info->die_num; i++) {
		struct vastai_fw_bmcu_xspi_version *bmcu_xspi_ver;
		bmcu_xspi_ver = (struct vastai_fw_bmcu_xspi_version *)
				pci_info->dies[i].fw_ver.bmcu_ver;
		
		ret += sprintf(buf + ret, "[die %d] [%s] [active]\n",
				i, bmcu_xspi_ver->xspi_sub_ver);

		bmcu_xspi_ver = (struct vastai_fw_bmcu_xspi_version *)
				pci_info->dies[i].fw_ver.bmcu_backup_ver;
		ret += sprintf(buf + ret, "[die %d] [%s] [backup]\n",
				i, bmcu_xspi_ver->xspi_sub_ver);
	}

	return ret;
}

static ssize_t version_read(struct file *filp, char __user *buffer,
			    size_t count, loff_t *offset)
{
	struct vastai_version_file *version_file = filp->private_data;
	struct vastai_pci_info *pci_info = version_file->pci_info;
	void *buf = vmalloc(TEMP_BUF_SIZE);
	int ret;

	if ((count < sizeof(pci_info->dies[0].fw_ver) * 2) || (*offset != 0))
		return 0;

	ret = version_sprint(buf, pci_info);
	*offset += ret;

	if (copy_to_user(buffer, buf, ret)) {
		return -EIO;
	}

	return ret;
}

static const struct file_operations fop = {
	.owner = THIS_MODULE,
	.open = version_open,
	.release = version_release,
	.read = version_read,
};

int vastai_version_file_init(struct vastai_pci_info *pci_info)
{
	struct vastai_version_file *version_file =
		vmalloc(sizeof(struct vastai_version_file));
	int ret;

	if (version_file == NULL) {
		return -ENOMEM;
	}
	ret = vastai_file_create(&version_file->file_info, pci_info, &fop,
				 "vastai%d_version", pci_info->dev_id);

	pci_info->version_file = version_file;
	version_file->pci_info = pci_info;

	return ret;
}

void vastai_version_file_deinit(struct vastai_pci_info *pci_info)
{
	vastai_file_destroy(&pci_info->version_file->file_info);
	vfree(pci_info->version_file);
}
