/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: frank.bai
 */
#include "vatools.h"
#include "vatools_log.h"

/******************************************************************************
 * vatools all file ops
 *
 *
 */

//根据app_category排列
static const struct file_operations app_fops[] = {
	{
		.owner = THIS_MODULE,
		.llseek = dummy_llseek,
		.read = dummy_read,
		.write = dummy_write,
		.write_iter = dummy_write_iter,
		.poll = dummy_poll,
		.unlocked_ioctl = dummy_ioctl,
		.compat_ioctl = dummy_ioctl,
		.mmap = dummy_mmap,
		.open = dummy_open,
		.release = dummy_release,
		.fasync = dummy_fasync,
	},
	{
		.owner = THIS_MODULE,
		.llseek = smi_llseek,
		.read = smi_read,
		.write = smi_write,
		.write_iter = smi_write_iter,
		.poll = smi_poll,
		.unlocked_ioctl = smi_ioctl,
		.compat_ioctl = smi_ioctl,
		.mmap = smi_mmap,
		.open = smi_open,
		.release = smi_release,
		.fasync = smi_fasync,
	},
	{
		.owner = THIS_MODULE,
		.llseek = profiler_llseek,
		.read = profiler_read,
		.write = profiler_write,
		.write_iter = profiler_write_iter,
		.poll = profiler_poll,
		.unlocked_ioctl = profiler_ioctl,
		.compat_ioctl = profiler_ioctl,
		.mmap = profiler_mmap,
		.open = profiler_open,
		.release = profiler_release,
		.fasync = profiler_fasync,
	},
	{
		.owner = THIS_MODULE,
		.llseek = debugger_llseek,
		.read = debugger_read,
		.write = debugger_write,
		.write_iter = debugger_write_iter,
		.poll = debugger_poll,
		.unlocked_ioctl = debugger_ioctl,
		.compat_ioctl = debugger_ioctl,
		.mmap = debugger_mmap,
		.open = debugger_open,
		.release = debugger_release,
		.fasync = debugger_fasync,
	},
	{
		.owner = THIS_MODULE,
		.llseek = logger_llseek,
		.read = logger_read,
		.write = logger_write,
		.write_iter = logger_write_iter,
		.poll = logger_poll,
		.unlocked_ioctl = logger_ioctl,
		.compat_ioctl = logger_ioctl,
		.mmap = logger_mmap,
		.open = logger_open,
		.release = logger_release,
		.fasync = logger_fasync,
	},
	{
		.owner = THIS_MODULE,
		.llseek = dummy_llseek,
		.read = dummy_read,
		.write = dummy_write,
		.write_iter = dummy_write_iter,
		.poll = dummy_poll,
		.unlocked_ioctl = dummy_ioctl,
		.compat_ioctl = dummy_ioctl,
		.mmap = dummy_mmap,
		.open = dummy_open,
		.release = dummy_release,
		.fasync = dummy_fasync,
	},
	{
		.owner = THIS_MODULE,
		.llseek = sharedmem_llseek,
		.read = sharedmem_read,
		.write = sharedmem_write,
		.write_iter = sharedmem_write_iter,
		.poll = sharedmem_poll,
		.unlocked_ioctl = sharedmem_ioctl,
		.compat_ioctl = sharedmem_ioctl,
		.mmap = sharedmem_mmap,
		.open = sharedmem_open,
		.release = sharedmem_release,
		.fasync = sharedmem_fasync,
	},
	{
		.owner = THIS_MODULE,
		.llseek = dummy_llseek,
		.read = dummy_read,
		.write = dummy_write,
		.write_iter = dummy_write_iter,
		.poll = dummy_poll,
		.unlocked_ioctl = dummy_ioctl,
		.compat_ioctl = dummy_ioctl,
		.mmap = dummy_mmap,
		.open = dummy_open,
		.release = dummy_release,
		.fasync = dummy_fasync,
	}
};

/******************************************************************************
 * get core id by core index
 *
 *
 */

static uint32_t _internal_parray_coreid[CORE_NUMBER] = {
	COREID_DEFINE_8(CORE_TYPE_LMCU),  COREID_DEFINE_8(CORE_TYPE_ODSP),
	COREID_DEFINE_4(CORE_TYPE_VEMCU), COREID_DEFINE_3(CORE_TYPE_VDMCU),
	(CORE_TYPE_SMCU << 6) | 0,	  COREID_DEFINE_4(CORE_TYPE_VDSP),
	(CORE_TYPE_CMCU << 6) | 0,
};

u32 vatools_get_core_id(u32 n_core_index)
{
	return _internal_parray_coreid[n_core_index];
}

static struct vatools_node *vatools_get_node_from_minor(int minor)
{
	struct vatools_node *node = NULL;

	list_for_each_entry (node, vatools_get_vastai_head(), list_nodes) {
		if (node->misc.minor == minor)
			return node;
	}
	return NULL;
}

struct vatools_node *vatools_get_node(void)
{
	struct vatools_node *node = NULL;

	list_for_each_entry (node, vatools_get_vastai_head(), list_nodes) {
		if (node != NULL)
			return node;
	}
	return NULL;
}

static long vatools_get_device_num(struct file *filp, unsigned int cmd,
				   unsigned long arg)
{
	long ret = -EINVAL;
	void __user *argp = (void __user *)arg;
	unsigned int device_num = 0;

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;

	device_num = vatools_get_vastai_pci_device_number();
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "device_num=%d", device_num);

	if (copy_to_user(argp, &device_num, sizeof(unsigned int))) {
		ret = -EFAULT;
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld", ret);
		goto out;
	}

	ret = 0;

out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld", ret);
	VATOOLS_FUNC_EXIT;

	return ret;
}

static long vatools_get_dev_info(struct file *filp, unsigned int cmd,
				 unsigned long arg)
{
	long ret = -EINVAL;
	void __user *argp = (void __user *)arg;
	struct vatools_node *node = vatools_file_get_node(filp);
	vatools_ioctl_arg_t *ioctl_arg;
	vatools_dev_info_t *dev_info_p = NULL;
	struct vastai_cdev *vacc_dev_p = NULL;
	struct vastai_cdev *vacc_dev_temp = NULL;
	struct vastai_render_file *render_file_p = NULL;
	struct vastai_render_file *render_file_temp = NULL;

	struct vastai_pci_info *pci_info = NULL;
	u32 device_num = 0;
	u32 dev = 0;
	int die = 0;
	void *buffer = NULL;
	int buffer_size = 0;

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;

	ioctl_arg = (vatools_ioctl_arg_t *)kvmalloc(sizeof(vatools_ioctl_arg_t),
						    GFP_KERNEL);
	if (ioctl_arg == NULL) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "kvmalloc error. size=%ld",
			    sizeof(vatools_ioctl_arg_t));
		ret = -ENOMEM;
		goto out;
	}

	ret = copy_from_user(ioctl_arg, argp, sizeof(vatools_ioctl_arg_t));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_from_user ret=%ld", ret);
		ret = -EFAULT;
		goto out_free_ioctl_arg;
	}

	device_num = vatools_get_vastai_pci_device_number();
	if (device_num == 0) {
		ret = -ENOENT;
		ioctl_arg->header.length = 0;
		if (copy_to_user(argp, ioctl_arg,
				 sizeof(vatools_ioctl_arg_t))) {
			ret = -EFAULT;
		}
		goto out_free_ioctl_arg;
	}

	buffer_size = device_num * sizeof(vatools_dev_info_t);

	buffer = kvmalloc(buffer_size, GFP_KERNEL);
	if (buffer == NULL) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "kvmalloc error. size=%d",
			    buffer_size);
		ret = -ENOMEM;
		goto out_free_ioctl_arg;
	}

	dev_info_p = (vatools_dev_info_t *)buffer;
	memset(buffer, 0, buffer_size);

	for (dev = 0; dev < device_num; dev++) {
		pci_info = get_vastai_pci_device_info((char)dev);
		if (pci_info == NULL) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID,
				    "priv is NULL, dev_id[%u] ", dev);
			continue;
		}
		if ((atomic_read(&pci_info->pci_state) !=
		     VASTAI_NORMAL_STATE)) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID,
				    "priv->pci_state= 0x%x\n",
				    atomic_read(&pci_info->pci_state));
			continue;
		}
		dev_info_p->dev_id = pci_info->dev_id;
		dev_info_p->die_num = pci_info->die_num;
		VATOOLS_DBG(NULL, DUMMY_DIE_ID,
			    "dev_id=%d die_num=%d buffer_size=%d",
			    dev_info_p->dev_id, dev_info_p->die_num,
			    buffer_size);

		// bar size
		memcpy(dev_info_p->bar_address, vatools_die_check,
		       sizeof(vatools_die_check));

		// char unid[ VASTAI_BUF_SIZE_32 ];       ///< SN号; get from common (host); not here
		// char card_type[ VASTAI_BUF_SIZE_32 ];  ///< 卡类型; get from common (host); not here

		dev_info_p->pciid.vender_id = pci_info->dev->vendor;
		dev_info_p->pciid.device_id = pci_info->dev->device;
		dev_info_p->pcisubid.sub_vender_id =
			pci_info->dev->subsystem_vendor;
		dev_info_p->pcisubid.sub_device_id =
			pci_info->dev->subsystem_device;

		memcpy(dev_info_p->pcie_slot, pci_name(pci_info->dev),
		       VASTAI_BUF_SIZE_32);

		dev_info_p->kchar.majormin.major =
			MAJOR(pci_info->p_char->file_info.dev_num);
		dev_info_p->kchar.majormin.minor =
			MINOR(pci_info->p_char->file_info.dev_num);

		memcpy(dev_info_p->kchar.name,
		       pci_info->p_char->file_info.file_name,
		       VASTAI_BUF_SIZE_32);

		dev_info_p->vastai_version.majormin.major =
			MAJOR(pci_info->version_file->file_info.dev_num);
		dev_info_p->vastai_version.majormin.minor =
			MINOR(pci_info->version_file->file_info.dev_num);
		memcpy(dev_info_p->vastai_version.name,
		       pci_info->version_file->file_info.file_name,
		       VASTAI_BUF_SIZE_32);

		dev_info_p->vastai_ctrl.majormin.major =
			MAJOR(pci_info->ctl_file->file_info.dev_num);
		dev_info_p->vastai_ctrl.majormin.minor =
			MINOR(pci_info->ctl_file->file_info.dev_num);

		memcpy(dev_info_p->vastai_ctrl.name,
		       pci_info->ctl_file->file_info.file_name,
		       VASTAI_BUF_SIZE_32);

		dev_info_p->vatools.majormin.minor = node->misc.minor;
		dev_info_p->vatools.majormin.major = 10;
		memcpy(dev_info_p->vatools.name, node->misc.name,
		       VASTAI_BUF_SIZE_32);

		dev_info_p->cap.ai_cap =
			dev_info_p->cap.ai_cap; ///< AI VIDEO 能力集
		dev_info_p->cap.video_cap =
			dev_info_p->cap.video_cap; ///< AI VIDEO 能力集

		for (die = 0; die < dev_info_p->die_num; die++) {
			dev_info_p->die_info[die].die_index.val =
				pci_info->dies[die].die_index;
			VATOOLS_DBG(NULL, DUMMY_DIE_ID,
				    "dies[%d] dev_info_p->die_index=%x", die,
				    dev_info_p->die_info[die].die_index.val);

			// vastai_video
			list_for_each_entry_safe (render_file_p,
						  render_file_temp,
						  &pci_info->render_head,
						  node) {
				if (render_file_p->die_index ==
				    pci_info->dies[die].die_index) {
					dev_info_p->die_info[die]
						.vastai_video.majormin.major =
						MAJOR(render_file_p->file_info
							      .dev_num);
					dev_info_p->die_info[die]
						.vastai_video.majormin.minor =
						MINOR(render_file_p->file_info
							      .dev_num);

					snprintf(dev_info_p->die_info[die]
							 .vastai_video.name,
						 VASTAI_BUF_SIZE_32,
						 "vastai_video%d",
						 dev_info_p->die_info[die]
							 .die_index.seq_num);
					break;
				}
			}
#ifdef CONFIG_DRM_DRIVER
			// renderD
			list_for_each_entry_safe (render_file_p,
						  render_file_temp,
						  &pci_info->render_head,
						  node) {
				if (render_file_p->die_index ==
				    pci_info->dies[die].die_index) {
					struct vastai_device_handle
						*device_handle;
					device_handle =
						(struct vastai_device_handle
							 *)(render_file_p
								    ->dec_private_data);

					dev_info_p->die_info[die]
						.render.majormin.major =
						226; // drm 设备有系统指定主设备号为 226, 不变.

					dev_info_p->die_info[die]
						.render.majormin.minor =
						device_handle->drm_dev->render
							->index;

					snprintf(dev_info_p->die_info[die]
							 .render.name,
						 VASTAI_BUF_SIZE_32,
						 "renderD%d",
						 pci_info->dies[die].render_id);
					break;
				}
			}
#endif
			// vacc
			list_for_each_entry_safe (vacc_dev_p, vacc_dev_temp,
						  &pci_info->vacc_dev_head,
						  vacc_dev_node) {
				if (vacc_dev_p->die_index ==
				    pci_info->dies[die].die_index) {
					dev_info_p->die_info[die]
						.vacc.majormin.major = MAJOR(
						vacc_dev_p->file_info.dev_num);
					dev_info_p->die_info[die]
						.vacc.majormin.minor = MINOR(
						vacc_dev_p->file_info.dev_num);

					snprintf(dev_info_p->die_info[die]
							 .vacc.name,
						 VASTAI_BUF_SIZE_32, "vacc%d",
						 dev_info_p->die_info[die]
							 .die_index.seq_num);
					break;
				}
			}
		}

		dev_info_p++;
	}

	ioctl_arg->header.err = 0;
	ioctl_arg->header.length = buffer_size;

	if (copy_to_user(argp, ioctl_arg, sizeof(vatools_ioctl_arg_t))) {
		ret = -EFAULT;
		goto out_free_buffer;
	}

	if (copy_to_user((u8 *)argp + sizeof(vatools_ioctl_arg_header_t),
			 buffer, buffer_size)) {
		ret = -EFAULT;
		goto out_free_buffer;
	}

	ret = 0;

out_free_buffer:
	kvfree(buffer);

out_free_ioctl_arg:
	kvfree(ioctl_arg);

out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld", ret);
	VATOOLS_FUNC_EXIT;
	return ret;
}

static long vatools_get_die_info(struct file *filp, unsigned int cmd,
				 unsigned long arg)
{
	long ret = -EINVAL;
	void __user *argp = (void __user *)arg;
	struct vastai_pci_info *priv = NULL;

	int die_info[MAX_DIE_NUM_PER_DEVICE + 2] = { 0 };
	int die_index[MAX_DIE_NUM_PER_DEVICE] = { 0 };
	int die_num = 0;
	int i = 0;

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;

	ret = copy_from_user(die_info, argp, sizeof(die_info));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_from_user ret=%ld", ret);
		ret = -EFAULT;
		goto out;
	}
	priv = vatools_get_vastai_pci_device_info((char)die_info[0]);
	if (priv == NULL) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "priv is NULL, dev_id[%u] ",
			    die_info[0]);
		return -ENODEV;
	}
	if ((atomic_read(&priv->pci_state) != VASTAI_NORMAL_STATE)) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "priv->pci_state= 0x%x\n",
			    atomic_read(&priv->pci_state));
		return -ENODEV;
	}
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "device_id: %d", die_info[0]);

	ret = vatools_get_vastai_die_info_via_device_id((char)die_info[0],
							die_index, &die_num);
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "device_id=%d, error %ld\n",
			    die_info[0], ret);
		goto out;
	}
	die_info[1] = die_num;
	for (i = 0; i < die_num; i++) {
		die_info[2 + i] = die_index[i];
	}

	if (copy_to_user(argp, die_info, sizeof(die_info))) {
		ret = -EFAULT;
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "ret=%ld", ret);
		goto out;
	}

	ret = 0;
out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld", ret);
	VATOOLS_FUNC_EXIT;

	return ret;
}

static long vatools_get_dev_pcie_bus_id(struct file *filp, unsigned int cmd,
					unsigned long arg)
{
	long ret = -EINVAL;
	void __user *argp = (void __user *)arg;

	struct vastai_pci_info *priv = NULL;
	TS_SMI_DEV_PCIE_BUS_ID dev_pcie_bus_id = { 0 };

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;

	memset(&dev_pcie_bus_id, 0x00, sizeof(TS_SMI_DEV_PCIE_BUS_ID));
	ret = copy_from_user(&dev_pcie_bus_id, argp,
			     sizeof(TS_SMI_DEV_PCIE_BUS_ID));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy from user ret=%ld", ret);
		ret = -EFAULT;
		goto out;
	}

	priv = vatools_get_vastai_pci_device_info((char)dev_pcie_bus_id.dev_id);
#if VATOOLS_INDEPENDANCE_DRIVER
	strcpy(dev_pcie_bus_id.pcie_port, "pcie-simu");
#else
	strcpy(dev_pcie_bus_id.pcie_port, pci_name(priv->dev));
#endif
	ret = copy_to_user(argp, &dev_pcie_bus_id,
			   sizeof(TS_SMI_DEV_PCIE_BUS_ID));
	if (ret) {
		ret = -EFAULT;
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "ret=%ld", ret);
		goto out;
	}

	ret = 0;

out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld", ret);
	VATOOLS_FUNC_EXIT;

	return ret;
}

static long vatools_get_dev_pcie_id(struct file *filp, unsigned int cmd,
				    unsigned long arg)
{
	long ret = -EINVAL;
	void __user *argp = (void __user *)arg;
	TS_SMI_DEV_PCIE_ID dev_pcie_id = { 0 };

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;

	memset(&dev_pcie_id, 0x00, sizeof(TS_SMI_DEV_PCIE_ID));
	ret = copy_from_user(&dev_pcie_id, argp, sizeof(TS_SMI_DEV_PCIE_ID));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy from user ret= %ld", ret);
		ret = -EFAULT;
		goto out;
	}

	dev_pcie_id.pcie_id = 0x1ec6 | 0x0100 << 16;

	ret = copy_to_user(argp, &dev_pcie_id, sizeof(TS_SMI_DEV_PCIE_ID));
	if (ret) {
		ret = -EFAULT;
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "ret=%ld", ret);
		goto out;
	}
	ret = 0;

out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld", ret);
	VATOOLS_FUNC_EXIT;

	return ret;
}

static long vatools_get_dev_pcie_sub_id(struct file *filp, unsigned int cmd,
					unsigned long arg)
{
	long ret = -EINVAL;
	void __user *argp = (void __user *)arg;
	TS_SMI_DEV_PCIE_SUB_ID dev_pcie_sub_id = { 0 };

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;

	memset(&dev_pcie_sub_id, 0x00, sizeof(TS_SMI_DEV_PCIE_SUB_ID));
	ret = copy_from_user(&dev_pcie_sub_id, argp,
			     sizeof(TS_SMI_DEV_PCIE_SUB_ID));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy from user ret=%ld", ret);
		ret = -EFAULT;
		goto out;
	}

	dev_pcie_sub_id.pcie_sub_id = 0xFFFF;

	ret = copy_to_user(argp, &dev_pcie_sub_id,
			   sizeof(TS_SMI_DEV_PCIE_SUB_ID));
	if (ret) {
		ret = -EFAULT;
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "ret=%ld", ret);
		goto out;
	}
	ret = 0;

out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld", ret);
	VATOOLS_FUNC_EXIT;

	return ret;
}

static long vatools_get_dev_major_minor_id(struct file *filp, unsigned int cmd,
					   unsigned long arg)
{
	long ret = -EINVAL;
	void __user *argp = (void __user *)arg;
	TS_SMI_DEV_MAJOR_MINOR dev_major_minor = { 0 };

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;

	memset(&dev_major_minor, 0x00, sizeof(TS_SMI_DEV_MAJOR_MINOR));
	ret = copy_from_user(&dev_major_minor, argp,
			     sizeof(TS_SMI_DEV_MAJOR_MINOR));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy from user ret=%ld", ret);
		ret = -EFAULT;
		goto out;
	}

	dev_major_minor.major_minor = 0x0;

	ret = copy_to_user(argp, &dev_major_minor,
			   sizeof(TS_SMI_DEV_MAJOR_MINOR));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy to user ret=%ld", ret);
		ret = -EFAULT;
		goto out;
	}
	ret = 0;

out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld", ret);
	VATOOLS_FUNC_EXIT;

	return ret;
}

static long vatools_get_die_render(struct file *filp, unsigned int cmd,
				   unsigned long arg)
{
	long ret = -EINVAL;
	void __user *argp = (void __user *)arg;
	struct vastai_pci_info *priv = NULL;
	TS_SMI_DIE_RENDER die_render = { 0 };

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;

	memset(&die_render, 0x00, sizeof(TS_SMI_DIE_RENDER));
	ret = copy_from_user(&die_render, argp, sizeof(die_render));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy from user ret=%ld", ret);
		ret = -EFAULT;
		goto out;
	}

	priv = vatools_get_vastai_pci_device_info((char)die_render.dev_id);
	if (priv == NULL) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "priv is NULL, dev_id[%u] ",
			    die_render.dev_id);
		ret = -ENODEV;
		goto out;
	}
	if ((atomic_read(&priv->pci_state) != VASTAI_NORMAL_STATE)) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "priv->pci_state= 0x%x\n",
			    atomic_read(&priv->pci_state));
		ret = -ENODEV;
		goto out;
	}
	snprintf(die_render.render, 16, "renderD%d",
		 priv->dies[die_render.die_id].render_id);
	ret = copy_to_user(argp, &die_render, sizeof(TS_SMI_DIE_RENDER));
	if (ret) {
		ret = -EFAULT;
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "ret=%ld", ret);
		goto out;
	}

	ret = 0;

out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld", ret);
	VATOOLS_FUNC_EXIT;

	return ret;
}

static long vatools_get_device_capability(struct file *filp, unsigned int cmd,
					  unsigned long arg)
{
	long ret = -EINVAL;
	void __user *argp = (void __user *)arg;
	struct vastai_pci_info *priv = NULL;
	SMI_DEVICE_CAPABILITY device_cap = { 0 };

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;

	memset(&device_cap, 0x00, sizeof(SMI_DEVICE_CAPABILITY));
	ret = copy_from_user(&device_cap, argp, sizeof(SMI_DEVICE_CAPABILITY));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy from user ret=%ld", ret);
		ret = -EFAULT;
		goto out;
	}

	priv = vatools_get_vastai_pci_device_info((char)device_cap.dev_id);
	if (priv == NULL) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "priv is NULL, dev_id[%u] ",
			    device_cap.dev_id);
		ret = -ENODEV;
		goto out;
	}
	if ((atomic_read(&priv->pci_state) != VASTAI_NORMAL_STATE)) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "priv->pci_state= 0x%x\n",
			    atomic_read(&priv->pci_state));
		ret = -ENODEV;
		goto out;
	}
	memcpy(&device_cap.capability, &priv->vadev,
	       sizeof(struct vastai_dev_info));
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld", ret);

	ret = copy_to_user(argp, &device_cap, sizeof(SMI_DEVICE_CAPABILITY));
	if (ret) {
		ret = -EFAULT;
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "ret=%ld", ret);
		goto out;
	}

	ret = 0;

out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld", ret);
	VATOOLS_FUNC_EXIT;

	return ret;
}

static long vatools_set_app_category(struct file *filp, unsigned int cmd,
				     unsigned long arg)
{
	struct vatools_reader *reader = vatools_file_get_reader(filp);
	long ret = -EINVAL;
	void __user *argp = (void __user *)arg;

	VATOOLS_FUNC_ENTERY;
	ret = copy_from_user(&(reader->trans_category), argp,
			     sizeof(reader->trans_category));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_from_user ret=%ld", ret);
		ret = -EFAULT;
		goto out;
	}

	//下面category需要同时绑定 device 信息. 在release可以释放对应device 的资源
	if ((reader->trans_category.app_category == VATOOLS_APP_SMI) ||
	    (reader->trans_category.app_category == VATOOLS_APP_PROFILER) ||
	    (reader->trans_category.app_category == VATOOLS_APP_DEBUGGER) ||
	    (reader->trans_category.app_category == VATOOLS_APP_MONITOR)) {
		reader->trans_category.device.die_index =
			vatools_get_vastai_pci_die_index(
				(char)reader->trans_category.device.dev_id,
				reader->trans_category.device.die_id);
	} else {
		//在共享内存状态下的数据由其他用处
		if (reader->trans_category.app_category !=
		    VATOOLS_APP_SHAREDMEM) {
			reader->trans_category.block_id = SMI_CMD_BLOCK_ID_0;
			reader->trans_category.app_category = VATOOLS_APP_SMI;
			reader->trans_category.device.dev_id = 0;
			reader->trans_category.device.die_id = 0;
			reader->trans_category.device.die_index = 0;
		} else {
			//用户态的pid 换成内核态的pid 保证在docker中能正确识别
			reader->trans_category.device.dev_id = current->tgid;
			reader->trans_category.device.die_id = 0;
			reader->trans_category.device.die_index = current->tgid;
		}
	}
	// VATOOLS_DBG
	VATOOLS_DBG(
		NULL, DUMMY_DIE_ID,
		"reader = %p,app_category= %d  block_id=%d  device_id=%d  die_id=%d  "
		"die_index=%d  ",
		reader, reader->trans_category.app_category,
		reader->trans_category.block_id,
		reader->trans_category.device.dev_id,
		reader->trans_category.device.die_id,
		reader->trans_category.device.die_index);

	switch (reader->trans_category.app_category) {
	case VATOOLS_APP_SHAREDMEM:
		sharedmem_set_app_category(filp, cmd, arg);
		break;
	case VATOOLS_APP_DEBUGGER:
		debugger_set_app_category(filp, cmd, arg);
		break;
	}

out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld", ret);
	VATOOLS_FUNC_EXIT;

	return ret;
}

static unsigned int vatools_get_app_category(struct file *filp)
{
	struct vatools_reader *reader = vatools_file_get_reader(filp);
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "app_category=%d",
		    reader->trans_category.app_category);

	return (unsigned int)reader->trans_category.app_category;
}

/******************************************************************************
 * smi
 *
 *
 */

static long vatools_driver_read(struct file *file, unsigned int cmd,
				unsigned long arg)
{
	return vatools_driver_smi_fetch(file, cmd, arg);
}

static long vatools_driver_write(struct file *file, unsigned int cmd,
				 unsigned long arg)
{
	return vatools_driver_smi_fetch(file, cmd, arg);
}

/******************************************************************************
 * pcie fops
 *
 *
 */
// pcie 命令接口, 直接调用pcie驱动函数. 不发送到smcu
static long vatools_pcie_ioctl(struct file *filp, unsigned int cmd,
			       unsigned long arg)
{
	// struct vatools_node*   node   = vatools_file_get_node( filp );
	// struct vatools_reader* reader = vatools_file_get_reader( filp );

	long ret = -EINVAL;
	void __user *argp = (void __user *)arg;
	int write_len = 0;
	int read_len = 0;
	T_SMI_IOCTL_TRANS_DATA kdata = { 0 };
	struct vastai_pci_info *priv = NULL;
	int die_index = 0;
	int ret2 = 0;
	u32 i = 0;

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;

	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "ioctl_cmd= 0x%x ", cmd);

	memset(&kdata, 0x00, sizeof(T_SMI_IOCTL_TRANS_DATA));
	ret = copy_from_user(&kdata, argp, sizeof(kdata));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_from_user ret=%ld", ret);
		ret = -EFAULT;
		goto out;
	}

	VATOOLS_DBG(NULL, DUMMY_DIE_ID,
		    "block_id=%d  device_id=%d  die_id=%d  flag= 0x%x  addr= "
		    "0x%llx ",
		    kdata.block_id, kdata.device.dev_id, kdata.device.die_id,
		    kdata.flag, kdata.address);
	VATOOLS_DBG(NULL, DUMMY_DIE_ID,
		    "read_len= %d  read_buf= 0x%p  write_len= %d  "
		    "write_buf= 0x%p",
		    kdata.output_buf.buf_size, kdata.output_buf.buf_addr,
		    kdata.input_buf.buf_size, kdata.input_buf.buf_addr);

	write_len = kdata.input_buf.buf_size;
	read_len = kdata.output_buf.buf_size;

	priv = vatools_get_vastai_pci_device_info((char)kdata.device.dev_id);

	// device reboot reset boot do not check status
	if (kdata.flag != PCIE_CMD_FLAG_RESET &&
	    kdata.flag != PCIE_CMD_FLAG_REBOOT &&
	    kdata.flag != PCIE_CMD_FLAG_BOOT &&
	    kdata.flag != PCIE_CMD_FLAG_REBOOT_DDRBW &&
	    kdata.flag != PCIE_CMD_FLAG_REBOOT_PCIEDMA) {
		if (priv == NULL) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID,
				    "priv is NULL, dev_id[%u] ",
				    kdata.device.dev_id);
			ret = -ENODEV;
			goto out;
		}
		if ((atomic_read(&priv->pci_state) != VASTAI_NORMAL_STATE)) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID,
				    "priv->pci_state= 0x%x\n",
				    atomic_read(&priv->pci_state));
			ret = -ENODEV;
			goto out;
		}
	}

	die_index = vatools_get_vastai_pci_die_index((char)kdata.device.dev_id,
						     kdata.device.die_id);

	kdata.errcode = 0;

	switch (kdata.flag) {
	case PCIE_CMD_FLAG_BANDWIDTH: {
		// u32 low_upstream_count   = 0;
		u64 upstream_count = 0;
		// u32 low_downstream_count = 0;
		u64 downstream_count = 0;
		ret = 0;
		/*
        // upstream
        ret = vatools_pci_mem_read( priv, die_index, 0x900002c, &low_upstream_count, sizeof( u32 ) );
        if ( ret < 0 )
        {
            ret = -EFAULT;
            goto out;
        }
        */
		ret = vatools_pci_mem_read(priv, die_index,
					   PCIE_BANDWIDTH_UPSTREAM,
					   &upstream_count, sizeof(u64));
		if (ret < 0) {
			ret = -EFAULT;
			goto out;
		}

		// upstream_count &= 0xFFFFFFFF00000000;
		// upstream_count |= low_upstream_count;

		VATOOLS_DBG(NULL, DUMMY_DIE_ID, "upstream_count: ret=%llx",
			    upstream_count);

		ret = copy_to_user((u8 *)kdata.output_buf.buf_addr +
					   sizeof(TS_SMI_CMD_REQ),
				   &upstream_count, sizeof(u64));
		if (ret < 0) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
			goto out;
		}
		/*
                // downstream
                ret = vatools_pci_mem_read( priv, die_index, 0x9000034, &low_downstream_count, sizeof( u32 ) );
                if ( ret < 0 )
                {
                    ret = -EFAULT;
                    goto out;
                }
        */
		ret = vatools_pci_mem_read(priv, die_index,
					   PCIE_BANDWIDTH_DOWNSTREAM,
					   &downstream_count, sizeof(u64));
		if (ret < 0) {
			ret = -EFAULT;
			goto out;
		}

		// downstream_count &= 0xFFFFFFFF00000000;
		// downstream_count |= low_downstream_count;
		VATOOLS_DBG(NULL, DUMMY_DIE_ID, "downstream_count: ret=%llx",
			    downstream_count);
		ret = copy_to_user((u8 *)kdata.output_buf.buf_addr +
					   sizeof(TS_SMI_CMD_REQ) + sizeof(u64),
				   &downstream_count, sizeof(u64));
		if (ret) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
			goto out;
		}
	} break;

	case PCIE_CMD_FLAG_RESET:
		VATOOLS_DBG(
			priv, die_index,
			"smi trigger die_index=%x : reset. current ddr_bw_test=%d\n",
			die_index, priv->ddr_bw_test);
		priv->ddr_bw_test = 0;
		ret = vastai_device_reset(priv, VASTAI_RESET_MSG_TRIGER);
		break;

	case PCIE_CMD_FLAG_BOOT:
		VATOOLS_DBG(
			priv, die_index,
			"smi trigger die_index=%x : boot. current ddr_bw_test=%d\n",
			die_index, priv->ddr_bw_test);
		priv->ddr_bw_test = 0;
		ret = vastai_device_reset(priv, VASTAI_RESET_BOOT_DEV);
		break;

	case PCIE_CMD_FLAG_REBOOT:
		VATOOLS_DBG(
			priv, die_index,
			"smi trigger die_index=%x : reboot. current ddr_bw_test=%d\n",
			die_index, priv->ddr_bw_test);
		priv->ddr_bw_test = 0;
		ret = vastai_device_reset(priv, VASTAI_RESET_MSG_TRIGER |
							VASTAI_RESET_BOOT_DEV);
		break;

	case PCIE_CMD_FLAG_FW_VERSION: {
		struct vastai_fw_version fw_ver;
		memcpy(&fw_ver, &(priv->dies[0].fw_ver),
		       sizeof(struct vastai_fw_version));
		VATOOLS_DUMP_BRIEF("fw_ver: ", &fw_ver,
				   sizeof(struct vastai_fw_version));
		ret = copy_to_user(kdata.output_buf.buf_addr, &fw_ver,
				   sizeof(struct vastai_fw_version));
		if (ret) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
			goto out;
		}
	} break;

	case PCIE_CMD_FLAG_DDR_ECC_1BIT: {
		struct vastai_fifo *ecc_fifo;
		u32 table_size = kdata.output_buf.buf_size;
		u8 *buffer;

		if (table_size > DDR_ECC_1BIT_SIZE) {
			table_size = DDR_ECC_1BIT_SIZE;
		}

		ret = 0;
		die_index = kdata.device.die_index;
		buffer = (u8 *)kvmalloc(table_size, GFP_KERNEL);
		if (buffer == NULL) {
			VATOOLS_ERR(
				NULL, DUMMY_DIE_ID,
				"PCIE_CMD_FLAG_DDR_ECC_1BIT kvmalloc error. size=%d",
				table_size);
			ret = -ENOMEM;
			goto out;
		}

		ret = vatools_pci_mem_read(priv, die_index, DDR_ECC_1BIT,
					   buffer, table_size);
		if (ret < 0) {
			ret = -EFAULT;
			goto out_free_write_buf_1bit;
		}
		ecc_fifo = (struct vastai_fifo *)buffer;
		VATOOLS_DUMP_BRIEF("ecc 1bit: ", buffer, table_size);
		// ecc_fifo->rd 返回到应用层, 对应的变量是entry_count
		if (ecc_fifo->elem_count == 0) {
			VATOOLS_DBG(NULL, DUMMY_DIE_ID,
				    "ecc1bit: ecc_fifo->elem_count = 0");
			ecc_fifo->rd = 0;
		} else {
			ecc_fifo->rd =
				(ecc_fifo->elem_count +
				 (((int)ecc_fifo->wr) - (int)(ecc_fifo->rd))) %
				ecc_fifo->elem_count;
		}

		ret = copy_to_user(kdata.output_buf.buf_addr, buffer,
				   table_size);
		if (ret) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
			goto out_free_write_buf_1bit;
		}

	out_free_write_buf_1bit:
		kvfree(buffer);
	} break;

	case PCIE_CMD_FLAG_DDR_ECC_2BIT: {
		struct vastai_fifo *ecc_fifo;
		u32 table_size = kdata.output_buf.buf_size;
		u8 *buffer;

		if (table_size > DDR_ECC_2BIT_SIZE) {
			table_size = DDR_ECC_2BIT_SIZE;
		}

		ret = 0;
		die_index = kdata.device.die_index;
		buffer = (u8 *)kvmalloc(table_size, GFP_KERNEL);
		if (buffer == NULL) {
			VATOOLS_ERR(
				NULL, DUMMY_DIE_ID,
				"PCIE_CMD_FLAG_DDR_ECC_1BIT kvmalloc error. size=%d",
				table_size);
			ret = -ENOMEM;
			goto out;
		}

		ret = vatools_pci_mem_read(priv, die_index, DDR_ECC_2BIT,
					   buffer, table_size);
		if (ret < 0) {
			ret = -EFAULT;
			goto out_free_write_buf_2bit;
		}
		ecc_fifo = (struct vastai_fifo *)buffer;
		VATOOLS_DUMP_BRIEF("ecc 2bit: ", buffer, table_size);
		// ecc_fifo->rd 返回到应用层, 对应的变量是entry_count
		if (ecc_fifo->elem_count == 0) {
			VATOOLS_DBG(NULL, DUMMY_DIE_ID,
				    "ecc2bit: ecc_fifo->elem_count = 0");
			ecc_fifo->rd = 0;
		} else {
			ecc_fifo->rd =
				(ecc_fifo->elem_count +
				 (((int)ecc_fifo->wr) - (int)(ecc_fifo->rd))) %
				ecc_fifo->elem_count;
		}

		ret = copy_to_user(kdata.output_buf.buf_addr, buffer,
				   table_size);
		if (ret) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
			goto out_free_write_buf_2bit;
		}

	out_free_write_buf_2bit:
		kvfree(buffer);
	} break;

	case PCIE_CMD_FLAG_VDSP_COUNT: {
		// u32 pcie_tx_vdsp_count[ 4 ];
		// u32 pcie_rx_vdsp_count[ 4 ];
		die_index = kdata.device.die_index;

		ret = copy_to_user(
			kdata.output_buf.buf_addr,
			priv->dies[vastai_pci_get_die_id(priv, die_index)]
				.pcie_tx_vdsp_count,
			kdata.output_buf.buf_size);
		if (ret) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
			goto out;
		}

		ret = copy_to_user(
			((u8 *)kdata.output_buf.buf_addr) + sizeof(u32) * 4,
			priv->dies[vastai_pci_get_die_id(priv, die_index)]
				.pcie_rx_vdsp_count,
			kdata.output_buf.buf_size);
		if (ret) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
			goto out;
		}
	} break;

	case PCIE_CMD_FLAG_AI_COUNT: {
		u8 *ai_count_buffer;
		u32 ai_count_buffer_size = 512;

		if (kdata.output_buf.buf_size < ai_count_buffer_size) {
			ret = -EFAULT;
			VATOOLS_ERR(NULL, DUMMY_DIE_ID,
				    "output buffer size < %d ret=%ld",
				    ai_count_buffer_size, ret);
			goto out;
		}

		ai_count_buffer =
			(u8 *)kvmalloc(ai_count_buffer_size, GFP_KERNEL);
		if (ai_count_buffer == NULL) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID,
				    "PCIE_CMD_FLAG_AI_COUNT kvmalloc. size=%d",
				    ai_count_buffer_size);
			ret = -ENOMEM;
			goto out;
		}

		die_index = kdata.device.die_index;
		ret = vatools_pci_mem_read(priv, die_index, PERFORMANCE_AI_BASE,
					   ai_count_buffer,
					   ai_count_buffer_size);
		if (ret < 0) {
			ret = -EFAULT;
			goto out_free_write_buf_ai_count;
		}

		ret = copy_to_user(kdata.output_buf.buf_addr, ai_count_buffer,
				   ai_count_buffer_size);
		if (ret) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
			goto out_free_write_buf_ai_count;
		}
	out_free_write_buf_ai_count:
		kvfree(ai_count_buffer);
	} break;

	case PCIE_CMD_FLAG_RING_BUF_STATISTIC: {
		T_RING_BUF_STATICS ring_buf_statics[] = {
			{ PERFORMANCE_HOST_2_VEMCU0 },
			{ PERFORMANCE_HOST_2_VEMCU1 },
			{ PERFORMANCE_HOST_2_VEMCU2 },
			{ PERFORMANCE_HOST_2_VEMCU3 },
			{ PERFORMANCE_SMCU_2_VDMCU0 },
			{ PERFORMANCE_VDSP0_2_VDMCU0 },
			{ PERFORMANCE_VDSP1_2_VDMCU0 },
			{ PERFORMANCE_VEMCU0_2_VDMCU0 },
			{ PERFORMANCE_VEMCU1_2_VDMCU0 },
			{ PERFORMANCE_SMCU_2_VDMCU1 },
			{ PERFORMANCE_VDSP1_2_VDMCU1 },
			{ PERFORMANCE_VDSP2_2_VDMCU1 },
			{ PERFORMANCE_VEMCU1_2_VDMCU1 },
			{ PERFORMANCE_VEMCU2_2_VDMCU1 },
			{ PERFORMANCE_SMCU_2_VDMCU2 },
			{ PERFORMANCE_VDSP2_2_VDMCU2 },
			{ PERFORMANCE_VDSP3_2_VDMCU2 },
			{ PERFORMANCE_VEMCU2_2_VDMCU2 },
			{ PERFORMANCE_VEMCU3_2_VDMCU2 },
			{ PERFORMANCE_VDSP0_2_CMCU },
			{ PERFORMANCE_VDSP1_2_CMCU },
			{ PERFORMANCE_VDSP2_2_CMCU },
			{ PERFORMANCE_VDSP3_2_CMCU },
			{ PERFORMANCE_CMCU_2_VDSP0 },
			{ PERFORMANCE_CMCU_2_VDSP1 },
			{ PERFORMANCE_CMCU_2_VDSP2 },
			{ PERFORMANCE_CMCU_2_VDSP3 },
			{ PERFORMANCE_HOST_2_VDSP0 },
			{ PERFORMANCE_HOST_2_VDSP1 },
			{ PERFORMANCE_HOST_2_VDSP2 },
			{ PERFORMANCE_HOST_2_VDSP3 },
			{ PERFORMANCE_SMCU0_2_HOST },
			{ PERFORMANCE_SMCU1_2_HOST },
			{ PERFORMANCE_VDMCU1_2_HOST },
			{ PERFORMANCE_VDMCU2_2_HOST },
			{ PERFORMANCE_VEMCU0_2_HOST },
			{ PERFORMANCE_VEMCU1_2_HOST },
			{ PERFORMANCE_VEMCU2_2_HOST },
			{ PERFORMANCE_VEMCU3_2_HOST },
			{ PERFORMANCE_VDSP0_2_HOST },
			{ PERFORMANCE_VDSP1_2_HOST },
			{ PERFORMANCE_VDSP2_2_HOST },
			{ PERFORMANCE_VDSP3_2_HOST },
			{ PERFORMANCE_DMA_HOST_2_DEV },
			{ PERFORMANCE_DMA_DEV_2_HOST },
			{ PERFORMANCE_DMA_RC_IN },
			{ PERFORMANCE_DMA_RC_OUT },
		};

		u32 static_buf_cnt =
			sizeof(ring_buf_statics) / sizeof(ring_buf_statics[0]);
		u32 static_buf_len =
			static_buf_cnt * sizeof(T_VASTAI_FIFO_HEADER);
		u8 *static_buf = NULL;
		if (kdata.output_buf.buf_size < static_buf_len) {
			ret = -EFAULT;
			VATOOLS_ERR(NULL, DUMMY_DIE_ID,
				    "output buffer size < %d ret=%ld",
				    static_buf_len, ret);
			goto out;
		}
		static_buf = (u8 *)kvmalloc(static_buf_len, GFP_KERNEL);
		if (NULL == static_buf) {
			VATOOLS_ERR(
				NULL, DUMMY_DIE_ID,
				"PCIE_CMD_FLAG_HOST_COUNT kvmalloc. size = %d",
				static_buf_len);
			ret = -ENOMEM;
			goto out;
		}

		for (i = 0; i < static_buf_cnt; i++) {
			ret = vatools_pci_mem_read(
				priv, kdata.device.die_index,
				ring_buf_statics[i].base_addr,
				(void *)(static_buf +
					 i * sizeof(T_VASTAI_FIFO_HEADER)),
				sizeof(T_VASTAI_FIFO_HEADER));
			if (ret < 0) {
				VATOOLS_ERR(
					NULL, DUMMY_DIE_ID,
					"Ring buf read memory base addr %x error\n",
					ring_buf_statics[i].base_addr);
				ret = -EFAULT;
				goto out_free_static_buf;
			}
		}

		ret = copy_to_user((u8 *)kdata.output_buf.buf_addr +
					   sizeof(TS_SMI_CMD_REQ),
				   static_buf, static_buf_len);
		if (ret) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
			goto out_free_static_buf;
		}

	out_free_static_buf:
		kvfree(static_buf);
	} break;

#ifdef CONFIG_VASTAI_VIDEO
	// 设置video使用multicore; 0: 不使用; 1: 使用
	case PCIE_CMD_FLAG_SET_VIDEO_MULTI_CORE: {
		u32 work_mode = (kdata.address & 0xFFFFFFFF);
		// call video callback function here
		VATOOLS_DBG(NULL, DUMMY_DIE_ID,
			    "die inex=%x set video multicore work_mode to %d\n",
			    kdata.device.die_index, work_mode);
		ret = hantroenc_set_work_mode(kdata.device.die_index,
					      (encoder_workmode_t)work_mode, 0);
		dec_work_mode_change(kdata.device.die_index,
				     (decoder_workmode_t)work_mode);
	} break;
	case PCIE_CMD_FLAG_GET_VIDEO_MULTI_CORE: {
		u32 work_mode = 0;
		int job_count = 0;
		// call video callback function here
		ret = hantroenc_get_work_mode(kdata.device.die_index,
					      (encoder_workmode_t *)&work_mode,
					      &job_count);
		kdata.address = work_mode;
		kdata.address = kdata.address << 32 | job_count;

		VATOOLS_DBG(
			NULL, DUMMY_DIE_ID,
			"die inex=%x get video multicore work_mode=%d, job_count=%d\n",
			kdata.device.die_index, work_mode, job_count);
		goto out;
	} break;
#endif
	// core heartbeat
	case PCIE_CMD_FLAG_HEART_BEAT: {
		core_heartbeat_t core_heartbeat;
		memset(&core_heartbeat, 0, sizeof(core_heartbeat_t));

		ret = vastai_pci_mem_read(priv, die_index, CORE_HEARTBEAT_ADDR,
					  &core_heartbeat, CORE_HEARTBEAT_LEN);

		VATOOLS_DBG(NULL, DUMMY_DIE_ID,
			    "die inex=%x read heartbeat ret=%ld\n",
			    kdata.device.die_index, ret);

		ret = copy_to_user(kdata.output_buf.buf_addr, &core_heartbeat,
				   CORE_HEARTBEAT_LEN);
		if (ret) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
		}
		goto out;
	} break;

	case PCIE_CMD_FLAG_BARRESET: {
		u32 bar_readdata = 0;
		ret = 0;
		ret = vatools_pci_mem_read(priv, die_index, DDR_PCIE_BARRESET,
					   &bar_readdata, sizeof(u32));
		if (ret < 0) {
			ret = -EFAULT;
			goto out;
		}
		VATOOLS_DBG(NULL, DUMMY_DIE_ID, "bar_readdata: ret=%x",
			    bar_readdata);

		ret = copy_to_user((u8 *)kdata.output_buf.buf_addr +
					   sizeof(TS_SMI_CMD_REQ),
				   &bar_readdata, sizeof(u32));
		if (ret < 0) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
			goto out;
		}
	} break;

	case PCIE_CMD_FLAG_BARSTATUS: {
		int i = 0;
		struct bar_info at_addr_array[VASTAI_PCI_BAR_NUM];
		memset(at_addr_array, 0x0,
		       sizeof(struct bar_info) * VASTAI_PCI_BAR_NUM);
		for (; i < VASTAI_PCI_BAR_NUM; i++) {
			VATOOLS_DBG(NULL, DUMMY_DIE_ID, "bar_addr:%lx",
				    priv->bar[i].at_addr);
			at_addr_array[i] = priv->bar[i];
		}
		ret = copy_to_user(kdata.output_buf.buf_addr, at_addr_array,
				   sizeof(struct bar_info) *
					   VASTAI_PCI_BAR_NUM);
		if (ret < 0) {
			VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld",
				    ret);
			ret = -EFAULT;
			goto out;
		}
	} break;

	case PCIE_CMD_FLAG_REBOOT_DDRBW:
		VATOOLS_DBG(
			priv, die_index,
			"smi trigger die_index=%x : reboot ddrbw. current ddr_bw_test=%d\n",
			die_index, priv->ddr_bw_test);
		priv->ddr_bw_test = 1;
		ret = vastai_device_reset(priv, VASTAI_RESET_MSG_TRIGER |
							VASTAI_RESET_BOOT_DEV);
		// priv->ddr_bw_test = 0;
		break;

	case PCIE_CMD_FLAG_REBOOT_PCIEDMA: {
		unsigned long long int dma_addr;
		struct dma_buf *dmabuf = NULL;
		unsigned int test_buf_len = 0x400000;
		unsigned int val_0x8bc0744 = 2;
		VATOOLS_DBG(
			priv, die_index,
			"smi trigger die_index=%x current ddr_bw_test=%d: pcie dma test\n",
			die_index, priv->ddr_bw_test);
		if (priv->ddr_bw_test == 1) {
			// 没有切换至 ddrbw test 模式, 不允许进行 pcie dma 测试
			if (iommu_is_enable(priv))
				dmabuf = vastai_alloc_dma_buf_sg(priv,
								 test_buf_len);
			else
				dmabuf = vastai_alloc_dma_buf(priv,
							      test_buf_len);
			if (dmabuf) {
				struct vastai_dma_buf *dm =
					get_vastai_dma_buf(dmabuf, 0);
				dma_addr = dm->dma_bus_addr;

				/* set 0x8bc0748 and 0x8bc0744 */
				ret = vastai_pci_mem_write_direct(priv,
								  die_index,
								  0x8bc0748,
								  &dma_addr, 4);
				if (ret < 0) {
					VATOOLS_ERR(
						priv, die_index,
						"ERROR: reset pcie fw failed. Please save dmesg for debug. (0x8bc0748)\n");
					VATOOLS_ERR(
						priv, die_index,
						"ERROR: Please poweroff server/pc for recover\n");
					VATOOLS_ERR(
						priv, die_index,
						"ERROR: die index %x errn %ld\n",
						die_index, ret);
					ret = -EFAULT;
					break;
				}

				ret = vastai_pci_mem_write_direct(
					priv, die_index, 0x8bc0744,
					&val_0x8bc0744, 4);
				if (ret < 0) {
					VATOOLS_ERR(
						priv, die_index,
						"ERROR: reset pcie fw failed. Please save dmesg for debug. (0x8bc0744n)");
					VATOOLS_ERR(
						priv, die_index,
						"ERROR: Please poweroff server/pc for recover\n");
					VATOOLS_ERR(
						priv, die_index,
						"ERROR: die index %x errn %ld\n",
						die_index, ret);
					ret = -EFAULT;
					break;
				}
			} else {
				ret = -ENOMEM;
				VATOOLS_ERR(
					priv, die_index,
					"ERROR: die index %x dma_addr is null. errn %ld\n",
					die_index, ret);
				break;
			}

		} else {
			ret = -ENODEV;
			break;
		}

	} break;

	default:
		VATOOLS_DBG(NULL, DUMMY_DIE_ID,
			    "Pcie cmd id %d don't support in current system\n",
			    kdata.flag);
		ret = -EINVAL;
		break;
	}

out:
	kdata.errcode = ret;
	ret2 = copy_to_user(argp, &kdata, sizeof(kdata));
	if (ret2) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user ret=%ld", ret);
		ret2 = -EFAULT;
	}
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "out: ret=%ld   ret2=%d", ret, ret2);
	VATOOLS_FUNC_EXIT;

	return ret2;
}

/// 内部测试使用
static int vatools_ioctl_test(struct file *filp, unsigned int cmd,
			      unsigned long arg)
{
	long ret = -EIO;
	// struct vatools_node*   node   = NULL;
	// struct vatools_reader* reader = NULL;
	// void __user*           argp   = ( void __user* )arg;

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	V_UNREFERENCE(arg);
	VATOOLS_FUNC_ENTERY;

	// node   = vatools_file_get_node( filp );
	// reader = vatools_file_get_reader( filp );

	return ret;
}

/******************************************************************************
 * vatools
 *
 *
 */

static loff_t vatools_llseek(struct file *filp, loff_t offset, int orig)
{
	loff_t ret = 0;
	struct vatools_node *node = vatools_file_get_node(filp);

	mutex_lock(&node->node_mutex);
	ret = app_fops[vatools_get_app_category(filp)].llseek(filp, offset,
							      orig);
	mutex_unlock(&node->node_mutex);
	return ret;
}

static ssize_t vatools_read(struct file *filp, char __user *buf, size_t size,
			    loff_t *pos)
{
	ssize_t ret = 0;
	struct vatools_node *node = vatools_file_get_node(filp);

	mutex_lock(&node->node_mutex);
	ret = app_fops[vatools_get_app_category(filp)].read(filp, buf, size,
							    pos);
	mutex_unlock(&node->node_mutex);
	return ret;
}

static ssize_t vatools_write(struct file *filp, const char __user *buf,
			     size_t size, loff_t *pos)
{
	ssize_t ret = 0;
	struct vatools_node *node = vatools_file_get_node(filp);

	mutex_lock(&node->node_mutex);
	ret = app_fops[vatools_get_app_category(filp)].write(filp, buf, size,
							     pos);
	mutex_unlock(&node->node_mutex);
	return ret;
}

static ssize_t vatools_write_iter(struct kiocb *iocb, struct iov_iter *from)
{
	ssize_t ret = 0;
	struct vatools_node *node = vatools_file_get_node(iocb->ki_filp);

	mutex_lock(&node->node_mutex);
	ret = app_fops[vatools_get_app_category(iocb->ki_filp)].write_iter(
		iocb, from);
	mutex_unlock(&node->node_mutex);
	return ret;
}

static unsigned int vatools_poll(struct file *filp,
				 struct poll_table_struct *wait)
{
	int ret = 0;
	struct vatools_node *node = vatools_file_get_node(filp);

	mutex_lock(&node->node_mutex);
	ret = app_fops[vatools_get_app_category(filp)].poll(filp, wait);
	mutex_unlock(&node->node_mutex);
	return ret;
}

static long vatools_ioctl(struct file *filp, unsigned int cmd,
			  unsigned long arg)
{
	long ret = -EIO;
	struct vatools_node *node = NULL;
	struct vatools_reader *reader = NULL;

	node = vatools_file_get_node(filp);
	reader = vatools_file_get_reader(filp);

	VATOOLS_DBG(NULL, DUMMY_DIE_ID,
		    "ioctl cmd=0x%x node=0x%p reader=0x%p app_category=%d", cmd,
		    node, reader, reader->trans_category.app_category);

	// 根据命令指定 app_category, 无论用户是否设置了 app_category
	switch (cmd) {
	case VATOOLS_GENERAL_IOCTL_LOG_STATUS:
		reader->trans_category.app_category = VATOOLS_APP_LOGGER;
		break;
#ifdef CONFIG_TOOLS_V2
	case VATOOLS_IOCTL_V2_SMI_CMD:
		reader->trans_category.app_category = VATOOLS_APP_SMI;
		break;
	case VATOOLS_IOCTL_V2_GET_DEVICE_INFO:
		reader->trans_category.app_category = VATOOLS_APP_SMI;
		break;
	case VATOOLS_IOCTL_V2_PROFILER:
		reader->trans_category.app_category = VATOOLS_APP_PROFILER;
		break;
	case VATOOLS_IOCTL_V2_FLASH:
		reader->trans_category.app_category = VATOOLS_APP_SMI;
		break;
	case VATOOLS_IOCTL_V2_GUARD:
		reader->trans_category.app_category = VATOOLS_APP_SMI;
		break;
	case VATOOLS_IOCTL_V2_SHARE_MEM:
		reader->trans_category.app_category = VATOOLS_APP_SHAREDMEM;
		break;
	case VATOOLS_IOCTL_V2_VIDEO_DEBUGGER:
		reader->trans_category.app_category = VATOOLS_APP_DEBUGGER;
		break;
	case VATOOLS_IOCTL_V2_AI_DEBUGGER:
		reader->trans_category.app_category = VATOOLS_APP_DEBUGGER;
		break;
	case VATOOLS_IOCTL_V2_DEVICE_EVENT:
		reader->trans_category.app_category = VATOOLS_APP_EVENT;
		break;
	case VATOOLS_IOCTL_V2_CORE_TIME_SYNC:
		reader->trans_category.app_category = VATOOLS_APP_PROFILER;
		break;
#endif
	default:
		// reader->trans_category.app_category = VATOOLS_APP_SMI;
		break;
	}

#ifdef CONFIG_TOOLS_V2
	//==========================================================
	/// V2 版本 ioctl
	ret = -0x0effffff; // 不可能的返回值
	switch (cmd) {
		/// 测试命令
	case VATOOLS_IOCTL_V2_TEST: {
		// ret = vatools_ioctl_test(filp, cmd, arg);
	} break;
		/// 进程锁机制, 优先处理
	case VATOOLS_IOCTL_V2_GUARD: {
		// ret = vatools_ioctl_process_mutex_cmd(filp, cmd, arg);
	} break;
		/// 设置打开的文件fd的类别
	case VATOOLS_IOCTL_V2_SET_APP_CATEGORY: {
		// ret = vatools_ioctl_set_app_category(filp, cmd, arg);
	} break;
		/// 获取device info
	case VATOOLS_IOCTL_V2_GET_DEVICE_INFO: {
		/// 读取后缓存, 在设置命令后, 清空缓存, 重新读取
		ret = vatools_ioctl_v2_get_device_info(filp, cmd, arg);
	} break;
		/// 通过 smi 读写数据, 先写,触发中断到smcu, 后读, 读写通过一次调用完成
	case VATOOLS_IOCTL_V2_SMI_CMD: {
		// ret = vatools_ioctl_smi_cmd(filp, cmd, arg);
	} break;
		/// 普通方式读取数据; 只使用vastai_pcie 读取; 地址转换判断在host/common里面完成
	case VATOOLS_IOCTL_V2_READ_BUF: {
		// ret = vatools_ioctl_read_buf(filp, cmd, arg);
	} break;
		/// 普通方式写入数据; 只使用vastai_pcie 写入; 地址转换判断在host/common里面完成
	case VATOOLS_IOCTL_V2_WRITE_BUF: {
		// ret = vatools_ioctl_write_buf(filp, cmd, arg);
	} break;
	case VATOOLS_IOCTL_V2_PCIE_CMD: {
		// ret = vatools_ioctl_pcie_cmd(filp, cmd, arg);
	} break;
	case VATOOLS_IOCTL_V2_PROFILER: {
		// ret = vatools_ioctl_profiler_cmd(filp, cmd, arg);
	} break;
	case VATOOLS_IOCTL_V2_FLASH: {
		// ret = vatools_ioctl_flash_cmd(filp, cmd, arg);
	} break;
	case VATOOLS_IOCTL_V2_SHARE_MEM: {
		// ret = vatools_ioctl_sharedmem_cmd(filp, cmd, arg);
	} break;
	case VATOOLS_IOCTL_V2_AI_DEBUGGER: {
		// ret = vatools_ioctl_ai_debugger_cmd(filp, cmd, arg);
	} break;
	case VATOOLS_IOCTL_V2_CORE_TIME_SYNC: {
		// ret = vatools_ioctl_core_time_sync_cmd(filp, cmd, arg);
	} break;
	case VATOOLS_IOCTL_V2_LOG_STATUS: {
		// ret = vatools_ioctl_get_log_status_cmd(filp, cmd, arg);
	} break;
	case VATOOLS_IOCTL_V2_DMA: {
		// ret = vatools_ioctl_dma_cmd(filp, cmd, arg);
	} break;

	default: {
		VATOOLS_DBG(NULL, DUMMY_DIE_ID, "unknown v2 ioctl cmd 0x%x ",
			    cmd);
		break;
	}
	}
	if (ret != -0x0effffff) {
		reader->ioctl_version = VATOOLS_IOCTL_V2_MAGIC;
		// exit ioctl v2 here
		VATOOLS_DBG(NULL, DUMMY_DIE_ID, "v2 ioctl cmd 0x%x return %ld",
			    cmd, ret);
		return ret;
	}
	/// V2 版本 ioctl 结束V
	//==========================================================
#endif

	mutex_lock(&node->node_mutex);
	mutex_lock(&node->node_sharedmem_mutex);
	switch (cmd) {
	case VATOOLS_GENERAL_IOCTL_TEST:
		ret = vatools_ioctl_test(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_SET_APP_CATEGORY:
		ret = vatools_set_app_category(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_GET_DEVICE_NUM:
		ret = vatools_get_device_num(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_GET_DEV_INFO:
		ret = vatools_get_dev_info(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_GET_DIE_INFO:
		ret = vatools_get_die_info(filp, cmd, arg);
		break;
	case VATOOLS_GENERAL_IOCTL_READ:
		ret = vatools_driver_read(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_WRITE:
		ret = vatools_driver_write(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_FETCH:
		ret = vatools_driver_smi_fetch(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_GET_DEV_PCIE_BUS_ID:
		ret = vatools_get_dev_pcie_bus_id(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_GET_PCIE_ID:
		ret = vatools_get_dev_pcie_id(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_GET_PCIE_SUB_ID:
		ret = vatools_get_dev_pcie_sub_id(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_GET_MAJOR_MINOR_ID:
		ret = vatools_get_dev_major_minor_id(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_GET_DIE_RENDER:
		ret = vatools_get_die_render(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_PCIE:
		ret = vatools_pcie_ioctl(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_DEVICE_CAPABILITY:
		ret = vatools_get_device_capability(filp, cmd, arg);
		break;

		//分配 DMA 内存
	case VATOOLS_GENERAL_IOCTL_DMA_ALLOC:
		mutex_unlock(&node->node_sharedmem_mutex);
		mutex_unlock(&node->node_mutex);
		ret = vatools_ioctl_dma_alloc(filp, cmd, arg);
		goto out;
		break;
		//开始 DMA 传输数据
	case VATOOLS_GENERAL_IOCTL_DMA_START:
		mutex_unlock(&node->node_sharedmem_mutex);
		mutex_unlock(&node->node_mutex);
		ret = vatools_ioctl_dma_start(filp, cmd, arg);
		goto out;
		break;
	case VATOOLS_GENERAL_IOCTL_DMA_TRANS:
		mutex_unlock(&node->node_sharedmem_mutex);
		mutex_unlock(&node->node_mutex);
		ret = vatools_ioctl_dma_trans(filp, cmd, arg);
		goto out;
		break;

		//普通方式读取数据
	case VATOOLS_GENERAL_IOCTL_READ_BUF:
		ret = vatools_ioctl_read_reg_buf(filp, cmd, arg);
		break;

		//普通方式写入数据
	case VATOOLS_GENERAL_IOCTL_WRITE_BUF:
		ret = vatools_ioctl_write_reg_buf(filp, cmd, arg);
		break;
	case VATOOLS_GENERAL_IOCTL_GET_BMCU_XSPI_VERSION:
		ret = vatools_ioctl_get_bmcu_xspi_ver(filp, cmd, arg);
		break;
	case VATOOLS_GENERAL_IOCTL_SRIOV_NUMBER:
		ret = vatools_ioctl_get_sriov_number(filp, cmd, arg);
		break;
	case VATOOLS_GENERAL_IOCTL_FLASH_BMCU_FW:
		ret = vatools_ioctl_flash_bmcu_fw(filp, cmd, arg);
		break;
	case VATOOLS_GENERAL_IOCTL_FLASH_XSPI_FW:
		ret = vatools_ioctl_flash_xspi_fw(filp, cmd, arg);
		break;
	case VATOOLS_GENERAL_IOCTL_GET_FLASH_STA:
		ret = vatools_ioctl_flash_sta(filp, cmd, arg);
		break;
	case VATOOLS_GENERAL_IOCTL_SWITCH_BMCU:
		ret = vatools_ioctl_bmcu_switch(filp, cmd, arg);
		break;

		//core timestamp sync
	case VATOOLS_GENERAL_IOCTL_CORE_TIME_SYNC:
		ret = vatools_ioctl_core_time_sync(filp, cmd, arg);
		break;

	case VATOOLS_GENERAL_IOCTL_DEVICE_EVENT:
		mutex_unlock(&node->node_sharedmem_mutex);
		mutex_unlock(&node->node_mutex);
		ret = vatools_ioctl_device_event(filp, cmd, arg);
		goto out;
		break;
	case VATOOLS_GENERAL_IOCTL_DEVICE_EVENT_SELECT:
		mutex_unlock(&node->node_sharedmem_mutex);
		mutex_unlock(&node->node_mutex);
		ret = vatools_ioctl_device_event_select(filp, cmd, arg);
		goto out;
		break;

	default:
		// 只有在shared_memory 时 使用内部锁控制，其他的保持原流程暂不考虑并发
		// logger 使用自己的 ioctl和 mutex
		if (reader->trans_category.app_category ==
			    VATOOLS_APP_SHAREDMEM ||
		    reader->trans_category.app_category == VATOOLS_APP_LOGGER ||
		    VATOOLS_GENERAL_IOCTL_LOG_STATUS == cmd) {
			mutex_unlock(&node->node_sharedmem_mutex);
			mutex_unlock(&node->node_mutex);
		}

		VATOOLS_DBG(NULL, DUMMY_DIE_ID,
			    "ioctl_cmd=0x%X run app_fops.ioctl ", cmd);

		ret = app_fops[vatools_get_app_category(filp)].unlocked_ioctl(
			filp, cmd, arg);
		if (reader->trans_category.app_category ==
			    VATOOLS_APP_SHAREDMEM ||
		    reader->trans_category.app_category == VATOOLS_APP_LOGGER ||
		    VATOOLS_GENERAL_IOCTL_LOG_STATUS == cmd) {
			VATOOLS_DBG(NULL, DUMMY_DIE_ID, "ioctl return %ld",
				    ret);
			return ret;
		} else {
			break;
		}
	}
	mutex_unlock(&node->node_sharedmem_mutex);
	mutex_unlock(&node->node_mutex);
out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "ioctl return %ld", ret);
	return ret;
}

static int vatools_mmap(struct file *filp, struct vm_area_struct *pvm)
{
	int ret = 0;
	struct vatools_node *node = vatools_file_get_node(filp);
	V_UNREFERENCE(pvm);

	mutex_lock(&node->node_mutex);
	ret = app_fops[vatools_get_app_category(filp)].mmap(filp, pvm);
	mutex_unlock(&node->node_mutex);
	return ret;
}

static int vatools_open(struct inode *inode, struct file *filp)
{
	struct vatools_node *node = NULL;
	int ret = 0;
	struct vatools_reader *reader = NULL;

	V_UNREFERENCE(inode);
	VATOOLS_FUNC_ENTERY;
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "filp->f_mode= 0x%x", filp->f_mode);

	ret = nonseekable_open(inode, filp);
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "nonseekable_open ret = %d",
			    ret);
		return ret;
	}

	node = vatools_get_node_from_minor(MINOR(inode->i_rdev));
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "vatools_node *node= 0x%p", node);

	if (!node) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, " node = NULL");
		return -ENODEV;
	}

	reader = (struct vatools_reader *)kvmalloc(
		sizeof(struct vatools_reader), GFP_KERNEL);
	if (!reader) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "kvmalloc error. size=%ld",
			    sizeof(struct vatools_reader));
		return -ENOMEM;
	}

	memset(reader, 0, sizeof(struct vatools_reader));

	reader->node = node;
	reader->r_ver = 0;
	reader->r_all = in_egroup_p(inode->i_gid) || capable(CAP_SYSLOG);
	reader->profiler_checkpoint_corebitmap = 0;
	reader->event_timestamp_ns = 0;

	// 设置默认值
	reader->trans_category.app_category = VATOOLS_APP_DUMMY;
	reader->trans_category.block_id = SMI_CMD_BLOCK_ID_0;
	reader->trans_category.device.dev_id = 0; //0x7FFFFFFF;
	reader->trans_category.device.die_id = 0; //0x7FFFFFFF;
	reader->trans_category.device.die_index = 0; //0x7FFFFFFF;
	reader->event_timestamp_ns = 0;
#ifdef CONFIG_TOOLS_V2
	mutex_lock(&node->node_reader_mutex);
#else
	mutex_lock(&node->node_mutex);
	mutex_lock(&node->node_sharedmem_mutex);
#endif
	INIT_LIST_HEAD(&reader->list_readers);
	list_add_tail(&reader->list_readers, &node->list_node_readers);
	filp->private_data = reader;
#ifdef CONFIG_TOOLS_V2
	mutex_unlock(&node->node_reader_mutex);
#else
	mutex_unlock(&node->node_sharedmem_mutex);
	mutex_unlock(&node->node_mutex);
#endif

	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "node= 0x%p  reader= 0x%p", node,
		    reader);

	VATOOLS_FUNC_EXIT;
	return 0;
}

static int vatools_release(struct inode *ignored, struct file *filp)
{
	struct vatools_node *node = vatools_file_get_node(filp);
	struct vatools_reader *reader = vatools_file_get_reader(filp);

	V_UNREFERENCE(ignored);
	VATOOLS_FUNC_ENTERY;
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "node= 0x%p , filp= 0x%p", node, filp);

#ifdef CONFIG_TOOLS_V2
	// if (reader->ioctl_version == VATOOLS_IOCTL_V2_MAGIC)
	{
		// TODO:  优先释放进程间锁  mutex
		// vatools_process_release_mutex_unlock(filp, 0x0, 0x0, NULL,
		// 				     NULL);
		mutex_lock(&node->node_mutex);
		mutex_lock(&node->node_sharedmem_mutex);
		mutex_lock(&node->node_reader_mutex);
		app_fops[vatools_get_app_category(filp)].release(ignored, filp);
		reader = filp->private_data;
		if (reader) {
			// pr_info( "vatools_release reader =%p ", reader );
			list_del(&reader->list_readers);
			kvfree(reader);
			filp->private_data = NULL;
		}
		mutex_unlock(&node->node_reader_mutex);
		mutex_unlock(&node->node_sharedmem_mutex);
		mutex_unlock(&node->node_mutex);
	}
#else
	{
		mutex_lock(&node->node_mutex);
		mutex_lock(&node->node_sharedmem_mutex);
		app_fops[vatools_get_app_category(filp)].release(ignored, filp);
		reader = (struct vatools_reader *)filp->private_data;
		if (reader) {
			// pr_info( "vatools_release reader = %p ", reader );
			list_del(&reader->list_readers);
			kvfree(reader);
			filp->private_data = NULL;
		}
		mutex_unlock(&node->node_sharedmem_mutex);
		mutex_unlock(&node->node_mutex);
	}
#endif
	VATOOLS_FUNC_EXIT;
	return 0;
}

static int vatools_fasync(int fd, struct file *filp, int mode)
{
	int ret = 0;
	struct vatools_node *node = vatools_file_get_node(filp);

	mutex_lock(&node->node_mutex);
	mutex_lock(&node->node_sharedmem_mutex);
	// pr_info( "vatools_fasync reader " );
	ret = app_fops[vatools_get_app_category(filp)].fasync(fd, filp, mode);
	mutex_unlock(&node->node_sharedmem_mutex);
	mutex_unlock(&node->node_mutex);
	return ret;
}

static const struct file_operations vatools_fops = {
	.owner = THIS_MODULE,
	.llseek = vatools_llseek,
	.read = vatools_read,
	.write = vatools_write,
	.write_iter = vatools_write_iter,
	.poll = vatools_poll,
	.unlocked_ioctl = vatools_ioctl,
	.compat_ioctl = vatools_ioctl,
	.mmap = vatools_mmap,
	.open = vatools_open,
	.release = vatools_release,
	.fasync = vatools_fasync,
};

/******************************************************************************
 *
 *
 *
 */
static int create_node(void *pci_dev, const char *node_name, int size)
{
	int ret = 0;
	struct vatools_node *node = NULL;
	unsigned char *buffer = NULL;
	struct vastai_fifo *fifo = NULL;
	char vatools_dev_name[VATOOLS_DEVICE_NAME_LEN] = { 0 };

	VATOOLS_FUNC_ENTERY;

	buffer = (unsigned char *)kvmalloc(size, GFP_KERNEL);
	if (buffer == NULL) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "kvmalloc error. size=%d",
			    size);
		return -ENOMEM;
	}

	node = (struct vatools_node *)kzalloc(sizeof(struct vatools_node),
					      GFP_KERNEL);
	if (node == NULL) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "node is NULL");
		ret = -ENOMEM;
		goto out_free_buffer;
	}

	memset(vatools_dev_name, 0, sizeof(char) * VATOOLS_DEVICE_NAME_LEN);
	sprintf(vatools_dev_name, "%s", node_name);

	node->node_buffer = buffer;
	node->size = size;
	node->pci_dev = pci_dev;

	node->misc.minor = MISC_DYNAMIC_MINOR;
	node->misc.name = kstrdup(vatools_dev_name, GFP_KERNEL);
	if (node->misc.name == NULL) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, " node->misc.name == NULL ");
		ret = -ENOMEM;
		goto out_free_node;
	}

	node->misc.fops = &vatools_fops;
	node->misc.parent = NULL;

	node->misc.mode = 0666;

	init_waitqueue_head(&node->wq);
	INIT_LIST_HEAD(&node->list_node_readers);
	mutex_init(&node->node_mutex);
	mutex_init(&node->node_sharedmem_mutex);
	mutex_init(&node->node_common_mutex);
	mutex_init(&node->node_profiler_mutex);
	mutex_init(&node->node_debugger_mutex);
	mutex_init(&node->node_logger_mutex);
	mutex_init(&node->node_smi_mutex);
#ifdef CONFIG_TOOLS_V2
	mutex_init(&node->node_dummy_mutex);
	mutex_init(&node->node_driver_mutex);
	mutex_init(&node->node_reader_mutex);
#endif
	mutex_init(&pcie_timeline_dma_mutex);
	mutex_init(&pcie_timeline_msg_mutex);

	// init timeline of pci dma
	node->profiler_dma_ringbuf =
		kvmalloc(PROFILER_PCI_DMA_FIFO_LENGTH, GFP_KERNEL);
	if (node->profiler_dma_ringbuf == NULL) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "profile dma ringbuf is NULL");
		ret = -ENOMEM;
		goto out_free_misc_name;
	}
	gp_profiler_dma_fifo = node->profiler_dma_ringbuf;
	fifo = (struct vastai_fifo *)node->profiler_dma_ringbuf;
	fifo->elem_count =
		(PROFILER_PCI_DMA_FIFO_LENGTH - sizeof(struct vastai_fifo)) /
		sizeof(profiler_dma_checkpoint_t);
	fifo->elem_size = sizeof(profiler_dma_checkpoint_t);
	//spin_lock_init(&node->profiler_dma_ringbuf_lock);

	// init timeline of pci msg
	node->profiler_msg_ringbuf =
		kvmalloc(PROFILER_PCI_MSG_FIFO_LENGTH, GFP_KERNEL);
	if (node->profiler_msg_ringbuf == NULL) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "profile msg ringbuf is NULL");
		ret = -ENOMEM;
		goto out_free_dma_fifo;
	}
	gp_profiler_msg_fifo = node->profiler_msg_ringbuf;
	fifo = (struct vastai_fifo *)node->profiler_msg_ringbuf;
	fifo->elem_count =
		(PROFILER_PCI_MSG_FIFO_LENGTH - sizeof(struct vastai_fifo)) /
		sizeof(profiler_msg_checkpoint_t);
	fifo->elem_size = sizeof(profiler_msg_checkpoint_t);
	//spin_lock_init(&node->profile_msg_ringbuf_lock);

	// init event ringbuf
	node->event_ringbuf = kvmalloc(EVENT_FIFO_LENGTH, GFP_KERNEL);
	if (node->event_ringbuf == NULL) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "event_ringbuf is NULL");
		ret = -ENOMEM;
		goto out_free_msg_fifo;
	}
	memset(node->event_ringbuf, 0, EVENT_FIFO_LENGTH);
	fifo = (struct vastai_fifo *)node->event_ringbuf;
	fifo->elem_count = (EVENT_FIFO_LENGTH - sizeof(struct vastai_fifo)) /
			   sizeof(device_event_t);
	fifo->elem_size = sizeof(device_event_t);
	mutex_init(&node->event_mutex);
	init_waitqueue_head(&node->event_wait_queue);

	INIT_LIST_HEAD(&node->list_nodes);

	INIT_LIST_HEAD(
		&node->list_node_process_mutex); // TODO:  使用内核mutex实现进程互斥保护

	// INIT_LIST_HEAD( &node->t_sharedmem_mem_info_head );
	// INIT_LIST_HEAD( &node->t_sharedmem_signal_info_head );

	list_add_tail(&node->list_nodes, vatools_get_vastai_head());

	/* finally, initialize the misc device for this node */
	ret = misc_register(&node->misc);
	if (unlikely(ret)) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID,
			    "failed to register misc device for node '%s'!\n",
			    node->misc.name);
		goto out_free_event_fifo;
	}

	VATOOLS_INFO(NULL, DUMMY_DIE_ID,
		     "[SUCCESS] pci_dev=%p created %lu bytes on node '%s'\n",
		     pci_dev, (unsigned long)node->size, node->misc.name);

	return 0;
out_free_event_fifo:
	kfree(node->event_ringbuf);

out_free_msg_fifo:
	kfree(node->profiler_dma_ringbuf);

out_free_dma_fifo:
	kfree(node->profiler_dma_ringbuf);

out_free_misc_name:
	kvfree((void *)node->misc.name);

out_free_node:
	kvfree(node);

out_free_buffer:
	kvfree(buffer);
	VATOOLS_FUNC_EXIT;
	VATOOLS_INFO(
		NULL, DUMMY_DIE_ID,
		"[FAILED] pci_dev=%p created %lu bytes on node '%s' failed! ret=%d \n",
		pci_dev, (unsigned long)size, node_name, ret);
	return ret;
}

/******************************************************************************
 *
 *
 *
 */

static int vatools_driver_init(void *priv)
{
	int ret = 0;
	VATOOLS_FUNC_ENTERY;
	ret = create_node(priv, VATOOLS_DEV_NAME, VATOOLS_BUFFER_SIZE);
	if (unlikely(ret)) {
		VATOOLS_INFO(NULL, DUMMY_DIE_ID,
			     "[VATOOLS] init %s failed. ret=%d\n",
			     VATOOLS_DEV_NAME, ret);
		goto out;
	}
	VATOOLS_INFO(NULL, DUMMY_DIE_ID, "[VATOOLS] init %s OK\n",
		     VATOOLS_DEV_NAME);
out:
	VATOOLS_FUNC_EXIT;
	return ret;
}

static void vatools_driver_exit(void *pci_dev)
{
	struct vatools_node *current_node = NULL, *next_node = NULL;
	V_UNREFERENCE(pci_dev);
	VATOOLS_FUNC_ENTERY;

	list_for_each_entry_safe (current_node, next_node,
				  vatools_get_vastai_head(), list_nodes) {
		VATOOLS_INFO(NULL, DUMMY_DIE_ID, "[VATOOLS] deinit %s OK\n",
			     current_node->misc.name);
		list_del(&current_node->list_nodes);
		/* we have to delete all the entry inside headlist */
		misc_deregister(&current_node->misc);
		kvfree(current_node->node_buffer);
		kvfree((void *)current_node->misc.name);

		// free profiler timeline buffer
		kvfree(current_node->profiler_dma_ringbuf);
		kvfree(current_node->profiler_msg_ringbuf);
		kvfree(current_node->event_ringbuf);
		gp_profiler_dma_fifo = NULL;
		gp_profiler_msg_fifo = NULL;
		current_node->event_ringbuf = NULL;

		// free node
		kvfree(current_node);
	}

	VATOOLS_FUNC_EXIT;
}

//*************************************************************************************
// 加载/卸载驱动时, 更新数据
int vatools_update_data_probe(void *priv)
{
	V_UNREFERENCE(priv);
	return 0;
}

int vatools_update_data_remove(void *priv)
{
	V_UNREFERENCE(priv);
	return 0;
}

//*************************************************************************************
// 对于所有的device,vatools只加载一次. 使用此变量标识device计数 init 时
// 如果等于0 加载驱动; exit 时,如果等于0, 卸载驱动.
static atomic_t _vatools_init_device_count = ATOMIC_INIT(0);

// pcie 驱动调用函数, 初始化vatools驱动. 只有一个节点,和device, die不相关.
int tools_init(void *priv)
{
	int ret = 0;

	VATOOLS_INFO(NULL, DUMMY_DIE_ID,
		     "priv=0x%p , current device_count=%d \n", priv,
		     atomic_read(&_vatools_init_device_count));

	if (0 == atomic_read(&_vatools_init_device_count)) {
		ret = vatools_driver_init(priv);
	}
	atomic_inc(&_vatools_init_device_count);
	VATOOLS_INFO(NULL, DUMMY_DIE_ID, "total device_count=%d \n",
		     atomic_read(&_vatools_init_device_count));
	ret = vatools_update_data_probe(priv);
	return ret;
}

void tools_exit(void *priv)
{
	atomic_dec(&_vatools_init_device_count);

	VATOOLS_INFO(NULL, DUMMY_DIE_ID, "priv=0x%p, device_count=%d\n", priv,
		     atomic_read(&_vatools_init_device_count));

	if (0 == atomic_read(&_vatools_init_device_count)) {
		vatools_driver_exit(priv);
	}
	vatools_update_data_remove(priv);
	return;
}

#if VATOOLS_INDEPENDANCE_DRIVER

static int __init vatools_init(void)
{
	return tools_init(NULL);
}
static void __exit vatools_exit(void)
{
	tools_exit(NULL);
}

module_init(vatools_init);
module_exit(vatools_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Frank Bai<frank.bai@vastaitech.com>");
MODULE_DESCRIPTION("vatools driver");
MODULE_VERSION("0.1");

#endif

#pragma pack()
