/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: frank.bai
 */
#include "vatools.h"
#include "vatools_log.h"


/******************************************************************************
 * logger
 *
 *
 */

//获取 firmware log 状态
int vatools_ioctl_get_log_status(struct file *filp, unsigned int cmd,
				 unsigned long arg)
{
	int ret = 0;
	// struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	struct vatools_cmd vcmd;
	void __user *argp = (void __user *)arg;
	struct vastai_pci_info *priv = NULL;
	int die_index = 0;
	union die_index_data did;
	u32 flag = 0;
	u32 timeout_water = 0;
	struct vastai_sv100_die *die = NULL;
	u8 buf[464] = { 0 };

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;

	memset(&vcmd, 0, sizeof(vcmd));
	ret = copy_from_user(&vcmd, (void *)argp, sizeof(vcmd));
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_from_user is error\n");
		ret = -EIO;
		goto out;
	}

	did.val = (u32)vcmd.cmd.log_poll_cmd.die_index;
	die_index = vcmd.cmd.log_poll_cmd.die_index;

	priv = vatools_get_vastai_pci_device_info((char)did.dev_id);

	if ((atomic_read(&priv->pci_state) != VASTAI_NORMAL_STATE)) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "priv->pci_state= 0x%x\n",
			    atomic_read(&priv->pci_state));
		ret = -ENODEV;
		goto out;
	}

	/* trigger smcu to poll log buf */
	flag = 0;
	timeout_water = vcmd.cmd.log_poll_cmd.timeout_water != 0x0 ?
				vcmd.cmd.log_poll_cmd.timeout_water :
				500;
	die = &(((struct vastai_pci_info *)priv)
			->dies[vastai_pci_get_die_id(priv, die_index)]);

	vastai_pci_mem_write(priv, die_index, LOG_MSG_FLAG, &flag,
			     sizeof(flag));
	/* wait complete */
	ret = wait_for_completion_interruptible_timeout(
		&(die->log_comp), HZ * timeout_water / 1000);
	reinit_completion(&(die->log_comp));

	if (ret == 0) {
		VATOOLS_DBG(NULL, DUMMY_DIE_ID, "timeout! (%d ms) ret=%d\n",
			    timeout_water, ret);
		vcmd.cmd.log_poll_cmd.is_timeout = 1;
		ret = -EIO;
		goto out;
	} else if (ret == -ERESTARTSYS || ret == -EINTR) {
		VATOOLS_DBG(NULL, DUMMY_DIE_ID, "process be killed! ret=%d\n",
			    ret);
		ret = -EINTR;
		goto out;
	} else if (ret < 0) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "Unknow err! ret=%d\n", ret);
		ret = -EIO;
		goto out;
	}

	/* get log ring ctl */
	/* we need size == 0x1D0 buf */
	ret = vastai_pci_mem_read(priv, vcmd.cmd.log_poll_cmd.die_index,
				  LOG_CTRL_INFO, buf, 0x1D0);
	ret = copy_to_user(vcmd.cmd.log_poll_cmd.log_ring_ctl, buf, 0x1D0);
	if (ret) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "copy_to_user is error\n");
		ret = -EIO;
		goto out;
	}
out:
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "ret=%d\n", ret);

	VATOOLS_FUNC_EXIT;

	return ret;
}

loff_t logger_llseek(struct file *filp, loff_t offset, int orig)
{
	V_UNREFERENCE(filp);
	V_UNREFERENCE(offset);
	V_UNREFERENCE(orig);

	return 0;
}

ssize_t logger_read(struct file *filp, char __user *buf, size_t size,
		    loff_t *pos)
{
	V_UNREFERENCE(filp);
	V_UNREFERENCE(buf);
	V_UNREFERENCE(size);
	V_UNREFERENCE(pos);

	return 0;
}

ssize_t logger_write(struct file *filp, const char __user *buf, size_t size,
		     loff_t *pos)
{
	V_UNREFERENCE(filp);
	V_UNREFERENCE(buf);
	V_UNREFERENCE(size);
	V_UNREFERENCE(pos);

	return 0;
}

ssize_t logger_write_iter(struct kiocb *iocb, struct iov_iter *from)
{
	V_UNREFERENCE(iocb);
	V_UNREFERENCE(from);

	return 0;
}

unsigned int logger_poll(struct file *filp, struct poll_table_struct *wait)
{
	V_UNREFERENCE(filp);
	V_UNREFERENCE(wait);

	return 0;
}

long logger_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	long ret = -EIO;
	struct vatools_node *node = NULL;
	struct vatools_reader *reader = NULL;

	VATOOLS_FUNC_ENTERY;

	node = vatools_file_get_node(filp);
	reader = vatools_file_get_reader(filp);

	VATOOLS_DBG(NULL, DUMMY_DIE_ID,
		    "ioctl cmd=0x%X node=0x%p reader=0x%p app_category=%d", cmd,
		    node, reader, reader->trans_category.app_category);
	mutex_lock(&node->node_logger_mutex);

	switch (cmd) {
		//获取 firmware log status
	case VATOOLS_GENERAL_IOCTL_LOG_STATUS:
		ret = vatools_ioctl_get_log_status(filp, cmd, arg);
		break;
	default:
		VATOOLS_DBG(
			NULL, DUMMY_DIE_ID,
			"ioctl logger unknown cmd=0x%X node=0x%p reader=0x%p app_category=%d",
			cmd, node, reader, reader->trans_category.app_category);
		break;
	}
	mutex_unlock(&node->node_logger_mutex);
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "ioctl ret=%ld", ret);
	VATOOLS_FUNC_EXIT;

	return ret;
}

int logger_mmap(struct file *filp, struct vm_area_struct *pvm)
{
	V_UNREFERENCE(filp);
	V_UNREFERENCE(pvm);

	return 0;
}
int logger_open(struct inode *inode, struct file *filp)
{
	V_UNREFERENCE(inode);
	V_UNREFERENCE(filp);

	return 0;
}

int logger_release(struct inode *ignored, struct file *filp)
{
	struct vatools_node *p_dev = NULL;
	struct vatools_reader *reader = NULL;

	V_UNREFERENCE(ignored);
	VATOOLS_FUNC_ENTERY;

	p_dev = vatools_file_get_node(filp);
	if (!p_dev) {
		VATOOLS_ERR(NULL, DUMMY_DIE_ID, "p_dev is null.");
		return -ENOMEM;
	}
	reader = vatools_file_get_reader(filp);
	if (reader) {
		list_del(&reader->list_readers);
		kvfree(reader);
		filp->private_data = NULL;
	} else {
	}

	VATOOLS_FUNC_EXIT;
	return 0;
}

int logger_fasync(int fd, struct file *filp, int mode)
{
	V_UNREFERENCE(fd);
	V_UNREFERENCE(filp);
	V_UNREFERENCE(mode);

	return 0;
}
